/* addfile.c
 *
 * Copyright (C) 1997 Angus J. C. Duggan
 *
 * Add a PC file to a BeebEm disc image.
 *
 */

#include <windows.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include <string.h>

#if defined(MSDOS) || defined(WINNT)
#include <fcntl.h>
#include <io.h>
#endif

#define DEBUG(x) /* nowt */

static char *program = NULL ;

void usage(void)
{
  fprintf(stderr, "Copyright (C) Angus J. C. Duggan, 1997\n") ;
  fprintf(stderr, "Usage: %s diskimage file...\n", program) ;
  exit(1) ;
}

void error(char *format, ...)
{
  va_list args ;
  char *scan = format, ch ;

  fprintf(stderr, "%s: ", program) ;
  va_start(args, format) ;
  while ( (ch = *scan++) != '\0' ) {
    if ( ch == '%' ) {
      switch ( *scan++ ) {
      case 'd':
	fprintf(stderr, "%d", va_arg(args, long)) ;
	break ;
      case 's':
	fprintf(stderr, "%s", va_arg(args, char *)) ;
	break ;
      case '%':
	fputc(ch, stderr) ;
	break ;
      default:
	fprintf(stderr, "Bad format string\n") ;
	exit(1) ;
      }
    } else
      fputc(ch, stderr) ;
  }
  fputc('\n', stderr) ;
  fflush(stderr) ;
  va_end(args) ;

  exit(1) ;
}

/* Extract catalogue information */
void cat(unsigned char *catalogue)
{
  int nfiles = catalogue[256 + 5] >> 3;
  long last = (catalogue[256 + 7] | ((catalogue[256 + 6] & 0x3) << 8));
  long maxgap = 0 ;
  int index ;

  printf("Sectors=%x\n", last) ;

  /* Catalogue stored in reverse sector order */
  for (index = 0 ; index < nfiles * 8 ; index += 8) {
    long load = catalogue[256 + 8 + index] |
      (catalogue[256 + 9 + index] << 8) |
      ((catalogue[256 + 14 + index] & 0x0c) << 14);
    long exec = catalogue[256 + 10 + index] |
      (catalogue[256 + 11 + index] << 8) |
      ((catalogue[256 + 14 + index] & 0xc0) << 10);
    long length = catalogue[256 + 12 + index] |
      (catalogue[256 + 13 + index] << 8) |
      ((catalogue[256 + 14 + index] & 0x30) << 12);
    long start = catalogue[256 + 15 + index] |
      ((catalogue[256 + 14 + index] & 0x03) << 8);
    long end = start + ((length + 255) >> 8) ;
    char filename[8] ;
    int j ;

    if ( load & 0x20000 ) /* extend load address */
      load |= 0xfffc0000 ;

    if ( exec & 0x20000 ) /* extend load address */
      exec |= 0xfffc0000 ;

    /* Filename is low 7 bits of first sector's index entry */
    for (j = 0 ; j < 7 ; ++j)
      filename[j] = (catalogue[8 + index + j] & 0x7f) ;
    filename[7] = '\0' ;

    printf("%c.%-7s %06x %06x %06x %c %03x-%03x\n",
	   (catalogue[15 + index] & 0x7f), filename,
	   (load & 0xffffff), (exec & 0xffffff), length,
	   (catalogue[15 + index] & 0x80) ? 'L' : ' ',
	   start, end) ;

    if ( last - end > maxgap )
      maxgap = last - end;

    last = start;
  }

  if ( last - 2 > maxgap )
    maxgap = last - 2;

  printf("Files=%d, Max gap=%x\n", nfiles, maxgap * 256) ;
}

/* Extract catalogue information */
void addtodisk(FILE *diskimage, unsigned char *catalogue, FILE *addfile,
	       char *beebname, long load, long exec, long len, int locked)
{
  /* verify beebname */
  char dir = '$', *leafname = beebname ;
  int sectors = (len + 255) >> 8 ;
  int nfiles = catalogue[256 + 5] >> 3;
  long last = (catalogue[256 + 7] | ((catalogue[256 + 6] & 0x3) << 8));
  long bestgap = last, bestindex = -1, beststart = 0 ;
  int index, namelen = strlen(beebname) ;

  for (index = 0 ; index < namelen ; ++index) {
    if (beebname[index] <= 32 || beebname[index] >= 127)
      error("Illegal character in BBC filename %s\n", beebname) ;
    switch (beebname[index]) {
    case '.':
      if (index == 1 && leafname == beebname) {
	dir = beebname[0] ;
	leafname = beebname + 2 ;
	break ;
      }
    case '*': case '#': case '"':
      error("Illegal character in BBC filename %s\n", beebname) ;
    }
  }

  if ( strlen(leafname) > 7 )
    error("BBC filename %%s longer than 7 characters\n", beebname) ;

  if ( nfiles >= 31 )
    error("No catalogue space to add file %s\n", beebname) ;

  /* Look for best-fit space; take exact fit or carve largest space. Also
     check filename for conflict. */

  /* Catalogue stored in reverse sector order */
  for (index = 0 ; index < nfiles * 8 ; index += 8) {
    long length = catalogue[256 + 12 + index] |
      (catalogue[256 + 13 + index] << 8) |
      ((catalogue[256 + 14 + index] & 0x30) << 12);
    long start = catalogue[256 + 15 + index] |
      ((catalogue[256 + 14 + index] & 0x03) << 8);
    long end = start + ((length + 255) >> 8) ;
    long gap = last - end ;
    char filename[8] ;
    int j ;

    /* Filename is low 7 bits of first sector's index entry */
    for (j = 0 ; j < 7 ; ++j) {
      if ( (filename[j] = (catalogue[8 + index + j] & 0x7f)) <= ' ' )
	filename[j] = '\0';
    }
    filename[7] = '\0' ;

    if ( _stricmp(filename, leafname) == 0 &&
	 tolower(catalogue[15 + index] & 0x7f) == tolower(dir) )
      error("File %s already on disc image\n", beebname) ;

    if ( gap >= sectors && gap < bestgap ) {
      bestindex = index ;
      bestgap = gap ;
      beststart = end ;
    }

    last = start;
  }

  last -= 2 ; /*  space from start to first file, 2 sectors for catalogue */

  if ( last >= sectors && last < bestgap ) {
    bestindex = index ;
    bestgap = last ;
    beststart = 2 ;
  }

  if ( bestindex < 0 )
    error("Not enough space to add file %s\n", beebname) ;

  printf("Adding %-9s %06x %06x %06x %c %03x-%03x\n",
	 beebname, (load & 0xffffff), (exec & 0xffffff),
	 len, locked ? 'L' : ' ', beststart, sectors + beststart) ;

  if ( bestindex < index ) { /* Shuffle up entries */
    memmove(catalogue + bestindex + 16, catalogue + bestindex + 8,
	    index - bestindex) ;
    memmove(catalogue + bestindex + 16 + 256, catalogue + bestindex + 8 + 256,
	    index - bestindex) ;
  }

  catalogue[256 + 5] += 8 ; /* Increase number of files by one */

  for (index = 0 ; index < 7 ; ++index) {
    catalogue[8 + bestindex + index] = *leafname ? *leafname++ : ' ' ;
  }
  catalogue[15 + bestindex] = locked ? (dir | 0x80) : dir ;
  catalogue[256 + 8 + bestindex] = (load & 0xff) ;
  catalogue[256 + 9 + bestindex] = (load & 0xff00) >> 8;
  catalogue[256 + 10 + bestindex] = (exec & 0xff) ;
  catalogue[256 + 11 + bestindex] = (exec & 0xff00) >> 8;
  catalogue[256 + 12 + bestindex] = (len & 0xff) ;
  catalogue[256 + 13 + bestindex] = (len & 0xff00) >> 8;
  catalogue[256 + 14 + bestindex] = ((beststart & 0x3) >> 8) |
    ((load & 0x30000) >> 14) | ((len & 0x30000) >> 12) |
    ((exec & 0x30000) >> 10);
  catalogue[256 + 15 + bestindex] = (beststart & 0xff);

  /* add file to diskimage */
  if ( fseek(diskimage, beststart * 256, SEEK_SET) == 0 &&
       fseek(addfile, 0, SEEK_SET) == 0 &&
       ftell(diskimage) == beststart * 256 ) {
    char sector[256] ;
    long nread = 0 ;

    while ( (nread = fread(sector, sizeof(char), 256, addfile)) > 0 ) {
      --sectors;
      len -= nread;
      if ( nread < 256 ) /* clear remainder of sector */
	memset(sector + nread, 0, 256 - nread);

      if ( fwrite(sector, sizeof(char), 256, diskimage) != 256 )
	error("I/O error writing file %s to diskimage\n", beebname) ;
    }
    if ( len != 0 || sectors != 0 )
      error("Length of %s written doesn't match expected\n", beebname) ;
  } else
    error("Can't position diskimage to add file %s\n", beebname) ;

  /* write catalogue to diskimage */
  if ( fseek(diskimage, 0, SEEK_SET) != 0 ||
       fwrite(catalogue, sizeof(char), 512, diskimage) != 512 )
    error("Write failed for diskimage catalogue\n") ;
}

int main(int argc, char *argv[])
{
  FILE *diskimage = NULL ;
  unsigned char catalogue[512] ; /* two 256 byte sectors */

  for (program = *argv++ ; --argc ; argv++ ) {
    if ( argv[0][0] == '-' ) {
      switch ( argv[0][1] ) {
      default:
	usage();
      }
    } else if (!diskimage) {
      if ((diskimage = fopen(*argv, "rb+")) == NULL)
	error("can't open disk image %s\n", *argv);
      if ( fread(catalogue, sizeof(char), 512, diskimage) != 512 )
	error("can't read catalogue from %s\n", *argv) ;
      cat(catalogue) ;
    } else {
      FILE *addfile = NULL, *inffile = NULL ;
      char infname[MAX_PATH] ;
      char beebname[32] ; /* For safety... */
      long exec = -1, load = -1, len = -1 ;
      int locked = FALSE ;

      strcpy(infname, *argv) ;
      strcat(infname, ".inf") ;

      if ((addfile = fopen(*argv, "rb")) == NULL)
	error("can't open additional file %s\n", *argv);

      if ( fseek(addfile, 0, SEEK_END) != 0 ||
	   (len = ftell(addfile)) < 0 ||
	   fseek(addfile, 0, SEEK_SET) != 0 )
	error("can't get length of file %s\n", *argv) ;

      if ((inffile = fopen(infname, "r")) != NULL) {
	unsigned long inflen ;
	char inflock[32] ;
	int nread = fscanf(inffile, "%s %lx %lx %lx %s", beebname,
			   &load, &exec, &inflen, inflock) ;

	switch (nread) {
	case 0:
	  error("INF file %s empty\n", infname);
	case 5: /* locked? */
	  if ( _stricmp(inflock, "locked") == 0 ||
	       _stricmp(inflock, "l") == 0 )
	    locked = TRUE ;
	case 4: /* len */
	  if ( inflen != len )
	    fprintf(stderr, "INF file length %x differs from file length %x\n",
		    infname, inflen, len) ;
	case 3: /* exec */
	case 2: /* load */
	case 1: /* beebname */
	  ;
	}

	fclose(inffile) ;
      } else { /* Find leaf filename */
	char *leaf = *argv;
	char *slash = strrchr(leaf, '\\') ;
	char *bslash = strrchr(leaf, '/') ;

	if (slash && bslash) {
	  if (bslash > slash)
	    leaf = bslash + 1;
	  else
	    leaf = slash + 1;
	} else if (slash) {
	  leaf = slash + 1;
	} else if (bslash) {
	  leaf = bslash + 1;
	}

	strcpy(beebname, leaf) ;
      }

      if (beebname[0] == '\0')
	error("Empty BBC filename for %s\n", *argv) ;

      addtodisk(diskimage, catalogue, addfile,
		beebname, load, exec, len, locked) ;

      fclose(addfile) ;
    }
  }

  if (diskimage) {
    cat(catalogue) ;
    fclose(diskimage) ;
  }

  return 0 ;
}
