/*
 * $Id: db.c,v 2.10 90/06/15 21:32:08 tynor Exp $
 *----------------------------------------------------------------------------
 *	FPLAN - Flight Planner
 *	Steve Tynor
 *	tynor@prism.gatech.edu
 *
 *	This program is in the public domain. Permission to copy,
 * distribute, modify this program is hearby given as long as this header
 * remains. If you redistribute this program after modifying it, please
 * document your changes so that I do not take the blame (or credit) for
 * those changes.  If you fix bugs or add features, please send me a
 * patch so that I can keep the 'official' version up-to-date.
 *
 *	Bug reports are welcome and I'll make an attempt to fix those
 * that are reported.
 *
 *	USE AT YOUR OWN RISK! I assume no responsibility for any
 * errors in this program, its database or documentation. I will make an
 * effort to fix bugs, but if you crash and burn because, for example,
 * fuel estimates in this program were inaccurate, it's your own fault
 * for trusting somebody else's code! Remember, as PIC, it's _your_
 * responsibility to do complete preflight planning. Use this program as
 * a flight planning aid, but verify its results before using them.
 *----------------------------------------------------------------------------
 */

#include "config.h"
#include <stdio.h>
#include <ctype.h>
#include "wp_info.h"
#include "proto.h"

static char rcsid[] = "$Id: db.c,v 2.10 90/06/15 21:32:08 tynor Exp $";

/*
 * the database file pointers:
 */
typedef struct {
   FILE  *fp;
   long  rec_size;
   long  num_recs;
   enum { COMMA_NAV, COMMA_APT, VOR, AIRPORT, INVALID_DB } kind ;
} DB_FILE;

DB_FILE pub_airports, pvt_airports, pub_vors, pvt_vors, comma_apt, comma_nav ;

#define FSEEK_FROM_BEGINNING 0
#define FSEEK_FROM_CURRENT   1
#define FSEEK_FROM_END       2

/*----------------------------------------------------------------------------*/
static long get_rec_size (char *buffer)
{
   long i;

   for (i = 0;; i++) {
      if (buffer[i] == '\n') {
	 return (i + 1L);
      }
   }
}

/*----------------------------------------------------------------------------*/
static long get_num_recs (long rec_size, FILE **fp, char *filename)
{
   long file_size;

   if (fseek (*fp, 0L, FSEEK_FROM_END)) {
      fprintf (stderr, "ERROR: could not seek in db file: %s\n", filename);
      *fp = NULL;
      return (0L);
   }
   file_size = ftell (*fp);

//#ifndef MSDOS
   /*
    * can't figure out why MSC doesn't do this properly - maybe it's an 
    * inconsistency in the way fseek works? For now, just assume that 
    * database is in the proper format.
    */
   if ((file_size % rec_size) != 0L) {
      fprintf (stderr, 
	       "ERROR: file size (%ld/%ld) inconsistency in db file: %s\n", 
	       file_size, rec_size, filename);
      *fp = NULL;
      return (0L);
   }
//#endif
   return (file_size / rec_size);
}

/*----------------------------------------------------------------------------*/
BOOLEAN open_dbs (void)
{
   char *pub_dir, *pvt_dir;
   BOOLEAN ok;

#define BUFSIZE 200
   char filename[BUFSIZE];
   char buffer[BUFSIZE];

#ifdef MSDOS
#define OPEN_MODE "rb"
#else
#define OPEN_MODE "r"
#endif

   if (! (pub_dir = getenv (NAV_PUBLIC)))
      pub_dir = DEFAULT_PUB_DIRECTORY;

   if (! (pvt_dir = getenv (NAV_PRIVATE)))
      pvt_dir = DEFAULT_PVT_DIRECTORY;

#ifndef MSDOS
   if (pvt_dir[0] == '~') {
      strcpy (filename, getenv (HOME));
      strcat (filename, &pvt_dir[1]);
      pvt_dir = strdup (filename);
   }
#endif

   strcpy (filename, pub_dir);
   strcat (filename, COMMA_DOT "apt");
   if ((comma_apt.fp = fopen (filename, OPEN_MODE)) != NULL) {
      comma_apt.rec_size = -1 ;
      comma_apt.num_recs = -1 ;
      comma_apt.kind = COMMA_APT ;
   } else
      comma_apt.kind = INVALID_DB ;

   strcpy (filename, pub_dir);
   strcat (filename, AIRPORTS_NAME);
   if ((pub_airports.fp = fopen (filename, OPEN_MODE)) != NULL) {
      fgets (buffer, BUFSIZE, pub_airports.fp);
      pub_airports.rec_size = get_rec_size (buffer);
      pub_airports.num_recs = get_num_recs (pub_airports.rec_size,
					    &pub_airports.fp, filename);
      pub_airports.kind = AIRPORT ;
   } else
      pub_airports.kind = INVALID_DB ;

   strcpy (filename, pvt_dir);
   strcat (filename, AIRPORTS_NAME);
   if ((pvt_airports.fp = fopen (filename, OPEN_MODE)) != NULL) {
      fgets (buffer, BUFSIZE, pvt_airports.fp);
      pvt_airports.rec_size = get_rec_size (buffer);
      pvt_airports.num_recs = get_num_recs (pvt_airports.rec_size,
					    &pvt_airports.fp, filename);
      pvt_airports.kind = AIRPORT ;
   } else
      pvt_airports.kind = INVALID_DB ;

   strcpy (filename, pub_dir);
   strcat (filename, COMMA_DOT "nav");
   if ((comma_nav.fp = fopen (filename, OPEN_MODE)) != NULL) {
      comma_nav.rec_size = -1 ;
      comma_nav.num_recs = -1 ;
      comma_nav.kind = COMMA_NAV ;
   } else
      comma_nav.kind = INVALID_DB ;

   strcpy (filename, pub_dir);
   strcat (filename, VORS_NAME);
   if ((pub_vors.fp = fopen (filename, OPEN_MODE)) != NULL) {
      fgets (buffer, BUFSIZE, pub_vors.fp);
      pub_vors.rec_size = get_rec_size (buffer);
      pub_vors.num_recs = get_num_recs (pub_vors.rec_size,
					&pub_vors.fp, filename);
      pub_vors.kind = VOR ;
   } else
      pub_vors.kind = INVALID_DB ;

   strcpy (filename, pvt_dir);
   strcat (filename, VORS_NAME);
   if ((pvt_vors.fp = fopen (filename, OPEN_MODE)) != NULL) {
      fgets (buffer, BUFSIZE, pvt_vors.fp);
      pvt_vors.rec_size = get_rec_size (buffer);
      pvt_vors.num_recs = get_num_recs (pvt_vors.rec_size,
					&pvt_vors.fp, filename);
      pvt_vors.kind = VOR ;
   } else
      pvt_vors.kind = INVALID_DB ;

   ok = TRUE;

   if (!comma_apt.fp && !pub_airports.fp && !pvt_airports.fp) {
      ok = FALSE;
      fprintf (stderr, "ERROR: neither public or private airports db found\n");
   }
   if (!comma_nav.fp && !pub_vors.fp && !pvt_vors.fp) {
      ok = FALSE;
      fprintf (stderr, "ERROR: neither public or private vors db found\n");
   }
   return (ok);
}

/*----------------------------------------------------------------------------*/
BOOLEAN close_dbs (void)
{
   if (comma_apt.fp)
      fclose (comma_apt.fp);
   if (pub_airports.fp)
      fclose (pub_airports.fp);
   if (pvt_airports.fp)
      fclose (pvt_airports.fp);
   if (comma_nav.fp)
      fclose (comma_nav.fp);
   if (pub_vors.fp)
      fclose (pub_vors.fp);
   if (pvt_vors.fp)
      fclose (pvt_vors.fp);

   return (TRUE);
}


/*----------------------------------------------------------------------------*/
static BOOLEAN lookup_desig_internal (DB_FILE *db, char *desig,
				      int buffer_size, char *buffer)
{
   long low, high, mid;
   int  cmp;
   if (! db->fp) 
      return FALSE;

   if ( db->kind == COMMA_NAV || db->kind == COMMA_APT ) {
     rewind (db->fp);   
     while (fgets (buffer, buffer_size, db->fp)) {
       if (!strcmp (desig, strtok (buffer, ",\n")))
	 return TRUE;
     }
   } else {
     low = 0L;
     high = db->num_recs;
     while (low <= high) {
       mid = (low + high) / 2L;
       fseek (db->fp, mid * db->rec_size, FSEEK_FROM_BEGINNING);
       fgets (buffer, buffer_size, db->fp);
       cmp =strcmp (desig, strtok (buffer, ":\n"));
       if (!cmp)
	 return TRUE;
       else if (cmp > 0)
	 low = mid + 1;
       else
	 high = mid - 1;
     }
   }

   return FALSE;
}

/*----------------------------------------------------------------------------*/
static void decode_common (DATABASE_INFO *db)
{
   double d1, d2;

   db->altitude.value =  atof (strtok ((char*)0, ":\n"));
   db->altitude.valid = TRUE;

   d1 = atof (strtok ((char*)0, ":\n"));
   d2 = atof (strtok ((char*)0, ":\n"));
   db->mag_variation = d1 + d2 / 60.0;

   d1 = atof (strtok ((char*)0, ":\n"));
   d2 = atof (strtok ((char*)0, ":\n"));
   db->latitude = d1 + d2 / 60.0;

   d1 = atof (strtok ((char*)0, ":\n"));
   d2 = atof (strtok ((char*)0, ":\n"));
   db->longitude = d1 + d2 / 60.0;
}

/*----------------------------------------------------------------------------*/
static void decode_vor (DATABASE_INFO *db)
{
   char *tok;

   db->name = strdup (strtok ((char*)0, ":\n"));

   tok = strdup (strtok ((char*)0, ":\n"));   
   if (tok[0] != '\0') {
      db->freq.valid = TRUE;
      db->freq.value = atof (tok);
   } else
      db->freq.valid = FALSE;

   decode_common (db);

   tok = strtok ((char*)0, ":\n");
   if (!strcmp ("NDB", tok))
      db->mode = WP_NDB;
   else if (!strcmp ("VOR", tok))
      db->mode = WP_VOR;
   else if (!strcmp ("DME", tok))
      db->mode = WP_DME;
   else if (!strcmp ("TAC", tok))
      db->mode = WP_TAC;
   else if (!strcmp ("ILS", tok))
      db->mode = WP_ILS;
   else if (!strcmp ("INT", tok))
      db->mode = WP_NAMED_INTERSECTION;
   else if (!strcmp ("WPT", tok))
      db->mode = WP_WPT;
   else if (!strcmp ("LOM", tok))
      db->mode = WP_LOM;
   else if (!strcmp ("LMM", tok))
      db->mode = WP_LMM;
   else 
      db->mode = WP_UNK;

   tok = strtok ((char*)0, ":\n");   
   if (tok)
      db->comment = strdup (tok);
   else
      db->comment = NULL;
}

/*----------------------------------------------------------------------------*/
static void decode_airport (DATABASE_INFO *db)
{
   char *tok;

   db->mode = WP_AIRPORT;
   db->city = strdup (strtok ((char*)0, ":\n"));   
   db->name = strdup (strtok ((char*)0, ":\n"));   

   decode_common (db);

   tok = strdup (strtok ((char*)0, ":\n"));   

   if (tok[0] != '\0') {
      db->freq.valid = TRUE;
      db->freq.value = atof (tok);
   } else
      db->freq.valid = FALSE;

   tok = strtok ((char*)0, ":\n");   
   if (tok)
      db->comment = strdup (tok);
   else
      db->comment = NULL;
}

/*----------------------------------------------------------------------------*/
static double decode_lat_long(char *dms, char *s)
{
  double result ;
  char *end ;

  if ( s && isdigit(*s) ) {
    result = strtod(s, &end) / 3600.0 ;
  } else if ( dms && isdigit(*dms) ) {
    result = strtod(dms, &end) ;
    if ( *end == '-' )
      result += strtod(end + 1, &end) / 60.0 ;
    if ( *end == '-' )
      result += strtod(end + 1, &end) / 3600.0 ;
  } else {
    fprintf(stderr, "No latitude or longitude found\n") ;
    return 360.0 ;
  }

  switch ( *end ) {
  case 'e': case 's': case 'E': case 'S':
    result = -result ;
    break ;
  case 'w': case 'n': case 'W': case 'N': case '\0':
    break;
  default:
    fprintf(stderr, "Bad lat/long compass point %c\n", *end) ;
  }
  return result ;
}

/*----------------------------------------------------------------------------*/
static void decode_comma_nav (DATABASE_INFO *db)
{
   char *tok, *prev ;

   /* Ignore irrelevant fields */
   tok = strtok((char *)0, ",\n") ;		/* Facility ID special chrs */

   tok = strtok ((char*)0, ",\n");		/* Facility type */
   if (!strcmp ("NDB", tok))
      db->mode = WP_NDB;
   else if (!strcmp ("VOR", tok))
      db->mode = WP_VOR;
   else if (!strcmp ("DME", tok))
      db->mode = WP_DME;
   else if (!strcmp ("VOR/DME", tok))
      db->mode = WP_DME;
   else if (!strcmp ("NDB/DME", tok))
      db->mode = WP_DME;
   else if (!strcmp ("TAC", tok))
      db->mode = WP_TAC;
   else if (!strcmp ("VORTAC", tok))
      db->mode = WP_VOR;
   else if (!strcmp ("ILS", tok))
      db->mode = WP_ILS;
   else if (!strcmp ("INT", tok))
      db->mode = WP_NAMED_INTERSECTION;
   else if (!strcmp ("WPT", tok))
      db->mode = WP_WPT;
   else if (!strcmp ("LOM", tok))
      db->mode = WP_LOM;
   else if (!strcmp ("LMM", tok))
      db->mode = WP_LMM;
   else 
      db->mode = WP_UNK;

   tok = strtok((char *)0, ",\n") ;		/* info date */

   db->name = strdup (strtok ((char*)0, ",\n")); /* Navaid name */

   db->city = strdup (strtok ((char*)0, ",\n")); /* City */

   tok = strtok((char *)0, ",\n") ;		/* State */
   tok = strtok((char *)0, ",\n") ;		/* Region */
   tok = strtok((char *)0, ",\n") ;		/* Country */
   tok = strtok((char *)0, ",\n") ;		/* Navaid system usage? */
   tok = strtok((char *)0, ",\n") ;		/* Public use? */

   db->comment = strdup(strtok((char *)0, ",\n")) ; /* Class */

   tok = strtok((char *)0, ",\n") ;		/* Hours of operation */
   tok = strtok((char *)0, ",\n") ;		/* ARTCC ID */

   prev = strtok((char *)0, ",\n") ;		/* Latitude H-M-S */
   tok = strtok((char *)0, ",\n") ;		/* Latitude S */
   db->latitude = decode_lat_long(prev, tok) ;

   prev = strtok((char *)0, ",\n") ;		/* Longitude H-M-S */
   tok = strtok((char *)0, ",\n") ;		/* Longitude S */
   db->longitude = decode_lat_long(prev, tok) ;

   tok = strtok((char *)0, ",\n") ;		/* Latitude of TACAN H-M-S */
   tok = strtok((char *)0, ",\n") ;		/* Longitude of TACAN H-M-S */

   tok = strtok((char *)0, ",\n") ;		/* Elevation */
   if (isdigit((int)*tok)) {
     db->altitude.value = atof(tok);
     db->altitude.valid = TRUE ;
   } else
     db->altitude.valid = FALSE;

   tok = strtok((char *)0, ",\n") ;		/* Radio callname */
   tok = strtok((char *)0, ",\n") ;		/* TACAN channel */

   tok = strtok((char *)0, ",\n") ;		/* Frequency */
   if (isdigit((int)*tok)) {
      db->freq.valid = TRUE;
      db->freq.value = atof (tok);
   } else
      db->freq.valid = FALSE;

   tok = strtok((char *)0, ",\n") ;		/* Xmit fan marker/ID */
   tok = strtok((char *)0, ",\n") ;		/* Bearing fan major axis */
   tok = strtok((char *)0, ",\n") ;		/* Protected freq. alt */
   tok = strtok((char *)0, ",\n") ;		/* Low alt facility? */
   tok = strtok((char *)0, ",\n") ;		/* NAVAID Z marker? */
   tok = strtok((char *)0, ",\n") ;		/* WX broadcast hours */
   tok = strtok((char *)0, ",\n") ;		/* WX phone # */
   tok = strtok((char *)0, ",\n") ;		/* Controlling FSS ID */
   tok = strtok((char *)0, ",\n") ;		/* Controlling FSS name */
   tok = strtok((char *)0, ",\n") ;		/* Controlling FSS hours */
   tok = strtok((char *)0, ",\n") ;		/* NOTAM account code ID */
   tok = strtok((char *)0, ",\n") ;		/* Quadrant/range leg bear */
}

/*----------------------------------------------------------------------------*/
static void decode_comma_apt (DATABASE_INFO *db)
{
   char *tok, *prev, *end ;

   db->mode = WP_AIRPORT;

   /* Ignore irrelevant fields */
   tok = strtok((char *)0, ",\n") ;		/* site number */
   tok = strtok((char *)0, ",\n") ;		/* site type */
   tok = strtok((char *)0, ",\n") ;		/* info date */
   tok = strtok((char *)0, ",\n") ;		/* FAA region code */
   tok = strtok((char *)0, ",\n") ;		/* FAA district code */
   tok = strtok((char *)0, ",\n") ;		/* State post office code */
   tok = strtok((char *)0, ",\n") ;		/* State */
   tok = strtok((char *)0, ",\n") ;		/* County */
   tok = strtok((char *)0, ",\n") ;		/* County's state */

   db->city = strdup (strtok ((char*)0, ",\n")); /* City */

   db->name = strdup (strtok ((char*)0, ",\n")); /* Facility name */

   tok = strtok((char *)0, ",\n") ;		/* Ownership type */
   tok = strtok((char *)0, ",\n") ;		/* Facility usage */
   tok = strtok((char *)0, ",\n") ;		/* Owner's phone */

   prev = strtok((char *)0, ",\n") ;		/* Latitude D-M-S */
   tok = strtok((char *)0, ",\n") ;		/* Latitude S */
   db->latitude = decode_lat_long(prev, tok) ;

   prev = strtok((char *)0, ",\n") ;		/* Longitude D-M-S */
   tok = strtok((char *)0, ",\n") ;		/* Longitude S */
   db->longitude = decode_lat_long(prev, tok) ;

   tok = strtok((char *)0, ",\n") ;		/* Reference point method */

   tok = strtok((char *)0, ",\n") ;		/* Elevation */
   if (isdigit((int)*tok)) {
     db->altitude.value = atof(tok);
     db->altitude.valid = TRUE ;
   } else
     db->altitude.valid = FALSE;

   tok = strtok((char *)0, ",\n") ;		/* Elevation method */

   tok = strtok((char *)0, ",\n") ;		/* Magnetic variation D */
   db->mag_variation = strtod(tok, &end);
   switch (*end) {
   case 'E': case 'e':
     db->mag_variation = -(db->mag_variation) ;
     break ;
   case 'W': case 'w':
     break ;
   default:
     fprintf(stderr, "Bad magnetic variation compass point\n") ;
   }

   tok = strtok((char *)0, ",\n") ;		/* Variation epoch */
   tok = strtok((char *)0, ",\n") ;		/* TPA */
   db->comment = strdup(strtok((char *)0, ",\n")) ; /* Sectional */
   tok = strtok((char *)0, ",\n") ;		/* Distance from biz centre */
   tok = strtok((char *)0, ",\n") ;		/* Direction from biz centre */
   tok = strtok((char *)0, ",\n") ;		/* Boundary ARTCC ID */
   tok = strtok((char *)0, ",\n") ;		/* Boundary ARTCC compute ID */
   tok = strtok((char *)0, ",\n") ;		/* Boundary ARTCC name */
   tok = strtok((char *)0, ",\n") ;		/* Responsible ARTCC ID */
   tok = strtok((char *)0, ",\n") ;		/* Responsible ARTCC comp ID */
   tok = strtok((char *)0, ",\n") ;		/* Responsible ARTCC name */
   tok = strtok((char *)0, ",\n") ;		/* FSS on field? */
   tok = strtok((char *)0, ",\n") ;		/* FSS ID */
   tok = strtok((char *)0, ",\n") ;		/* FSS name */
   tok = strtok((char *)0, ",\n") ;		/* FSS phone */
   tok = strtok((char *)0, ",\n") ;		/* Alt FSS ID */
   tok = strtok((char *)0, ",\n") ;		/* Alt FSS name */
   tok = strtok((char *)0, ",\n") ;		/* Alt FSS phone */
   tok = strtok((char *)0, ",\n") ;		/* WX info ID */
   tok = strtok((char *)0, ",\n") ;		/* NOTAM 'D' service? */
   tok = strtok((char *)0, ",\n") ;		/* Fuel types */
   tok = strtok((char *)0, ",\n") ;		/* Frame repair service? */
   tok = strtok((char *)0, ",\n") ;		/* Engine repair service? */
   tok = strtok((char *)0, ",\n") ;		/* Type of bottled oxygen */
   tok = strtok((char *)0, ",\n") ;		/* Type of bulk oxygen */
   tok = strtok((char *)0, ",\n") ;		/* Airport attendance */
   tok = strtok((char *)0, ",\n") ;		/* Airport lighting */
   tok = strtok((char *)0, ",\n") ;		/* Tower? */
   tok = strtok((char *)0, ",\n") ;		/* UNICOM */

   tok = strtok((char *)0, ",\n") ;		/* CTAF */
   if (isdigit((int)*tok)) {
     db->freq.valid = TRUE;
     db->freq.value = atof(tok) ;
   } else
     db->freq.valid = FALSE;

   tok = strtok((char *)0, ",\n") ;		/* Segmented circle? */
   tok = strtok((char *)0, ",\n") ;		/* Beacon colour */
}

/*----------------------------------------------------------------------------*/
static void str_upcase (char *str)
{
   int i;

   for (i = 0; str[i] != '\0'; i++)
      if (islower ((int)str[i]))
	  str[i] = toupper (str[i]);
}

/*----------------------------------------------------------------------------*/
static BOOLEAN in_cache (WAYPOINT_MODE mode, char *desig)
{
   int i;

   for (i = 0; i < num_cached; i++) {
      if ((!strcmp (desig, cache[i]->desig)) && 
	  (mode == cache[i]->mode)) {
	 return TRUE;
      }
   }
   return FALSE;
}

/*----------------------------------------------------------------------------*/
BOOLEAN lookup_desig (WAYPOINT_KIND kind, char *desig, DATABASE_INFO **db)
{
#undef BUFSIZE
#define BUFSIZE 1024
   char buffer [BUFSIZE];
   DB_FILE *fps[6];
   int i;
   int db_kind = INVALID_DB ;

   str_upcase (desig);

   if (kind == WP_VIA) {
      /* 
       * prefer the VOR databases
       */
      fps[0] = &pvt_vors;
      fps[1] = &comma_nav;
      fps[2] = &pub_vors;
      fps[3] = &pvt_airports;
      fps[4] = &comma_apt;
      fps[5] = &pub_airports;
   } else {
      /* 
       * prefer the AIRPORT databases 
       */
      fps[0] = &pvt_airports;
      fps[1] = &comma_apt;
      fps[2] = &pub_airports;
      fps[3] = &pvt_vors;
      fps[4] = &comma_nav;
      fps[5] = &pub_vors;
   }

   for (i = 0; i < 6; i++) {
      if ( fps[i]->kind != INVALID_DB &&
	   lookup_desig_internal (fps[i], desig, BUFSIZE, buffer) ) {
	db_kind = fps[i]->kind;
#if 0
	printf ("%s found w/db %d[%ld] (%s)\n",desig,i,fps[i]->fp,
		(db_kind == AIRPORT || db_kind == COMMA_APT ? "APT" : "VOR"));
#endif
	break;
      }
   }
   if ( db_kind == INVALID_DB ) {
      fprintf (stderr, "ERROR: could not find %s in any database\n", desig);
      return FALSE;
   }

   /*
    * now, convert it into a db record...
    */

   *db = (DATABASE_INFO*) malloc (sizeof (DATABASE_INFO));
   if (! *db)
      yyerror ("unable to allocate space for database element");

   (*db)->desig = desig;	/* no need to strdup - the parser already did */

   switch (db_kind) {
   case VOR:
      decode_vor (*db);
      break ;
   case AIRPORT:
      decode_airport (*db);
      break ;
   case COMMA_NAV:
      decode_comma_nav (*db);
      break ;
   case COMMA_APT:
      decode_comma_apt (*db);
      break ;
   default:
      yyerror ("unknown database type");
   }

   /*
    * append to the cache: (NOTE: it might already be there if we didn't grab
    * it out of the cach because of incompatible modes - e.g. the cached wp is
    * an airport, but this is a via node - we forced a recheck in case there was
    * a navaid with the same desig)
    */
   if (num_cached < CACHE_SIZE)
      if (! in_cache ((*db)->mode, desig)) {
	 cache[num_cached++] = *db;
      }
   return TRUE;
}

/*----------------------------------------------------------------------------*/
static int dbcmp (const void *p1, const void *p2)
{
   const DATABASE_INFO **db1 = p1, **db2 = p2 ;
   return (strcmp ((*db1)->desig, (*db2)->desig));
}

/*----------------------------------------------------------------------------*/
void put_db_summary (FILE *out_fp)
{
   int i;

   /*
    * first, sort the cache 
    */
   qsort ((char*) &cache[0], num_cached, sizeof(DATABASE_INFO*), dbcmp);

   /*
    * now, print it out.
    */
   fprintf (out_fp, "\f\n");
   for (i = 0; i < num_cached; i++)
      put_db (out_fp, cache[i]);
}


/*----------------------------------------------------------------------------*/
void min_max_lat_long (double *min_lat, double *max_lat,
		       double *min_long, double *max_long)
{
   int i;

   *min_lat  =  190.0;
   *max_lat  = -190.0;
   *min_long =  190.0;
   *max_long = -190.0;

   for (i = 0; i < num_cached; i++) {
      *min_lat = MIN (cache[i]->latitude, *min_lat);
      *max_lat = MAX (cache[i]->latitude, *max_lat);
      *min_long = MIN (cache[i]->longitude, *min_long);
      *max_long = MAX (cache[i]->longitude, *max_long);
   }
}
