;;; Simple POP (RFC 1939) client for VM
;;; Copyright (C) 1993, 1994, 1997, 1998 Kyle E. Jones
;;;
;;; Hacked to support non-UIDL download for VM-6.76 by Angus J. C. Duggan, 2000
;;;
;;; UNSUPPORTED NON-OFFICIAL VERSION. DO NOT RELY ON THIS CODE.
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(provide 'vm-pop)

(if (fboundp 'define-error)
    (progn
      (define-error 'vm-cant-uidl "Can't use UIDL")
      (define-error 'vm-dele-failed "DELE command failed")
      (define-error 'vm-uidl-failed "UIDL command failed"))
  (put 'vm-cant-uidl 'error-conditions '(vm-cant-uidl error))
  (put 'vm-cant-uidl 'error-message "Can't use UIDL")
  (put 'vm-dele-failed 'error-conditions '(vm-dele-failed error))
  (put 'vm-dele-failed 'error-message "DELE command failed")
  (put 'vm-uidl-failed 'error-conditions '(vm-uidl-failed error))
  (put 'vm-uidl-failed 'error-message "UIDL command failed"))

;; Our goal is to drag the mail from the POP maildrop to the crash box.
;; just as if we were using movemail on a spool file.
;; We remember which messages we have retrieved so that we can
;; leave the message in the mailbox, and yet not retrieve the
;; same messages again and again.

(defvar vm-pop-local-cache t "\
Non-nil means mail is cached on the local disk, but held on the POP server.
Mail is deleted from the server when the synchronised and marked as deleted.")

(defvar vm-pop-debug nil "\
If non-nil, leave full transcript in trace file.")

(defvar vm-pop-x-uidl-regexp "^X-UIDL:[ \t]*\\([^ \t\n]+\\)" "\
Regexp for normal X-UIDL headers.")

(defvar vm-pop-message-id-regexp "^Message-ID: *<\\([^>]+\\)>" "\
Regexp for normal message-ids.")

(defvar vm-pop-use-uidl t "\
Use POP3 UIDL command to cross-reference cached messages and server messages")

(defvar vm-pop-action-file-name (expand-file-name "~/.mailkill")
  "Name of kill file. Kill file contains a lisp-format list, of the form:

     ((match-expression... action) ...)

If there are multiple match expressions, then all of the expressions must
succeed for the action to be processed. Match expressions may be either a
regexp matching a header, like:

     \"^Subject:[ \t]*\\([Rr]e: \\)*subject-regexp[ \t]*$\"

or an expression of the form:

    (and match-expression match-expression ...) or
    (or match-expression match-expression ...) or
    (not match-expression)

The action should be either kill, ignore, deliver or an expression which will
be evaluated to return one of these symbols. An example action file which
kills mail with the subject \"Testing\" unless it came from user test is:

((and (not \"^From:.*<test>\") \"^Subject:[ \t]*Testing[ \t]*$\") kill))
"
)

(defvar vm-pop-log-file-name (expand-file-name "~/.maillog")
  "Name of a file to log killed message headers in.")

(defun vm-pop-move-mail (source destination)
  (let* ((process nil)
	 (folder-type vm-folder-type)
	 (previous-buffer (current-buffer))
	 (m-per-session vm-pop-messages-per-session)
	 (b-per-session vm-pop-bytes-per-session)
	 (handler (and (fboundp 'find-file-name-handler)
		       (condition-case ()
			   (find-file-name-handler source 'vm-pop-move-mail)
			 (wrong-number-of-arguments
			  (find-file-name-handler source)))))
	 (popdrop (vm-safe-popdrop-string source))
	 (statblob nil)
	 (have-list (vm-pop-message-list nil))
	 (messages-cache (vm-pop-cached-messages have-list))
	 (action-list (vm-pop-load-action-file vm-pop-action-file-name))
	 messages-server
	 size-alist uidl-alist different
	 (pop-retrieved-messages
	  (vm-pop-select-retrieved
	   (vm-popdrop-sans-password source) t vm-pop-retrieved-messages))
	 new-retrieved-messages
	 auto-expunge x
	 mailbox-count mailbox-size message-size response
	 n nn retrieved retrieved-bytes process-buffer uidl)
    (setq auto-expunge (cond (vm-pop-local-cache nil)
			     ((setq x (assoc source vm-pop-auto-expunge-alist))
			      (cdr x))
			     ((setq x (assoc (vm-popdrop-sans-password source)
					     vm-pop-auto-expunge-alist))
			      (cdr x))
			     (t vm-pop-expunge-after-retrieving)))
    (unwind-protect
	(catch 'done
	  (if handler
	      (throw 'done
		     (funcall handler 'vm-pop-move-mail source destination)))
	  (setq process (vm-pop-make-session source))
	  (or process (throw 'done nil))
	  (setq process-buffer (process-buffer process))
	  (save-excursion
	    (set-buffer process-buffer)
	    (setq vm-folder-type (or folder-type vm-default-folder-type))
	    ;; find out how many messages are in the box.
	    (vm-pop-send-command process "STAT")
	    (setq response (vm-pop-read-stat-response process)
		  mailbox-count (nth 0 response)
		  mailbox-size (nth 1 response))
	    ;; forget it if the command fails
	    ;; or if there are no messages present.
	    (if (or (null mailbox-count)
		    (< mailbox-count 1))
		(throw 'done nil))
	    ;; forget it if there are no new messages
	    (if (and (equal mailbox-count (length pop-retrieved-messages))
		     (equal mailbox-size
			    (apply '+ (mapcar '(lambda (x) (nth 3 x))
					      pop-retrieved-messages))))
		;; sync if cached messages marked as deleted, otherwise done
		(let ((nowt t) (cached-list have-list) msg)
		  (while (and cached-list nowt)
		    (setq msg (car cached-list)
			  nowt (not (vm-deleted-flag msg))
			  cached-list (cdr cached-list)))
		  (if nowt
		      (throw 'done nil))))
	    (message "Synchronising %d messages with %s..."
		     mailbox-count popdrop)
	    ;; loop through the maildrop retrieving and deleting
	    ;; messages as we go.
	    (setq n 1 nn 1 retrieved 0 retrieved-bytes 0)
	    (setq statblob (vm-pop-start-status-timer))
	    (vm-set-pop-stat-x-box statblob popdrop)
	    (vm-set-pop-stat-x-maxmsg statblob mailbox-count)
	    ;; get server list of sizes using LIST, and uidls using UIDL
	    (vm-pop-send-command process "LIST")
	    (setq size-alist (vm-pop-read-list-long-response process))
	    (if vm-pop-use-uidl
		(progn
		  (vm-pop-send-command process "UIDL")
		  (setq uidl-alist (vm-pop-read-uidl-long-response process))))
	    (while (and (<= n mailbox-count)
			(or (not (natnump m-per-session))
			    (< retrieved m-per-session))
			(or (not (natnump b-per-session))
			    (< retrieved-bytes b-per-session)))
	      (let* ((message-size (cdr (assoc n size-alist)))
		     (message-id (cdr (assoc n uidl-alist)))
		     (info (assoc n (mapcar 'reverse
					    pop-retrieved-messages)))
		     message have-headers headers action)
		(if (not message-size)
		    (progn
		      (vm-pop-send-command process (format "LIST %d" n))
		      (setq message-size (vm-pop-read-list-response process))))
		;; while there are no prior differences, assume matching size
		;; is matching message; find message-id for it
		(and (not different)
		     (not (setq different (/= (or (nth 1 info) -1)
					      message-size)))
		     (if message-id
			 (setq different (equal (nth 4 info) message-id))
		       (setq message-id (nth 4 info))))
		;; get headers and extract id if UIDL not available
		(if (not message-id)
		    (setq headers (vm-pop-get-message-headers process n)
			  message-id (and headers
					  (vm-pop-extract-message-uidl headers))))
		(vm-set-pop-stat-x-currmsg statblob n)
		(setq message (cdr (assoc message-id messages-cache))
		      messages-server (cons (cons message-id n)
					    messages-server))
		(cond ((and vm-pop-local-cache message)
		       ;; caching, and got the message already
		       (if (vm-deleted-flag message)
			   (progn
			     (message "Deleting message %d..." n)
			     (vm-pop-send-command process (format "DELE %d" n))
			     (and (null (vm-pop-read-response process))
				  (throw 'done (not (equal retrieved 0)))))
			 ;; update info about message
			 (setq new-retrieved-messages
			       (cons (list message-id
					   (vm-popdrop-sans-password source)
					   'uidl message-size nn)
				     new-retrieved-messages))
			 (vm-increment nn)))
		      (message
		       ;; not caching, but seen message already
		       (if vm-pop-ok-to-ask
			   (message
			    "Skipping message %d (of %d) from %s (retrieved already)..."
			    n mailbox-count popdrop)))
		      (t
		       ;; not caching, or don't have it in the cache
		       ;; headers are now needed for kill file and size limit
		       (if (and action-list (not headers))
			   (setq headers (vm-pop-get-message-headers process n)))
		       (if (and action-list headers)
			   (setq action (vm-pop-get-message-action headers
								   action-list)))
		       (vm-set-pop-stat-x-need statblob message-size)
		       ;; No auto-kill or skip, test if message is too large
		       (if (not action)
			   (setq action
				 (and (integerp vm-pop-max-message-size)
				      (> message-size vm-pop-max-message-size)
				      (if vm-pop-ok-to-ask
					  (vm-pop-ask-about-large-message
					   (or headers
					       (vm-pop-get-message-headers process n))
					   message-size n)
					'skip))))
		  (cond
		   ;; delete message without reading it
		   ((eq action 'delete)
		    (message "Deleting message %d..." n)
		    (vm-pop-send-command process (format "DELE %d" n))
		    (and (null (vm-pop-read-response process))
			 (throw 'done (not (equal retrieved 0)))))
		   ;; delete message without retrieving anything
		   ((eq action 'kill)
		    (message "Killing message %d..." n)
		    (vm-pop-send-command process (format "DELE %d" n))
		    (and (null (vm-pop-read-response process))
			 (throw 'done (not (equal retrieved 0))))
		    (vm-pop-kill-log vm-pop-log-file-name popdrop headers))
		   ;; skip message, but leave in mailbox
		   ((eq action 'skip)
		    (if vm-pop-ok-to-ask
			(message "Skipping message %d..." n)
		      (message "Skipping message %d in %s, too large (%d > %d)..."
			       n popdrop message-size vm-pop-max-message-size))
		    (vm-increment nn))
		   (t
		    ;; retrieve message, removing from mailbox if not caching
		    (message "Retrieving message %d (of %d) from %s..."
			     n mailbox-count popdrop)
		    (vm-pop-send-command process (format "RETR %d" n))
		    (and (null (vm-pop-read-response process))
			 (throw 'done (not (equal retrieved 0))))
		    (and (null (vm-pop-retrieve-to-crashbox process destination
							    statblob
							    message-size
							    message-id
							    popdrop))
			 (throw 'done (not (equal retrieved 0))))
		    (vm-increment retrieved)
		    (and b-per-session
			 (setq retrieved-bytes (+ retrieved-bytes message-size)))
		    (if (not auto-expunge)
			(progn
			  (setq new-retrieved-messages
				(cons (list message-id
					    (vm-popdrop-sans-password source)
					    'uidl message-size nn)
				      new-retrieved-messages))
			  (vm-increment nn))
		      ;; The user doesn't want the messages kept in 
		      ;; the mailbox. Delete the message now.
		      (vm-pop-send-command process (format "DELE %d" n))
		      ;; DELE can't fail but Emacs or this code might
		      ;; blow a gasket and spew filth down the
		      ;; connection, so...
		      (and (null (vm-pop-read-response process))
			   (throw 'done (not (equal retrieved 0)))))))))
		(vm-increment n)))
	    (if vm-pop-local-cache
		(progn
		  ;; expunge marked messages
		  (save-excursion
		    (set-buffer previous-buffer)
		    (vm-expunge-folder t))
		  ;; mark non-existant messages as deleted
		  (while messages-cache
		    (let ((uidl (car (car messages-cache)))
			  (msg (cdr (car messages-cache))))
		      (if (not (assoc uidl messages-server))
			  (vm-set-deleted-flag msg t)))
		    (setq messages-cache (cdr messages-cache)))))
	    (not (equal retrieved 0)) )
	  (setq vm-pop-retrieved-messages
		(append new-retrieved-messages
			(vm-pop-select-retrieved
			 (vm-popdrop-sans-password source)
			 nil
			 vm-pop-retrieved-messages))) )
      (and statblob (vm-pop-stop-status-timer statblob))
      (if process
	  (vm-pop-end-session process)))))

(defun vm-pop-check-mail (source)
  (let* ((process nil)
	 (handler (and (fboundp 'find-file-name-handler)
		       (condition-case ()
			   (find-file-name-handler source 'vm-pop-check-mail)
			 (wrong-number-of-arguments
			  (find-file-name-handler source)))))
	 (popdrop (vm-popdrop-sans-password source))
	 (retrieved (vm-pop-select-retrieved popdrop t vm-pop-retrieved-messages))
	 x response)
    (unwind-protect
	(save-excursion
	  (catch 'done
	    (if handler
		(throw 'done
		       (funcall handler 'vm-pop-check-mail source)))
	    (setq process (vm-pop-make-session source))
	    (or process (throw 'done nil))
	    (set-buffer (process-buffer process))
	    (if vm-pop-use-uidl
		(progn
		  (vm-pop-send-command process "UIDL")
		  (setq response (vm-pop-read-uidl-long-response process))))
	    (if (null response)
		;; server doesn't understand UIDL
		nil
	      (if (null (car response))
		  ;; (nil . nil) is returned if there are no
		  ;; messages in the mailbox.
		  (throw 'done nil)
		(while response
		  (if (not (and (setq x (assoc (cdr (car response)) retrieved))
				(equal (nth 1 x) popdrop)
				(eq (nth 2 x) 'uidl)))
		      (throw 'done t))
		  (setq response (cdr response))))
	      ;; all messages in the mailbox have already been retrieved
	      (throw 'done nil))
	    (vm-pop-send-command process "STAT")
	    (setq response (vm-pop-read-stat-response process))
	    (if (null response)
		nil
	      (if vm-pop-local-cache
		  (and (equal (car response)
			      (length retrieved))
		       (equal (nth 1 response)
			      (apply '+ (mapcar '(lambda (x) (nth 3 x))
						retrieved))))
		(not (equal 0 (car response)))))))
      (and process (vm-pop-end-session process)))))

(defun vm-expunge-pop-messages ()
  "Deletes all messages from POP mailbox that have already been retrieved
into the current folder.  VM sends POP DELE commands to all the
relevant POP servers to remove the messages."
  (interactive)
  (vm-follow-summary-cursor)
  (vm-select-folder-buffer)
  (vm-check-for-killed-summary)
  (vm-error-if-virtual-folder)
  (if (not vm-pop-local-cache)
      (let ((process nil)
	    (source nil)
	    (trouble nil)
	    (delete-count 0)
	    (vm-block-new-mail t)
	    (vm-pop-ok-to-ask t)
	    popdrop uidl-alist data mp match)
	(unwind-protect
	    (save-excursion
	      (setq vm-pop-retrieved-messages
		    (sort vm-pop-retrieved-messages
			  (function (lambda (a b)
				      (cond ((string-lessp (nth 1 a) (nth 1 b)) t)
					    ((string-lessp (nth 1 b)
							   (nth 1 a))
					     nil)
					    ((string-lessp (car a) (car b)) t)
					    (t nil))))))
	      (setq mp vm-pop-retrieved-messages)
	      (while mp
		(condition-case nil
		    (catch 'replay
		      (setq data (car mp))
		      (if (not (equal source (nth 1 data)))
			  (progn
			    (if process
				(progn
				  (vm-pop-end-session process)
				  (setq process nil)))
			    (setq source (nth 1 data))
			    (setq popdrop (vm-safe-popdrop-string source))
			    (condition-case nil
				(progn
				  (message "Opening POP session to %s..." popdrop)
				  (setq process (vm-pop-make-session source))
				  (if (null process)
				      (signal 'error nil))
				  (message "Expunging messages in %s..." popdrop))
			      (error
			       (message
				"Couldn't open POP session to %s, skipping..."
				popdrop)
			       (setq trouble (cons popdrop trouble))
			       (sleep-for 2)
			       (while (equal (nth 1 (car mp)) source)
				 (setq mp (cdr mp)))
			       (throw 'replay t)))
			    (set-buffer (process-buffer process))
			    (if vm-pop-use-uidl
				(progn
				  (vm-pop-send-command process "UIDL")
				  (setq uidl-alist (vm-pop-read-uidl-long-response process))))
			    ))
		      ;; If we have uidl alist, look up number in list. If not,
		      ;; check if message is still where we retrieved it from,
		      ;; and finally search for it if it isn't there.
		      (if (setq match (if uidl-alist
					  (rassoc (car data) uidl-alist)
					(catch 'found
					  (let* (count
						 (headers (vm-pop-get-message-headers process (nth 4 mp)))
						 (message-id (and headers
								  (vm-pop-extract-message-uidl headers))))
					    (if (equal message-id (car mp))
						(reverse mp)
					      ;; else search popdrop for UIDL
					      (vm-pop-send-command process "STAT")
					      (setq count (car (vm-pop-read-stat-response process)))
					      (while (> count 0)
						(setq headers (vm-pop-get-message-headers process count)
						      message-id (and headers
								      (vm-pop-extract-message-uidl headers)))
						(if (equal message-id (car mp))
						    (throw 'found
							   (cons count nil)))
						(setq count (- count 1)))
					      nil)))))
			  (progn
			    (vm-pop-send-command process
						 (format "DELE %d" (car match)))
			    (if (null (vm-pop-read-response process))
				(signal 'vm-dele-failed nil))
			    (vm-increment delete-count)))
		      (setq mp (cdr mp)))
		  (vm-dele-failed
		   (message "DELE %d failed on %s, skipping rest of mailbox..."
			    (car match) popdrop)
		   (setq trouble (cons popdrop trouble))
		   (sleep-for 2)
		   (while (equal (nth 1 (car mp)) source)
		     (setq mp (cdr mp)))
		   (throw 'replay t))
		  (vm-uidl-failed
		   (message "UIDL %d failed on %s, skipping this mailbox..."
			    (car match) popdrop)
		   (setq trouble (cons popdrop trouble))
		   (sleep-for 2)
		   (while (equal (nth 1 (car mp)) source)
		     (setq mp (cdr mp)))
		   (throw 'replay t))))
	      (if trouble
		  (progn
		    (set-buffer (get-buffer-create "*POP Expunge Trouble*"))
		    (setq buffer-read-only nil)
		    (erase-buffer)
		    (insert (format "%s POP message%s expunged.\n\n"
				    (if (zerop delete-count) "No" delete-count)
				    (if (= delete-count 1) "" "s")))
		    (insert "VM had problems expunging messages from:\n")
		    (nreverse trouble)
		    (setq mp trouble)
		    (while mp
		      (insert "   " (car mp) "\n")
		      (setq mp (cdr mp)))
		    (setq buffer-read-only t)
		    (display-buffer (current-buffer)))
		(message "%s POP message%s expunged."
			 (if (zerop delete-count) "No" delete-count)
			 (if (= delete-count 1) "" "s"))))
	  (and process (vm-pop-end-session process)))
	(or trouble (setq vm-pop-retrieved-messages nil)))))

(defun vm-pop-make-session (source)
  (let ((process-to-shutdown nil)
	process
	(popdrop (vm-safe-popdrop-string source))
	(coding-system-for-read (vm-binary-coding-system))
	(coding-system-for-write (vm-binary-coding-system))
	greeting timestamp
	host port auth user pass source-list process-buffer source-nopwd)
    (unwind-protect
	(catch 'done
	  ;; parse the maildrop
	  (setq source-list (vm-parse source "\\([^:]+\\):?")
		host (nth 0 source-list)
		port (nth 1 source-list)
		auth (nth 2 source-list)
		user (nth 3 source-list)
		pass (nth 4 source-list)
		source-nopwd (vm-popdrop-sans-password source))
	  ;; carp if parts are missing
	  (if (null host)
	      (error "No host in POP maildrop specification, \"%s\""
		     source))
	  (if (null port)
	      (error "No port in POP maildrop specification, \"%s\""
		     source))
	  (if (string-match "^[0-9]+$" port)
	      (setq port (string-to-int port)))
	  (if (null auth)
	      (error
	       "No authentication method in POP maildrop specification, \"%s\""
	       source))
	  (if (null user)
	      (error "No user in POP maildrop specification, \"%s\""
		     source))
	  (if (null pass)
	      (error "No password in POP maildrop specification, \"%s\""
		     source))
	  (if (equal pass "*")
	      (progn
		(setq pass (car (cdr (assoc source-nopwd vm-pop-passwords))))
		(if (null pass)
		    (if (null vm-pop-ok-to-ask)
			(progn (message "Need password for %s" popdrop)
			       (throw 'done nil))
		      (setq pass
			    (vm-read-password
			     (format "POP password for %s: "
				     popdrop)))))))
	  ;; save the password for the sake of
	  ;; vm-expunge-pop-messages, which passes password-less
	  ;; popdrop specifications to vm-make-pop-session.
	  (if (null (assoc source-nopwd vm-pop-passwords))
	      (setq vm-pop-passwords (cons (list source-nopwd pass)
					   vm-pop-passwords)))
	  ;; get the trace buffer
	  (setq process-buffer
		(generate-new-buffer (format "trace of POP session to %s"
					     host)))
	  (save-excursion
	    (set-buffer process-buffer)
	    (buffer-disable-undo process-buffer)
	    ;; clear the trace buffer of old output
	    (erase-buffer)
	    ;; Tell MULE not to mess with the text.
	    (if (or vm-xemacs-mule-p vm-fsfemacs-mule-p)
		(set-buffer-file-coding-system (vm-binary-coding-system) t))
	    (insert "starting POP session " (current-time-string) "\n")
	    (insert (format "connecting to %s:%s\n" host port))
	    ;; open the connection to the server
	    (setq process (open-network-stream "POP" process-buffer host port))
	    (and (null process) (throw 'done nil))
	    (insert "connected\n")
	    (process-kill-without-query process)
	    (make-local-variable 'vm-pop-read-point)
	    (setq vm-pop-read-point (point))
	    (if (null (setq greeting (vm-pop-read-response process t)))
		(progn (delete-process process)
		       (throw 'done nil)))
	    (setq process-to-shutdown process)
	    ;; authentication
	    (cond ((equal auth "pass")
		   (vm-pop-send-command process (format "USER %s" user))
		   (and (null (vm-pop-read-response process))
			(throw 'done nil))
		   (vm-pop-send-command process (format "PASS %s" pass))
		   (if (null (vm-pop-read-response process))
		       (progn
			 (setq vm-pop-passwords
			       (delete (list source-nopwd pass)
				       vm-pop-passwords))
			 (message "POP password for %s incorrect" popdrop)
			 (sleep-for 2)
			 (throw 'done nil))))
		  ((equal auth "rpop")
		   (vm-pop-send-command process (format "USER %s" user))
		   (and (null (vm-pop-read-response process))
			(throw 'done nil))
		   (vm-pop-send-command process (format "RPOP %s" pass))
		   (and (null (vm-pop-read-response process))
			(throw 'done nil)))
		  ((equal auth "apop")
		   (setq timestamp (vm-parse greeting "[^<]+\\(<[^>]+>\\)")
			 timestamp (car timestamp))
		   (if (null timestamp)
		       (progn
			 (goto-char (point-max))
   (insert-before-markers "<<< ooops, no timestamp found in greeting! >>>\n")
                         (message "Server of %s does not support APOP" popdrop)
			 (sleep-for 2)
			 (throw 'done nil)))
		   (vm-pop-send-command
		    process
		    (format "APOP %s %s"
			    user
			    (vm-pop-md5 (concat timestamp pass))))
		   (and (null (vm-pop-read-response process))
			(throw 'done nil)))
		  (t (error "Don't know how to authenticate using %s" auth)))
	    (setq process-to-shutdown nil)
	    process ))
      (if process-to-shutdown
	  (vm-pop-end-session process-to-shutdown t)))))

(defun vm-pop-end-session (process &optional keep-buffer)
  (save-excursion
    (set-buffer (process-buffer process))
    (vm-pop-send-command process "QUIT")
    ;; we don't care about the response
    ;;(vm-pop-read-response process)
    (if (and (not keep-buffer) (not vm-pop-debug))
	(kill-buffer (process-buffer process))
      (save-excursions
       (set-buffer (process-buffer process))
       (rename-buffer (concat "saved " (buffer-name)) t)
       (vm-keep-some-buffers (current-buffer) 'vm-kept-pop-buffers
			     vm-pop-keep-failed-trace-buffers)))
    (if (fboundp 'add-async-timeout)
	(add-async-timeout 2 'delete-process process)
      (run-at-time 2 nil 'delete-process process))))

(defun vm-pop-stat-timer (o) (aref o 0))
(defun vm-pop-stat-x-box (o) (aref o 1))
(defun vm-pop-stat-x-currmsg (o) (aref o 2))
(defun vm-pop-stat-x-maxmsg (o) (aref o 3))
(defun vm-pop-stat-x-got (o) (aref o 4))
(defun vm-pop-stat-x-need (o) (aref o 5))
(defun vm-pop-stat-y-box (o) (aref o 6))
(defun vm-pop-stat-y-currmsg (o) (aref o 7))
(defun vm-pop-stat-y-maxmsg (o) (aref o 8))
(defun vm-pop-stat-y-got (o) (aref o 9))
(defun vm-pop-stat-y-need (o) (aref o 10))

(defun vm-set-pop-stat-timer (o val) (aset o 0 val))
(defun vm-set-pop-stat-x-box (o val) (aset o 1 val))
(defun vm-set-pop-stat-x-currmsg (o val) (aset o 2 val))
(defun vm-set-pop-stat-x-maxmsg (o val) (aset o 3 val))
(defun vm-set-pop-stat-x-got (o val) (aset o 4 val))
(defun vm-set-pop-stat-x-need (o val) (aset o 5 val))
(defun vm-set-pop-stat-y-box (o val) (aset o 6 val))
(defun vm-set-pop-stat-y-currmsg (o val) (aset o 7 val))
(defun vm-set-pop-stat-y-maxmsg (o val) (aset o 8 val))
(defun vm-set-pop-stat-y-got (o val) (aset o 9 val))
(defun vm-set-pop-stat-y-need (o val) (aset o 10 val))

(defun vm-pop-start-status-timer ()
  (let ((blob (make-vector 11 nil))
	timer)
    (setq timer (add-timeout 5 'vm-pop-report-retrieval-status blob 5))
    (vm-set-pop-stat-timer blob timer)
    blob ))

(defun vm-pop-stop-status-timer (status-blob)
  (if (fboundp 'disable-timeout)
      (disable-timeout (vm-pop-stat-timer status-blob))
    (cancel-timer (vm-pop-stat-timer status-blob))))

(defun vm-pop-report-retrieval-status (o)
  (cond ((null (vm-pop-stat-x-got o)) t)
	;; should not be possible, but better safe...
	((not (eq (vm-pop-stat-x-box o) (vm-pop-stat-y-box o))) t)
	((not (eq (vm-pop-stat-x-currmsg o) (vm-pop-stat-y-currmsg o))) t)
	(t (message "Retrieving message %d (of %d) from %s, %s..."
		    (vm-pop-stat-x-currmsg o)
		    (vm-pop-stat-x-maxmsg o)
		    (vm-pop-stat-x-box o)
		    (format "%d%s of %d%s"
			    (vm-pop-stat-x-got o)
			    (if (> (vm-pop-stat-x-got o)
				   (vm-pop-stat-x-need o))
				"!"
			      "")
			    (vm-pop-stat-x-need o)
			    (if (eq (vm-pop-stat-x-got o)
				    (vm-pop-stat-y-got o))
				" (stalled)"
			      "")))))
  (vm-set-pop-stat-y-box o (vm-pop-stat-x-box o))
  (vm-set-pop-stat-y-currmsg o (vm-pop-stat-x-currmsg o))
  (vm-set-pop-stat-y-maxmsg o (vm-pop-stat-x-maxmsg o))
  (vm-set-pop-stat-y-got o (vm-pop-stat-x-got o))
  (vm-set-pop-stat-y-need o (vm-pop-stat-x-need o)))

(defun vm-pop-send-command (process command)
  (goto-char (point-max))
  (if (= (aref command 0) ?P)
      (insert-before-markers "PASS <omitted>\r\n")
    (insert-before-markers command "\r\n"))
  (setq vm-pop-read-point (point))
  (process-send-string process command)
  (process-send-string process "\r\n"))

(defun vm-pop-read-response (process &optional return-response-string)
  (let ((case-fold-search nil)
	 match-end)
    (goto-char vm-pop-read-point)
    (while (not (search-forward "\r\n" nil t))
      (accept-process-output process)
      (goto-char vm-pop-read-point))
    (setq match-end (point))
    (goto-char vm-pop-read-point)
    (if (not (looking-at "+OK"))
	(progn (setq vm-pop-read-point match-end) nil)
      (setq vm-pop-read-point match-end)
      (if return-response-string
	  (buffer-substring (point) match-end)
	t ))))

(defun vm-pop-read-past-dot-sentinel-line (process)
  (let ((case-fold-search nil))
    (goto-char vm-pop-read-point)
    (while (not (re-search-forward "^\\.\r\n" nil 0))
      (beginning-of-line)
      ;; save-excursion doesn't work right
      (let ((opoint (point)))
	(accept-process-output process)
	(goto-char opoint)))
    (setq vm-pop-read-point (point))))

(defun vm-pop-read-stat-response (process)
  (let ((response (vm-pop-read-response process t))
	list)
    (setq list (vm-parse response "\\([^ ]+\\) *"))
    (list (string-to-int (nth 1 list)) (string-to-int (nth 2 list)))))

(defun vm-pop-read-list-response (process)
  (let ((response (vm-pop-read-response process t)))
    (string-to-int (nth 2 (vm-parse response "\\([^ ]+\\) *")))))

(defun vm-pop-read-list-long-response (process)
  (let ((start vm-pop-read-point)
	(list nil)
	n size)
    (catch 'done
      (goto-char start)
      (while (not (re-search-forward "^\\.\r\n" nil 0))
	(beginning-of-line)
	;; save-excursion doesn't work right
	(let ((opoint (point)))
	  (accept-process-output process)
	  (goto-char opoint)))
      (setq vm-pop-read-point (point-marker))
      (goto-char start)
      ;; no size support, bail.
      (if (not (looking-at "\\+OK"))
	  (throw 'done nil))
      (forward-line 1)
      (while (not (eq (char-after (point)) ?.))
	;; not loking at a number, bail.
	(if (not (looking-at "[0-9]"))
	    (throw 'done nil))
	(setq n (read (current-buffer)))
	(skip-chars-forward " ")
	(if (not (looking-at "[0-9]"))
	    (throw 'done nil))
	(setq size (read (current-buffer)))
	(setq list (cons (cons n size) list))
	(forward-line 1))
      ;; returning nil means the size command failed so don't
      ;; return nil if there aren't any messages.
      (if (null list)
	  (cons nil nil)
	list ))))

(defun vm-pop-read-uidl-long-response (process)
  (let ((start vm-pop-read-point)
	(list nil)
	n uidl)
    (catch 'done
      (goto-char start)
      (while (not (re-search-forward "^\\.\r\n\\|^-ERR .*$" nil 0))
	(beginning-of-line)
	;; save-excursion doesn't work right
	(let ((opoint (point)))
	  (accept-process-output process)
	  (goto-char opoint)))
      (setq vm-pop-read-point (point-marker))
      (goto-char start)
      ;; no uidl support, bail.
      (if (not (looking-at "\\+OK"))
	  (throw 'done nil))
      (forward-line 1)
      (while (not (eq (char-after (point)) ?.))
	;; not loking at a number, bail.
	(if (not (looking-at "[0-9]"))
	    (throw 'done nil))
	(setq n (read (current-buffer)))
	(skip-chars-forward " ")
	(setq start (point))
	(skip-chars-forward "\041-\176")
	;; no tag after the message number, bail.
	(if (= start (point))
	    (throw 'done nil))
	(setq uidl (buffer-substring start (point)))
	(setq list (cons (cons n uidl) list))
	(forward-line 1))
      ;; returning nil means the uidl command failed so don't
      ;; return nil if there aren't any messages.
      (if (null list)
	  (cons nil nil)
	list ))))

(defun vm-pop-ask-about-large-message (headers size n)
  (let ((work-buffer nil)
	(pop-buffer (current-buffer)))
    (unwind-protect
	(save-excursion
	  (save-window-excursion
	    (if headers
		(progn
		  (setq work-buffer (generate-new-buffer
				     (format "*headers of message %d*" n)))
		  (set-buffer work-buffer)
		  (insert headers)
		  (vm-display-buffer work-buffer)
		  (setq minibuffer-scroll-window (selected-window))
		  (goto-char (point-min))
		  (if (re-search-forward "^Received:" nil t)
		      (progn
			(goto-char (match-beginning 0))
			(vm-reorder-message-headers
			 nil vm-visible-headers
			 vm-invisible-header-regexp)))
		  (set-window-point (selected-window) (point))))
	    (if (y-or-n-p (format "Message %d, size = %d, retrieve? " n size))
		'retrieve
	      (if (y-or-n-p (format "Delete message %d from popdrop? " n size))
		  'delete
		'skip))))
      (and work-buffer (kill-buffer work-buffer)))))

(defun vm-pop-retrieve-to-crashbox (process crash statblob size uidl popdrop)
  (let ((start vm-pop-read-point) end)
    (goto-char start)
    (vm-set-pop-stat-x-got statblob 0)
    (while (not (re-search-forward "^\\.\r\n" nil 0))
      (beginning-of-line)
      ;; save-excursion doesn't work right
      (let* ((opoint (point))
	     (func
	      (function
	       (lambda (beg end len)
		 (if vm-pop-read-point
		     (progn
		       (vm-set-pop-stat-x-got statblob (- end start))
		       (if (zerop (% (random) 10))
			   (vm-pop-report-retrieval-status statblob)))))))
	     (after-change-functions (cons func after-change-functions)))
	(accept-process-output process)
	(goto-char opoint)))
    (vm-set-pop-stat-x-got statblob nil)
    (setq vm-pop-read-point (point-marker))
    (goto-char (match-beginning 0))
    (setq end (point-marker))
    (vm-pop-cleanup-region start end)
    ;; Some POP servers strip leading and trailing message
    ;; separators, some don't.  Figure out what kind we're
    ;; talking to and do the right thing.
    (if (eq (vm-get-folder-type nil start end) 'unknown)
	(progn
	  (vm-munge-message-separators vm-folder-type start end)
	  (goto-char start)
	  ;; avoid the consing and stat() call for all but babyl
	  ;; files, since this will probably slow things down.
	  ;; only babyl files have the folder header, and we
	  ;; should only insert it if the crash box is empty.
	  (if (and (eq vm-folder-type 'babyl)
		   (let ((attrs (file-attributes crash)))
		     (or (null attrs) (equal 0 (nth 7 attrs)))))
	      (let ((opoint (point)))
		(vm-convert-folder-header nil vm-folder-type)
		;; if start is a marker, then it was moved
		;; forward by the insertion.  restore it.
		(setq start opoint)
		(goto-char start)
		(vm-skip-past-folder-header)))
	  (insert (vm-leading-message-separator))
	  ;; this will not find the trailing message separator but
	  ;; for the Content-Length stuff counting from eob is
	  ;; the same thing in this case.
	  (vm-convert-folder-type-headers nil vm-folder-type)
	  (goto-char end)
	  (insert-before-markers (vm-trailing-message-separator))))
    ;; Set file type to binary for DOS/Windows.  I don't know if
    ;; this is correct to do or not; it depends on whether the
    ;; the CRLF or the LF newline convention is used on the inbox
    ;; associated with this crashbox.  This setting assumes the LF
    ;; newline convention is used.
    (let ((buffer-file-type t)
	  (selective-display nil))
      (write-region start end crash t 0))
    (if (not vm-pop-debug)
	(delete-region start end))
    t ))

(defun vm-pop-cleanup-region (start end)
  (if (> (- end start) 30000)
      (message "CRLF conversion and char unstuffing..."))
  (setq end (vm-marker end))
  (save-excursion
    (goto-char start)
    ;; CRLF -> LF
    (while (and (< (point) end) (search-forward "\r\n"  end t))
      (replace-match "\n" t t))
    (goto-char start)
    ;; chop leading dots
    (while (and (< (point) end) (re-search-forward "^\\."  end t))
      (replace-match "" t t)
      (forward-char)))
  (if (> (- end start) 30000)
      (message "CRLF conversion and char unstuffing... done"))
  (set-marker end nil))

(defun vm-pop-md5 (string)
  (let ((buffer nil))
    (unwind-protect
	(save-excursion
	  (setq buffer (generate-new-buffer "*vm-work*"))
	  (set-buffer buffer)
	  ;; call-process-region calls write-region.
	  ;; don't let it do CR -> LF translation.
	  (setq selective-display nil)
	  (insert string)
	  (if (fboundp 'md5)
	      (progn
		(goto-char (point-min))
		(insert (md5 buffer (point-min) (point-max)))
		(delete-region (point) (point-max)))
	    (call-process-region (point-min) (point-max)
				 (or shell-file-name "/bin/sh") t buffer nil
				 shell-command-switch vm-pop-md5-program))
	  ;; MD5 digest is 32 chars long
	  ;; mddriver adds a newline to make neaten output for tty
	  ;; viewing, make sure we leave it behind.
	  (vm-buffer-substring-no-properties (point-min) (+ (point-min) 32)))
      (and buffer (kill-buffer buffer)))))

(defun vm-popdrop-sans-password (source)
  (let (source-list)
    (setq source-list (vm-parse source "\\([^:]+\\):?"))
    (concat (nth 0 source-list) ":"
	    (nth 1 source-list) ":"
	    (nth 2 source-list) ":"
	    (nth 3 source-list) ":*")))

;; Return default message list
(defun vm-pop-message-list (message-list)
  (save-excursion
    (vm-select-folder-buffer)
    (or message-list vm-message-list)))

;;; Return an associative list of POP message ids to VM message vectors.
(defun vm-pop-cached-messages (message-list)
  (save-excursion
    (let ((mp (or message-list vm-message-list))
	  idlist uidl headers)
      (while mp
	(save-restriction
	  (set-buffer (vm-buffer-of (car mp)))
	  (widen)
	  (setq headers (buffer-substring
			 (vm-headers-of (car mp))
			 (vm-text-of (car mp))))
	  (while (string-match "\n[ \t]+" headers)
	    (setq headers (replace-match " " t t headers)))
	  (setq uidl (vm-pop-extract-message-uidl headers))
	  (if uidl
	      (setq idlist
		    (cons (cons uidl (car mp)) idlist)))
	  (setq mp (cdr mp))))
      idlist)))

;;; extract UIDL from message headers, if present
(defun vm-pop-extract-message-uidl (headers)
  (vm-pop-zap-space
   (if (or (string-match vm-pop-x-uidl-regexp headers)
	   (string-match vm-pop-message-id-regexp headers))
       (let ((msgid (substring headers (match-beginning 1) (match-end 1))))
	 (if (equal msgid "")
	     "<no-message-id>" ; <> are out of band for good msgid
	   msgid))
     ;; else fake it from Date and From
     (let (date from)
       (if (string-match "From: *\\([^\r\n]+\\)" headers)
	   (setq from (substring headers (match-beginning 1) (match-end 1))))
       (if (string-match "Date: *\\([^\r\n]+\\)" headers)
	   (setq date (substring headers (match-beginning 1) (match-end 1))))
       (concat "<" date "," from ">")) ; <> are out-of-band for good msgid
     )))

;;; evaluate action match, including and/or/not clauses
(defun vm-pop-match-action (headers action)
  (cond
   ((stringp action)
    (string-match action headers))
   ((and action (listp action))
    (let ((head (car action)) (tail (cdr action)) (match t))
      (cond
       ((eq head 'and) ; match clauses until one fails
	(while (and match tail)
	  (setq match (vm-pop-match-action headers (car tail))
		tail (cdr tail)))
	match)
       ((eq head 'or) ; match clauses until one succeeds
	(while (and match tail)
	  (setq match (not (vm-pop-match-action headers (car tail)))
		tail (cdr tail)))
	(not match))
       ((eq head 'not)
	(not (vm-pop-match-action headers (car tail))))
       )))
   (action) ;; else return action
   )
  )

;;; get message action; action symbol will be 'delete, 'skip, or nil
;;; (indicating that the message should be downloaded).
(defun vm-pop-get-message-action (headers action-list)
  (catch 'act
    (while action-list
      (let ((match (vm-pop-match-action headers (car (car action-list))))
	    (tail (cdr (car action-list)))
	    (next (cdr action-list))) ; next action
	(while (and match tail (cdr tail)) ; more matches?
	  (setq match (vm-pop-match-action headers (car tail))
		tail (cdr tail)))
	(if match ; if we matched an action, do the corresponding
	    (throw 'act
		   (let ((value (car tail)))
		     (if (listp value)
		       (eval value)
		       value))))
	(setq action-list next)))
    nil)
  )

(defun vm-pop-load-action-file (filename)
  (save-excursion
    (let ((standard-input (set-buffer (find-file-noselect filename t)))
	  actions)
      (goto-char (point-min))
      (setq actions (if (looking-at "[ \t]*(") (read) nil))
      (kill-buffer standard-input)
      actions))
  )

(defun vm-pop-save-action-file (filename action-list)
  (save-excursion
    (let ((standard-output (set-buffer (find-file-noselect filename t))))
      (delete-region (point-min) (point-max))
      (prin1 action-list)
      (basic-save-buffer)
      (kill-buffer standard-output)))
  )

;;; log killed messages
(defun vm-pop-kill-log (filename host headers)
  (if filename
      (save-excursion
	(let ((log (set-buffer (find-file-noselect filename t))))
	  (goto-char (point-max))
	  (insert "KILL " host " at " (current-time-string) "\n\n"
		  headers "\n")
	  (basic-save-buffer)
	  (kill-buffer log))))
  )

;;; get headers for a message as a string
(defun vm-pop-get-message-headers (process n)
  (vm-pop-send-command process (format "TOP %d 0" n))
  (and
   (vm-pop-read-response process)
   (vm-pop-snarf-message-headers process))
  )

(defun vm-pop-snarf-message-headers (process)
  (let ((case-fold-search nil)
	(start vm-pop-read-point)
	end headers)
    (goto-char start)
    (while (not (re-search-forward "^\\.\r\n" nil 0))
      (beginning-of-line)
      ;; save-excursion doesn't work right
      (let ((opoint (point)))
	(accept-process-output process)
	(goto-char opoint)))
    (setq vm-pop-read-point (point-marker))
    (goto-char (match-beginning 0))
    (setq end (point-marker))
    (vm-pop-cleanup-region start end)
    (setq headers (buffer-substring start end))
    (if (not vm-pop-debug)
	(delete-region start end))
    (while (string-match "\n[ \t]+" headers)
      (setq headers (replace-match " " t t headers)))
    headers))

;;; change spaces to something else
(defun vm-pop-zap-space (str)
  (concat (mapcar '(lambda (arg) (if (= arg 32) ?_ arg)) str)))

;;; Extract messages for this popdrop from retrieved list
(defun vm-pop-select-retrieved (source same retrieved)
  (let (result info)
    (while retrieved
      (setq info (car retrieved)
	    retrieved (cdr retrieved))
      (if (eq (null same) (null (equal source (nth 1 info))))
	  (setq result (cons info result))))
    result))
