/* $Id: block.c,v 1.1 1999/11/28 18:03:05 angus Exp $
 *
 * Address blocking management
 *
 * Copyright (C) Angus J. C. Duggan 1999
 *
 */

#include "mailcheck.h"
#include <string.h>
#include <ctype.h>

/* Hash function for strings */
static unsigned long stringhash(char *str)
{
  unsigned long hash = 0 ;

  while ( *str ) {
    hash *= hash * 2 + 1 ; /* Make high bits depend on all bits */
    hash = (hash >> 31) ^ (hash >> 1) ^ *str++ ;
  }

  return hash ;
}

/*****************************************************************************/

#define DOMAIN_TABLE_SIZE 16

typedef struct domain_entry {
  struct domain_entry *next ;
  char *domain ;
  bool block ;
} domain_entry ;

static domain_entry *domain_table[DOMAIN_TABLE_SIZE] = { NULL } ;

void add_domain(char *domain, bool block)
{
  unsigned long hash = stringhash(domain) % DOMAIN_TABLE_SIZE ;
  domain_entry **bucket = &domain_table[hash], *here ;

  while ( (here = *bucket) != NULL ) {
    if ( strcmp(here->domain, domain) == 0 ) {
      if ( debug && here->block != block )
        fprintf(stderr, "*** %s domain %s\n",
                block ? "Testing" : "Protecting", domain) ;
      here->block = block ;
      return ;
    }
    bucket = &here->next ;
  }

  if ( debug )
    fprintf(stderr, "*** %s domain %s\n",
            block ? "Testing" : "Protecting", domain) ;

  if ( (here = *bucket = malloc(sizeof(domain_entry))) == NULL ||
       (here->domain = strdup(domain)) == NULL )
    error("out of memory in domain table") ;

  here->block = block ;
  here->next = NULL ;
}

bool block_domain(char *domain)
{
  unsigned long hash = stringhash(domain) % DOMAIN_TABLE_SIZE ;
  domain_entry *bucket = domain_table[hash] ;

  while ( bucket ) {
    if ( strcmp(bucket->domain, domain) == 0 )
      return bucket->block ;
    bucket = bucket->next ;
  }

  return false ;
}

/*****************************************************************************/

#define USER_TABLE_SIZE 16

typedef struct user_entry {
  struct user_entry *next ;
  char *user ;
  bool block ;
} user_entry ;

static user_entry *user_table[USER_TABLE_SIZE] = { NULL } ;

void add_user(char *user, bool block)
{
  unsigned long hash = stringhash(user) % USER_TABLE_SIZE ;
  user_entry **bucket = &user_table[hash], *here ;

  while ( (here = *bucket) != NULL ) {
    if ( strcmp(here->user, user) == 0 ) {
      if ( debug && here->block != block )
        fprintf(stderr, "*** %s user %s\n",
                block ? "Testing" : "Protecting", user) ;
      here->block = block ;
      return ;
    }
    bucket = &here->next ;
  }

  if ( debug )
    fprintf(stderr, "*** %s user %s\n",
            block ? "Testing" : "Protecting", user) ;

  if ( (here = *bucket = malloc(sizeof(user_entry))) == NULL ||
       (here->user = strdup(user)) == NULL )
    error("out of memory in user table") ;

  here->block = block ;
  here->next = NULL ;
}

bool block_user(char *user)
{
  unsigned long hash = stringhash(user) % USER_TABLE_SIZE ;
  user_entry *bucket = user_table[hash] ;

  while ( bucket ) {
    if ( strcmp(bucket->user, user) == 0 )
      return bucket->block ;
    bucket = bucket->next ;
  }

  return false ;
}

/*****************************************************************************/

#define ADDRESS_TABLE_SIZE 16

typedef struct address_entry {
  struct address_entry *next ;
  unsigned long address ;
  bool block ;
} address_entry ;

static address_entry *address_table[ADDRESS_TABLE_SIZE] = { NULL } ;

static unsigned long pack_address(uint a, uint b, uint c, uint d)
{
  return ((unsigned char)a << 24) | ((unsigned char)b << 16) |
    ((unsigned char)c << 8) | (unsigned char)d ;
}

static void unpack_address(unsigned long address,
                           uint *aa, uint *bb, uint *cc, uint *dd)
{
  *aa = address >> 24 ;
  *bb = (unsigned char)(address >> 16) ;
  *cc = (unsigned char)(address >> 8) ;
  *dd = (unsigned char)address ;
}

void add_address(uint a, uint b, uint c, uint d, bool block)
{
  unsigned long address = pack_address(a, b, c, d) ;
  unsigned long hash = address % ADDRESS_TABLE_SIZE ;
  address_entry **bucket = &address_table[hash], *here ;

  while ( (here = *bucket) != NULL ) {
    if ( here->address == address ) {
      if ( debug && here->block == block )
        fprintf(stderr, "*** %s address %u.%u.%u.%u\n",
                block ? "Testing" : "Protecting", a, b, c, d) ;

      here->block = block ;
      return ;
    }
    bucket = &here->next ;
  }

  if ( debug )
    fprintf(stderr, "*** %s address %u.%u.%u.%u\n",
            block ? "Testing" : "Protecting", a, b, c, d) ;

  if ( (here = *bucket = malloc(sizeof(address_entry))) == NULL )
    error("out of memory in address table") ;

  here->address = address ;
  here->block = block ;
  here->next = NULL ;
}

bool block_address(uint a, uint b, uint c, uint d)
{
  unsigned long address = pack_address(a, b, c, d) ;
  unsigned long hash = address % ADDRESS_TABLE_SIZE ;
  address_entry *bucket = address_table[hash] ;

  while ( bucket ) {
    if ( bucket->address == address )
      return bucket->block ;
    bucket = bucket->next ;
  }

  return false ;
}

typedef struct {
  address_entry *addr ;
  int nbucket ;
} address_iterator ;

bool iterate_address(void **iterator, uint *aa, uint *bb, uint *cc, uint *dd)
{
  address_iterator *here = *(address_iterator **)iterator ;

  if ( here == NULL ) { /* Initialise iterator */
    if ( (here = malloc(sizeof(address_iterator))) == NULL )
      error("out of memory iterating addresses") ;
    here->nbucket = ADDRESS_TABLE_SIZE ;
    here->addr = NULL ;
  }

  do {
    if ( here->addr ) /* Still in old bucket */
      here->addr = here->addr->next ;
    else if ( !here->nbucket-- ) { /* Done all buckets */
      free(here) ;
      *iterator = NULL ;
      return false ;
    } else /* Start new bucket */
      here->addr = address_table[here->nbucket] ;
  } while ( here->addr == NULL || !here->addr->block ) ;

  unpack_address(here->addr->address, aa, bb, cc, dd) ;
  *iterator = (void *)here ;

  return true ;
}

/*
$Log: block.c,v $
Revision 1.1  1999/11/28 18:03:05  angus
Add Unix name resolution, blocked address update, etc.

*/
