/* $Id: mailcheck.c,v 1.5 1999/11/28 19:53:17 angus Exp $
 *
 * mailcheck
 *
 * Copyright (C) Angus J. C. Duggan 1999
 *
 * Program to check for spam email. Parses headers, extracting domains, dotted
 * quads, usernames, etc., checks these against blocked list. Checks RFC 822
 * compliance and reports discrepancies, looks up domains and dotted quads and
 * checks with MAPS real time blackhole.
 */

#include "mailcheck.h"
#include <string.h>
#include <ctype.h>

int debug = 0 ; /* no debugging info */

/* Check internet real-time blackhole. Here's how to use the RBL, from the
   FAQ:

   Direct usage via DNS (no subscription needed)

   The simplest way to get started using the MAPS RBL to protect your mail
   relay against theft of service by spammers is to arrange for it to make
   a DNS query (of a stylized name) whenever you receive an incoming mail
   message from a host whose spam status you do not know.

   The theory of operation is simple. Given a host address in its
   dotted-quad form, reverse the octets and check for the existence of an
   ``A RR'' at that node under the rbl.maps.vix.com node. So if you get an
   SMTP session from [192.5.5.1] you would check for the existence of:

     1.5.5.192.rbl.maps.vix.com. IN A 127.0.0.2

   We chose to use an ``A RR'' because that's what Sendmail makes easy to
   do. The choice of [127.0.0.2] as the target address was arbitary but
   will not change. As it happens, we supply a bogus MAPS RBL entry for
   [127.0.0.2] so that mail transport developers have something to test
   against.

   If an ``A RR'' is found by this mechanism, then there will also be a
   ``TXT RR'' at the same DNS node. The text of this record will be
   suitable for use as a reason text for a bounced mail notification.
   Currently the text is constant and currently there is no way to use it
   from Sendmail, but there it is anyway. */

#if defined(WIN32)
#include <fcntl.h>
#include <io.h>
#include <winsock.h>

#define MAXDNAME 1025 /* Maximum domain name size */

/* Version for testing on Win32 uses gethostbyname rather than resolver
   query, because MicroSoft's WSALookupService* interface is so painful to
   use. */
static bool winsock_ready = false ;

static bool blackhole_lookup(char *name)
{
  struct hostent *host ;

  if ( !winsock_ready ) {
    WSADATA impl;

    if ( WSAStartup(MAKEWORD(1,1), &impl) != 0 )
      return false ;

    winsock_ready = true ;
  }

  if ( debug >= 2 ) {
    name = "localhost" ;
    fprintf(stderr, "*** lookup changed to localhost to avoid timeout\n") ;
  }

  if ( (host = gethostbyname(name)) != NULL ) {
    struct in_addr **addrs = (struct in_addr **)host->h_addr_list ;
    unsigned long loopback2 = htonl(0x7f000002) ; /* 127.0.0.2 */

    if ( host->h_addrtype == AF_INET ) {
      while ( *addrs ) {
        if ( (**addrs).s_addr == loopback2 )
          return true ;
        ++addrs ;
      }
    }
  }

  return false ;
}

static void cleanup(void)
{
  (void)WSACleanup() ;
}
#else
/* Include resolver header files */
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

static void skip_name(unsigned char **where)
{
  unsigned char *here = *where ;
  unsigned int n ;

  while ( (n = *here++) != 0 ) {
    if ( (n & 0xc0) == 0xc0 ) { /* Compressed name offset */
      here += 1 ;
      break ;
    } else
      here += n ;
  }

  *where = here ;
}

static bool blackhole_lookup(char *name)
{
  int len ;
  union {
    unsigned char bytes[PACKETSZ] ; /* Maximum packet */
    HEADER header ;
  } answer ;

  if ( (len = res_query(name, C_IN, T_A, answer.bytes, PACKETSZ)) >= 0 ) {
    unsigned char *here = answer.bytes + 12 ; /* HEADER size */
    int i ;

    for ( i = 0 ; i < answer.header.qdcount ; ++i ) { /* Skip question */
      skip_name(&here) ; /* Ignore name */
      here += 4 ; /* Ignore type and class */
    }

    for ( i = 0 ; i < answer.header.ancount ; ++i ) { /* Test answers */
      skip_name(&here) ; /* Ignore name */
      here += 8 ; /* Ignore type, class, and time to live */
      len = here[0] << 8 | here[1] ;
      if ( len == 4 &&
           here[2] == 127 && here[3] == 0 && here[4] == 0 && here[5] == 2 )
        return true ;
    }
  }

  return false ;
}

static void cleanup(void)
{
  /* Nothing needed to cleanup */
}
#endif

static char *program = NULL ;

static void usage(void)
{
  fprintf(stderr, "Copyright (C) Angus J. C. Duggan, 1997-1999\n"
          "Usage: %s [-u] [-b file] [-r keyword,...] [-l file] [-h server] [-w] [-a] [-m] [infile]\n"
          "\tinfile\tfile to read message from (default stdin)\n"
	  "\t-l file\tsends error and information output to file (default stderr)\n"
	  "\t-b file\tfile containing blocked users/addresses/dotted quads\n"
	  "\t-h server\tuses Realtime Blackhole Server\n"
	  "\t-u\tupdates blocked file with blackholed addresses\n"
	  "\t-m\tsend failed messages to stdout\n"
	  "\t-w\twarn about RFC822 compliance rather than fail\n"
	  "\t-a\tfind all failures rather than stop after first failure\n"
	  "\t-821\tdisallows RFC821 Received header lexemes\n"
	  "\t-r keyword,...\tRelaxes RFC822 checking. Keywords are:\n"
	  "\t   for\tallows mailbox addresses without '@' in Received lines\n"
	  "\t   path\tallows addresses without '@' in Return-Path headers\n"
	  "\t   to\tallows addresses without '@' in To headers\n"
	  "\t   at\tallows all mailboxes to omit '@'\n"
	  "\t   mbox\tallows mailboxes to omit angle brackets\n"
	  "\t   semi\tallows optional semicolons in Received headers\n"
	  "\t   ms\trelaxes Received 'with' syntax for Microsoft MAPI\n"
	  "\t   id\trelaxes Received 'id' rules for Lotus Notes\n"
	  "\t   ref\trelaxes References and In-Reply-To checking\n"
	  "\t   from\tignores From without colon, as passed by procmail\n",
          program) ;
  exit(0) ; /* Exit with zero to stop error from dropping messages */
}

void error(char *format, ...)
{
  va_list args ;
  char *scan = format, ch ;

  fprintf(stderr, "%s: ", program) ;
  va_start(args, format) ;
  while ( (ch = *scan++) != '\0' ) {
    if ( ch == '%' ) {
      switch ( *scan++ ) {
      case 'd':
	fprintf(stderr, "%d", va_arg(args, long)) ;
	break ;
      case 's':
	fprintf(stderr, "%s", va_arg(args, char *)) ;
	break ;
      case '%':
	fputc(ch, stderr) ;
	break ;
      default:
	fprintf(stderr, "Bad format string\n") ;
	exit(0) ; /* Exit with zero to stop error from dropping messages */
      }
    } else
      fputc(ch, stderr) ;
  }
  fputc('\n', stderr) ;
  fflush(stderr) ;
  va_end(args) ;

  exit(0) ; /* Exit with zero to stop error from dropping messages */
}

/* The top-level of parsing calls back to a parse function for each matched
   header. Store the header-callback correspondence */
static struct headerinfo {
  char *name ;
  bool (*fn)(char *) ;
} headers[] = {
  {"received", &received},
  {"old-received", &received},
  {"message-id", &messageid},
  {"resent-message-id", &messageid},
  {"return-path", &returnpath},
  {"date", &dateheader},
  {"resent-date", &dateheader},
  {"from", &mailboxes},
  {"resent-from", &mailboxes},
  {"sender", &mailbox},
  {"resent-sender", &mailbox},
  {"reply-to", &addresses},
  {"resent-reply-to", &addresses},
  {"to", &addresses},
  {"resent-to", &addresses},
  {"cc", &addresses},
  {"resent-cc", &addresses},
  {"references", &references},
  {"in-reply-to", &references},
} ;

static struct relax_entry {
  char *name ;
  int flags ;
} relax_table[] = {
  {"for", RelaxFor},
  {"path", RelaxReturnPath},
  {"to", RelaxTo},
  {"at", RelaxTo|RelaxFor|RelaxReturnPath},
  {"mbox", RelaxMailbox},
  {"semi", RelaxSemi},
  {"ms", RelaxMicrosoft},
  {"id", RelaxId},
  {"ref", RelaxReferences},
  {"from", RelaxFrom},
  {"all", -1},
} ;

/* Prevent own domain and machine from being blocked, by adding with false
   flag before any other domains are added */
void home_domain(void)
{
  char hostname[MAXDNAME], *dmn, *dot ;

  if ( gethostname(hostname, MAXDNAME) != 0 )
    return ;

  for ( dmn = hostname ; (dot = strchr(dmn, '.')) != NULL ; dmn = dot + 1 )
    add_domain(dmn, false) ;
}

int main(int argc, char *argv[])
{
  FILE *infile = stdin, *blocked = NULL ;
  char *msgh = NULL ;
  long linelen = 0 ;
  int i, result = 0, rfc822 = 0, *warn = &result, errors = 0 ;
  bool all = false, message = false, update = false ;
  char *blackhole = NULL ;

  if ( (program = strrchr(*argv, '/')) == NULL &&
       (program = strrchr(*argv, '\\')) == NULL )
    program = *argv ;
  else
    ++program ;

  for (++argv ; --argc ; argv++ ) {
    if ( argv[0][0] == '-' ) {
      char *optarg = NULL, ch = argv[0][1] ;

      if ( strchr("hlbrd", ch) ) {
        if ( argv[0][2] )
          optarg = &argv[0][2] ;
        else if ( --argc )
          optarg = *++argv ;
        else
          usage() ;
      }

      switch ( ch ) {
      case '8': /* disable RFC821 lexemes */
	rfc821 = false ;
	break ;
      case 'r': /* Relax keyword */
        while ( optarg ) {
          char *next = strchr(optarg, ',') ;

          if ( next )
            *next++ = '\0' ;

          for ( i = 0 ; i < sizeof(relax_table)/sizeof(struct relax_entry) ; ++i ) {
            if ( strcmp(relax_table[i].name, optarg) == 0 ) {
              relax |= relax_table[i].flags ;
              break ;
            }
          }

          optarg = next ;
        }
        if ( i >= sizeof(relax_table)/sizeof(struct relax_entry) )
          usage() ;
	break ;
      case 'l': /* Log */
	if ( strcmp(optarg, "-") != 0 ) {
	  if ( !freopen(optarg, "ab", stdout) )
	    error("Can't reopen log file %s for stdout", optarg) ;
	}
	break ;
      case 'b': /* Read blocked file of addresses */
        if ( (blocked = fopen(optarg, update ? "rw" : "r")) == NULL ) {
          if ( !update || /* Try again, read only */
               (blocked = fopen(optarg, "r")) == NULL )
            error("can't open block file %s", optarg) ;
          update = false ; /* Can't update file, it's read-only */
        }
	break ;
      case 'h': /* Set Realtime Blackhole server */
        blackhole = optarg ;
        break ;
      case 'w': /* Warn about compliance */
        warn = &rfc822 ;
        break ;
      case 'a': /* All failures */
        all = true ;
        break ;
      case 'u': /* Update blocked file from blackhole */
        if ( blocked )
          error("update flag must be before blocked file") ;
        update = true ;
        break ;
      case 'm': /* Message to stdout */
        message = true ;
        break ;
      case 'd': /* Debug info */
        debug = strtol(optarg, &optarg, 0) ;
        break ;
      default:
	usage();
      }
    } else if (infile == stdin) {
      if ((infile = fopen(*argv, "rb")) == NULL)
	error("can't open input file %s", *argv);
    } else usage();
  }

#if defined(MSDOS) || defined(WINNT)
  if ( infile == stdin ) {
    int fd = fileno(stdin) ;
    if ( setmode(fd, O_BINARY) < 0 )
      error("can't reset stdin to binary mode\n");
  }
#endif

  home_domain() ; /* Own domain cannot be blocked */

  /* Now; get headers, storing original headers in buffer and creating folded
     canonical version for parser to read */
  while ( (msgh = readline(infile, &linelen, true)) != NULL ) {
    for ( i = 0 ; i < sizeof(headers)/sizeof(struct headerinfo) ; ++i ) {
      char *start = msgh ;
      if ( keyword(&msgh, headers[i].name) ) {
	if ( !(*headers[i].fn)(msgh) ) {
	  long len = errorpos - start ;
          if ( !errors++ )
            puts("ERRORS:\n") ;
	  puts(start) ;
	  while ( len-- > 0 )
	    putchar(' ') ;
	  printf("^ %s\n", errorstr) ;
	  ++*warn ;
	}
	break ;
      }
    }
    if ( result && !all )
      break ;
  }

  if ( blocked ) {
    if ( !result || all ) {
      char buffer[BUFSIZ] ;

      while ( fgets(buffer, BUFSIZ, blocked) ) {
        char *scan = buffer, *end ;
        bool at = false, block = true ;
        uint a, b, c, d, ha, hb, hc, hd ;
        int count ;

        while ( isspace(*scan) ) ++scan ;
        if ( *scan == '#' || *scan == '\0' ) /* Skip blanks and comments */
          continue ;

        if ( *scan == '!' ) { /* This shouldn't be blocked */
          ++scan ;
          block = false ;
        }

        for ( end = scan ; !isspace(*end) ; ++end ) { /* Find end */
          if ( *end == '@' )
            at = true ;
        }
        *end = '\0' ;

        if ( sscanf(scan, "%u.%u.%u.%u%n", &a, &b, &c, &d, &count) == 4 &&
             scan[count] == '\0' ) {
          if ( !block )
            add_address(a, b, c, d, false) ;
          else if ( block_address(a, b, c, d) ) {
            if ( !errors++ )
              puts("ERRORS:\n") ;
            printf("Blocked file address %s matched\n", scan) ;
            ++result ;
          }
        } else if ( sscanf(scan, "%u.%u.%u.%u-%u.%u.%u.%u%n", &a, &b, &c, &d,
                           &ha, &hb, &hc, &hd, &count) == 8 &&
                    scan[count] == '\0' ) { /* Class C */
          fprintf(stderr, "Blocked ranges NYI\n") ;
        } else if ( at ) {
          if ( !block )
            add_user(scan, false) ;
          else if ( block_user(scan) ) {
            if ( !errors++ )
              puts("ERRORS:\n") ;
            printf("Blocked file user %s matched\n", scan) ;
            ++result ;
          }
        } else {
          if ( !block )
            add_domain(scan, false) ;
          else if ( block_domain(scan) ) {
            if ( !errors++ )
              puts("ERRORS:\n") ;
            printf("Blocked file domain %s matched\n", scan) ;
            ++result ;
          }
        }

        if ( result && !all )
          break ;
      }
    } else {
      puts("Skipping block file, message rejected already") ;
    }
  }

  /* We only test explicit dotted quad addresses, rather than looking up all
     domains in the message and using them. This is probably the right thing
     to do, since various sendmails along the route will have noted the
     incoming IP addresses in the Received lines. If we're updating the
     blocked file, it will now be positioned at EOF, or we will be omitting
     the following code altogether. */
  if ( blackhole ) {
    if ( !result || all ) {
      void *iterator = NULL ;
      uint a, b, c, d ;

      while ( iterate_address(&iterator, &a, &b, &c, &d) ) {
        char name[MAXDNAME] ; /* Maximum domain name */

        sprintf(name, "%u.%u.%u.%u.%s", d, c, b, a, blackhole) ;

        if ( debug )
          fprintf(stderr, "*** Testing blackhole name %s\n", name) ;

        if ( blackhole_lookup(name) ) {
          if ( update ) /* blocked should be at EOF */
            fprintf(blocked, "%u.%u.%u.%u\n", a, b, c, d) ;
          if ( !errors++ )
            puts("ERRORS:\n") ;
          printf("Address %u.%u.%u.%u found on blackhole server %s\n",
                 a, b, c, d, blackhole);
          ++result ;
          if ( !all ) /* Terminate if not showing all failures */
            break ;
        }
      }
    } else {
      puts("Skipping blackhole check, message rejected already");
    }
  }


  if ( result || rfc822 ) {
    puts("\nHEADERS:\n") ;
    flushlines(stdout) ;
    if ( message ) {
      puts("\nMESSAGE:\n") ;
      while ( (msgh = readline(infile, &linelen, false)) != NULL )
        flushlines(stdout) ;
    }
    printf("%d failure%s\n", result, result == 1 ? "" : "s") ;
  }

  fclose(blocked) ;
  fclose(infile) ;

  cleanup() ;

  return result ;
}

/*
$Log: mailcheck.c,v $
Revision 1.5  1999/11/28 19:53:17  angus
Test RRs properly

Revision 1.4  1999/11/28 18:03:05  angus
Add Unix name resolution, blocked address update, etc.

Revision 1.3  1999/11/25 17:18:48  angus
Makefile for Linux now OK

Revision 1.2  1999/11/24 05:17:10  angus
Avoid blocking own domain

Revision 1.1.1.1  1999/11/24 04:18:51  angus
Initial import

*/
