/* $Id: rfc822.c,v 1.3 1999/11/28 18:03:05 angus Exp $
 *
 * RFC822 parser.
 *
 * Copyright (C) Angus J. C. Duggan 1997-1999
 *
 */

#include "mailcheck.h"

#include <stdlib.h>
#include <string.h>

#if defined(MSDOS) || defined(WIN32)
#define strnicmp _strnicmp
#endif

#ifdef NO_STRNICMP
/* For those without, define a case-insensitive compare */
int strnicmp(char *left, char *right, int n)
{
  while ( --n >= 0 ) {
    int l = tolower(*left++), r = tolower(*right++) ;
    if ( l < r )
      return -1 ;
    else if ( l > r )
      return 1 ;
    else if ( l == 0 )
      return 0 ;
  }

  return 0 ;
}
#endif

char *errorpos ; /* Pointer to where error occurred */
char *errorstr ; /* Reason for error */
int relax = 0 ;  /* How lax are we? */
bool rfc821 = true ;    /* Allow RFC821 lexemes */

/* Null-terminated buffer structure for accumulating user and domain names. */
typedef struct strbuf {
  char *buffer ;
  long len, current ;
} strbuf ;

void addbuffer(char *data, long len, strbuf *buffer)
{
  int current = buffer->current ;

  if ( buffer->len <= len + current ) {
    buffer->len += len * 2 + 1 ;
    if ( (buffer->buffer = realloc(buffer->buffer, buffer->len)) == NULL )
      error("out of memory allocating string buffer") ;
  }

  while ( len-- > 0 )
    buffer->buffer[current++] = tolower(*data++) ;

  buffer->current = current ;
  buffer->buffer[current] = '\0' ;
}

/* String buffer for accumulating user and domain names */
static strbuf domainuser  = { NULL, 0, 0 } ;

/* Basic lexical units for structured fields are (according to RFC 822):
                     -  individual special characters
                     -  quoted-strings
                     -  domain-literals
                     -  comments
                     -  atoms
  They may be separated by whitespace, so the following definitions include
  comments in whitespace, and allow whitespace before any of these units.

  Special characters are treated individually where they occur.
  */

/* Boolean used to mark that this comment may contain domain names, and
   helper function to add these names. */
static bool inreceived = false ;

static void comment_domain(char *start, long len)
{
  char *dot, *name ;

  static strbuf commentbuf = { NULL, 0, 0 } ;

  while ( *start == ' ' || *start == '\t' ) {
    if ( --len == 0 )
      return ;
    ++start ;
  }

  commentbuf.current = 0 ;
  addbuffer(start, len, &commentbuf) ;

  for ( name = commentbuf.buffer ;
        (dot = strchr(name, '.')) != NULL ;
        name = dot + 1 )
    add_domain(name, true) ;
}

static bool comment(char **lineptr)
{
  char *line = *lineptr, *start ;
  int nesting = 1 ;
  bool isdomain = inreceived, isdotquad = inreceived ;

  if ( *(errorpos = line++) != '(' ) {
    errorstr = "Invalid start of comment" ;
    return false ;
  }

  start = line ;

  for (;;) {
    char ch = *line ;
    if ( ch == '\\' ) { /* quoted-pair */
      isdomain = isdotquad = false ;
      ++line ;
    } else if ( ch == '\r' || ch == '\n' ) { /* not ctext */
      break ;
    } else if ( ch == '(' ) { /* nested comment */
      isdomain = isdotquad = false ;
      ++nesting ;
    } else if ( ch == ')' && --nesting == 0 ) { /* end comment */
      if ( isdomain ) {
        comment_domain(start, line - start) ;
      } else if ( isdotquad ) {
        unsigned int a, b, c, d ;

        if ( sscanf(start, "[%u.%u.%u.%u]", &a, &b, &c, &d) == 4 )
          add_address(a, b, c, d, true) ;
      }
      break ;
    } else {
      switch ( ch ) {
      case ' ': case '\t':
        if ( nesting == 1 ) {
          if ( isdomain ) {
            comment_domain(start, line - start) ;
          } else if ( isdotquad ) {
            unsigned int a, b, c, d ;

            if ( sscanf(start, "[%u.%u.%u.%u]", &a, &b, &c, &d) == 4 )
              add_address(a, b, c, d, true) ;
          }
          start = line + 1 ;
          isdomain = isdotquad = inreceived ;
        }
        break ;
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
      case '.':
        break ; /* Digits and dot are OK in both dotquad and domain */
      case '<': case '>': case '@': case ',': case ';':
      case ':': case '"': case '(': case ')':
        isdomain = isdotquad = false ; /* OK in neither */
        break ;
      case '[': case ']': /* OK in dotquad */
        isdomain = false ;
        break ;
      default: /* All printable characters are OK in domain but not dotquad */
        isdotquad = false ;
        if ( ch < 32 || ch > 126 )
          isdomain = false ;
        break ;
      }
    }
    ++line ;
  }

  if ( *(errorpos = line++) != ')' ) {
    errorstr = "Invalid end of comment" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

static bool white(char **lineptr)
{
  char *line = *lineptr ;

  for (;;) {
    if ( *line == ' ' || *line == '\t' )
      line++ ;
    else if ( !comment(&line) )
      break ;
  }

  *lineptr = line ;

  return true ;
}

static bool atom(char **lineptr)
{
  char *line = *lineptr, *prev, *start ;

  if ( !white(&line) )
    return false ;

  start = line ;

  do {
    char ch = *line ;

    prev = line ;

    if ( ch > 31 && ch < 127 )
      switch ( ch ) {
      case ' ': case '(': case ')': case '<': case '>': case '@':
      case ',': case ';': case ':': case '\\': case '"': case '.':
      case '[': case ']':
	break ;
      default:
	++line ;
      }
  } while ( prev != line ) ;

  if ( line == start ) { /* Must be at least one */
    errorpos = line ;
    errorstr = "Atom contains no valid characters" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

static bool quotedstring(char **lineptr)
{
  char *line = *lineptr ;

  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != '"' ) {
    errorstr = "Invalid quote at start of string" ;
    return false ;
  }

  for (;;) {
    char ch = *line ;
    if ( ch == '"' || ch == '\\' || ch == '\r' || ch == '\n' )
      break ;
    ++line ;
  }

  if ( *(errorpos = line++) != '"' ) {
    errorstr = "Invalid quote at end of string" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

static bool domainliteral(char **lineptr)
{
  char *line = *lineptr ;
  unsigned int a, b, c, d ;

  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != '[' ) {
    errorstr = "Invalid bracket at start of domain literal" ;
    return false ;
  }

  for (;;) {
    char ch = *line ;
    if ( ch == '[' || ch == ']' || ch == '\r' || ch == '\n' ) /* not dtext */
      break ;
    if ( ch == '\\' ) /* quoted-pair */
      ++line ;
    ++line ;
  }

  if ( sscanf(errorpos, "[%u.%u.%u.%u]", &a, &b, &c, &d) == 4 )
    add_address(a, b, c, d, true) ;

  if ( *(errorpos = line++) != ']' ) {
    errorstr = "Invalid bracket at end of domain literal" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

/* String as defined by RFC821 */
static bool rfc821string(char **lineptr)
{
  char *line = *lineptr, *prev, *start ;

  if ( !white(&line) )
    return false ;

  start = line ;

  do {
    char ch = *line ;

    prev = line ;

    if ( ch > 31 && ch < 127 )
      switch ( ch ) {
      case '.':
	if ( relax & RelaxId )
	  ++line ;
      case ' ': case '(': case ')': case '<': case '>': case '@':
      case ',': case ';': case ':': case '"':
      case '[': case ']':
	break ;
      case '\\':
	++line ; /* FALLTHRU */
      default:
	++line ;
      }
  } while ( prev != line ) ;

  if ( line == start ) { /* Must be at least one */
    errorpos = line ;
    errorstr = "String contains no characters" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

/* Delimiters */
bool delimiter(char delim)
{
  switch (delim) {
  case ' ': case '\t': case '\r': case '\n':
  case '(': case ')': case '<': case '>': case '@': case ',': case ';':
  case ':': case '\\': case '"': case '.': case '[': case ']':
    return true ;
  }

  return false ;
}

/* Note that keyword does NOT allow whitespace beforehand */
bool keyword(char **lineptr, char *key)
{
  char *line = *lineptr ;
  int len = strlen(key) ;

  if ( strnicmp(line, key, len) != 0 || !delimiter(line[len]) ) {
    errorpos = line ;
    errorstr = "Unmatched keyword" ;
    return false ;
  }

  *lineptr = line + len ;

  return true ;
}

/* Optional character. Return true if char was present, false otherwise */
static bool optional(char **lineptr, char ch)
{
  char *line = *lineptr ;

  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ch ) {
    errorstr = "Optional character wasn't optional" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

/* End of line checks that there's nothing more but whitespace */
static bool eol(char **lineptr)
{
  char *line = *lineptr ;

  return *line == '\0' ;
}

/* Higher level units for structured fields are composed out of the basic units
 */
static bool word(char **lineptr)
{
  char *line = *lineptr ;

  if ( !atom(&line) && !quotedstring(&line) )
    return false ;

  *lineptr = line ;

  return true ;
}

static bool localpart(char **lineptr, strbuf *name)
{
  char *line = *lineptr, *start ;

  /* Dispose of whitespace/comments so that accumulation of atom into name
     won't take a copy */
  if ( !white(&line) )
    return false ;

  start = line ;

  if ( !word(&line) )
    return false ;

  addbuffer(start, line - start, name) ;

  while ( white(&line) && *line == '.' ) {
    start = ++line ;
    addbuffer(".", 1, name) ;
    if ( !word(&line) )
      return false ;
    addbuffer(start, line - start, name) ;
  }

  *lineptr = line ;
  
  return true ;
}

/* Subdomain parsing affects isliteral flag only if a domain literal is
   found */
static bool subdomain(char **lineptr, strbuf *name, bool *isliteral)
{
  char *line = *lineptr, *start ;

  /* Dispose of whitespace/comments so that accumulation of atom into name
     won't take a copy */
  if ( !white(&line) )
    return false ;

  start = line ;

  if ( !atom(&line) ) {
    if ( !domainliteral(&line) )
      return false ;
    *isliteral = true ;
  }

  if ( name ) addbuffer(start, line - start, name) ;

  *lineptr = line ;

  return true ;
}

static bool domain(char **lineptr, strbuf *name)
{
  char *line = *lineptr ;
  bool isliteral = false ;
  long current ;

  if ( name ) current = name->current ;

  if ( !subdomain(&line, name, &isliteral) )
    return false ;

  while ( white(&line) && *line == '.' ) {
    if ( name ) addbuffer(".", 1, name) ;
    ++line ;
    if ( !subdomain(&line, name, &isliteral) )
      return false ;
  }

  *lineptr = line ;

  /* If it's not a literal (in which case we'll have added dotted quads to
     the domain list), add all domains to the hash tables. */
  if ( !isliteral && name ) {
    char *dot, *dmn ;
    for ( dmn = name->buffer + current ;
          (dot = strchr(dmn, '.')) != NULL ;
          dmn = dot + 1 )
      add_domain(dmn, true) ;
  }

  return true ;
}

static bool addrspec(char **lineptr, int relaxed)
{
  char *line = *lineptr ;

  domainuser.current = 0 ;

  if ( !localpart(&line, &domainuser) || !white(&line) )
    return false ;

  if ( *line == '@' ) {
    ++line ;
    addbuffer("@", 1, &domainuser) ;
    if ( !domain(&line, &domainuser) )
      return false ;
  } else if ( !relaxed ) {
    errorpos = line ;
    errorstr = "Missing @ in address spec" ;
    return false ;
  }

  add_user(domainuser.buffer, true) ;

  *lineptr = line ;
  
  return true ;
}

static bool rfc821path(char **lineptr, int relaxed)
{
  char *line = *lineptr, *prev ;

  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != '<' ) {
    errorstr = "Invalid angle at start of address path" ;
    return false ;
  }

  if ( white(&line) && *line == '@' ) {
    do {
      if ( *(errorpos = line++) != '@' ) {
	errorstr = "Missing @ in address path" ;
	return false ;
      }
      if ( !domain(&line, NULL) )
	return false ;
    } while ( white(&line) && *line != ':' ) ;
    ++line ;
  }

  if ( !addrspec(&line, relaxed) )
    return false ;

  if ( *(errorpos = line++) != '>' ) {
    errorstr = "Invalid angle at end of address path" ;
    return false ;
  }

  *lineptr = line ;
  
  return true ;
}

static bool msgid(char **lineptr)
{
  char *line = *lineptr ;

  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != '<' ) {
    errorstr = "Invalid angle at start of message-id" ;
    return false ;
  }

  if ( !addrspec(&line, true) )
    return false ;

  if ( *(errorpos = line++) != '>' ) {
    errorstr = "Invalid angle at end of message-id" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

static bool day(char **lineptr)
{
  static char *days[] = { "mon", "tue", "wed", "thu", "fri", "sat", "sun" } ;

  char *line = *lineptr ;
  int index = 0 ;
  int found = false ;

  if ( !white(&line) )
    return false ;

  for ( index = 0 ; index < sizeof(days)/sizeof(char *) ; ++index ) {
    if ( keyword(&line, days[index]) ) {
      found = true ;
      break ;
    }
  }

  if ( !found ) {
    errorpos = line ;
    errorstr = "Three-letter day abbreviation not found" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

static bool month(char **lineptr)
{
  static char *months[] = { "jan", "feb", "mar", "apr", "may", "jun",
			    "jul", "aug", "sep", "oct", "nov", "dec" } ;

  char *line = *lineptr ;
  int index = 0 ;
  int found = false ;

  if ( !white(&line) )
    return false ;

  for ( index = 0 ; index < sizeof(months)/sizeof(char *) ; ++index ) {
    if ( keyword(&line, months[index]) ) {
      found = true ;
      break ;
    }
  }

  if ( !found ) {
    errorpos = line ;
    errorstr = "Three-letter month abbreviation not found" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

static bool date(char **lineptr)
{
  char *line = *lineptr ;

  if ( !white(&line) )
    return false ;

  if ( !isdigit(*(errorpos = line)) ) {
    errorstr = "Date does not start with number" ;
    return false ;
  }

  ++line ;

  if ( isdigit(*line) )
    ++line ;

  if ( !delimiter(*line) || !month(&line) || !white(&line) )
    return false ;

  if ( !isdigit(line[0]) || !isdigit(line[1]) ) {
    errorpos = line ;
    errorstr = "Two or more digits required for year" ;
    return false ;
  }

  line += 2 ;

  if ( isdigit(*line) )
    ++line ;

  if ( isdigit(*line) )
    ++line ;

  if ( !delimiter(*(errorpos = line)) ) {
    errorstr = "Invalid delimiter at end of date" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

static bool time(char **lineptr)
{
  static char *zones[] = { "ut", "gmt", "est", "edt", "cst", "cdt",
			   "mst", "mdt", "pst", "pdt",
			   /* strictly speaking not allowed */ "bst" } ;

  char *line = *lineptr ;
  int index = 0 ;
  int found = false ;

  if ( !white(&line) )
    return false ;

  if ( !isdigit(line[0]) || !isdigit(line[1]) ||
       line[2] != ':' || !isdigit(line[3]) || !isdigit(line[4]) ) {
    errorpos = line ;
    errorstr = "Time format HH:MM required" ;
    return false ;
  }

  line += 5 ;

  if ( line[0] == ':' ) {
    if ( !isdigit(line[1]) || !isdigit(line[2]) ) {
      errorpos = line ;
      errorstr = "Time format HH:MM:SS not correct" ;
      return false ;
    }
    line += 3 ;
  }

  if ( !white(&line) )
    return false ;

  for ( index = 0 ; index < sizeof(zones)/sizeof(char *) ; ++index ) {
    if ( keyword(&line, zones[index]) ) {
      found = true ;
      break ;
    }
  }

  if ( !found && isalpha(line[0]) && delimiter(line[1]) ) {
    ++line ;
    found = true ;
  }

  if ( !found ) {
    if ( line[0] != '+' && line[0] != '-' ) {
      errorpos = line ;
      errorstr = "Numeric offset timezone not correct" ;
      return false ;
    }

    ++line ;

    if ( !white(&line) )
      return false ;

    if ( !isdigit(line[0]) || !isdigit(line[1]) ||
	 !isdigit(line[2]) || !isdigit(line[3]) ) {
      errorpos = line ;
      errorstr = "Four digits required for numeric offset timezone" ;
      return false ;
    }

    line += 4 ;
  }

  *lineptr = line ;

  return true ;
}

static bool datetime(char **lineptr)
{
  char *line = *lineptr ;

  if ( day(&line) && white(&line) ) {
    if ( *(errorpos = line++) != ',' ) {
      errorstr = "Comma required between day and date" ;
      return false ;
    }
  }

  if ( !date(&line) || !time(&line) )
    return false ;

  *lineptr = line ;

  return true ;
}

static bool phrase(char **lineptr)
{
  char *line = *lineptr ;
  int count = 0 ;

  while ( word(&line) )
    ++count ;

  if ( count == 0 ) {
    errorpos = line ;
    errorstr = "No words in phrase" ;
    return false ;
  }

  *lineptr = line ;

  return true ;
}

static bool mbox(char **lineptr)
{
  char *line = *lineptr ;

  if ( !addrspec(&line, false) ) {
    if ( (!phrase(&line) && !(relax & RelaxMailbox)) ||
	 !rfc821path(&line, (relax & RelaxTo)) )
      return false ;
  }

  *lineptr = line ;

  return true ;
}

/* Header functions. These include EOL checks. */
bool received(char *line)
{
  int gotsemi = false ;

  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ':' ) {
    errorpos = "Received header missing initial colon" ;
    return false ;
  }

  if ( white(&line) && keyword(&line, "from") ) {
    inreceived = true ;
    domainuser.current = 0 ;
    if ( !domain(&line, &domainuser) ) {
      inreceived = false ;
      return false ;
    }
    inreceived = false ;
    if ( relax & RelaxSemi )
      gotsemi = optional(&line, ';') ;
  }

  if ( white(&line) && keyword(&line, "by") ) {
    domainuser.current = 0 ;
    if ( !domain(&line, &domainuser) )
      return false ;
    gotsemi = false ;
    if ( relax & RelaxSemi )
      gotsemi = optional(&line, ';') ;
  }

  if ( white(&line) && keyword(&line, "via") ) {
    if ( !atom(&line) )
      return false ;
    gotsemi = false ;
    if ( relax & RelaxSemi )
      gotsemi = optional(&line, ';') ;
  }

  while ( white(&line) && keyword(&line, "with") ) {
    if ( (relax & RelaxMicrosoft) && keyword(&line, "Microsoft") ) {
      (void)keyword(&line, "MAPI") ; /* Ignore MAPI, Microsoft is an atom */
    } else if ( !atom(&line) )
      return false ;
    gotsemi = false ;
    if ( relax & RelaxSemi )
      gotsemi = optional(&line, ';') ;
  }

  if ( white(&line) && keyword(&line, "id") ) {
    if ( !msgid(&line) && (!rfc821 || !rfc821string(&line)) )
      return false ;
    /* RFC821 also allows <string>
       string is any sequence of chars not including special or space
       */
    gotsemi = false ;
    if ( relax & RelaxSemi )
      gotsemi = optional(&line, ';') ;
  }

  if ( white(&line) && keyword(&line, "for") ) {
    if ( !addrspec(&line, (relax & RelaxFor)) &&
	 (!rfc821 || !rfc821path(&line, (relax & RelaxFor))) )
      return false ;
    /* RFC821 also allows <path>, which is
       "<" [ #( "@" domain ) ":" ] local-part "@" domain ">"
       RFC1123 extends this with a list of paths
       */
    gotsemi = false ;
  }

  if ( !gotsemi ) {
    if ( !white(&line) )
      return false ;
    if ( *(errorpos = line++) != ';' ) {
      errorstr = "Semi-colon required before date" ;
      return false ;
    }
  }

  if ( !datetime(&line) || !eol(&line) )
    return false ;

  return true ;
}

bool messageid(char *line)
{
  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ':' ) {
    errorstr = "Message-ID header missing initial colon" ;
    return false ;
  }

  if ( !msgid(&line) || !eol(&line) )
    return false ;

  return true ;
}

bool returnpath(char *line)
{
  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ':' ) {
    errorstr = "Return-Path header missing initial colon" ;
    return false ;
  }

  if ( !rfc821path(&line, (relax & RelaxReturnPath)) || !eol(&line) )
    return false ;

  return true ;
}

bool dateheader(char *line)
{
  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ':' ) {
    errorstr = "Date header missing initial colon" ;
    return false ;
  }

  if ( !datetime(&line) || !eol(&line) )
    return false ;

  return true ;
}

bool mailbox(char *line)
{
  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ':' ) {
    errorstr = "Sender header missing initial colon" ;
    return false ;
  }

  if ( !mbox(&line) || !eol(&line) )
    return false ;

  return true ;
}

bool mailboxes(char *line)
{
  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ':' ) {
    if ( relax & RelaxFrom )
      return true ;
    errorstr = "From header missing initial colon" ;
    return false ;
  }

  do {
    if ( !mbox(&line) || !white(&line) )
      if ( *line != ',' ) /* Allow null list elements */
	return false ;
  } while ( *line == ',' && ++line ) ;

  if ( !eol(&line) )
    return false ;

  return true ;
}

bool addresses(char *line)
{
  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ':' ) {
    errorstr = "To/Reply-To/CC header missing initial colon" ;
    return false ;
  }

  do {
    if ( !mbox(&line) ) {
      if ( !phrase(&line) || !white(&line) )
	return false ;

      if ( *(errorpos = line++) != ':' ) {
	errorstr = "Colon required between phrase and mailboxes" ;
	return false ;
      }

      do {
	(void)mbox(&line) ; /* Allowed null elements, ignore return value */
	if ( !white(&line) )
	  return false ;
      } while ( *line == ',' && ++line ) ;

      if ( *(errorpos = line++) != ';' ) {
	errorstr = "Semi-colon required after group" ;
	return false ;
      }
    }

    if ( !white(&line) )
      return false ;
  } while ( *line == ',' && ++line ) ;

  if ( !eol(&line) )
    return false ;

  return true ;
}

bool references(char *line)
{
  if ( relax & RelaxReferences )
    return true ;

  if ( !white(&line) )
    return false ;

  if ( *(errorpos = line++) != ':' ) {
    errorstr = "References header missing initial colon" ;
    return false ;
  }

  for (;;)
    if ( !phrase(&line) && !msgid(&line) )
      break ;

  if ( !eol(&line) )
    return false ;

  return true ;
}

/*
$Log: rfc822.c,v $
Revision 1.3  1999/11/28 18:03:05  angus
Add Unix name resolution, blocked address update, etc.

Revision 1.2  1999/11/24 05:17:10  angus
Avoid blocking own domain

Revision 1.1.1.1  1999/11/24 04:18:51  angus
Initial import

*/
