/* article.c $Id: article.c,v 1.2 2000/01/14 03:40:04 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997-1998.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Manage article database for NNTP buffer
 */

#include "nntp.h"
#include "messages.h"

static DBM *article_db = NULL ;

int InitArticleDB(void)
{
  char filename[MAX_PATH] ;
  int result = TRUE ;

  if ( !Lock(ARTICLE_MUTEX, INFINITE) )
    return FALSE ;

  assert (confdir) ;

  strcpy(filename, confdir) ;
  strcat(filename, "\\") ;
  strcat(filename, ARTICLE_DB_NAME) ;

  debug(DEBUG_ARTICLES, "Opening article DBM %s\n", filename) ;

  if ( (article_db = dbm_open(filename, O_RDWR|O_CREAT, 0644)) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_FILEOPEN, "database", filename,
	    ErrorText(GetLastError()), NULL) ;
    result = FALSE ;
  }

  Unlock(ARTICLE_MUTEX) ;

  return result ;
}

void CloseArticleDB(void)
{
  (void)Lock(ARTICLE_MUTEX, INFINITE) ;	/* Plough on regardless of result */

  if ( article_db != NULL ) {
    debug(DEBUG_ARTICLES, "Closing article DBM\n") ;
    dbm_close(article_db) ;
  }

  Unlock(ARTICLE_MUTEX) ;
}

Article *ArticleDBGet(char *name)
{
  Article *result = NULL ;

  if ( Lock(ARTICLE_MUTEX, INFINITE) ) {
    datum articlekey, articleval ;

    articlekey.dptr = name ;
    articlekey.dsize = strlen(name) + 1 ; /* include \0 in key */

    /* Check article DB for article record */
    debug(DEBUG_ARTICLES, "Get article %s (%d)\n",
          articlekey.dptr, articlekey.dsize) ;
    articleval = dbm_fetch(article_db, articlekey) ;
    debug(DEBUG_ARTICLES, "Get article result=%s (%d)\n",
          articleval.dptr ? articleval.dptr : "NULL",
          articleval.dsize) ;
    result = (Article *)articleval.dptr ;

    Unlock(ARTICLE_MUTEX) ;
  }

  return result ;
}

int ArticleDBPut(char *name, Article *value)
{
  int result = FALSE ;

  if ( Lock(ARTICLE_MUTEX, INFINITE) ) {
    datum articlekey, articleval ;
    char *scan = value->names ;
    int len ;

    do {
      scan += (len = strlen(scan)) + 1 ;
    } while (len != 0) ;

    len = scan - (char *)value ;

    articlekey.dptr = name ;
    articlekey.dsize = strlen(name) + 1 ; /* include \0 in key */
    articleval.dptr = (char *)value ;
    articleval.dsize = len ;	/* Include all names */

    debug(DEBUG_ARTICLES, "PutArticle %s (%d) => %x (%d)\n", articlekey.dptr,
	  articlekey.dsize, articleval.dptr, articleval.dsize) ;

    /* Check article DB for article record */
    if ( dbm_store(article_db, articlekey, articleval, DBM_REPLACE) >= 0 )
      result = TRUE ;

    debug(DEBUG_ARTICLES, "PutArticle result=%d\n", result) ;

    Unlock(ARTICLE_MUTEX) ;
  }

  return result ;
}

int ArticleDBRemove(char *name)
{
  int result = FALSE ;
  if ( Lock(ARTICLE_MUTEX, INFINITE) ) {
    datum articlekey ;

    articlekey.dptr = name ;
    articlekey.dsize = strlen(name) + 1 ; /* include \0 in key */

    debug(DEBUG_ARTICLES, "Removing article %s (%d)\n", articlekey.dptr,
          articlekey.dsize) ;
    if ( dbm_delete(article_db, articlekey) >= 0 )
      result = TRUE ;

    Unlock(ARTICLE_MUTEX) ;
  }

  return result ;
}

Article *NewArticle(char *article, ArticleState istate, long size)
{
  Article newarticle ;

  debug(DEBUG_ARTICLES, "Article %s is new\n", article) ;

  newarticle.state = istate ;
  newarticle.size = size ;
  newarticle.mark = 0 ;
  newarticle.names[0] = '\0' ;
  GetSystemTimeAsFileTime(&newarticle.download) ;
  memset(&newarticle.expires, 0, sizeof(FILETIME)) ;

  debug(DEBUG_ARTICLES, "Adding article %s\n", article) ;

  if ( !ArticleDBPut(article, &newarticle) )
    return NULL ;

  return ArticleDBGet(article) ;	/* Now look it up and return pointer */
}

int ForAllArticles(int (*fn)(char *msgid, void *passthrough),
		   void *passthrough)
{
  int result = TRUE ;
  char *names = NULL ;

  static long allocsize = 0 ;	/* size of initial alloc for names */

  if ( Lock(ARTICLE_MUTEX, INFINITE) && (names = malloc(allocsize)) != NULL ) {
    datum artkey = dbm_firstkey(article_db) ;
    long used = 0 ;

    while ( artkey.dptr ) {
      if ( names == NULL || used + artkey.dsize >= allocsize ) {
	allocsize += 4096 ;
	if ( (names = realloc(names, allocsize)) == NULL )
	  break ;
      }

      strcpy(names + used, artkey.dptr) ;
      used += artkey.dsize ;
      
      artkey = dbm_nextkey(article_db) ;
    }

    if ( names )
      names[used] = '\0' ;

    Unlock(ARTICLE_MUTEX) ;
  }

  if ( names ) {
    char *here = names ;	/* reset here pointer */

    while ( result && *here ) {
      result = (*fn)(here, passthrough) ; /* Callback is not locked */
      here += strlen(here) + 1 ;
    }

    free(names) ;
  }

  return result ;
}

void SetCurrentArticle(Client *client, long number)
{
  debug(DEBUG_ARTICLES, "Client set article %d\n", number) ;
  client->curr_artno = number ;
}

long CurrentArticleNum(Client *client)
{
  return client->curr_artno ;
}

/* reserved names
   CON, AUX, COM1, COM2, COM3, COM4, LPT1, LPT2, LPT3, PRN, NUL
   */ 

char *ArticleFileName(char *group, long artno, char *buffer, int bufsize)
{
  char *out = buffer ;
  int len = strlen(spooldir) ;

  static char *three[] = {"CON", "AUX", "PRN", "NUL"} ;
  static char *four[] = {"COM1", "COM2", "COM3", "COM4", "LPT1", "LPT2", "LPT3"} ;

  if ( bufsize < len + 2 )
    return NULL ;

  strcpy(out, spooldir) ;
  out += len ;
  *out++ = '\\' ;
  bufsize -= len + 1 ;

  for (;;) {
    char *dot = strchr(group, '.') ;
    int len = dot ? (dot - group) : strlen(group) ;
    int index ;

    if ( bufsize < len + 2 )	/* One for \0 or \\, one for possible _ */
      return NULL ;

    if ( isdigit(*group) ) {	/* Prevent group and article confusion */
      *out++ = '_' ;
      bufsize-- ;
    } else if ( len == 3 ) {
      for ( index = 0 ; index < sizeof(three)/sizeof(char *) ; index++ )
	if ( _strnicmp(group, three[index], 3) == 0 ) {
	  *out++ = '_' ;
	  bufsize-- ;
	  break ;
	}
    } else if ( len == 4 ) {
      for ( index = 0 ; index < sizeof(four)/sizeof(char *) ; index++ )
	if ( _strnicmp(group, four[index], 4) == 0 ) {
	  *out++ = '_' ;
	  bufsize-- ;
	  break ;
	}
    }

    strncpy(out, group, len) ;	/* copy component into name */
    out += len ;
    group += len ;
    bufsize -= len ;

    if ( !dot )	/* finished, no more components */
      break ;

    *out++ = '\\' ;
    group++ ;	/* skip dot */
  }

  if ( artno >= 0 ) {
    if ( bufsize < 12 )	/* 10 digits in 0x7fffffff plus \\ and \0 */
      return NULL ;

    sprintf(out, "\\%d", artno) ;
  } else
    *out = '\0' ;

  return buffer ;
}

/* Create article directories */
int MakeArticleDirs(char *filename)
{
  char *scan = filename, *slash ;
  int result = TRUE ;

  /* Start at C:\dir or \\host\dir */
  if ( isalpha(scan[0]) && scan[1] == ':' && scan[2] == '\\' )
    scan += 3 ;
  else if ( scan[0] == '\\' && scan[1] == '\\' ) {
    if ( (scan = strchr(scan + 2, '\\')) == NULL )
      return FALSE ;
    scan += 1 ;
  }

  while ( result && (slash = strchr(scan, '\\')) != NULL ) {
    HANDLE dirh ;

    *slash = '\0' ;

    debug(DEBUG_ARTICLES, "Create article directory %s\n", filename) ;

    if ( (dirh = CreateFile(filename,
			    0,
			    FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
			    NULL,
			    OPEN_EXISTING,
			    FILE_FLAG_BACKUP_SEMANTICS,
			    NULL)) != INVALID_HANDLE_VALUE )
      CloseHandle(dirh) ;
    else
      result = CreateDirectory(filename, NULL) ;

    *slash = '\\' ;
    scan = slash + 1 ;
  }

  return result ;
}

/* Add a new group:number pair to an article */
Article *AddArticleName(char *artname, char *group, long number, Article *art)
{
  int len, namelen ;
  char *scan ;
  Article *new, *result = NULL ;
  char name[MAX_PATH] ;

  if ( art == NULL && (art = ArticleDBGet(artname)) == NULL )
    return NULL ;

  sprintf(name, "%s:%d", group, number) ;
  namelen = strlen(name) ;

  scan = art->names ;

  do {
    if ( _stricmp(scan, name) == 0 )	/* test if name already there */
      return art ;
    scan += (len = strlen(scan)) + 1 ;	/* skip over string */
  } while (len != 0) ;

  len = scan - (char *)art ;

  if ( (new = malloc(len + namelen + 1)) == NULL )
    return NULL ;

  memcpy((char *)new, (char *)art, len) ;
  strcpy((char *)new + len - 1, name) ;
  ((char *)new)[len + namelen] = '\0' ;

  debug(DEBUG_ARTICLES, "Add article name %s %s:%d\n", artname, group, number) ;

  if ( ArticleDBPut(artname, new) &&	/* Update article */
       MessageAdd(group, number, artname) )
    result = ArticleDBGet(artname) ;

  free(new) ;
  return result ;
}

/* Remove a filename from an article */
Article *RemoveArticleName(char *artname, char *group, long artno, Article *art)
{
  char *scan, *found = NULL ;
  Article *new, *result = NULL ;
  int namelen, len ;
  char name[MAX_PATH] ;

  sprintf(name, "%s:%d", group, artno) ;
  namelen = strlen(name) ;

  if ( art == NULL && (art = ArticleDBGet(artname)) == NULL )
    return NULL ;

  scan = art->names ;

  do {
    if ( _stricmp(scan, name) == 0 )
      found = scan ;	/* Note old name position */
    scan += (len = strlen(scan)) + 1 ;	/* skip over string */
  } while (len != 0) ;

  if ( found == NULL )	/* Not in this article */
    return art ;

  len = scan - (char *)art ;

  if ( (new = malloc(len - namelen - 1)) == NULL )
    return NULL ;

  len = found - (char *)art ;

  memcpy((char *)new, (char *)art, len) ;	/* low section */
  memcpy((char *)new + len, found + namelen + 1,	/* high section */
	 scan - found - namelen - 1) ;

  if ( ArticleDBPut(artname, new) )	/* Update article */
    result = ArticleDBGet(artname) ;

  free(new) ;

  return result ;
}

/* Check if group name followed by number appears in article already */
int ArticleInGroup(Article *art, char *group)
{
  char *scan = art->names ;
  int len = strlen(group) ;

  while ( *scan ) {
    if ( _strnicmp(group, scan, len) == 0 && scan[len] == ':') {
      scan += len + 1 ;
      while ( isdigit(*scan) ) {
	if ( *++scan == '\0' )
	  return TRUE ;
      }
    }
    scan += strlen(scan) + 1 ;
  }

  return FALSE ;
}

/* Get Message id of file */
char *MessageIDFile(char *filename, char *messageid, int idbufsize)
{
  Stream *arts = NULL, *lines = NULL, *heads = NULL ;
  char *result = NULL ;

  /* Scan for Message-ID line, copy into buffer */
  if ( (arts = MakeFileStream(filename, FILESTREAM_READ, stream_error)) != NULL &&
       (lines = MakeLineStream(arts, "\n")) != NULL &&
       (heads = MakeHeadStream(lines)) != NULL ) {
    long rbytes ;
    char *line, *msgid = NULL ;

    while ( result == NULL && (rbytes = ReadStream(heads, &line, 0)) > 0 ) {
      if ( rbytes > 11 && _strnicmp(line, "Message-ID:", 11) == 0 ) {
	for ( line += 11, rbytes -= 11 ; rbytes ; line++, rbytes-- ) {
	  char ch = *line ;

	  if ( ch == '<' ) {
	    if ( msgid != NULL ) /* error, badly formed message-id */
	      goto msgid_exit ;

	    msgid = line + 1 ;
	  } else if ( ch == '>' ) {
	    int len ;

	    if ( msgid == NULL ) /* error, badly formed message-id */
	      goto msgid_exit ;

	    len = line - msgid ;
	    if ( len >= idbufsize ) /* error, too long for buffer */
	      goto msgid_exit ;

	    strncpy(messageid, msgid, len) ;
	    messageid[len] = '\0' ;
	    result = messageid ;
	    break ;
	  } else if ( msgid != NULL && isspace(ch) )	/* error, bad char */
	    goto msgid_exit ;
	}
      }
    }
  }

msgid_exit:
  if ( heads )
    DestroyStream(heads) ;

  if ( lines )
    DestroyStream(lines) ;

  if ( arts )
    DestroyStream(arts) ;

  return result ;
}

/*
 * $Log: article.c,v $
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
