/* client.c $Id: client.c,v 1.3 2000/01/14 21:22:34 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Client code for NNTP buffer.
 */

#include "nntp.h"
#include "messages.h"

/* Chop off next argument and return it. Return NULL if no arg found. */
static char *NextArg(char **arg)
{
  char *start, *end ;

  for (start = *arg; *start && isspace(*start); start++ ) /* skip whitespace */ ;
  for (end = start ; *end && !isspace(*end) ; end++ ) /* skip non-blank */ ;

  if ( *start ) {
    if ( *end != '\0' )	/* Must be a space */
      *end++ = '\0' ;
    *arg = end ;
    return start ;
  }

  return NULL ;
}

enum { NNTP_ARTICLE, NNTP_BODY, NNTP_HEAD, NNTP_STAT, NNTP_GROUP,
       NNTP_HELP, NNTP_IHAVE, NNTP_LAST, NNTP_LIST, NNTP_NEWGROUPS,
       NNTP_NEWNEWS, NNTP_NEXT, NNTP_POST, NNTP_QUIT, NNTP_SLAVE, NNTP_WHAT } ;

static int NNTP_Invalid(Client *client, int cmd, char *args)
{
  char *warning ;

  switch (cmd) {
  case NNTP_IHAVE: warning = "502 Use POST instead\r\n" ; break ;
  case NNTP_SLAVE: warning = "202 Sing for me, Aida\r\n" ; break ;
  case NNTP_WHAT: warning = "500 Tell me another one\r\n"; break ;
  default: warning = "501 Garbage in, garbage out\r\n"; break ;
  }

  if ( SendStream(client->stream, warning) < 0 )
    return FALSE ;

  return TRUE ;
}

/* Write Xref header if necessary */
int WriteXref(Stream *stream, char *names)
{
  char *next = names + strlen(names) + 1 ;

  if ( *names && *next ) {
    static char hostname[MAX_HOSTLEN] ;
    static int hostnamelen = 0 ;

    if ( hostnamelen == 0 ) {
      if ( gethostname(hostname, MAX_HOSTLEN) != 0 )
	return FALSE ;
      hostnamelen = strlen(hostname) ;
    }

    if ( !WriteStreamAll(stream, "Xref: ", 6) ||
	 !WriteStreamAll(stream, hostname, hostnamelen) )
      return FALSE ;

    while ( *names ) {
      int len = strlen(names) ;

      if ( !WriteStreamAll(stream, " ", 1) ||
	   !WriteStreamAll(stream, names, len) )
	return FALSE ;

      names += len + 1 ;
    }

    if ( !WriteStreamAll(stream, "\r\n", 2) )
      return FALSE ;
  }

  return TRUE ;
}

static int NNTP_Article(Client *client, int cmd, char *args)
{
  char *art = NextArg(&args) ;
  int artno = 0 ;
  Article *article = NULL ;
  Stream *strm = NULL ;
  char filename[MAX_PATH] ;
  char messageid[MAX_MESSAGEID] ;
  int result = TRUE ;
  int xref = FALSE ;

  filename[0] = '\0' ;
  messageid[0] = '\0' ;

  if ( !Lock(GROUP_MUTEX, INFINITE) ) {
    (void)SendStream(client->stream, "503 Can't lock group\r\n") ;
    return FALSE ;
  }

  if ( !Lock(ARTICLE_MUTEX, INFINITE) ) {
    (void)SendStream(client->stream, "503 Can't lock group\r\n") ;
    result = FALSE ;
    goto article_unlock ;
  }

  if ( art ) {
    if ( isdigit(*art) ) {	/* Get article number, check its an int */
      char *scan ;

      for (scan = art ; isdigit(*scan) ; *scan++ )
	artno = artno * 10 + *scan - '0' ;

      if ( *scan || (art = NextArg(&args)) != NULL ) {
	result = NNTP_Invalid(client, cmd, art) ;
	goto article_exit ;
      }
    } else { /* Message-ID parameter. Check syntax first */
      int len = strlen(art) ;
      char *msgid = art + 1 ;

      if ( len < 2 || art[0] != '<' || art[len-1] != '>' ) {
	result = NNTP_Invalid(client, cmd, art) ;
	goto article_exit ;
      }

      art[len-1] = '\0' ;

      if ( (art = NextArg(&args)) != NULL ) {
	result = NNTP_Invalid(client, cmd, art) ;
	goto article_exit ;
      }

      if ( len > MAX_MESSAGEID || (article = ArticleDBGet(msgid)) == NULL ||
	   article->state != ARTICLE_BODY ) {
	if ( SendStream(client->stream,
                        "430 No article matching message-id %s\r\n",
                        msgid) < 0 )
	  result = FALSE ;
	goto article_exit ;
      }

      strcpy(messageid, msgid) ;	/* Save message-id */
    }
  } else if ( (artno = CurrentArticleNum(client)) < 0 ) { /* Current article */
    if ( SendStream(client->stream, "420 No current article selected\r\n") < 0 )
      result = FALSE ;
    goto article_exit ;
  }

  if ( article == NULL ) {	/* Find article based on group,number */
    Group *group = CurrentGroup(client) ;

    if ( group == NULL ) {
      if ( SendStream(client->stream, "412 No newsgroup selected\r\n") < 0 )
	result = FALSE ;
      goto article_exit ;
    }

    if ( artno >= group->low && artno <= group->high ) {
      char *gname = CurrentGroupName(client) ;

      if ( ArticleFileName(gname, artno, filename, MAX_PATH) != NULL &&
	   MessageID(gname, artno, messageid, MAX_MESSAGEID) != NULL )
	article = ArticleDBGet(messageid) ;
    }

    if ( article == NULL || article->state != ARTICLE_BODY ) {
      if ( SendStream(client->stream, "423 No such article in group\r\n") < 0 )
	result = FALSE ;
      goto article_exit ;
    }

    /* Set current article number */
    SetCurrentArticle(client, artno) ;
  }

  /* At this point, we have article and message-id, but not necessarily
     file-name if doing a <message-id> lookup. Responses are:
     220 n <a> - article follows
     221 n <a> - head follows
     222 n <a> - body follows
     223 n <a> - text separately
     430 no such article found
     */

  if ( filename[0] == '\0' ) /* Pick first filename */
    MessageFileName(article->names, filename, MAX_PATH) ;

  {
    char *response = NULL ;
    Stream *line = NULL, *file = NULL ;

    switch ( cmd ) {
    case NNTP_ARTICLE:
      if ( (file = MakeFileStream(filename, FILESTREAM_READ, stream_error)) == NULL ||
	   (strm = MakeLineStream(file, "\r\n")) == NULL ) {
	if ( file )
	  DestroyStream(file) ;
      } else
	response = "220 %d <%s> Article follows\r\n" ;
      break ;
    case NNTP_HEAD:
      if ( (file = MakeFileStream(filename, FILESTREAM_READ, stream_error)) == NULL ||
	   (line = MakeLineStream(file, "\r\n")) == NULL ||
	   (strm = MakeHeadStream(line)) == NULL ) {
	if ( file )
	  DestroyStream(file) ;
	if ( line )
	  DestroyStream(line) ;
      } else
	response = "221 %d <%s> Head follows\r\n" ;
      break ;
    case NNTP_BODY:
      if ( (file = MakeFileStream(filename, FILESTREAM_READ, stream_error)) == NULL ||
	   (line = MakeLineStream(file, "\r\n")) == NULL ||
	   (strm = MakeBodyStream(line)) == NULL ) {
	if ( file )
	  DestroyStream(file) ;
	if ( line )
	  DestroyStream(line) ;
      } else
	response = "222 %d <%s> Body follows\r\n" ;
      xref = TRUE ; /* No Xref substitution */
      break ;
    case NNTP_STAT:
      response = "223 %d <%s> Request text\r\n" ;
      break ;
    }

    if ( response == NULL ) {
      if ( SendStream(client->stream, "430 no such article found\r\n") < 0 )
	result = FALSE ;
      goto article_exit ;
    }

    if ( SendStream(client->stream, response, artno, messageid) < 0 ) {
      result = FALSE ;
      goto article_exit ;
    }
  }

  if ( strm != NULL ) {
    long rbytes ;
    char *buffer ;

    IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT, DEBUG_CLIENT_DATA) ;

    while ( (rbytes = ReadStream(strm, &buffer, 0)) > 0 ) {
      if ( !xref && _strnicmp(buffer, "xref:", 5) == 0 ) {
	xref = TRUE ;	/* Substitute Xref: line */
	if ( !WriteXref(client->stream, article->names) ) {
	  result = FALSE ;
	  break ;
	}
      } else {
	if ( !xref && rbytes == 2 && buffer[0] == '\r' && buffer[1] == '\n' ) {
	  xref = TRUE ;	/* Add Xref: line */
	  if ( !WriteXref(client->stream, article->names) ) {
	    result = FALSE ;
	    break ;
	  }
	}

	if ( !WriteStreamAll(client->stream, buffer, rbytes) ) {
	  result = FALSE ;
	  break ;
	}
      }
    }

    if ( !xref && !WriteXref(client->stream, article->names) )
      result = FALSE ;

    IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT_DATA, DEBUG_CLIENT) ;

    if ( SendStream(client->stream, ".\r\n") < 0 )
      result = FALSE ;

    if ( rbytes < 0 )
      result = FALSE ;

    do {
      strm = DestroyStream(strm) ;
    } while ( strm != NULL ) ;
  }

article_exit:
  Unlock(ARTICLE_MUTEX) ;

article_unlock:
  Unlock(GROUP_MUTEX) ;

  return result ;
}

static int NNTP_Group(Client *client, int cmd, char *args)
{
  char *group, *rest ;
  Server *server ;
  Pattern *patt ;
  Connection *conn ;
  int result = TRUE ;
  
  if ( (group = NextArg(&args)) == NULL || (rest = NextArg(&args)) != NULL )
    return NNTP_Invalid(client, cmd, group ? rest : group) ;

  if ( FindGroupServer(group, &server, &patt) ) {
    Group data, *gptr ;

    debug(DEBUG_GROUPS, "Feed for group %s is from %s\n", group, server->hostname) ;
    debug(DEBUG_GROUPS, "Expiry for group %s is %d/%d/%d\n", group, patt->min, patt->normal, patt->purge) ;
    debug(DEBUG_GROUPS, "Limit for group %s is %d\n", group, patt->limit) ;

    if ( !Lock(GROUP_MUTEX, INFINITE) ) {
      (void)SendStream(client->stream, "503 Can't lock group\r\n") ;
      return FALSE ;
    }

    if ( (gptr = GroupDBGet(group)) == NULL &&
	 (gptr = NewGroup(group, GROUP_ACTIVE, 0, 'y', patt->normal)) == NULL ) {
      if ( SendStream(client->stream, "411 Can't create group %s\r\n", group) < 0 )
	result = FALSE ;
      goto group_unlock ;
    }

    data = *gptr ;
    if ( gptr->state != GROUP_ACTIVE ) {
      data.state = GROUP_ACTIVE ;	/* Active unless verified bogus */

      if ( !Lock(SERVER_MUTEX, INFINITE) ) {
	(void)SendStream(client->stream, "503 Can't lock server %s\r\n",
			 server->hostname) ;
	result = FALSE ;
	goto group_unlock ;
      }

      /* Now check group against server */
      if ( (conn = ConnectServer(server)) != NULL ) {
	long code ;

	if ( SendStream(conn->stream, "GROUP %s\r\n", group) >= 0 &&
	     (code = Response(conn->stream)) >= 0 ) {
	  if ( code == 411 ) {	/* No such group */
            char reason[BUFSIZ] ;
            if ( Reason(conn->stream, reason, BUFSIZ) > 0 )
              debug(DEBUG_CLIENT, "TO Client: %s\n", reason) ;
	    data.state = GROUP_BOGUS ;
          }
	} else
	  DisconnectServer(server) ;
      }

      Unlock(SERVER_MUTEX) ;
    }

    /* Update group if status changed or selected. */
    if ( data.state != gptr->state || gptr->state != GROUP_BOGUS ) {
      GetSystemTimeAsFileTime(&data.selected) ;	/* Set select time */
      if ( GroupDBPut(group, &data) )
	gptr = GroupDBGet(group) ;
    }

    /* refuse bogus groups */
    if ( gptr == NULL || gptr->state == GROUP_BOGUS ) {
      if ( SendStream(client->stream, "411 Group %s is bogus\r\n", group) < 0 )
	result = FALSE ;
      goto group_unlock ;
    }

    SetCurrentGroup(client, group) ;
    SetCurrentArticle(client, gptr->low) ;

    if ( SendStream(client->stream, "211 %d %d %d %s group selected\r\n",
                    gptr->articles, gptr->low, gptr->high, group) < 0 )
      result = FALSE ;

  group_unlock:
    Unlock(GROUP_MUTEX) ;
  } else if ( SendStream(client->stream,
                         "411 No server for group %s\r\n", group) < 0 )
    return FALSE ;

  return result ;
}

static int NNTP_Help(Client *client, int cmd, char *args)
{
  char *helpstrings[] = {
    "100 You want help? Call the NewsBusters!\r\n",
    "This is " SERVICE_DISPLAYNAME " Version " VERSION_STRING "\r\n",
    "All RFC977 commands are supported, but SLAVE doesn't\r\n",
    "do anything, and IHAVE access is disallowed. Only one\r\n",
    "client is permitted at once. Group management is dynamic.\r\n",
    "\r\n",
    "Config directory is:\r\n",
    "\r\n",
    "Timeouts are:\r\n",
    "\r\n",
    ".\r\n",
    NULL	/* Must terminate with NULL */
  } ;
  char *rest, **help = helpstrings ;

  if ( (rest = NextArg(&args)) != NULL )
    return NNTP_Invalid(client, cmd, rest) ;

  while ( *help ) {
    if ( SendStream(client->stream, *help) < 0 )
      return FALSE ;
    help++ ;
  }

  return TRUE ;
}

static int NNTP_Select(Client *client, int cmd, char *args)
{
  char *rest ;
  Group *gptr ;
  long artno ;
  int result = TRUE ;

  if ( (rest = NextArg(&args)) != NULL )
    return NNTP_Invalid(client, cmd, rest) ;

  if ( !Lock(GROUP_MUTEX, INFINITE) ) {
    (void)SendStream(client->stream, "503 Can't lock group\r\n") ;
    return FALSE ;
  }

  if ( (gptr = CurrentGroup(client)) == NULL ) {
    if ( SendStream(client->stream, "412 No newsgroup selected\r\n") < 0 )
      result = FALSE ;
    goto select_exit ;
  }

  if ( (artno = CurrentArticleNum(client)) < 0 ) {
    if ( SendStream(client->stream, "420 No current article selected\r\n") < 0 )
      result = FALSE ;
    goto select_exit ;
  }

  for (;;) {
    if ( cmd == NNTP_LAST && artno <= gptr->low ) {
      if ( SendStream(client->stream, "422 No previous article in group\r\n") < 0 )
	result = FALSE ;
      goto select_exit ;
    } else if ( cmd == NNTP_NEXT && artno >= gptr->high ) {
      if ( SendStream(client->stream, "421 No next article in group\r\n") < 0 )
	result = FALSE ;
      goto select_exit ;
    }

    artno += (cmd == NNTP_LAST) ? -1 : 1 ;

    {
      char *name, messageid[MAX_MESSAGEID] ;

      if ( (name = CurrentGroupName(client)) != NULL &&
	   MessageID(name, artno, messageid, MAX_MESSAGEID) != NULL ) {
	if ( SendStream(client->stream, "223 %d %s article selected\r\n",
                        artno, messageid) < 0 )
	  result = FALSE ;
	else {
	  SetCurrentArticle(client, artno) ;
	  result = TRUE ;
	}
	goto select_exit ;
      }
    }
  }

  result = FALSE ; /* Bizarre internal error */

select_exit:
  Unlock(GROUP_MUTEX) ;

  return result ;
}

typedef struct {
  Client *client ;
  DWORD mark ;
} ListHelperArgs ;

static int NNTP_ListRest(char *group, void *passthrough)
{
  Client *client = ((ListHelperArgs *)passthrough)->client ;
  DWORD mark = ((ListHelperArgs *)passthrough)->mark ;
  int result = TRUE ;

  if ( Lock(GROUP_MUTEX, INFINITE) ) {
    Group *gptr = GroupDBGet(group) ;
    if ( gptr && gptr->state != GROUP_BOGUS && gptr->mark != mark ) {
      Group data = *gptr ;
      data.mark = mark ;
      if ( GroupDBPut(group, &data) &&
	   SendStream(client->stream, "%s %d %d %c\r\n", group, data.high,
                      data.low, data.post) < 0 )
	result = FALSE ;
    }
    Unlock(GROUP_MUTEX) ;
  }

  return result ;
}

static int NNTP_ListServer(Server *server, void *passthrough)
{
  Connection *conn = server->connection ;
  Client *client = ((ListHelperArgs *)passthrough)->client ;
  DWORD mark = ((ListHelperArgs *)passthrough)->mark ;
  long code ;
  Stream *dots ;

  debug(DEBUG_CLIENT, "ListServer connected to %s\n", server->hostname) ;

  if ( !Lock(GROUP_MUTEX, INFINITE) ) {
    (void)SendStream(client->stream, "503 Can't lock groups\r\n") ;
    return FALSE ;
  }

  if ( !Lock(SERVER_MUTEX, INFINITE) ) {
    (void)SendStream(client->stream, "503 Can't lock server %s\r\n",
		     server->hostname) ;
    goto list_serverfail ;
  }

  if ( SendStream(conn->stream, "LIST\r\n") < 0 ||
       (code = Response(conn->stream)) < 0 )
    goto list_disconnect ;

  if ( code != 215 ) {	/* Not getting anything from this one */
    char reason[BUFSIZ] ;
    if ( Reason(conn->stream, reason, BUFSIZ) > 0 )
      debug(DEBUG_CLIENT, "TO Client: %s\n", reason) ;
    goto list_exit ;
  }

  if ( (dots = MakeDotStream(conn->stream)) == NULL )
    goto list_disconnect ;

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_SERVER, DEBUG_DOWNLOAD) ;

  for (;;) {
    Pattern *patt ;
    long rbytes ;
    char *line, *group ;

    if ( (rbytes = ReadStream(dots, &line, 0)) < 0 )
      goto list_rmdots ;

    if ( rbytes == 0 )	/* Finished reading dot stream */
      break ;

    line[rbytes - 2] = '\0' ;	/* Terminate CRLF line */

    /* Pass groups through that match our pattern for this server */
    if ( (group = NextArg(&line)) != NULL &&
	 (patt = FindGroupPattern(group, server->patterns, server->nPatterns)) != NULL ) {
      char *first, *last, *post ;
      Group *gptr = NULL ;

      if ( (last = NextArg(&line)) == NULL ||
	   (first = NextArg(&line)) == NULL ||
	   (post = NextArg(&line)) == NULL )
	goto list_rmdots ;

      if ( (gptr = GroupDBGet(group)) != NULL ) {
	if ( gptr->mark != mark ) {	/* Not seen this one before */
	  Group data = *gptr ;
	  
	  data.mark = mark ;	/* Mark we've seen group */
	  data.post = *post ;	/* Update posting info */

	  if ( gptr->state == GROUP_BOGUS )	/* Can't be bogus, it exists */
	    data.state = GROUP_INACTIVE ;

	  if ( GroupDBPut(group, &data) )
	    gptr = GroupDBGet(group) ;
	} else
	  gptr = NULL ;	/* seen before, ignore it */
      } else /* didn't know about that group, create it */
	gptr = NewGroup(group, GROUP_INACTIVE, mark, *post, patt->normal) ;

      /* Forward response for groups which we haven't already touched */
      if ( gptr != NULL && gptr->state != GROUP_BOGUS &&
	   SendStream(client->stream, "%s %d %d %c\r\n", group, gptr->high,
                      gptr->low, gptr->post) < 0 )
	goto list_rmdots ;
    }
  }

  DestroyStream(dots) ;

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_DOWNLOAD, DEBUG_SERVER) ;

list_exit:
  Unlock(SERVER_MUTEX) ;
  Unlock(GROUP_MUTEX) ;

  return TRUE ;

list_rmdots:
  DestroyStream(dots) ;

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_DOWNLOAD, DEBUG_SERVER) ;

list_disconnect:
  DisconnectServer(server) ;
  Unlock(SERVER_MUTEX) ;

list_serverfail:
  Unlock(GROUP_MUTEX) ;

  return FALSE ;
}

static int NNTP_List(Client *client, int cmd, char *args)
{
  char *rest ;
  ListHelperArgs helpargs ;

  if ( (rest = NextArg(&args)) != NULL )
    return NNTP_Invalid(client, cmd, rest) ;

  LogString(LOG_STATUS, "Client listing groups\n") ;

  /* return known newsgroups */
  if ( SendStream(client->stream, "215 Listing newsgroups\r\n") < 0 )
    return FALSE ;

  helpargs.client = client ;
  helpargs.mark = Mark() ;	/* Mark to see which groups I've touched */

  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT, DEBUG_CLIENT_DATA) ;

  ForAllServers(NNTP_ListServer, &helpargs, TRUE) ;
  ForAllGroups(NNTP_ListRest, &helpargs) ;

  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT_DATA, DEBUG_CLIENT) ;

  if ( SendStream(client->stream, ".\r\n") < 0 )
    return FALSE ;

  return TRUE ;
}

typedef struct {
  Client *client ;
  DWORD mark ;
  FILETIME since ;
} NewgroupHelperArgs ;

static int NNTP_NewgroupRest(char *group, void *passthrough)
{
  Client *client = ((NewgroupHelperArgs *)passthrough)->client ;
  DWORD mark = ((NewgroupHelperArgs *)passthrough)->mark ;
  LPFILETIME since = &((NewgroupHelperArgs *)passthrough)->since ;
  int result = TRUE ;

  if ( Lock(GROUP_MUTEX, INFINITE) ) {
    Group *gptr = GroupDBGet(group) ;
    if ( gptr && gptr->state != GROUP_BOGUS && gptr->mark != mark &&
	 CompareFileTime(&gptr->created, since) >= 0 ) {
      Group data = *gptr ;
      data.mark = mark ;
      if ( GroupDBPut(group, &data) &&
	   SendStream(client->stream, "%s %d %d %c\r\n", group, data.high,
                      data.low, data.post) < 0 )
	result = FALSE ;
    }
    Unlock(GROUP_MUTEX) ;
  }

  return TRUE ;
}

static int NNTP_NewgroupServer(Server *server, void *passthrough)
{
  Connection *conn = server->connection ;
  Client *client = ((NewgroupHelperArgs *)passthrough)->client ;
  DWORD mark = ((NewgroupHelperArgs *)passthrough)->mark ;
  LPFILETIME since = &((NewgroupHelperArgs *)passthrough)->since ;
  long code ;
  Stream *dots ;
  char regname[MAX_REGISTRYNAME] ;
  FILETIME when ;

  debug(DEBUG_CLIENT, "NewgroupServer connected to %s\n", server->hostname) ;

  strcpy(regname, SERVICE_REGISTRY_KEY) ;
  strcat(regname, "\\") ;
  strcat(regname, conn->name) ;

  if ( !GetRegFiletime(regname, NEWGROUPS_REGNAME, &when) )
    memset(&when, 0, sizeof(FILETIME)) ;

  if ( !Lock(GROUP_MUTEX, INFINITE) ) {
    (void)SendStream(client->stream, "503 Can't lock groups\r\n") ;
    return FALSE ;
  }

  if ( !Lock(SERVER_MUTEX, INFINITE) ) {
    (void)SendStream(client->stream, "503 Can't lock server %s\r\n",
		     server->hostname) ;
    goto newgroup_serverfail ;
  }

  if ( SendStream(conn->stream, "NEWGROUPS %t GMT\r\n", &when) < 0 ||
       (code = Response(conn->stream)) < 0 )
    goto newgroup_disconnect ;

  if ( code != 231 ) {	/* Not getting anything from this one */
    char reason[BUFSIZ] ;
    if ( Reason(conn->stream, reason, BUFSIZ) > 0 )
      debug(DEBUG_CLIENT, "TO Client: %s\n", reason) ;
    goto newgroup_exit ;
  }

  if ( (dots = MakeDotStream(conn->stream)) == NULL )
    goto newgroup_disconnect ;

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_SERVER, DEBUG_DOWNLOAD) ;

  for (;;) {
    Pattern *patt ;
    long rbytes ;
    char *line, *group ;

    if ( (rbytes = ReadStream(dots, &line, 0)) < 0 )
      goto newgroup_rmdots ;

    if ( rbytes == 0 )	/* Finished reading dot stream */
      break ;

    line[rbytes - 2] = '\0' ;	/* Terminate CRLF line */

    /* Pass groups through that match our pattern for this server */
    if ( (group = NextArg(&line)) != NULL &&
	 (patt = FindGroupPattern(group, server->patterns, server->nPatterns)) != NULL ) {
      char *first, *last, *post ;
      Group *gptr = NULL ;

      if ( (last = NextArg(&line)) == NULL ||
	   (first = NextArg(&line)) == NULL ||
	   (post = NextArg(&line)) == NULL )
	goto newgroup_rmdots ;

      if ( (gptr = GroupDBGet(group)) != NULL ) {
	if ( gptr->mark != mark ) {	/* Not seen this one before */
	  Group data = *gptr ;

	  data.mark = mark ;	/* Mark we've seen group */
	  data.post = *post ;	/* Update posting info */
	  
	  if ( gptr->state == GROUP_BOGUS )	/* Can't be bogus, it exists */
	    data.state = GROUP_INACTIVE ;

	  if ( GroupDBPut(group, &data) )
	    gptr = GroupDBGet(group) ;
	} else
	  gptr = NULL ;	/* seen before, ignore it */
      } else /* didn't know about that group, create it */
	gptr = NewGroup(group, GROUP_INACTIVE, mark, *post, patt->normal) ;

      /* Forward response for groups which we haven't already touched */
      if ( gptr != NULL && gptr->state != GROUP_BOGUS &&
	   CompareFileTime(&gptr->created, since) >= 0 &&
	   SendStream(client->stream, "%s %d %d %c\r\n", group, gptr->high,
                      gptr->low, gptr->post) < 0 )
	goto newgroup_rmdots ;
    }
  }

  DestroyStream(dots) ;

  /* update last newgroup time, since that was successful */
  GetSystemTimeAsFileTime(&when) ;
  SetRegFiletime(regname, NEWGROUPS_REGNAME, &when) ;

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_DOWNLOAD, DEBUG_SERVER) ;

newgroup_exit:
  Unlock(SERVER_MUTEX) ;
  Unlock(GROUP_MUTEX) ;

  return TRUE ;

newgroup_rmdots:
  DestroyStream(dots) ;

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_DOWNLOAD, DEBUG_SERVER) ;

newgroup_disconnect:
  DisconnectServer(server) ;
  Unlock(SERVER_MUTEX) ;

newgroup_serverfail:
  Unlock(GROUP_MUTEX) ;

  return FALSE ;
}

static int NNTP_Newgroups(Client *client, int cmd, char *args)
{
  char *gdate, *gtime, *gmt = NULL, *distrib = NULL ;
  NewgroupHelperArgs helpargs ;

  if ( (gdate = NextArg(&args)) == NULL )
    return NNTP_Invalid(client, cmd, gdate) ;

  if ( (gtime = NextArg(&args)) == NULL )
    return NNTP_Invalid(client, cmd, gtime) ;

  if ( (gmt = NextArg(&args)) != NULL ) {
    if ( _stricmp(gmt, "gmt") != 0 )
      return NNTP_Invalid(client, cmd, gmt) ;
    if ( (distrib = NextArg(&args)) != NULL ) {
      char *rest ;
      if ( (rest = NextArg(&args)) != NULL )
	return NNTP_Invalid(client, cmd, rest) ;
    }
  }

  helpargs.client = client ;
  helpargs.mark = Mark() ;	/* Mark to see which groups I've touched */
  if ( !DateTimeToFiletime(&helpargs.since, gdate, gtime) ) {
    if ( SendStream(client->stream, "501 Funny date %s %s\r\n", gdate, gtime) < 0 )
      return FALSE ;
    return TRUE ;
  }

  LogString(LOG_STATUS, "Client asking for new groups\n") ;

  /* return newgroups since time */
  if ( SendStream(client->stream, "231 New newsgroups follow\r\n") < 0 )
    return FALSE ;

  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT, DEBUG_CLIENT_DATA) ;

  ForAllServers(NNTP_NewgroupServer, &helpargs, TRUE) ;
  ForAllGroups(NNTP_NewgroupRest, &helpargs) ;

  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT_DATA, DEBUG_CLIENT) ;

  if ( SendStream(client->stream, ".\r\n") < 0 )
    return FALSE ;

  return TRUE ;
}

typedef struct {
  DWORD mark ;
  Client *client ;
  FILETIME now ;
  FILETIME since ;
  char *match ;
  char *group ;
} NewnewsHelperArgs ;

static int NNTP_NewnewsFile(char *filename, void *passthrough)
{
  NewnewsHelperArgs *args = passthrough ;
  char messageid[MAX_MESSAGEID] ;
  Article *art ;
  int result = TRUE ;
  char *slash = strrchr(filename, '\\') ;

  if ( slash != NULL &&
       MessageID(args->group, atoi(slash + 1), messageid, MAX_MESSAGEID) != NULL &&
       Lock(ARTICLE_MUTEX, INFINITE) ) {
    if ( (art = ArticleDBGet(messageid)) != NULL &&
	 art->mark != args->mark &&
	 CompareFileTime(&args->since, &art->download) >= 0 ) {
      Article *clone = malloc(art->size) ;

      if ( clone != NULL ) {
	memcpy(clone, art, art->size) ;
	clone->mark = args->mark ;
	(void)ArticleDBPut(messageid, clone) ;
	if ( SendStream(args->client->stream, "<%s>\r\n", messageid) < 0 )
	  result = FALSE ;
	free(clone) ;
      } else
	message(EVENTLOG_ERROR_TYPE, MSG_MEMORY, "NEWNEWS mark", NULL) ;
    }
    Unlock(ARTICLE_MUTEX) ;
  }

  return result ;
}

static int NNTP_NewnewsGroup(char *group, void *passthrough)
{
  NewnewsHelperArgs *args = passthrough ;
  int result = TRUE ;
  char *match = args->match ;

  while ( result && match != NULL ) {
    char *comma = strchr(match, ',') ;
    int negate = (*match == '!') && match++ != NULL ;

    if ( comma )
      *comma = '\0' ;

    /* If matches and negated or doesn't match and not negated, this
       pattern doesn't match */
    if ( Match(match, group) ? negate : !negate )
      result = FALSE ;

    if ( comma ) {
      *comma = ',' ;
      match = comma + 1 ;
    } else
      match = NULL ;
  }

  /* If group matches the pattern, test for new news */
  if ( result && Lock(GROUP_MUTEX, INFINITE) ) {
    char dirname[MAX_PATH] ;
    Group *gptr = GroupDBGet(group) ;
    
    if ( gptr && ArticleFileName(group, -1, dirname, MAX_PATH) ) {
      Group data = *gptr ;

      args->group = group ;
      ForAllFiles(dirname, "", NNTP_NewnewsFile, passthrough) ;

      data.selected = args->now ;	/* Update group selected time */
      (void)GroupDBPut(group, &data) ;
    }
    Unlock(GROUP_MUTEX) ;
  }

  return result ;
}

static int NNTP_Newnews(Client *client, int cmd, char *args)
{
  char *gdate, *gtime, *gmt = NULL, *distrib = NULL ;
  NewnewsHelperArgs helpargs ;

  if ( (helpargs.match = NextArg(&args)) == NULL )
    return NNTP_Invalid(client, cmd, helpargs.match) ;

  if ( (gdate = NextArg(&args)) == NULL )
    return NNTP_Invalid(client, cmd, gdate) ;

  if ( (gtime = NextArg(&args)) == NULL )
    return NNTP_Invalid(client, cmd, gtime) ;

  if ( (gmt = NextArg(&args)) != NULL ) {
    if ( _stricmp(gmt, "gmt") != 0 )
      return NNTP_Invalid(client, cmd, gmt) ;
    if ( (distrib = NextArg(&args)) != NULL ) {
      char *rest ;
      if ( (rest = NextArg(&args)) != NULL )
	return NNTP_Invalid(client, cmd, rest) ;
    }
  }

  helpargs.client = client ;
  helpargs.mark = Mark() ;
  GetSystemTimeAsFileTime(&helpargs.now) ;

  if ( !DateTimeToFiletime(&helpargs.since, gdate, gtime) ) {
    if ( SendStream(client->stream, "501 Funny date %s %s\r\n", gdate, gtime) < 0 )
      return FALSE ;
    return TRUE ;
  }

  Log(LOG_STATUS, "Client asking for new news in %s\n", helpargs.match) ;

  /* return news since time */
  if ( SendStream(client->stream, "230 New news follows\r\n") < 0 )
    return FALSE ;

  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT, DEBUG_CLIENT_DATA) ;

  /* Not currently forwarding to server, just query local groups */
  ForAllGroups(NNTP_NewnewsGroup, &helpargs) ;

  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT_DATA, DEBUG_CLIENT) ;

  if ( SendStream(client->stream, ".\r\n") < 0 )
    return FALSE ;

  return TRUE ;
}

static int NNTP_Post(Client *client, int cmd, char *args)
{
  char *rest, last = '\0', lastlast = '\0' ;
  int result = FALSE ;
  Stream *dots = NULL, *posts = NULL ;
  char postname[MAX_PATH], summary[MAX_SUMMARY], reason[BUFSIZ] ;
  
  if ( (rest = NextArg(&args)) != NULL )
    return NNTP_Invalid(client, cmd, rest) ;

  if ( !Lock(POST_MUTEX, INFINITE) ) {
    (void)SendStream(client->stream, "503 Can't lock posting dir\r\n") ;
    return FALSE ;
  }

  if ( PostingName(postname, MAX_PATH) == NULL ||
       (posts = MakeFileStream(postname, FILESTREAM_WRITE, stream_error)) == NULL ) {
    if ( SendStream(client->stream, "503 Can't create posting file\r\n") >= 0 )
      result = TRUE ;
    goto post_exit ;
  }

  LogString(LOG_STATUS, "Client posting article\n") ;

  if ( SendStream(client->stream, "340 Have your say, <CRLF>.<CRLF> to finish\r\n") < 0 )
    goto post_rmexit ;

  /* Set terminator to <LF> to match <LF>.<CRLF> without preceding <CR> */
  if ( IoctlStream(client->stream, LINESTREAM_TERM, "\n") < 0 )
    goto post_rmexit ;

  if ( (dots = MakeDotStream(client->stream)) == NULL )
    goto post_rmexit ;

  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_CLIENT, DEBUG_POSTING) ;

  /* Read from client to <LF>.<CRLF>, store in posting file. */
  for (;;) {
    char *line ;
    long rbytes ;

    LogString(LOG_STATUS, "Client connected, waiting for command\n") ;

    if ( (rbytes = ReadStream(dots, &line, 0)) < 0 ) {
      (void)SendStream(client->stream, "503 Error reading posting\r\n") ;
      goto post_rmexit ;
    }

    if ( rbytes == 0 )
      break ;

    if ( !WriteStreamAll(posts, line, rbytes) ) {
      (void)SendStream(client->stream, "503 Error saving posting\r\n") ;
      goto post_rmexit ;
    }
  }

  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_POSTING, DEBUG_CLIENT) ;

  /* Reset stream termination */
  if ( IoctlStream(client->stream, LINESTREAM_TERM, "\r\n") < 0 )
    goto post_rmexit ;

  if ( posts )
    DestroyStream(posts) ;

  if ( dots )
    DestroyStream(dots) ;

  /* Try to contact appropriate server(s) */
  /* if contacted, post there and return response code */
  /* otherwise return 240 OK status */

  switch (PostFile(postname, summary, MAX_SUMMARY, reason, BUFSIZ)) {
  case 240:	/* Posting sent successfully */
    DeleteFile(postname) ;
    if ( SendStream(client->stream, "240 Posting accepted by server\r\n") >= 0 )
      result = TRUE ;
    break ;
  case 440:	/* Posting refused by server */
    message(EVENTLOG_INFORMATION_TYPE, MSG_POST_REJECTED, summary, reason,
	    NULL) ;
    {
      char errname[MAX_PATH] ;

      strcpy(errname, postname) ;
      strcat(errname, POST_FAIL_EXTN) ;

      if ( !MoveFile(postname, errname) )
	DeleteFile(postname) ;	/* Can't rename, so delete instead */

      if ( SendStream(client->stream, "440 Posting refused by server, saved as %s\r\n", errname) >= 0 )
	result = TRUE ;
    }
    break ;
  case 441:	/* Posting failed, try again */
  default:	/* Result indeterminate, or not connected */
    message(EVENTLOG_INFORMATION_TYPE, MSG_POST_QUEUED, summary,
	    NULL) ;
    if ( SendStream(client->stream, "240 Posting queued for server\r\n") >= 0 )
      result = TRUE ;
  }

  Unlock(POST_MUTEX) ;

  return result ;

post_rmexit:
  IoctlStream(client->stream, STREAM_DEBUG, DEBUG_POSTING, DEBUG_CLIENT) ;

  /* Reset stream termination */
  (void)IoctlStream(client->stream, LINESTREAM_TERM, "\r\n") ;

  if ( dots )
    DestroyStream(dots) ;

  if ( posts )
    DestroyStream(posts) ;

  DeleteFile(postname) ;

post_exit:
  Unlock(POST_MUTEX) ;

  return FALSE ;
}

static int NNTP_Quit(Client *client, int cmd, char *args)
{
  char *rest ;
  if ( (rest = NextArg(&args)) != NULL )
    return NNTP_Invalid(client, cmd, rest) ;
  
  /* say bye-bye */
  (void)SendStream(client->stream, "205 Goodbye and thanks for all the fish\r\n") ;

  return FALSE ;
}

static int ClientCommand(Client *client, char *command)
{
  int index ;

  static struct command_entry {
    char *name ;
    int namelen ;
    int cmd ;
    int (*fn)(Client *client, int cmd, char *args) ;
  } nntp_commands[] = {
    { "ARTICLE", 7, NNTP_ARTICLE, NNTP_Article},
    { "BODY", 4, NNTP_BODY, NNTP_Article},
    { "HEAD", 4, NNTP_HEAD, NNTP_Article},
    { "STAT", 4, NNTP_STAT, NNTP_Article},
    { "GROUP", 5, NNTP_GROUP, NNTP_Group},
    { "HELP", 4, NNTP_HELP, NNTP_Help},
    { "IHAVE", 5, NNTP_IHAVE, NNTP_Invalid},
    { "LAST", 4, NNTP_LAST, NNTP_Select},
    { "LIST", 4, NNTP_LIST, NNTP_List},
    { "NEWGROUPS", 9, NNTP_NEWGROUPS, NNTP_Newgroups},
    { "NEWNEWS", 7, NNTP_NEWGROUPS, NNTP_Newnews},
    { "NEXT", 4, NNTP_NEXT, NNTP_Select},
    { "POST", 4, NNTP_POST, NNTP_Post},
    { "QUIT", 4, NNTP_QUIT, NNTP_Quit},
    { "SLAVE", 5, NNTP_SLAVE, NNTP_Invalid},
  } ;

  for ( index = 0 ; index < sizeof(nntp_commands)/sizeof(struct command_entry) ; index++ ) {
    struct command_entry *this = &nntp_commands[index] ;
    int cmdlen = this->namelen ;
    if ( _strnicmp(command, this->name, cmdlen) == 0 &&
	 (command[cmdlen] == '\0' || isspace(command[cmdlen])) ) {
      char *args = command + cmdlen ;

      while (*args && isspace(*args))	/* search for whitespace end */
	args++ ;

      return (this->fn)(client, this->cmd, args) ;
    }
  }

  return NNTP_Invalid(client, NNTP_WHAT, command) ;
}

DWORD WINAPI ClientThread(LPVOID info)
{
  SOCKET clientsock = *((SOCKET *)info) ;
  unsigned long nonblock = 0 ;
  Stream *stream ;
  Client client ;

  ThreadType(CLIENT_THREAD) ;

  LogString(LOG_STATUS, "Client connected\n") ;

  if ( WSAEventSelect(clientsock, NULL, 0) == SOCKET_ERROR ) {
    message(EVENTLOG_ERROR_TYPE, MSG_CLEAREVENT,
	    ErrorText(WSAGetLastError()), NULL) ;
    return 1 ;
  }

  if ( ioctlsocket(clientsock, FIONBIO, &nonblock) == SOCKET_ERROR ) {
    message(EVENTLOG_ERROR_TYPE, MSG_NONBLOCK,
	    ErrorText(WSAGetLastError()), NULL) ;
    return 1 ;
  }
  
  if ( (stream = MakeSocketStream(clientsock, stream_error)) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_STREAM, ErrorText(GetLastError()), NULL) ;
    return 1 ;
  }

  if ( (client.stream = MakeLineStream(stream, "\r\n")) == NULL ) {
    DestroyStream(stream) ;
    message(EVENTLOG_ERROR_TYPE, MSG_STREAM, ErrorText(GetLastError()), NULL) ;
    return 1 ;
  }

  client.curr_groupname[0] = '\0' ;
  client.curr_artno = -1 ;

  IoctlStream(client.stream, STREAM_DEBUG, LINESTREAM_DEBUG_READ, DEBUG_CLIENT) ;
  IoctlStream(client.stream, STREAM_DEBUG, LINESTREAM_DEBUG_WRITE, DEBUG_CLIENT) ;

  /* Greet client, then serve commands */
  if ( SendStream(client.stream, "200 " SERVICE_DISPLAYNAME
		  " Version " VERSION_STRING " ready, posting OK\r\n") >= 0 ) {
    FILETIME when ;
    char *command ;
    long len = 0 ;

    message(EVENTLOG_INFORMATION_TYPE, MSG_CLIENT_CONNECT, NULL) ;

    while ( (len = ReadStream(client.stream, &command, 0)) > 0 ) {
      if ( len > 2 ) {
	command[len - 2] = '\0' ; /* cut off CRLF */

	if ( !ClientCommand(&client, command) )
	  break ;
      } else	/* didn't read enough, quit (should never happen) */
	break ;
    }

    if ( len >= 0 ) {    /* Client quit cleanly, note finish time */
      GetSystemTimeAsFileTime(&when) ;
      SetRegFiletime(SERVICE_REGISTRY_KEY, LAST_CLIENT, &when) ;
    }

    message(EVENTLOG_INFORMATION_TYPE, MSG_CLIENT_DISCONNECT, NULL) ;
  }
  
  stream = client.stream ;
  while ( stream != NULL )
    stream = DestroyStream(stream) ;

  LogString(LOG_STATUS, "Client disconnected\n") ;

  return 0 ;
}

/*
 * $Log: client.c,v $
 * Revision 1.3  2000/01/14 21:22:34  angus
 * More changes to logging. Use new streams 1.31 ioctl to specify debugging info
 * for client & server connections. Make fmtout more secure against buffer
 * overruns, and add a new format specifier for binary strings with visible
 * control characters. Change logging interface to add a varargs log, and use
 * that for stream debugging. Modify debug flags, add more status returns.
 *
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
