/* expire.c $Id: expire.c,v 1.3 2000/01/14 21:22:34 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997-1998.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Manage expiring of news for NNTP buffer
 */

#include "nntp.h"
#include "messages.h"

/* Max time limit to remember messages */
long remember_max = 0, remember_min = 0 ;
static FILETIME remember_max_time, remember_min_time ;

typedef struct {
  FILETIME now, before ;
  char *group ;
  long number ;
  DWORD low, high, articles ;
  DWORD interval ;
} ExpireArgs ;

int ExpireFile(char *filename, void *passthrough)
{
  ExpireArgs *args = passthrough ;
  char msgid[MAX_MESSAGEID] ;
  long artno = 0, remove = FALSE ;
  Group *gptr ;

  /* Check if should quit */
  switch ( WaitForMultipleObjects(ALL_CONTROLS, hEvents[SIGNAL_EXPIRE], FALSE, 0) ) {
  case WAIT_OBJECT_0 + CONTROL_START:
  case WAIT_TIMEOUT:
    break ;
  default: /* Abort on client connect, quit, stop, or error */
    return FALSE ;
  }

  if ( !Lock(GROUP_MUTEX, INFINITE) )
    return FALSE ;

  if ( !Lock(ARTICLE_MUTEX, INFINITE) ) {
    Unlock(GROUP_MUTEX) ;
    return FALSE ;
  }

  { /* Get article number */
    char *slash = strrchr(filename, '\\') ;
    artno = atoi(slash + 1) ;
  }

  if ( (gptr = GroupDBGet(args->group)) != NULL ) {
    if ( MessageID(args->group, artno, msgid, MAX_MESSAGEID) ) {
      Article *art = ArticleDBGet(msgid) ;

      if ( art != NULL ) {
	if ( CompareFileTime(&art->download, &args->before) <= 0 ) {
	  debug(DEBUG_EXPIRERM, "Removing article file %s, expired\n",
		filename) ;

	  remove = TRUE ;

	  if ( (art = RemoveArticleName(msgid, args->group, artno, art)) != NULL &&
	       MessageRemove(args->group, artno) )
	    if ( art->names[0] == '\0' ) {
	      if ( CompareFileTime(&art->download, &remember_min_time) > 0 ) {
		Article data = *art ;

		debug(DEBUG_EXPIRERM, "Marking article %s as expired\n",
		      msgid) ;
		data.state = ARTICLE_EXPIRED ;
		(void)ArticleDBPut(msgid, &data) ;
	      } else {
		debug(DEBUG_EXPIRERM, "Removing article %s\n", msgid) ;
		ArticleDBRemove(msgid) ;	/* No more names, remove article */
	      }
	    }
	} else { /* Keep article */
	  debug(DEBUG_EXPIRERM, "Keeping article file %s (%d)\n", filename,
		artno) ;

	  args->articles++ ;
	  if ( artno < args->low )
	    args->low = artno ;
	  if ( artno > args->high )
	    args->high = artno ;
	}
      } else {	/* Orphaned file */
	debug(DEBUG_EXPIRERM, "Removing article file %s, no article\n",
	      filename) ;
	remove = TRUE ;
      }
    } else { /* No message-id in file */
      debug(DEBUG_EXPIRERM, "Removing article file %s, no Message-ID\n",
	    filename) ;
      remove = TRUE ;
    }

    if ( remove ) {	/* Remove article from disc & group */
      Group data = *gptr ;

      args->number++ ;
      DeleteFile(filename) ;

      data.articles-- ;
      if ( artno == data.low )
	data.low++ ;
      (void)GroupDBPut(args->group, &data) ;
    }
  } else {	/* Orphaned group */
    debug(DEBUG_EXPIRERM, "Removing article file %s, no group\n", filename) ;
    DeleteFile(filename) ;
    args->number++ ;
  }

  Unlock(ARTICLE_MUTEX) ;
  Unlock(GROUP_MUTEX) ;

  Sleep(0) ; /* Yield control to other processes if possible; be nice */

  return TRUE ;
}

int ExpireGroup(char *group, void *passthrough)
{
  ExpireArgs *args = passthrough ;
  char dirname[MAX_PATH] ;
  int result = TRUE ;

  if ( !ArticleFileName(group, -1, dirname, MAX_PATH) )
    return FALSE ;

  Log(LOG_STATUS, "Expiring group %s", group) ;

  { /* Ignore group if directory does not exist */
    HANDLE dirh ;
    if ( (dirh = CreateFile(dirname,
			    0,
			    FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
			    NULL,
			    OPEN_EXISTING,
			    FILE_FLAG_BACKUP_SEMANTICS,
			    NULL)) == INVALID_HANDLE_VALUE )
      return TRUE ;
    CloseHandle(dirh) ;
  }

  if ( !Lock(EXPIRE_MUTEX, INFINITE) )
    return FALSE ;

  if ( Lock(GROUP_MUTEX, INFINITE) ) {
    Group *gptr = GroupDBGet(group) ;

    if ( gptr ) {
      long seconds = 0 ;
      Server *server ;
      Pattern *pattern ;

      debug(DEBUG_EXPIRETHREAD, "Expiring group %s (%d..%d -> %d)\n",
	    group, gptr->low, gptr->high, gptr->articles) ;

      if ( FindGroupServer(group, &server, &pattern) )
	seconds = pattern->normal * 60 * 60 * 24 ;

      if ( seconds < args->interval )	/* No point being less than this */
	seconds = args->interval ;

      args->group = group ;
      args->before = TimeAdd(&args->now, -seconds) ;
      args->low = MAX_ARTICLE_NO ;
      args->high = 0 ;
      args->articles = 0 ;
    } else
      result = FALSE ;

    Unlock(GROUP_MUTEX) ;
  }

  if ( result &&
       ForAllFiles(dirname, "", ExpireFile, args) &&
       Lock(GROUP_MUTEX, INFINITE) ) {
    Group *gptr ;

    if ( (gptr = GroupDBGet(group)) != NULL ) {
      Group data = *gptr ;

      if ( args->articles == 0 ) /* Renumber from zero if no articles */
	args->low = args->high = 0 ;

      if ( data.low != args->low ||
	   data.high != args->high ||
	   data.articles != args->articles ) {	/* Update group info */
	debug(DEBUG_EXPIRETHREAD, "Expired group %s (%d..%d -> %d)\n",
	      group, args->low, args->high, args->articles) ;
	data.low = args->low ;
	data.high = args->high ;
	data.articles = args->articles ;
	data.expired = args->now ;
	(void)GroupDBPut(group, &data) ;
      }
    } else
      result = FALSE ;
    Unlock(GROUP_MUTEX) ;
  } else
    result = FALSE ;

  Unlock(EXPIRE_MUTEX) ;

  Sleep(0) ; /* Yield control to other processes if possible; be nice */

  return result ;
}

/* Trawl for orphaned article entries and ones past sell-by date */
int ExpireArticle(char *msgid, void *passthrough)
{
  ExpireArgs *args = passthrough ;
  Article *art ;

  /* Check if should quit */
  switch ( WaitForMultipleObjects(ALL_CONTROLS, hEvents[SIGNAL_EXPIRE], FALSE, 0) ) {
  case WAIT_OBJECT_0 + CONTROL_START:
  case WAIT_TIMEOUT:
    break ;
  default: /* Abort on client connect, quit, stop, or error */
    return FALSE ;
  }

  if ( !Lock(ARTICLE_MUTEX, INFINITE) )
    return FALSE ;

  if ( (art = ArticleDBGet(msgid)) != NULL ) {
    if ( (art->state == ARTICLE_EXPIRED &&
	  CompareFileTime(&art->download, &remember_min_time) <= 0) ||
	 CompareFileTime(&art->download, &remember_max_time) <= 0 ) {
      char *name = art->names ;

      /* Remove all files for article */
      while ( *name ) {
	char filename[MAX_PATH] ;
	if ( MessageFileName(name, filename, MAX_PATH) != NULL )
	  (void)DeleteFile(filename) ;	/* Don't care about result */
	name += strlen(name) + 1 ;
      }

      ArticleDBRemove(msgid) ;	/* Remove article itself */
    }
  }

  Unlock(ARTICLE_MUTEX) ;

  return TRUE ;
}

/* Trawl for orphaned message entries */
int ExpireMessage(char *group, long artno, void *passthrough)
{
  char messageid[MAX_MESSAGEID] ;
  Article *art ;

  /* Check if should quit */
  switch ( WaitForMultipleObjects(ALL_CONTROLS, hEvents[SIGNAL_EXPIRE], FALSE, 0) ) {
  case WAIT_OBJECT_0 + CONTROL_START:
  case WAIT_TIMEOUT:
    break ;
  default: /* Abort on client connect, quit, stop, or error */
    return FALSE ;
  }

  if ( MessageID(group, artno, messageid, MAX_MESSAGEID) != NULL &&
       (art = ArticleDBGet(messageid)) == NULL )
    MessageRemove(group, artno) ;

  return TRUE ;
}

/* Expiry thread. On startup, this waits a minimum time in order to let
   the slurp and post threads to operate, and to give a client a chance to
   start up. Each expire test trawls the groups, looking at the articles in
   the group, removing them if downloaded before the current time minus the
   expiry for that group. Before running the guts of an expire, a lock is
   taken on EXPIRE_MUTEX, which other threads can take to prevent expires
   while they are operating. */
DWORD WINAPI ExpireThread(LPVOID info)
{
  DWORD wait, retry, interval, memo ;
  FILETIME then, now ;

  ThreadType(EXPIRE_THREAD) ;

  if ( (debuglevel & DEBUG_NOEXPIRE) != 0 )
    return 1 ;

  debug(DEBUG_EXPIRETHREAD, "Started expire thread\n") ;

  wait = GetRegDWORD(SERVICE_REGISTRY_KEY, EXPIRE_WAIT_NAME) ;
  if ( wait <= 0 )	/* Default if silly value */
    wait = EXPIRE_WAIT_DEFAULT ;

  retry = GetRegDWORD(SERVICE_REGISTRY_KEY, EXPIRE_RETRY_NAME) ;
  if ( retry <= 0 )	/* Default if silly value */
    retry = EXPIRE_RETRY_DEFAULT ;
  debug(DEBUG_EXPIRETHREAD, "Expire retry time is %d seconds\n", retry) ;

  interval = GetRegDWORD(SERVICE_REGISTRY_KEY, EXPIRE_INTERVAL_NAME) ;
  if ( interval <= 0 )	/* Default if stupid value */
    interval = EXPIRE_INTERVAL_DEFAULT ;
  else if ( interval <= EXPIRE_INTERVAL_MIN )	/* Default if silly value */
    interval = EXPIRE_INTERVAL_MIN ;
  debug(DEBUG_EXPIRETHREAD, "Expire interval time is %d seconds\n", interval) ;

  if ( GetRegFiletime(SERVICE_REGISTRY_KEY, LAST_EXPIRE, &then) ) {
    long diff ;

    GetSystemTimeAsFileTime(&now) ;
    diff = TimeDiff(&then, &now) ;
    if ( diff + wait < interval ) /* Take max of remaining interval & wait */
      wait = interval - diff ;
  }

  debug(DEBUG_EXPIRETHREAD, "Expire wait time is %d seconds\n", wait) ;

  memo = GetRegDWORD(SERVICE_REGISTRY_KEY, REMEMBER_MAX_NAME) ;
  if ( memo <= 0 )	/* Default if silly value */
    memo = REMEMBER_MAX_DEFAULT ;
  if ( remember_max <= memo )
    remember_max = memo ;

  debug(DEBUG_EXPIRETHREAD, "Max remember time is %d seconds\n", remember_max) ;

  remember_min = GetRegDWORD(SERVICE_REGISTRY_KEY, REMEMBER_MIN_NAME) ;
  if ( remember_min <= 0 )	/* Default if silly value */
    remember_min = REMEMBER_MIN_DEFAULT ;

  debug(DEBUG_EXPIRETHREAD, "Min remember time is %d seconds\n", remember_min) ;

  SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_LOWEST) ;

  for (;;) {
    DWORD delay = wait ;
    ExpireArgs args ;
    char status[20] ;	/* Status code for failure */

    GetSystemTimeAsFileTime(&now) ;
    then = TimeAdd(&now, delay) ;

    do {
      if ( suspended[SIGNAL_EXPIRE] ) {
        LogString(LOG_STATUS, "Expiry suspended\n") ;
      } else if ( FileTimeToLocalFileTime(&then, &now) ) {
        char buffer[80] ;
        Log(LOG_STATUS, "Next expire is at %t\n", &now) ;
      }

      switch ( WaitForMultipleObjects(LOOP_CONTROLS, hEvents[SIGNAL_EXPIRE],
                                      FALSE, delay * 1000) ) {
      case WAIT_TIMEOUT:
      case WAIT_OBJECT_0 + CONTROL_START:
        delay = 0 ;
        break ;
      case WAIT_OBJECT_0 + CONTROL_QUIT:
        goto expire_quit ;
      default:
        GetSystemTimeAsFileTime(&now) ;
        delay = TimeDiff(&now, &then) ;
        break ;
      }
    } while ( delay > 0 ) ;

    if ( suspended[SIGNAL_EXPIRE] )
      continue ;

    /* Get current time */
    args.number = 0 ;
    args.interval = interval ;
    GetSystemTimeAsFileTime(&args.now) ;
    remember_min_time = TimeAdd(&args.now, -remember_min) ;
    remember_max_time = TimeAdd(&args.now, -remember_max) ;

    message(EVENTLOG_INFORMATION_TYPE, MSG_EXPIRE_START, NULL) ;

    if ( ForAllGroups(ExpireGroup, &args) &&
	 ForAllArticles(ExpireArticle, &args) &&
	 ForAllMessages(ExpireMessage, &args) ) {
      GetSystemTimeAsFileTime(&now) ;
      SetRegFiletime(SERVICE_REGISTRY_KEY, LAST_EXPIRE, &now) ;
      wait = interval ;
      strcpy(status, "OK") ;
    } else {
      wait = retry ;
      sprintf(status, "%d", GetLastError()) ;
    }

    {
      char strnum[20] ;
      sprintf(strnum, "%d", args.number) ;
      message(EVENTLOG_INFORMATION_TYPE, MSG_EXPIRE_FINISH, strnum, status,
	      NULL) ;
    }
  }

expire_quit:
  LogString(LOG_STATUS, "Expire thread terminated\n") ;

  debug(DEBUG_EXPIRETHREAD, "Expire thread terminating\n") ;

  return 0 ;
}

/*
 * $Log: expire.c,v $
 * Revision 1.3  2000/01/14 21:22:34  angus
 * More changes to logging. Use new streams 1.31 ioctl to specify debugging info
 * for client & server connections. Make fmtout more secure against buffer
 * overruns, and add a new format specifier for binary strings with visible
 * control characters. Change logging interface to add a varargs log, and use
 * that for stream debugging. Modify debug flags, add more status returns.
 *
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
