/* group.c $Id: group.c,v 1.2 2000/01/14 03:40:04 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997-1998.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Manage group database for NNTP buffer
 */

#include "nntp.h"
#include "messages.h"

static DBM *group_db = NULL ;

int InitGroupDB(void)
{
  char filename[MAX_PATH] ;
  int result = TRUE ;

  /* create group mutex */
  if ( !Lock(GROUP_MUTEX, INFINITE) )
    return FALSE ;

  assert (confdir) ;

  strcpy(filename, confdir) ;
  strcat(filename, "\\") ;
  strcat(filename, GROUP_DB_NAME) ;

  debug(DEBUG_GROUPS, "Opening group DBM %s\n", filename) ;

  if ( (group_db = dbm_open(filename, O_RDWR|O_CREAT, 0644)) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_FILEOPEN, "database", filename,
	    ErrorText(GetLastError()), NULL) ;
    result = FALSE ;
  }

  Unlock(GROUP_MUTEX) ;

  return result ;
}

void CloseGroupDB(void)
{
  (void)Lock(GROUP_MUTEX, INFINITE) ; /* Plough on regardless of result */

  if ( group_db != NULL ) {
    debug(DEBUG_GROUPS, "Closing group DBM\n") ;
    dbm_close(group_db) ;
  }

  Unlock(GROUP_MUTEX) ;
}

Group *GroupDBGet(char *name)
{
  Group *result = NULL ;

  if (Lock(GROUP_MUTEX, INFINITE)) {
    datum groupkey, groupval ;

    groupkey.dptr = name ;
    groupkey.dsize = strlen(name) + 1 ; /* include \0 in key */

    /* Check group DB for group record */
    if ( groupkey.dsize < MAX_GROUPLEN ) {
      debug(DEBUG_GROUPS, "Get group %s (%d)\n", groupkey.dptr,
            groupkey.dsize) ;
      groupval = dbm_fetch(group_db, groupkey) ;
      debug(DEBUG_GROUPS, "Get group result=%s (%d)\n",
            groupval.dptr ? groupval.dptr : "NULL",
            groupval.dsize) ;
      result = (Group *)groupval.dptr ;
    }

    Unlock(GROUP_MUTEX) ;
  }

  return result ;
}

int GroupDBPut(char *name, Group *value)
{
  int result = FALSE ;

  if (Lock(GROUP_MUTEX, INFINITE)) {
    datum groupkey, groupval ;

    groupkey.dptr = name ;
    groupkey.dsize = strlen(name) + 1 ; /* include \0 in key */
    groupval.dptr = (char *)value ;
    groupval.dsize = sizeof(Group) ;

    debug(DEBUG_GROUPS, "PutGroup %s (%d) => %x (%d)\n", groupkey.dptr,
	  groupkey.dsize, groupval.dptr, groupval.dsize) ;

    /* Check group DB for group record */
    if ( groupkey.dsize < MAX_GROUPLEN &&
         dbm_store(group_db, groupkey, groupval, DBM_REPLACE) >= 0 )
      result = TRUE ;

    debug(DEBUG_GROUPS, "PutGroup result=%d\n", result) ;

    Unlock(GROUP_MUTEX) ;
  }

  return result ;
}

int GroupDBRemove(char *name)
{
  int result = FALSE ;

  if ( Lock(GROUP_MUTEX, INFINITE) ) {
    datum groupkey ;

    groupkey.dptr = name ;
    groupkey.dsize = strlen(name) + 1 ; /* include \0 in key */

    debug(DEBUG_GROUPS, "Removing group %s (%d)\n", groupkey.dptr,
          groupkey.dsize) ;
    if ( dbm_delete(group_db, groupkey) >= 0 )
      result = TRUE ;

    Unlock(GROUP_MUTEX) ;
  }

  return result ;
}

Group *NewGroup(char *group, GroupState istate, DWORD mark, int post, long days)
{
  Group newgroup ;

  newgroup.state = istate ;
  newgroup.low = 1 ;
  newgroup.high = 0 ;
  newgroup.articles = 0 ;
  newgroup.post = post ;
  newgroup.mark = mark ;
  GetSystemTimeAsFileTime(&newgroup.created) ;
  newgroup.expired = newgroup.created ;
  newgroup.newnews = TimeAdd(&newgroup.created, -days * 60 * 60 * 24) ;

  debug(DEBUG_GROUPS, "Adding group %s\n", group) ;

  if ( !GroupDBPut(group, &newgroup) )
    return NULL ;

  return GroupDBGet(group) ;	/* Now look it up and return pointer */
}

int ForAllGroups(int (*fn)(char *group, void *passthrough),
		 void *passthrough)
{
  int result = TRUE ;
  char *names = NULL ;

  static long allocsize = 0 ;	/* size of initial alloc for names */

  if ( Lock(GROUP_MUTEX, INFINITE) && (names = malloc(allocsize)) != NULL ) {
    datum groupkey  = dbm_firstkey(group_db) ;
    long used = 0 ;

    while ( groupkey.dptr ) {
      if ( names == NULL || used + groupkey.dsize >= allocsize ) {
	allocsize += 1024 ;
	if ( (names = realloc(names, allocsize)) == NULL )
	  break ;
      }

      strcpy(names + used, groupkey.dptr) ;
      used += groupkey.dsize ;
      
      groupkey = dbm_nextkey(group_db) ;
    }

    if ( names )
      names[used] = '\0' ;

    Unlock(GROUP_MUTEX) ; /* Allow other processes in */
  }

  if ( names ) {
    char *here = names ;	/* reset here pointer */

    while ( result && *here ) {
      result = (*fn)(here, passthrough) ; /* Callbacks is not locked */
      here += strlen(here) + 1 ;
    }

    free(names) ;
  }

  return result ;
}

void SetCurrentGroup(Client *client, char *name)
{
  debug(DEBUG_GROUPS, "Client set group %s\n", name) ;
  strcpy(client->curr_groupname, name) ;
}

Group *CurrentGroup(Client *client)
{
  if ( client->curr_groupname[0] == '\0' )
    return NULL ;

  debug(DEBUG_GROUPS, "Client get current group %s\n", client->curr_groupname) ;

  return GroupDBGet(client->curr_groupname) ;
}

char *CurrentGroupName(Client *client)
{
  if ( client->curr_groupname[0] == '\0' )
    return NULL ;

  return client->curr_groupname ;
}

/*
 * $Log: group.c,v $
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
