/* log.c $Id: log.c,v 1.2 2000/01/14 21:22:34 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 2000
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Logging to named pipe for debugging NNTP service for Windows NT.
 */

#include "nntp.h"
#include "messages.h"

#define LOG_PIPE_NAME ("\\\\.\\pipe\\" SERVICE_APPNAME "_log")

/* How many old entries to keep when not writing to log */
#define KEEP_LOG 100
static char *keep[KEEP_LOG] ;
static int keep_start = 0, keep_end = 0 ;

static DWORD threads[N_THREAD_TYPES] ;
static char *status[N_THREAD_TYPES + 1] ;
static char *thread_prefix[N_THREAD_TYPES + 1] = {
  "[main]   ", 
  "[log]    ",
  "[post]   ",
  "[slurp]  ",
  "[expire] ",
  "[client] ",
  "[unknown]",
} ;
static int thread_prefix_len[N_THREAD_TYPES + 1] = {
  9, 9, 9, 9, 9, 9, 9,
} ;

static HANDLE logpipe = NULL, logevent = NULL ;
static BOOL connected = FALSE ;

static CRITICAL_SECTION loglock ;

static int PostQueue(char *filename, void *passthrough)
{
  char *format = passthrough ;
  char summary[MAX_SUMMARY] ;

  if ( Summary(filename, summary, MAX_SUMMARY) )
    Log(LOG_QUEUE, format, summary) ;
  else
    Log(LOG_QUEUE, format, filename) ;

  return TRUE ;
}

/* This thread is responsible for managing a named pipe connection
   for debugging and status information. It creates the pipe, and
   waits for a connection. It then reads an integer debug value
   from the pipe. If this is zero, the last status reported by
   each thread is sent to the pipe. If it is not zero, the statuses
   are sent and then the connection is maintained, and all new
   debug and status messages matching the debug mask are sent to the
   pipe. */
DWORD WINAPI LogThread(LPVOID info)
{
  SECURITY_DESCRIPTOR sdesc ;
  SECURITY_ATTRIBUTES sattr = { sizeof(SECURITY_ATTRIBUTES), (LPVOID)&sdesc, FALSE } ;

  ThreadType(LOG_THREAD) ;

  InitializeCriticalSection(&loglock) ;

  if ( (logevent = CreateEvent(NULL, TRUE, FALSE, NULL)) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_CREATEEVENT,
	    ErrorText(GetLastError()), NULL) ;
    return 1 ;
  }

  /* Set discretionary ACL to grant object access */
  if ( InitializeSecurityDescriptor(&sdesc, SECURITY_DESCRIPTOR_REVISION) ) {
    DWORD len = GetSidLengthRequired(1) ;
    PSID sid ;

    if ( (sid = (PSID)malloc(len)) != NULL ) {
      PACL acl ;

      len += sizeof(ACL) + sizeof(ACCESS_ALLOWED_ACE) - sizeof (DWORD) ;

      if ( (acl = (PACL)malloc(len)) != NULL ) {      
        SID_IDENTIFIER_AUTHORITY siaWorld = SECURITY_WORLD_SID_AUTHORITY ;
        if ( InitializeAcl(acl, len, ACL_REVISION) &&
             InitializeSid(sid, &siaWorld, 1) ) {
          *(GetSidSubAuthority(sid, 0)) = SECURITY_WORLD_RID ;

          if ( AddAccessAllowedAce(acl, ACL_REVISION,
                                   GENERIC_ALL|STANDARD_RIGHTS_ALL|SPECIFIC_RIGHTS_ALL,
                                   sid) &&
               SetSecurityDescriptorDacl(&sdesc, TRUE, acl, FALSE) ) {
            logpipe = CreateNamedPipe(LOG_PIPE_NAME,
                                      PIPE_ACCESS_DUPLEX|FILE_FLAG_WRITE_THROUGH, /* synchronous for now */
                                      PIPE_TYPE_MESSAGE|PIPE_READMODE_MESSAGE|PIPE_WAIT,
                                      1, /* Only one pipe instance allowed */
                                      BUFSIZ, /* output buffer size, in bytes */
                                      100,    /* input buffer size, in bytes */
                                      100,    /* 0.1s time-out */
                                      &sattr) ; /* Security attributes */
          }
        }
        free(acl) ;
      }
      free(sid) ;
    }
  }

  if ( logpipe == INVALID_HANDLE_VALUE ) { /* Failed to create named pipe */
    message(EVENTLOG_ERROR_TYPE, MSG_CREATEPIPE, LOG_PIPE_NAME,
	    ErrorText(GetLastError()), NULL) ;
    return 1 ;
  }

  /* Wait for connection on named pipe. NYI shutdown request. */
  for (;;) {
    if ( ConnectNamedPipe(logpipe, NULL) ) {
      char buffer[15] ;
      DWORD n ;

      if ( ReadFile(logpipe, buffer, 15, &n, NULL) )
        debuglevel = strtol(buffer, NULL, 0) ;

      connected = TRUE ;

      /* Print existing status to pipe */
      EnterCriticalSection(&loglock) ;
      if ( Lock(POST_MUTEX, INFINITE) ) {
        ForAllFiles(postdir, "", PostQueue, "Queued article %s\n") ;
        ForAllFiles(postdir, POST_FAIL_EXTN, PostQueue, "Failed article %s\n") ;
        Unlock(POST_MUTEX) ;
      } else
        LogString(LOG_QUEUE, "Error, can't read NNTP queue\n") ;

      /* Output status */
      for ( n = 0 ; n < N_THREAD_TYPES ; ++n ) {
        if ( status[n] != NULL )
          LogString(LOG_QUEUE, status[n]) ;
      }

      /* Output most recent messages */
      while ( keep_start != keep_end ) {
        if ( keep[keep_start] ) {
          LogString(LOG_QUEUE, keep[keep_start]) ;
          keep[keep_start] = NULL ;
        }
        keep_start = (keep_start + 1) % KEEP_LOG ;
      }

      LeaveCriticalSection(&loglock) ;

      /* Wait for client to disconnect if debugging */
      if ( debuglevel != 0 )
        WaitForSingleObject(logevent, INFINITE) ;

      ResetEvent(logevent) ;
      connected = FALSE ;

      DisconnectNamedPipe(logpipe) ;
    }
  }

  CloseHandle(logevent) ;
  CloseHandle(logpipe) ;

  DeleteCriticalSection(&loglock) ;

  return 0 ;
}

/* Function called by all threads to access named pipe. Should use
   critical section to interleave access. */
void LogString(int type, char *str)
{
  DWORD tid = GetCurrentThreadId() ;
  int tt, len = strlen(str), plen = 0 ;
  char *prefix = "" ;

  for ( tt = 0 ; tt < N_THREAD_TYPES ; ++tt )
    if ( threads[tt] == tid )
      break ;

  /* logevent is used as sanity check that we have set up critical section.
     Time between initialising critical section and creating event is
     irrelevant, just ignore messages. */
  if ( type == LOG_STATUS && logevent ) { /* Save status for each thread type */
    EnterCriticalSection(&loglock) ;

    if ( status[tt] != NULL )
      free(status[tt]) ;

    if ( (status[tt] = malloc(len + 1)) != NULL )
      strcpy(status[tt], str) ;

    LeaveCriticalSection(&loglock) ;
  } else if ( type == LOG_DEBUG ) {
    prefix = thread_prefix[tt] ;
    plen = thread_prefix_len[tt] ;
  }
  
  if ( connected && logpipe && logevent ) {
    if ( !WriteFile(logpipe, prefix, plen, &plen, NULL) ||
         !WriteFile(logpipe, str, len, &len, NULL) ) {
      SetEvent(logevent) ;
      connected = FALSE ;
    }
  } else if ( !is_service ) {
    fputs(str, stderr) ;
    fflush(stderr) ;
  } else if ( logevent && type == LOG_DEBUG &&
              (debuglevel & DEBUG_KEPT) != 0 ) {
    char *currend ; /* After the fact debugging */

    EnterCriticalSection(&loglock) ;
    if ( (currend = keep[keep_end]) != NULL )
      free(currend) ;

    if ( (currend = keep[keep_end] = malloc(len + 1 + plen)) != NULL ) {
      strcpy(currend, prefix) ;
      strcpy(currend + plen, str) ;
    }

    keep_end = (keep_end + 1) % KEEP_LOG ;
    if ( keep_end == keep_start ) { /* Oops, caught my own tail. Make space. */
      if ( keep[keep_start] )
        free(keep[keep_start]) ;

      keep[keep_start] = NULL ;
      keep_start = (keep_start + 1) % KEEP_LOG ;
    }
    LeaveCriticalSection(&loglock) ;
  }
}

/* Thread name for log */
void ThreadType(int type)
{
  threads[type] = GetCurrentThreadId() ;
}

static BOOL WINAPI LogControlHandler(DWORD dwCtrlType)
{
  switch( dwCtrlType ) {
  case CTRL_BREAK_EVENT:  /* use Ctrl+C or Ctrl+Break to simulate */
  case CTRL_C_EVENT:      /* SERVICE_CONTROL_STOP in debug mode */
    CloseHandle(logpipe) ;
    return TRUE;
  }
  return FALSE;
}

/* Corresponding client side of named pipe */
BOOL LogClient(int debuglevel)
{
  char buffer[BUFSIZ] ;
  DWORD bytes, wlen ;

  wlen = sprintf(buffer, "%d", debuglevel) + 1 ;

  SetConsoleCtrlHandler(LogControlHandler, TRUE) ;

  if ( (logpipe = CreateFile(LOG_PIPE_NAME,
                             GENERIC_READ|GENERIC_WRITE,
                             0, 
                             NULL,
                             OPEN_EXISTING,
                             FILE_ATTRIBUTE_NORMAL,
                             NULL)) == INVALID_HANDLE_VALUE )
    return FALSE ;

  if ( WriteFile(logpipe, buffer, wlen, &bytes, NULL) && wlen == bytes ) {
    while ( ReadFile(logpipe, buffer, BUFSIZ, &bytes, NULL) ) {
      buffer[bytes] = '\0' ;
      fputs(buffer, stdout) ;
    }
  }
  CloseHandle(logpipe) ;

  return TRUE ;
}

/*
$Log: log.c,v $
Revision 1.2  2000/01/14 21:22:34  angus
More changes to logging. Use new streams 1.31 ioctl to specify debugging info
for client & server connections. Make fmtout more secure against buffer
overruns, and add a new format specifier for binary strings with visible
control characters. Change logging interface to add a varargs log, and use
that for stream debugging. Modify debug flags, add more status returns.

Revision 1.1  2000/01/14 03:40:04  angus
Many changes; added named pipe logging, sending output to client debug
process for in-situ service debugging. Use this mechanism for queue status,
and enable after-the-fact debugging by saving 100 logged messages. Add
symbollic names for debug flags, and Mutex debugging.

*/
