/* message.c $Id: message.c,v 1.2 2000/01/14 03:40:04 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Manage message database for NNTP buffer
 */

#include "nntp.h"
#include "messages.h"

static DBM *message_db = NULL ;

int InitMessageDB(void)
{
  char filename[MAX_PATH] ;
  int result = TRUE ;

  if ( !Lock(MESSAGE_MUTEX, INFINITE) )
    return FALSE ;

  assert (confdir) ;

  strcpy(filename, confdir) ;
  strcat(filename, "\\") ;
  strcat(filename, MESSAGE_DB_NAME) ;

  debug(DEBUG_MESSAGES, "Opening message DBM %s\n", filename) ;

  if ( (message_db = dbm_open(filename, O_RDWR|O_CREAT, 0644)) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_FILEOPEN, "database", filename,
	    ErrorText(GetLastError()), NULL) ;
    result = FALSE ;
  }

  Unlock(MESSAGE_MUTEX) ;

  return result ;
}

void CloseMessageDB(void)
{
  (void)Lock(MESSAGE_MUTEX, INFINITE) ;	/* Plough on regardless of result */

  if ( message_db != NULL ) {
    debug(DEBUG_MESSAGES, "Closing message DBM\n") ;
    dbm_close(message_db) ;
  }

  Unlock(MESSAGE_MUTEX) ;
}

/* You must have the mutex in order to do this */
char *MessageID(char *group, long artno, char *buffer, int bufsize)
{
  char namenum[MAX_PATH] ;
  char *result = NULL ;

  sprintf(namenum, "%s:%d", group, artno) ;

  if ( Lock(MESSAGE_MUTEX, INFINITE) ) {
    datum messagekey, messageval ;
    Message *message = NULL ;

    messagekey.dptr = namenum ;
    messagekey.dsize = strlen(namenum) + 1 ; /* include \0 in key */

    /* Check message DB for message record */
    debug(DEBUG_MESSAGES, "MessageID %s (%d)\n", messagekey.dptr,
          messagekey.dsize) ;
    messageval = dbm_fetch(message_db, messagekey) ;

    debug(DEBUG_MESSAGES, "MessageID result=%s (%d)\n",
          messageval.dptr ? messageval.dptr : "NULL",
          messageval.dsize) ;

    if ( (message = (Message *)messageval.dptr) != NULL &&
	 strlen(message->messageid) < bufsize ) {
      strcpy(buffer, message->messageid) ;
      result = buffer ;
    }

    Unlock(MESSAGE_MUTEX) ;
  }

  return result ;
}

/* You must have the mutex in order to do this */
int MessageAdd(char *group, long artno, char *messageid)
{
  char namenum[MAX_PATH] ;
  Message *message = (Message *)malloc(strlen(messageid) + sizeof(Message)) ;
  int result = FALSE ;

  if ( message == NULL )
    return FALSE ;

  sprintf(namenum, "%s:%d", group, artno) ;

  strcpy(message->messageid, messageid) ; /* For struct alignment */

  debug(DEBUG_MESSAGES, "Adding message %s %s\n", namenum, messageid) ;

  if ( Lock(MESSAGE_MUTEX, INFINITE) ) {
    datum messagekey, messageval ;

    messagekey.dptr = namenum ;
    messagekey.dsize = strlen(namenum) + 1 ; /* include \0 in key */
    messageval.dptr = (char *)message ;
    messageval.dsize = strlen(message->messageid) + sizeof(Message) ;

    /* Check message DB for message record */
    if ( dbm_store(message_db, messagekey, messageval, DBM_REPLACE) >= 0 )
      result = TRUE ;

    Unlock(MESSAGE_MUTEX) ;
  }

  free(message) ;

  return result ;
}

/* You must have message mutexes in order to do this */
int MessageRemove(char *group, long artno)
{
  char namenum[MAX_PATH] ;
  int result = FALSE ;

  sprintf(namenum, "%s:%d", namenum, artno) ;

  if ( Lock(MESSAGE_MUTEX, INFINITE) ) {
    datum messagekey ;

    messagekey.dptr = namenum ;
    messagekey.dsize = strlen(namenum) + 1 ; /* include \0 in key */

    debug(DEBUG_MESSAGES, "Removing message %s\n", namenum) ;

    if ( dbm_delete(message_db, messagekey) >= 0 )
      result = TRUE ;

    Unlock(MESSAGE_MUTEX) ;
  }

  return result ;
}

int ForAllMessages(int (*fn)(char *group, long artno, void *passthrough),
		   void *passthrough)
{
  int result = TRUE ;
  char *names = NULL ;

  static long allocsize = 0 ;	/* size of initial alloc for names */

  if ( Lock(MESSAGE_MUTEX, INFINITE) && (names = malloc(allocsize)) != NULL ) {
    datum artkey = dbm_firstkey(message_db) ;
    long used = 0 ;

    while ( artkey.dptr ) {
      if ( names == NULL || used + artkey.dsize >= allocsize ) {
	allocsize += 4096 ;
	if ( (names = realloc(names, allocsize)) == NULL )
	  break ;
      }

      strcpy(names + used, artkey.dptr) ;
      used += artkey.dsize ;
      
      artkey = dbm_nextkey(message_db) ;
    }

    if ( names )
      names[used] = '\0' ;

    Unlock(MESSAGE_MUTEX) ;
  }

  if ( names ) {
    char *here = names ;	/* reset here pointer */

    while ( result && *here ) {
      char *group = here, *colon = strchr(group, ':') ;
      here += strlen(here) + 1 ;

      if ( colon ) {
	*colon = '\0' ;
	result = (*fn)(group, atoi(colon + 1), passthrough) ; /* Callback is not locked */
      }
    }

    free(names) ;
  }

  return result ;
}

char *MessageFileName(char *namenum, char *buffer, int bufsize)
{
  char *colon = strrchr(namenum, ':') ;
  char *result = NULL ;

  if ( colon ) {
    char group[MAX_GROUPLEN] ;
    int len = colon - namenum ;
    long artno = atoi(colon + 1) ;

    strncpy(group, namenum, len) ;
    group[len] = '\0' ;

    result = ArticleFileName(group, artno, buffer, bufsize) ;
  }

  return result ;
}

/*
 * $Log: message.c,v $
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
