/* nntp.c $Id: nntp.c,v 1.3 2000/01/14 21:22:34 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * NNTP outgoing buffer for NT. Implements dynamic news server for single-user
 * use, feeding to and from multiple news servers.
 */

#include "nntp.h"
#include "messages.h"

char *program ;
int debuglevel ;

char *confdir = NULL ;
char *spooldir = NULL ;
char *postdir = NULL ;
unsigned short inport = 0, outport = 0 ;

int code = 0 ;

static void usage(void)
{
  message(EVENTLOG_INFORMATION_TYPE, MSG_USAGE, SERVICE_DISPLAYNAME,
	  VERSION_STRING, SERVICE_NAME, NULL) ;

  ExitProcess(0) ;
}

static int PostKill(char *filename, void *passthrough)
{
  char *arg = passthrough ;
  char msgid[MAX_MESSAGEID] ;
  char *tail = strrchr(filename, '\\') ; /* tail end of filename */

  if ( strcmp(arg, tail + 1) == 0 || /* Same message number */
       (MessageIDFile(filename, msgid, MAX_MESSAGEID) && Match(arg, msgid)) ) {
    char summary[MAX_SUMMARY] ;

    if ( Summary(filename, summary, MAX_SUMMARY) )
      printf("Dequeing posting %s\n", summary) ;
    else
      printf("Dequeing posting %s\n", filename) ;
    (void)DeleteFile(filename) ;
  }

  return TRUE ;
}

static int KillGroupFiles(char *filename, void *passthrough)
{
  return DeleteFile(filename) ;
}

BOOL debug_flags(char *arg, int *debug)
{
  char *next ;

  do {
    static struct {
      char *name ;
      int value ;
    } table[] = {
      { "keep",         0x0000001 },
      { "mutex",        0x0000002 },
      { "registry",     0x0000004 },
      { "client",       0x0000008 },
      { "client-data",  0x0000010 },
      { "server",       0x0000020 },
      { "group",        0x0000040 },
      { "article",      0x0000080 },
      { "socket",       0x0000100 },
      { "file",         0x0000200 },
      { "read",         0x0000400 },
      { "write",        0x0000800 },
      { "noslurp",      0x0001000 },
      { "slurp-info",   0x0002000 },
      { "slurp-data",   0x0004000 },
      { "slurp",        0x0006000 },
      { "nopost",       0x0010000 },
      { "post-info",    0x0020000 },
      { "post-data",    0x0040000 },
      { "post",         0x0060000 },
      { "noexpire",     0x0100000 },
      { "expire-info",  0x0200000 },
      { "expire-data",  0x0400000 },
      { "expire",       0x0600000 },
      { "msgid",        0x1000000 },
      { "all",          -1 },
    } ;
    int i ;

    if ( (next = strchr(arg, ',')) != NULL )
      *next++ = '\0' ;

    for ( i = 0 ; i < sizeof(table)/sizeof(table[0]) ; ++i ) {
      if ( strcmp(table[i].name, arg) == 0 ) {
        *debug |= table[i].value ;
        break ;
      }
    }
    if ( i >= sizeof(table)/sizeof(table[0]) )
      return FALSE ;
  } while ( (arg = next) != NULL ) ;

  return TRUE ;
}

int parse_args(int argc, char *argv[])
{
  int action = ACTION_START ;
  int pathlen ;
  char path[MAX_PATH], *arg ;

  static int control_table[3][3] = {
    { NNTP_CONTROL_START_POST, NNTP_CONTROL_ABORT_POST, NNTP_CONTROL_STOP_POST, },
    { NNTP_CONTROL_START_SLURP, NNTP_CONTROL_ABORT_SLURP, NNTP_CONTROL_STOP_SLURP, },
    { NNTP_CONTROL_START_EXPIRE, NNTP_CONTROL_ABORT_EXPIRE, NNTP_CONTROL_STOP_EXPIRE, },
  } ;

  pathlen = GetModuleFileName(NULL, path, MAX_PATH) ;
  if ( pathlen ) {
    char *slash = strrchr(path, '\\') ;

    if ( slash ) {
      pathlen = slash - path ;
      *slash = '\0' ;
    }
  }

  for ( program = *argv++ ; --argc ; argv++ ) {
    if (argv[0][0] == '-') {
      int index = 0 ;
      char ch = argv[0][1] ;

      /* Get argument if necessary */
      if ( strchr("YNAIKUESPdsopcD", ch) != NULL ) {
	if ( argv[0][2] != '\0' )
	  arg = &argv[0][2] ;
	else if ( --argc )
	  arg = *++argv ;
	else
	  return ACTION_USAGE ;
      } else if ( argv[0][2] != '\0' )
        return ACTION_USAGE ;

      switch (ch) {
      case 'i':	/* install service */
	action = ACTION_INSTALL ;
	break;
      case 'r':	/* remove service */
	action = ACTION_UNINSTALL ;
	break;
      case 'q':	/* report on queue */
        if ( !DefaultDirectories(path, pathlen) )
          return ACTION_USAGE ;

        /* Create named pipe instance to read from */
        if ( !LogClient(debuglevel) ) {
          printf("Error, can't read NNTP queue, is server started?\n") ;
        }

	action = ACTION_DONE ;
	break;
      case 'Y':	/* permit posting */
      case 'N':	/* deny posting */
      case 'A':	/* activate group */
      case 'I':	/* inactivate group */
        if ( !DefaultDirectories(path, pathlen) )
          return ACTION_USAGE ;

        if ( Lock(GROUP_MUTEX, INFINITE) ) {
          if ( InitGroupDB() ) {
            Group *gptr = GroupDBGet(arg) ;

            if ( !gptr ) { /* Create new group */
              Server *server ;
              Pattern *patt ;
              
              if ( FindGroupServer(arg, &server, &patt) )
                gptr = NewGroup(arg, GROUP_ACTIVE, 0, 'y', patt->normal) ;
            }

            if ( gptr ) {
              Group gdata = *gptr ;

              if ( ch == 'Y' || ch == 'N' )
                gdata.post = tolower(ch) ;
              else if ( ch == 'A' )
                gdata.state = GROUP_ACTIVE ;
              else if ( ch == 'I' )
                gdata.state = GROUP_INACTIVE ;

              if ( GroupDBPut(arg, &gdata) ) {
                switch (gdata.state) {
                case GROUP_ACTIVE:
                  printf("  (ACTIVE)") ; break ;
                case GROUP_INACTIVE:
                  printf("  (INACTIVE)") ; break ;
                case GROUP_BOGUS:
                  printf("  (BOGUS)") ; break ;
                case GROUP_NEW:
                  printf("  (NEW)") ; break ;
                default:
                  printf("  (CORRUPT)") ; break ;
                }
                printf("  %d-%d (%d)", gdata.low, gdata.high, gdata.articles) ;
                printf("  Post %c", gdata.post) ;
                printf("  Mark %d\n", gdata.mark) ;
                printf("  Created %s", TimeAsString(&gdata.created)) ;
                printf("  Newnews %s\n", TimeAsString(&gdata.newnews)) ;
                printf("  Expired %s", TimeAsString(&gdata.expired)) ;
                printf("  Selected %s\n", TimeAsString(&gdata.selected)) ;
              } else
                printf("Error, can't change data for group %s\n", arg) ;
            } else
              printf("Error, can't create group %s\n", arg) ;
            
            CloseGroupDB() ;
          } else
            printf("Error, can't access group database") ;
          Unlock(GROUP_MUTEX) ;
        } else
          printf("Error, can't lock group database") ;
        action = ACTION_DONE ;
        break ;
      case 'U':	/* unsubscribe (delete) group */
        if ( !DefaultDirectories(path, pathlen) )
          return ACTION_USAGE ;

        if ( Lock(GROUP_MUTEX, INFINITE) ) {
          if ( InitGroupDB() ) {
            Group *gptr = GroupDBGet(arg) ;

            if ( gptr ) {
              char dirname[MAX_PATH] ;

              if ( ArticleFileName(arg, -1, dirname, MAX_PATH) ) {
                if ( !ForAllFiles(dirname, "", KillGroupFiles, NULL) ||
                     !RemoveDirectory(dirname) )
                  printf("Error, can't remove directory for group %s\n", arg) ;
                if ( !GroupDBRemove(arg) )
                  printf("Error, can't remove entry for group %s\n", arg) ;
              } else
                printf("Error, can't get directory for group %s\n", arg) ;
              
            } else
              printf("No group %s to remove\n", arg) ;
              
            CloseGroupDB() ;
          } else
            printf("Error, can't access group database") ;
          Unlock(GROUP_MUTEX) ;
        } else
          printf("Error, can't lock group database") ;
        action = ACTION_DONE ;
        break ;
      case 'K':	/* kill posting */
        if ( !DefaultDirectories(path, pathlen) )
          return ACTION_USAGE ;

        if ( Lock(POST_MUTEX, INFINITE) ) {
          ForAllFiles(postdir, "", PostKill, arg) ;
          ForAllFiles(postdir, POST_FAIL_EXTN, PostKill, arg) ;
          Unlock(POST_MUTEX) ;
        } else
          printf("Error, can't read NNTP queue\n") ;
        action = ACTION_DONE ;
        break ;
      case 'd':	/* set config directory */
        confdir = arg ;
        break ;
      case 's':	/* set spool directory */
        spooldir = arg ;
        break ;
      case 'o':	/* set outgoing directory */
        postdir = arg ;
        break ;
      case 'c':	/* client port number */
        inport = htons((unsigned short)atoi(arg)) ;
        break ;
      case 'p':	/* server port number */
        outport = htons((unsigned short)atoi(arg)) ;
        break ;
      case 'D':	/* debug level service */
        if ( isdigit(*arg) )
          debuglevel |= strtol(arg, &arg, 0) ;
        else if ( !debug_flags(arg, &debuglevel) )
          return ACTION_USAGE ;
        if ( debuglevel == 0 )
          return ACTION_USAGE ;
        action = ACTION_DEBUG ;
        break ;
      case 'x': /* Stop service */
        code = SERVICE_CONTROL_STOP ;
        action = ACTION_CONTROL ;
        break ;
      case 'E':	/* expire control */
        ++index ; /* FALLTHRU */
      case 'S':	/* slurp control */
        ++index ; /* FALLTHRU */
      case 'P':	/* posting control */
        if ( strcmp(arg, "start") == 0 ) {
          code = control_table[index][0] ;
        } else if ( strcmp(arg, "abort") == 0 ) {
          code = control_table[index][1] ;
        } else if ( strcmp(arg, "stop") == 0 ) {
          code = control_table[index][2] ;
        } else
          return ACTION_USAGE ;
        action = ACTION_CONTROL ;
        break ;
      default:
        return ACTION_USAGE ;
      }
    } else
      return ACTION_USAGE ;
  }

  if ( program == NULL )
    return ACTION_USAGE ;

  if ( !DefaultDirectories(path, pathlen) )
    return ACTION_USAGE ;

  return action ;
}

int main(int argc, char *argv[])
{
  SERVICE_TABLE_ENTRY dispatchTable[] = {
    { TEXT(SERVICE_NAME), (LPSERVICE_MAIN_FUNCTION)service_main },
    { NULL, NULL }
  } ;

  switch ( parse_args(argc, argv) ) {
  case ACTION_DONE:
    break ;
  case ACTION_CONTROL:
    if ( !UserService(code) )
      return 1 ;
    break ;
  case ACTION_INSTALL:
    if ( !InstallService() )
      return 1 ;
    break ;
  case ACTION_UNINSTALL:
    if ( !RemoveService() )
      return 2 ;
    break ;
  case ACTION_START:
    if ( !StartServiceCtrlDispatcher(dispatchTable) ) {
      if ( GetLastError() != ERROR_FAILED_SERVICE_CONTROLLER_CONNECT ||
           !RestartService() ) {
        message(EVENTLOG_ERROR_TYPE, MSG_STARTFAILED,
                ErrorText(GetLastError()), NULL) ;
        return 3 ;
      }
    }
    break ;
  case ACTION_DEBUG:
    SetConsoleCtrlHandler(ControlHandler, TRUE) ;
    service_main((DWORD)argc, (LPTSTR *)argv) ;
    break ;
  case ACTION_USAGE:
    usage() ;
  }

  return 0 ;
}

/*
 * $Log: nntp.c,v $
 * Revision 1.3  2000/01/14 21:22:34  angus
 * More changes to logging. Use new streams 1.31 ioctl to specify debugging info
 * for client & server connections. Make fmtout more secure against buffer
 * overruns, and add a new format specifier for binary strings with visible
 * control characters. Change logging interface to add a varargs log, and use
 * that for stream debugging. Modify debug flags, add more status returns.
 *
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
