#ifndef NNTP_H
#define NNTP_H

/* nntp.h $Id: nntp.h,v 1.4 2000/01/14 22:17:55 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997-2000.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Common definitions for NNTP buffer service for Windows NT
 *
 */

#include <winsock2.h>
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>

/* Include my Streams library */
#include <streams.h>

/***** generally-accessible variables *****/
extern char *program, *confdir, *spooldir, *postdir ;
extern int debuglevel ;
extern unsigned short inport, outport ;

#define SERVICE_APPNAME "NNTP"
#define SERVICE_NAME "NNTPBuffer"
#define SERVICE_DISPLAYNAME "NNTP Buffer Service"
#define SERVICE_DEPENDENCIES "EventLog\0\0" /* "dep1\0\dep2\0\0" */
#define SERVICE_REGISTRY_KEY ("System\\CurrentControlSet\\Services\\" SERVICE_NAME)
#define EVENTLOG_REGISTRY_KEY ("System\\CurrentControlSet\\Services\\EventLog\\Application\\" SERVICE_NAME)

#define VERSION_STRING "1.11"

/* directory names and registry entries */
#define POST_REGNAME "PostingDir"
#define POST_DIRNAME "\\outgoing"
#define SPOOL_REGNAME "SpoolDir"
#define SPOOL_DIRNAME "\\spool"
#define CONFIG_REGNAME "ConfigDir"
#define CONFIG_DIRNAME "\\config"

#define NEWSFEEDS "newsfeeds"

/* Default timeouts. The post, slurp and expiry threads have three time periods
   associated with them. The ...Wait period is the minimum time the thread
   waits before trying any action. The ...Retry period is the minimum time the
   thread waits between failed actions, and the ...Interval is the minimum
   time the thread waits between successful actions. */
#define CONNECT_TIMEOUT (5) /* 5 seconds allowed to make connection */
#define CONNECT_TIMEOUT_NAME "ConnectTimeout"
#define RECONNECT_TIMEOUT (60 * 5) /* 5 minutes wait before re-connecting */
#define RECONNECT_TIMEOUT_NAME "ReconnectTimeout"
#define DISCONNECT_TIMEOUT (60 * 10) /* Disconnect servers if idle 10 mins */
#define DISCONNECT_TIMEOUT_NAME "DisconnectTimeout"
#define EXPIRE_WAIT_NAME "ExpireWait"
#define EXPIRE_WAIT_DEFAULT (60 * 5) /* 5 minutes wait before first expire */
#define EXPIRE_RETRY_NAME "ExpireRetry"
#define EXPIRE_RETRY_DEFAULT (60 * 5) /* Not sure what a failed expire is */
#define EXPIRE_INTERVAL_NAME "ExpireInterval"
#define EXPIRE_INTERVAL_DEFAULT (60 * 60 * 24) /* 24 hours */
#define EXPIRE_INTERVAL_MIN (60 * 60) /* once an hour maximum frequency */
#define POSTING_WAIT_NAME "PostingWait"
#define POSTING_WAIT_DEFAULT (60 * 5) /* 5 minutes wait before posting */
#define POSTING_RETRY_NAME "PostingRetry"
#define POSTING_RETRY_DEFAULT (60 * 5) /* 5 minutes retry when not connected */
#define POSTING_INTERVAL_NAME "PostingInterval"
#define POSTING_INTERVAL_DEFAULT (60 * 30) /* half hour when working OK */
#define DOWNLOAD_WAIT_NAME "DownloadWait"
#define DOWNLOAD_WAIT_DEFAULT (60 * 10) /* Start downloading after 10 minutes */
#define DOWNLOAD_RETRY_NAME "DownloadRetry"
#define DOWNLOAD_RETRY_DEFAULT (60 * 15) /* Keep trying to download */
#define DOWNLOAD_INTERVAL_NAME "DownloadInterval"
#define DOWNLOAD_INTERVAL_DEFAULT (60 * 60 * 6) /* Check for new news every 6 hours */
#define DOWNLOAD_INACTIVATE_NAME "InactivateDelay"
#define DOWNLOAD_INACTIVATE_DEFAULT (60 * 60 * 24 * 7) /* One week minimum to inactivate */
#define REMEMBER_MIN_NAME "RememberMin"
#define REMEMBER_MIN_DEFAULT (60 * 60 * 24 * 1) /* One week minimum message-id memory */
#define REMEMBER_MAX_NAME "RememberMax"
#define REMEMBER_MAX_DEFAULT (60 * 60 * 24 * 14) /* Two weeks max message-id memory */

/* Time-related registry entries */
#define NEWGROUPS_REGNAME "Newgroups"	/* Subkey is dotted quad name */
#define LAST_EXPIRE "LastExpire" /* Last successful client command */
#define LAST_CLIENT "LastClient" /* Last successful client completion */
#define LAST_POSTING "LastPosting" /* Last successful post */
#define LAST_DOWNLOAD "LastDownload" /* Last successful download */

/* Mutex names */
typedef enum {MARK_MUTEX, EVENT_MUTEX, ARTICLE_MUTEX, MESSAGE_MUTEX,
	      GROUP_MUTEX, SERVER_MUTEX, EXPIRE_MUTEX, SLURP_MUTEX,
	      POST_MUTEX, CONTROL_MUTEX, N_MUTEXES} LockID ;

typedef enum {LockFailed = 0, LockOK = 1, LockTimeout = 2} LockResult ;

/* Other miscellaneous arbitrary limits */
#define MAX_SUMMARY 512		/* Max length for summary in messages */
#define MAX_HOSTLEN 256		/* guess */
#define MAX_MESSAGEID 256	/* guess */
#define MAX_REGISTRYNAME 256	/* guess */
#define MAX_GROUPLEN MAX_PATH	/* guess, but fits naming convention */

#define MAX_ARTICLE_NO 0x7fffffff

enum { DEBUG_KEPT = 0x1, DEBUG_MUTEX = 0x2,
       DEBUG_REGISTRY = 0x4,
       DEBUG_CLIENT = 0x8, DEBUG_CLIENT_DATA = 0x10,
       DEBUG_SERVER = 0x20, DEBUG_CONNECT = 0x20,
       DEBUG_GROUPS = 0x40, DEBUG_ARTICLES = 0x80,
       DEBUG_SOCKETS = 0x100, DEBUG_FILES = 0x200,
       DEBUG_READLINE = 0x400, DEBUG_WRITELINE = 0x800,
       DEBUG_NOSLURP = 0x1000, DEBUG_SLURPTHREAD = 0x2000,
       DEBUG_DOWNLOAD = 0x4000,
       DEBUG_NOPOST = 0x10000, DEBUG_POSTTHREAD = 0x20000,
       DEBUG_POSTING = 0x40000,
       DEBUG_NOEXPIRE = 0x100000, DEBUG_EXPIRETHREAD = 0x200000,
       DEBUG_EXPIRERM = 0x400000,
       DEBUG_MESSAGES = 0x1000000,
       } ;


/***** Forward references *****/
typedef struct _Client Client ;
typedef struct _Group Group ;
typedef struct _Article Article ;
typedef struct _Message Message ;

/***** Functions defined in service.c *****/
int InstallService(void) ;
int RemoveService(void) ;
int UserService(int code) ;
int RestartService(void) ;

/* Private Service control codes */
enum { NNTP_CONTROL_START_POST = 128,
       NNTP_CONTROL_ABORT_POST,
       NNTP_CONTROL_STOP_POST,
       NNTP_CONTROL_START_EXPIRE,
       NNTP_CONTROL_ABORT_EXPIRE,
       NNTP_CONTROL_STOP_EXPIRE,
       NNTP_CONTROL_START_SLURP,
       NNTP_CONTROL_ABORT_SLURP,
       NNTP_CONTROL_STOP_SLURP
} ;

enum { SIGNAL_POST, SIGNAL_EXPIRE, SIGNAL_SLURP, N_SIGNALS } ;
enum { CONTROL_START, CONTROL_STOP, CONTROL_QUIT,
       CONTROL_CLIENT, LOOP_CONTROLS = CONTROL_CLIENT,
       ALL_CONTROLS } ;

extern HANDLE hEvents[N_SIGNALS][ALL_CONTROLS] ;
extern int suspended[N_SIGNALS] ;

extern BOOL is_service ;

VOID WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv)  ;

BOOL WINAPI ControlHandler(DWORD dwCtrlType) ;

/***** Functions defined in common.c *****/
char *fmtout(char *buffer, int bufsize, char *format, va_list args) ;
char *formatstr(char *buffer, int bufsize, char *format, ...) ;
void debug(int level, char *format, ...) ;
void message(WORD type, DWORD id, ...) ;
void Log(int type, char *format, ...) ;

/* Type is
   EVENTLOG_ERROR_TYPE	Error event
   EVENTLOG_WARNING_TYPE	Warning event
   EVENTLOG_INFORMATION_TYPE	Information event
   */

int ForAllFiles(char *dir, char *extn,
		int (*fn)(char *filename, void *passthrough),
		void *passthrough) ;

char *ErrorText(DWORD err) ;

DWORD Mark(void) ;

LockResult Lock(LockID lid, DWORD timeout) ;
void Unlock(LockID lid) ;

int DefaultDirectories(char *path, int pathlen) ;
int Match(char *pattern, char *string) ;

/***** Functions defined in log.c *****/
DWORD WINAPI LogThread(LPVOID info) ;

enum { LOG_QUEUE, LOG_DEBUG, LOG_STATUS } ;
enum { MAIN_THREAD, LOG_THREAD, POST_THREAD,
       SLURP_THREAD, EXPIRE_THREAD, CLIENT_THREAD,
       N_THREAD_TYPES } ;
void LogString(int type, char *str) ;
BOOL LogClient(int debuglevel) ;
void ThreadType(int type) ;

/***** Functions defined in registry.c *****/

char *GetRegString(char *keyname, char *name) ;
int SetRegString(char *keyname, char *name, char *value) ;
FILETIME *GetRegFiletime(char *keyname, char *name, FILETIME *tptr) ;
int SetRegFiletime(char *keyname, char *name, FILETIME *tptr) ;
DWORD GetRegDWORD(char *keyname, char *name) ;
int SetRegDWORD(char *keyname, char *name, DWORD value) ;

/***** Functions defined in timedate.c *****/

FILETIME TimeAdd(LPFILETIME when, long seconds) ;
long TimeDiff(LPFILETIME from, LPFILETIME to) ;
LPFILETIME DateTimeToFiletime(LPFILETIME when, char *nntpdate, char *nntptime) ;
char *TimeAsString(LPFILETIME when) ;

/***** Functions defined in nntp.c *****/

enum { ACTION_INSTALL, ACTION_UNINSTALL, ACTION_DEBUG, ACTION_DONE,
       ACTION_START, ACTION_USAGE, ACTION_CONTROL } ;

int parse_args(int argc, char *argv[]) ;

/***** Functions defined in events.c *****/

void DestroyEvents(void) ;
HANDLE WaitForEvent(DWORD timeout) ;
HANDLE WithdrawEvent(HANDLE hEvent) ;
HANDLE RegisterEvent(HANDLE hEvent) ;

/***** Database access and variables *****/
#define H_PERL /* ignore most of SDBM redefinitions */
#include <sdbm.h>
#undef H_PERL

#define GROUP_DB_NAME "groups"
#define ARTICLE_DB_NAME "articles"
#define MESSAGE_DB_NAME "messages"

/***** Functions defined in group.c *****/

typedef enum { GROUP_ACTIVE, GROUP_INACTIVE, GROUP_BOGUS,
	       GROUP_NEW} GroupState ;

struct _Group {
  GroupState state ;
  long low, high, articles ;
  FILETIME newnews ;	/* Time last NEWNEWS issued to server */
  FILETIME created ;	/* In case client send us NEWGROUPS */
  FILETIME expired ;	/* First time for next expire */
  FILETIME selected ;	/* Last time group was selected, us for inactivate */
  DWORD mark ;		/* Mark for mark and sweep algorithms */
  int post ;		/* char 'y' 'n' or 'm' */
  /* Group directory? */
} ;

int InitGroupDB(void) ;
void CloseGroupDB(void) ;
Group *GroupDBGet(char *name) ;
int GroupDBPut(char *name, Group *value) ;
int GroupDBRemove(char *name) ;
Group *NewGroup(char *group, GroupState istate, DWORD mark, int post, long days) ;
int ForAllGroups(int (*fn)(char *group, void *passthrough), void *passthrough) ;

void SetCurrentGroup(Client *client, char *name) ;
Group *CurrentGroup(Client *client) ;
char *CurrentGroupName(Client *client) ;

/***** Functions defined in article.c *****/

typedef enum { ARTICLE_BODY, ARTICLE_HEAD, ARTICLE_EXPIRED,
	       ARTICLE_NEW} ArticleState ;

struct _Article {
  ArticleState state ;
  long size ;
  FILETIME download ;
  FILETIME expires ;
  DWORD mark ;	/* Mark for mark & sweep algorithms */
  char names[1] ; /* group:num stored as extension; strlen == 0 to terminate */
} ;

int InitArticleDB(void) ;
void CloseArticleDB(void) ;
Article *ArticleDBGet(char *name) ;
int ArticleDBPut(char *name, Article *value) ;
int ArticleDBRemove(char *name) ;
Article *NewArticle(char *msgid, ArticleState istate, long size) ;
int ForAllArticles(int (*fn)(char *msgid, void *passthrough),
		   void *passthrough) ;

void SetCurrentArticle(Client *client, long number) ;
Article *CurrentArticle(Client *client) ;
long CurrentArticleNum(Client *client) ;

int ArticleInGroup(Article *art, char *group) ;
Article *AddArticleName(char *artname, char *group, long number,
			Article *art) ;
Article *RemoveArticleName(char *artname, char *group, long artno, Article *art) ;

char *ArticleFileName(char *group, long artno, char *buffer, int bufsize) ;
int MakeArticleDirs(char *filename) ;
char *MessageIDFile(char *filename, char *messageid, int idbufsize) ;

/***** Functions defined in message.c *****/

struct _Message {
  char messageid[1] ;
} ;

int InitMessageDB(void) ;
void CloseMessageDB(void) ;
char *MessageID(char *group, long artno, char *buffer, int bufsize) ;
int MessageAdd(char *group, long artno, char *messageid) ;
int MessageRemove(char *group, long artno) ;
int ForAllMessages(int (*fn)(char *group, long artno, void *passthrough),
		   void *passthrough) ;
char *MessageFileName(char *group, char *buffer, int bufsize) ;

/***** Functions defined in servers.c *****/

typedef struct {
  struct in_addr sin_addr ;	/* Address */
  char *name ;			/* canonical name (dotted quad) */
  Stream *stream ;		/* Stream for connection */
  DWORD mark ;			/* Marker for mark & sweep algorithms */
  FILETIME newgroups ;		/* Time when NEWGROUPS command last finished */
  int postok ;
} Connection ;

typedef struct {
  char *match ;			/* group name */
  int min, normal, purge ;	/* expiry times */
  long limit ;			/* size limit */
  int negate ;			/* negate pattern? */
} Pattern ;

typedef struct {
  char hostname[MAX_HOSTLEN] ;	/* Server name */
  Pattern **patterns ;		/* Pattern array */
  int nPatterns ;		/* Number of patterns */
  int maxPatterns ;		/* Pattern slots allocated */
  FILETIME connecttime ;	/* Time last connect tried */
  Connection *connection ;	/* Connection pointer */
  DWORD mark ;			/* Marker for mark & sweep algorithms */
} Server ;

int InitNewsFeeds(char *name) ;
int FindGroupServer(char *group, Server **server, Pattern **pattern) ;
Pattern *FindGroupPattern(char *group, Pattern **pattern, int nPatterns) ;
int ForAllServers(int (*fn)(Server *server, void *passthrough),
		  void *passthrough, int online) ;
Connection *ConnectServer(Server *server) ;
int DisconnectServer(Server *server) ;

/***** Functions defined in post.c *****/

#define POST_FAIL_EXTN ".err"

int Summary(char *filename, char *summary, int sumlen) ;
char *PostingName(char *buffer, int bufsize) ;
int PostFile(char *filename, char *summary, int sumsize, char *reason,
	     int rsnsize) ;
DWORD WINAPI PostThread(LPVOID info) ;

/***** Functions defined in slurp.c *****/

DWORD WINAPI SlurpThread(LPVOID info) ;

/***** Functions defined in expire.c *****/

DWORD WINAPI ExpireThread(LPVOID info) ;

extern long remember_max ; /* Max time to remember messageid */

/***** Functions defined in client.c *****/

struct _Client {
  Stream *stream ;
  long curr_artno ;
  char curr_groupname[MAX_GROUPLEN] ;
} ;

DWORD WINAPI ClientThread(LPVOID info) ;

/***** Functions defined in netlib.c *****/

void stream_error(Stream *this, long err, va_list args) ;

/*
 * $Log: nntp.h,v $
 * Revision 1.4  2000/01/14 22:17:55  angus
 * Increase version number
 *
 * Revision 1.3  2000/01/14 21:22:34  angus
 * More changes to logging. Use new streams 1.31 ioctl to specify debugging info
 * for client & server connections. Make fmtout more secure against buffer
 * overruns, and add a new format specifier for binary strings with visible
 * control characters. Change logging interface to add a varargs log, and use
 * that for stream debugging. Modify debug flags, add more status returns.
 *
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
#endif /* NNTP_H */
