/* post.c $Id: post.c,v 1.3 2000/01/14 21:22:34 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Manage posting for NNTP buffer
 */

#include "nntp.h"
#include "messages.h"

char *PostingName(char *buffer, int bufsize)
{
  char *result = NULL ;

  if ( bufsize > strlen(postdir) + 12 && Lock(POST_MUTEX, INFINITE) ) {
    for (;;) {
      HANDLE fileh ;

      sprintf(buffer, "%s\\%d", postdir, Mark()) ;
    
      if ( (fileh = CreateFile(buffer,		/* Read failed message */
			       GENERIC_READ,	/* Access */
			       0,			/* Share mode */
			       NULL,		/* Security inheritance */
			       OPEN_EXISTING,	/* Creation/distribution */
			       FILE_FLAG_SEQUENTIAL_SCAN, /* Flags */
			       NULL))		/* Template */
	   == INVALID_HANDLE_VALUE )
	break ;	/* Found an unused one */

      CloseHandle(fileh) ;	/* Close and try again */
    }

    result = buffer ;

    Unlock(POST_MUTEX) ;
  }

  return result ;
}

typedef struct _PostGroup {
  char name[MAX_GROUPLEN] ;
  struct _PostGroup *next ;
} PostGroup ;

static int PostFileServer(Stream *posts, Server *server, char *reason,
			  int rsnsize)
{
  Stream *outs = server->connection->stream ;
  long code ;

  if ( !Lock(SERVER_MUTEX, INFINITE) )
    return FALSE ;

  if ( SendStream(outs, "POST\r\n") < 0 ||
       (code = Response(outs)) < 0 ) {
    debug(DEBUG_POSTTHREAD, "No initial response from posting server\n") ;
    goto postserv_exit ;
  }

  if ( code == 340 && /* Go ahead and post on server */
       IoctlStream(outs, LINESTREAM_TERM, "\n") >= 0 ) {
    IoctlStream(outs, STREAM_DEBUG, DEBUG_SERVER, DEBUG_POSTING) ;
    for (;;) {
      char *line ;
      long rbytes = ReadStream(posts, &line, 0) ;

      if ( rbytes < 0 ) {
	debug(DEBUG_POSTTHREAD, "Can't read posting stream\n") ;
	code = 441 ;
	goto postserv_exit ;
      }

      if ( rbytes == 0 )	/* EOF, all done */
	break ;

      if ( !WriteStreamAll(outs, line, rbytes) ) {
	debug(DEBUG_POSTTHREAD, "Can't send to posting server\n") ;
	code = 441 ;
	goto postserv_exit ;
      }
    }

    if ( SendStream(outs, ".\r\n") < 0 ) {
      debug(DEBUG_POSTTHREAD, "Can't send termination to posting server\n") ;
      code = 441 ;
      goto postserv_exit ;
    }

    if ( (code = Response(outs)) <= 0 ) {
      debug(DEBUG_POSTTHREAD, "No final response from posting server\n") ;
      code = 441 ;
      goto postserv_exit ;
    }
  }
  
postserv_exit:
  IoctlStream(outs, STREAM_DEBUG, DEBUG_POSTING, DEBUG_SERVER) ;

  if ( code != 240 ) {
    if ( Reason(outs, reason, rsnsize) > 0 )
      debug(DEBUG_POSTTHREAD, "FROM Server: %s\n", reason) ;
    DisconnectServer(server) ;
  } else
    (void)IoctlStream(outs, LINESTREAM_TERM, "\r\n") ;

  Unlock(SERVER_MUTEX) ;

  return code ;
}

void Summarise(char *buffer, long bufsize, char *add, long addlen)
{
  int currlen = strlen(buffer) ;

  if ( currlen > 0 ) {
    int remainder = bufsize - currlen ;

    if ( addlen + 2 >= remainder ) {
      if ( remainder < 20 ) {
	currlen = MAX_SUMMARY - 20 ;
	remainder = 20 ;
	strcpy(buffer + currlen - 3, "...") ;
      }
      if ( addlen + 2 >= remainder )
	addlen = remainder - 3 ;
    }

    strcpy(buffer + currlen, "  ") ;
    currlen += 2 ;
  } else if ( addlen >= MAX_SUMMARY )
    addlen = MAX_SUMMARY - 1 ;

  strncpy(buffer + currlen, add, addlen) ;
  currlen += addlen ;
  buffer[currlen] = '\0' ;
}

/* Get Summary of outgoing file */
int Summary(char *postname, char *summary, int sumsize)
{
  Stream *posts = NULL, *lines = NULL ;
  int result = FALSE ;

  summary[0] = '\0' ;

  if ( Lock(POST_MUTEX, INFINITE) ) {
    if ( (posts = MakeFileStream(postname, FILESTREAM_READ, stream_error)) != NULL &&
	 (lines = MakeLineStream(posts, "\n")) != NULL ) {
      long rbytes = 0 ;
      char *line ;

      /* Scan for Newsgroups: and Subject: line. Doesn't take into account
	 continuations */
      while ( (rbytes = ReadStream(lines, &line, 0)) > 0 ) {
	if ( rbytes == 1 || (rbytes == 2 && line[0] == '\r') ) {
	  result = TRUE ;
	  break ; /* End of headers */
	} else if ( (rbytes > 8 && _strnicmp(line, "Subject:", 8) == 0) ||
		    (rbytes > 11 &&
		     (_strnicmp(line, "Newsgroups:", 11) == 0 ||
		      _strnicmp(line, "Message-ID:", 11) == 0)) ) {
	  long sumlen = rbytes - 1;	/* Line without LF */

	  if ( sumlen > 0 && line[sumlen - 1] == '\r' )
	    sumlen-- ;

	  Summarise(summary, sumsize, line, sumlen) ;
	}
      }
    }

    if ( lines )
      DestroyStream(lines) ;

    if ( posts )
      DestroyStream(posts) ;

    Unlock(POST_MUTEX) ;
  }

  return result ;
}

int PostFile(char *postname, char *summary, int sumsize, char *reason,
	     int rsnsize)
{
  int result = 0 ;	/* Neither accepted nor denied */

  summary[0] = '\0' ;
  reason[0] = '\0' ;

  if ( !suspended[SIGNAL_POST] && Lock(POST_MUTEX, INFINITE) ) {
    Stream *posts = NULL, *lines = NULL ;
    PostGroup *newsgroups = NULL ;
    DWORD mark = Mark() ;

    debug(DEBUG_POSTTHREAD, "Trying to post file %s\n", postname) ;

    Log(LOG_STATUS, "Posting file %s\n", postname) ;

    /* Scan for newsgroups line, cons list of newsgroups */
    if ( (posts = MakeFileStream(postname, FILESTREAM_READ, stream_error)) != NULL &&
	 (lines = MakeLineStream(posts, "\n")) != NULL ) {
      long rbytes = 0 ;
      char *line ;

      /* Scan for Newsgroups: and Subject: line. Doesn't take into account
	 continuations */
      while ( (rbytes = ReadStream(lines, &line, 0)) > 0 ) {
	if ( rbytes == 1 || (rbytes == 2 && line[0] == '\r') ) {
	  break ; /* End of headers */
	} else if ( (rbytes > 8 && _strnicmp(line, "Subject:", 8) == 0) ||
		    (rbytes > 11 && _strnicmp(line, "Message-ID:", 11) == 0) ) {
	  long sumlen = rbytes - 1;

	  if ( sumlen > 0 && line[sumlen - 1] == '\r' )
	    sumlen-- ;

	  Summarise(summary, sumsize, line, sumlen) ;
	} else if ( rbytes > 11 && _strnicmp(line, "Newsgroups:", 11) == 0 ) {
	  char *gname = NULL ;
	  long sumlen = rbytes - 1;

	  if ( sumlen > 0 && line[sumlen - 1] == '\r' )
	    sumlen-- ;

	  Summarise(summary, sumsize, line, sumlen) ;

	  for ( line += 11, rbytes -= 11 ; rbytes ; line++, rbytes-- ) {
	    char ch = *line ;

	    if ( ch == ',' || isspace(ch) ) {
	      if ( gname ) {
		PostGroup *this ;

		if ( (this = malloc(sizeof(PostGroup))) != NULL ) {
		  int len = line - gname ;
		  strncpy(this->name, gname, len) ;
		  this->name[len] = '\0' ;
		  this->next = newsgroups ;
		  newsgroups = this ;
		  debug(DEBUG_POSTTHREAD, "Posting %s to group %s (%d)\n", postname,
			this->name, len) ;
		}

		gname = NULL ;
	      }
	    } else if ( gname == NULL )
	      gname = line ;
	  }
	}
      }

      Log(LOG_STATUS, "Posting article %s\n", summary) ;

      if ( rbytes <= 0 )
	goto postfile_exit ;

      /* Limited to 4GB postings by SetFilePointer! */
      while ( newsgroups ) {		/* For all newsgroups found */
	PostGroup *next = newsgroups->next ;

	if ( IoctlStream(lines, STREAM_REWIND) >= 0 ) {
	  Server *server ;
	  Pattern *patt ;
	  Connection *conn ;

	  /* Send to servers for all groups (once only per server) */
	  if ( FindGroupServer(newsgroups->name, &server, &patt) &&
	       (conn = ConnectServer(server)) != NULL &&
	       conn->mark != mark ) {
	    long response = PostFileServer(lines, server, reason, rsnsize) ;

	    debug(DEBUG_POSTTHREAD, "Post %s to %s responds %d\n",
		  newsgroups->name, server->hostname, response) ;

	    switch ( response ) {
	    default:	/* Unknown response, take first one */
	      if ( result != 0 )
		break ;
	    case 440:	/* Take it if result is not 240 or 441 */
	      if ( result == 441 )
		break ;
	    case 441:	/* Take it if result is not 240 */
	      if ( result == 240 )
		break ;
	    case 240:	/* Success overrides all else */
	      result = response ;
	    case 0:	/* Do nothing, any other result is better */
	      break ;
	    }

	    if ( result == 240 ) {
	      message(EVENTLOG_INFORMATION_TYPE, MSG_POST_SENT,
		      summary, server->hostname, NULL) ;
	    }

	    conn->mark = mark ;	/* Mark as attempted on connection */
	  }
	} else if ( result != 240 )
	  result = 440 ;	/* error if can't set file pointer */

	free(newsgroups) ;
	newsgroups = next ;
      }
    } else
      debug(DEBUG_POSTTHREAD, "Can't open streams on %s, %d\n", postname,
	    GetLastError()) ;

  postfile_exit:
    if ( posts )
      DestroyStream(posts) ;
    if ( lines )
      DestroyStream(lines) ;

    while ( newsgroups ) {
      PostGroup *next = newsgroups->next ;
      free(newsgroups) ;
      newsgroups = next ;
    }

    Unlock(POST_MUTEX) ;
  }

  return result ;
}

static int ForAllPostings(int (*fn)(char *filename, void *passthrough),
			  void *passthrough)
{
  int result = TRUE ;

  if (Lock(POST_MUTEX, INFINITE)) {
    result = ForAllFiles(postdir, "", fn, passthrough) ;
    Unlock(POST_MUTEX) ;
  }

  return result ;
}

/* Helper function for ForAllPostings and post thread. ForAllPostings has
   already got Mutex */
static int PostAll(char *postname, void *passthrough)
{
  int result = TRUE ;
  char summary[MAX_SUMMARY], reason[BUFSIZ] ;

  /* Signalled to quit */
  switch ( WaitForMultipleObjects(ALL_CONTROLS, hEvents[SIGNAL_POST], FALSE, 0) ) {
  case WAIT_OBJECT_0 + CONTROL_START:
  case WAIT_TIMEOUT:
    break ;
  default: /* Abort on client connect, quit, stop, or error */
    return FALSE ;
  }

  switch (PostFile(postname, summary, MAX_SUMMARY, reason, BUFSIZ)) {
  case 440:	/* Posting refused by server */
    message(EVENTLOG_INFORMATION_TYPE, MSG_POST_REJECTED, summary, reason,
	    NULL) ;
    {
      char errname[MAX_PATH] ;

      strcpy(errname, postname) ;
      strcat(errname, POST_FAIL_EXTN) ;

      if ( !MoveFile(postname, errname) )
	DeleteFile(postname) ;	/* Can't rename, so delete instead */
    }
    break ;
  case 240:	/* Posting sent successfully */
    DeleteFile(postname) ;
    break ;
  case 441:	/* Posting failed, try again */
    break ;	/* Try next posting anyway */
  default:	/* Result indeterminate, or not connected, so stop here */
    result = FALSE ;
    break ;
  }

  return result ;
}

DWORD WINAPI PostThread(LPVOID info)
{
  DWORD wait, retry, interval ;
  FILETIME then, now ;

  ThreadType(POST_THREAD) ;

  if ( (debuglevel & DEBUG_NOPOST) != 0 )
    return 1 ;

  debug(DEBUG_POSTTHREAD, "Started posting thread\n") ;

  wait = GetRegDWORD(SERVICE_REGISTRY_KEY, POSTING_WAIT_NAME) ;
  if ( wait <= 0 )	/* Default if silly value */
    wait = POSTING_WAIT_DEFAULT ;

  retry = GetRegDWORD(SERVICE_REGISTRY_KEY, POSTING_RETRY_NAME) ;
  if ( retry <= 0 )	/* Default if silly value */
    retry = POSTING_RETRY_DEFAULT ;
  debug(DEBUG_POSTTHREAD, "Posting retry time is %d seconds\n", retry) ;

  interval = GetRegDWORD(SERVICE_REGISTRY_KEY, POSTING_INTERVAL_NAME) ;
  if ( interval <= 0 )	/* Default if silly value */
    interval = POSTING_INTERVAL_DEFAULT ;
  debug(DEBUG_POSTTHREAD, "Posting interval time is %d seconds\n", interval) ;

  if ( GetRegFiletime(SERVICE_REGISTRY_KEY, LAST_POSTING, &then) ) {
    long diff ;

    GetSystemTimeAsFileTime(&now) ;
    diff = TimeDiff(&then, &now) ;
    if ( diff + wait < interval ) /* Take max of remaining interval & wait */
      wait = interval - diff ;
  }

  debug(DEBUG_POSTTHREAD, "Posting wait time is %d seconds\n", wait) ;

  SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL) ;

  for (;;) {
    DWORD delay = wait ;

    GetSystemTimeAsFileTime(&now) ;
    then = TimeAdd(&now, delay) ;
    do {
      if ( suspended[SIGNAL_POST] ) {
        LogString(LOG_STATUS, "Posting suspended\n") ;
      } else if ( FileTimeToLocalFileTime(&then, &now) ) {
        Log(LOG_STATUS, "Next post is at %t\n", &now) ;
      }

      switch ( WaitForMultipleObjects(LOOP_CONTROLS, hEvents[SIGNAL_POST],
                                      FALSE, delay * 1000) ) {
      case WAIT_TIMEOUT:
      case WAIT_OBJECT_0 + CONTROL_START:
        delay = 0 ;
        break ;
      case WAIT_OBJECT_0 + CONTROL_QUIT:
        goto post_quit ;
      default:
        GetSystemTimeAsFileTime(&now) ;
        delay = TimeDiff(&now, &then) ;
        break ;
      }
    } while ( delay > 0 ) ;

    if ( suspended[SIGNAL_POST] )
      continue ;

    if ( ForAllPostings(PostAll, NULL) ) {	/* Successful posting */
      GetSystemTimeAsFileTime(&now) ;
      SetRegFiletime(SERVICE_REGISTRY_KEY, LAST_POSTING, &now) ;
      wait = interval ;
    } else
      wait = retry ;
  }

post_quit:
  LogString(LOG_STATUS, "Post thread terminated\n") ;

  return 0 ;
}

/*
 * $Log: post.c,v $
 * Revision 1.3  2000/01/14 21:22:34  angus
 * More changes to logging. Use new streams 1.31 ioctl to specify debugging info
 * for client & server connections. Make fmtout more secure against buffer
 * overruns, and add a new format specifier for binary strings with visible
 * control characters. Change logging interface to add a varargs log, and use
 * that for stream debugging. Modify debug flags, add more status returns.
 *
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
