/* rebuild.c $Id: rebuild.c,v 1.3 2000/01/14 21:22:34 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1998.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Database rebuild utility for NNTP Buffer.
 */

#include "nntp.h"
#include "messages.h"

char *program ;
int debuglevel = 0x80000000 ;

long remember_max = 0 ;

BOOL is_service = FALSE ;

char *confdir = NULL ;
char *spooldir = NULL ;
char *postdir = NULL ;
unsigned short inport, outport ;

static void usage(void)
{
  fprintf(stderr, "Usage: %s [-dconfigdir] [-sspooldir]\n", program) ;
  ExitProcess(0) ;
}

int parse_args(int argc, char *argv[])
{
  int pathlen ;
  char path[MAX_PATH], *arg ;

  pathlen = GetModuleFileName(NULL, path, MAX_PATH) ;
  if ( pathlen ) {
    char *slash = strrchr(path, '\\') ;

    if ( slash ) {
      pathlen = slash - path ;
      *slash = '\0' ;
    }
  }

  for ( program = *argv++ ; --argc ; argv++ ) {
    if (argv[0][0] == '-') {
      char ch = argv[0][1] ;

      if ( strchr("dsD", ch) != NULL ) {
	if ( argv[0][2] != '\0' )
	  arg = &argv[0][2] ;
	else if ( --argc )
	  arg = *++argv ;
	else
	  return 0 ;
      } else if ( argv[0][2] != '\0' )
        return 0 ;

      switch (ch) {
      case 'd':	/* set config directory */
        confdir = arg ;
	break;
      case 's':	/* set spool directory */
        spooldir = arg ;
	break;
      case 'D':	/* debug level service */
        debuglevel |= strtol(arg, &arg, 0) ;
	if ( debuglevel == 0 )
	  return 0 ;
        break ;
      default:
	return 0 ;
      }
    } else
      return 0 ;
  }

  if ( program == NULL )
    return 0 ;

  if ( !DefaultDirectories(path, pathlen) )
    return 0 ;

  return 1 ;
}

void LogString(int type, char *str)
{
  fputs(str, stderr) ;
  fflush(stderr) ;
}

int RebuildDB(char *directory, char *groupname)
{
  char pattern[MAX_PATH] ;
  WIN32_FIND_DATA fdata ;
  HANDLE it ;
  int dirlen = strlen(directory) ;
  int grouplen = strlen(groupname) ;
  int result = TRUE ;
  int32 reported = FALSE ;

  strcpy(pattern, directory) ;
  strcpy(pattern + dirlen, "\\*") ;

  if ( (it = FindFirstFile(pattern, &fdata)) != INVALID_HANDLE_VALUE ) {
    do {
      if ( fdata.cFileName[0] != '.' ) {
        strcpy(pattern + dirlen + 1, fdata.cFileName) ; /* Full file name */

        if ( (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 ) {
          char *scan = fdata.cFileName ;
          int len = grouplen ;

          if ( len > 0 )
            groupname[len++] = '.' ;
          while ( *scan == '_' ) /* Ignore leading _ in groupname */
            scan++ ;
          strcpy(groupname + len, scan) ;
          RebuildDB(pattern, groupname) ;
        } else { /* File */
          char *scan = fdata.cFileName ;

          while ( isdigit(*scan) ) ++scan ;

          if ( *scan == '\0' ) { /* Ignore non-numeric files */
            char msgid[MAX_MESSAGEID] ;
            Article *art ;
            Group gdata, *group = GroupDBGet(groupname) ;
            int artno = atoi(fdata.cFileName) ; /* article number */

            if ( !group ) { /* If current group does not exist, create it */
              Pattern *patt ;
              Server *server ;

              printf("Rebuilding group %s...", groupname) ;
              reported = TRUE ;

              if ( FindGroupServer(groupname, &server, &patt) )
                group = NewGroup(groupname, GROUP_ACTIVE, 0, 'y', patt->normal) ;
              if ( !group ) {
                printf("failed\n") ;
                continue ;
              } else
                printf("OK\n") ;
            } else if ( !reported ) {
              reported = TRUE ;
              printf("Reusing group %s\n", groupname) ;
            }

            gdata = *group ;

            /* Extract message-id from article */
            if ( !MessageIDFile(pattern, msgid, MAX_MESSAGEID) ) {
              printf("Can't extract message-id from file %s\n", pattern) ;
              continue ;
            }

            printf(" %d", artno) ;

            if ( gdata.high < artno )
              gdata.high = artno ;

            if ( (art = ArticleDBGet(msgid)) != NULL ) {
              /* Article exists, check if it is known in this group */
              if ( !ArticleInGroup(art, groupname) ) {
                if ( (art = AddArticleName(msgid, groupname, artno, art)) != NULL ) {
                  gdata.articles++ ;
                } else {
                  printf("Failed to add article %d group %s message-id %s\n",
                         artno, groupname, msgid) ;
                  continue ;
                }
              }
            } else { /* Article does not exist, add it */
              if ( (art = NewArticle(msgid, ARTICLE_BODY, fdata.nFileSizeLow)) != NULL &&
                   AddArticleName(msgid, groupname, artno, art) != NULL ) {
                gdata.articles++ ;
              } else {
                printf("Failed to add article %d msgid %s to group %s\n",
                       artno, msgid, groupname) ;
                continue ;
              }
            }

            /* Write back changed information */
            if ( !GroupDBPut(groupname, &gdata) ) {
              printf("Failed to save data for group %s\n", groupname) ;
            }
          } else {
            printf("Non article file %s found in group %s spool directory %s\n",
                   fdata.cFileName, groupname, directory) ;
          }
        }
      }
    } while (FindNextFile(it, &fdata)) ;
    if ( GetLastError() != ERROR_NO_MORE_FILES )
      result = FALSE ;
    FindClose(it) ;
  } else	/* FindFirstFile failed */
    result = FALSE ;

  if ( reported )
    printf("\n") ;

  return result ;
}

int main(int argc, char *argv[])
{
  char groupname[MAX_PATH] = { '\0' } ;

  if ( !parse_args(argc, argv) )
    usage() ;

  if ( !InitNewsFeeds(NEWSFEEDS) ) {
    fprintf(stderr, "Can't initialise newfeeds") ;
    return 0 ;
  }

  if ( !InitGroupDB() ) {
    fprintf(stderr, "Can't initialise group database") ;
    return 0 ;
  }

  if ( !InitArticleDB() ) {
    fprintf(stderr, "Can't initialise article database") ;
    return 0 ;
  }

  if ( !InitMessageDB() ) {
    fprintf(stderr, "Can't initialise message database") ;
    return 0 ;
  }

  RebuildDB(spooldir, groupname) ;

  CloseMessageDB() ;
  CloseArticleDB() ;
  CloseGroupDB() ;

  return 0 ;
}

/*
 * $Log: rebuild.c,v $
 * Revision 1.3  2000/01/14 21:22:34  angus
 * More changes to logging. Use new streams 1.31 ioctl to specify debugging info
 * for client & server connections. Make fmtout more secure against buffer
 * overruns, and add a new format specifier for binary strings with visible
 * control characters. Change logging interface to add a varargs log, and use
 * that for stream debugging. Modify debug flags, add more status returns.
 *
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
