/* servers.c $Id: servers.c,v 1.3 2000/01/14 21:22:34 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Manage server connections for NNTP buffer
 */

#include "nntp.h"
#include "messages.h"

Server **serverTable = NULL ;
int nServers = 0 ;
int maxServers = 0 ;

/* Parse expiry info for /number */
static int PrefixNumber(char **str, char init, int *expiry)
{
  char *scan = *str ;
  int value = 0 ;

  if ( init != '\0' && *scan++ != init )
    return FALSE ;

  if ( isdigit(*scan) ) {
    while ( isdigit(*scan) ) {
      value = value * 10 + *scan - '0' ;
      scan++ ;
    }
  } else if ( *scan++ == '*' )
    value = -1 ;
  else
    return FALSE ;

  *expiry = value ;
  *str = scan ;

  return TRUE ;
}

int InitNewsFeeds(char *name)
{
  char filename[MAX_PATH] ;
  int result = FALSE ;
  Stream *files = NULL, *lines = NULL ;

  assert(confdir) ;

  if ( !Lock(SERVER_MUTEX, INFINITE) )
    return FALSE ;

  strcpy(filename, confdir) ;
  strcat(filename, "\\") ;
  strcat(filename, name) ;

  if ( (files = MakeFileStream(filename, FILESTREAM_READ, stream_error)) != NULL &&
       (lines = MakeLineStream(files, "\n")) != NULL ) {
    Server *current = NULL ;
    int lineno = 0 ;

    for (;;) {
      char *line, *scan ;
      long rbytes = ReadStream(lines, &line, 0) ;

      if ( rbytes < 0 ) {
	message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		filename, "error reading file", lineno, "",
		NULL) ;
	break ;
      }

      if ( rbytes == 0 ) {	/* EOF */
	if ( nServers == 0 ) {
	  message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		  filename, "no servers in file", lineno, "",
		  NULL) ;
	} else
	  result = TRUE ;	/* OK if we got here with servers */
	break ;
      }

      line[rbytes - 1] = '\0' ;
      lineno++ ;

      for (scan = line ; isspace(*scan) ; scan++ ) /* skip whitespace */ ;

      if ( *scan == '\0' ) {	/* Blank line terminates current */
	if ( current ) {
	  message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		  filename, "blank line in pattern", lineno, line,
		  NULL) ;
	  goto initfeed_exit ;
	}
      } else if ( *scan != '#' ) {	/* Not blank or comment */
	int min, normal, purge, limit ;	/* default expiry times */ 

	if ( current == NULL ) {	/* No current, create one */
	  char *colon = strchr(scan, ':') ;

	  if ( colon == NULL ) {	/* No hostname */
	    message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		    filename, "hostname not terminated", lineno, line,
		    NULL) ;
	    goto initfeed_exit ;
	  }

	  if ( nServers >= maxServers ) {
	    maxServers = maxServers * 2 + 8 ;
	    serverTable = realloc(serverTable,
				  maxServers * sizeof(Server *)) ;
	  }

	  if ( serverTable == NULL || 
	       (current = malloc(sizeof(Server))) == NULL ) {
	    message(EVENTLOG_ERROR_TYPE, MSG_MEMORY,
		    "reading newsfeed file (loading server)", NULL) ;
	    goto initfeed_exit ;
	  }

	  current->mark = 0 ;
	  current->connection = NULL ;
	  memset(&current->connecttime, 0, sizeof(FILETIME)) ;
	  current->nPatterns = 0 ;
	  current->maxPatterns = 0 ;
	  strncpy(current->hostname, scan, colon - scan) ;
	  current->hostname[colon - scan] = '\0' ;
	  scan = colon + 1 ;

	  serverTable[nServers++] = current ;

	  min = 0 ;		/* default expiry times */ 
	  normal = 7 ;
	  purge = 7 ;
	  limit = -1 ;
	}

	/* Now have a current server, fill in patterns */
	do {
	  long memo = 0 ;
	  int negate = FALSE ;
	  char *patend ;
	  Pattern *this ;

	  switch ( *scan ) {
	  case '/':
	    if ( !PrefixNumber(&scan, '/', &min) ||
		 !PrefixNumber(&scan, '/', &normal) ||
		 !PrefixNumber(&scan, '/', &purge) ) {
	      message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		      filename, "bad expiry pattern", lineno, line,
		      NULL) ;
	      goto initfeed_exit ;
	    }
	    break ;
	  case '<':
	    if ( !PrefixNumber(&scan, '<', &limit) ) {
	      message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		      filename, "bad size limit", lineno, line,
		      NULL) ;
	      goto initfeed_exit ;
	    }
	    break ;
	  case '!':
	    negate = TRUE ;
	    scan++ ;	/* FALLTHRU */
	  default:
	    for (patend = scan ; /* search for end of pattern */
		 *patend && *patend != ',' && !isspace(*patend) ;
		 patend++) ;
	    if ( patend == scan ) {
	      message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		      filename, "empty group pattern", lineno, line,
		      NULL) ;
	      goto initfeed_exit ;
	    }

	    if ( current->nPatterns >= current->maxPatterns ) {
	      current->maxPatterns = current->maxPatterns * 2 + 8 ;
	      current->patterns = realloc(current->patterns,
					  current->maxPatterns * sizeof(Pattern *)) ;
	    }

	    if ( current->patterns == NULL ||
		 (this = malloc(sizeof(Pattern))) == NULL ||
		 (this->match = malloc(patend - scan + 1)) == NULL ) {
	      message(EVENTLOG_ERROR_TYPE, MSG_MEMORY,
		      "reading newsfeed file (loading patterns)", NULL) ;
	      goto initfeed_exit ;
	    }

	    strncpy(this->match, scan, patend - scan) ;
	    this->match[patend - scan] = '\0' ;
	    this->min = min ;
	    this->normal = normal ;
	    this->purge = purge ;
	    this->limit = limit ;
	    this->negate = negate ;

	    /* Set max remember time limit */
	    if ( purge > normal )
	      memo = purge ;
	    else
	      memo = normal ;

	    memo *= 24 * 60 * 60 ;

	    if ( memo > remember_max )
	      remember_max = memo ;

	    current->patterns[current->nPatterns++] = this ;
		
	    scan = patend ;
	  }

	  while ( isspace(*scan) ) scan++ ; /* skip whitespace */

	  /* Continue or quit */
	  if ( *scan == '\0' ) {
	    if ( current->nPatterns == 0 ) {
	      message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		      filename, "no group patterns", lineno, line,
		      NULL) ;
	      goto initfeed_exit ;
	    }
	    current = NULL ;
	  } else if ( *scan++ != ',' ) {
	    message(EVENTLOG_ERROR_TYPE, MSG_FEEDSYNTAX,
		    filename, "bad group separator", lineno, line,
		    NULL) ;
	    goto initfeed_exit ;
	  } else {	/* skip whitespace after comma */
	    while ( isspace(*scan) ) scan++ ; /* skip whitespace */
	  }
	} while (*scan) ;
      }
    }
  } else {
    message(EVENTLOG_ERROR_TYPE, MSG_FILEOPEN, "newsfeeds", filename,
	    ErrorText(GetLastError()), NULL) ;
  }

initfeed_exit:
  if ( lines )
    DestroyStream(lines)  ;

  if ( files )
    DestroyStream(files) ;

  Unlock(SERVER_MUTEX) ;

  return result ;
}

Pattern *FindGroupPattern(char *group, Pattern **pattern, int nPatterns)
{
  int patind ;

  for (patind = 0 ; patind < nPatterns ; patind++ ) {
    Pattern *thispat = pattern[patind] ;

    if ( (Match(thispat->match, group) ? !thispat->negate : thispat->negate) )
      return thispat ;
  }

  return NULL ;
}

/* Find the newsfeed for a particular group */
int FindGroupServer(char *group, Server **server, Pattern **pattern)
{
  int servind ;

  for (servind = 0 ; servind < nServers ; servind++ ) {
    Server *thisserv = serverTable[servind] ;
    Pattern *thispat = FindGroupPattern(group, thisserv->patterns,
					thisserv->nPatterns) ;

    if ( thispat ) {
      *server = thisserv ;
      *pattern = thispat ;
      return TRUE ;
    }
  }

  return FALSE ;
}

/* Call async gethostbyname, timing out after 5 secs */
struct hostent *GetHostByName(char *name)
{
  MSG msg ;
  HANDLE hcall ;
  DWORD timeout ;

  static char buffer[MAXGETHOSTSTRUCT] ;
  static HWND hWnd ;
  static ATOM hClass ;

#ifndef WSAGETASYNCERROR
#define WSAGETASYNCERROR(lParam)            HIWORD(lParam)
#endif
#ifndef WSAGETASYNCBUFLEN
#define WSAGETASYNCBUFLEN(lParam)           LOWORD(lParam)
#endif
#ifndef MWMO_INPUTAVAILABLE
#define MWMO_INPUTAVAILABLE 0
#endif

  if ( hClass == 0 &&
       (hClass = GlobalAddAtom("Async Message Class")) == 0 ) {
    message(EVENTLOG_ERROR_TYPE, MSG_ATOMFAILED, name, ErrorText(GetLastError()), NULL) ;
    return NULL ;
  }

  if ( hWnd == NULL &&
       (hWnd = CreateWindow((LPCTSTR)hClass, /* class name */
                            NULL, /* window name */
                            WS_DISABLED, /* style */
                            0, /* x */
                            0, /* y */
                            0, /* width */
                            0, /* height */
                            NULL, /* parent */
                            NULL, /* menu */
                            NULL, /* instance */
                            NULL /* param */
                            )) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_WINDOWFAILED, name, ErrorText(GetLastError()), NULL) ;
    return NULL ;
  }

  timeout = GetRegDWORD(SERVICE_REGISTRY_KEY, CONNECT_TIMEOUT_NAME) ;
  if ( timeout <= 0 ) /* Default if silly value */
    timeout = CONNECT_TIMEOUT ;

  if ( PeekMessage(&msg, (HWND)hWnd, WM_USER, WM_USER, PM_REMOVE) ) {
    debug(DEBUG_CONNECT, "Message ready before WSAAsyncGetHostByName\n") ;
  }

  if ( (hcall = WSAAsyncGetHostByName((HWND)hWnd, WM_USER, name, buffer, MAXGETHOSTSTRUCT)) != NULL ) {
  again:
    debug(DEBUG_CONNECT, "Waiting for message from WSAAsyncGetHostByName for %s\n", name) ;
    switch ( MsgWaitForMultipleObjectsEx(0, NULL, timeout * 1000, QS_ALLINPUT,
                                         MWMO_INPUTAVAILABLE)) {
    case WAIT_OBJECT_0: /* Message waiting */
      if ( PeekMessage(&msg, (HWND)hWnd, WM_USER, WM_USER, PM_REMOVE) ) {
        debug(DEBUG_CONNECT, "Message is WM_USER (%d)\n", WSAGETASYNCERROR(msg.lParam)) ;
	if ( WSAGETASYNCERROR(msg.lParam) == 0 )
	  return (struct hostent *)buffer ;
	/* Error returned by gethostbyname */
	return NULL ;
      }
      debug(DEBUG_CONNECT, "Message not WM_USER\n") ;
      /* Other message */      
      if ( PeekMessage(&msg, (HWND)hWnd, 0, 0, PM_REMOVE) )
	goto again ;
      /* Strange error, fallthrough, cancel & quit */
    default:
      debug(DEBUG_CONNECT,
            "Strange error from MsgWaitForMultipleObjectsEx %d\n",
            GetLastError()) ;
    case WAIT_TIMEOUT:      /* Timed out, cancel request and return */
      (void)WSACancelAsyncRequest(hcall) ;
      break ;
    }
  } else
    message(EVENTLOG_ERROR_TYPE, MSG_ASYNCFAILED, name, ErrorText(GetLastError()), NULL) ;

  return NULL ;
}

Connection *ConnectServer(Server *server)
{
  Connection *conn = NULL ;
  Stream *socks = NULL, *lines = NULL ;
  struct hostent *hostentry = NULL ;
  struct sockaddr_in sremote ;
  char *dotquad ;
  int servind ;
  SOCKET sock ;
  int postok ;

  debug(DEBUG_CONNECT, "Connecting server %s\n", server->hostname) ;

  if ( !Lock(SERVER_MUTEX, INFINITE) )
    return NULL ;

  if ( (conn = server->connection) != NULL ) {	/* Are we already connected? */
    debug(DEBUG_CONNECT, "Already connected to %s\n", server->hostname) ;
    GetSystemTimeAsFileTime(&server->connecttime) ;
    goto connect_exit ;
  }

  /* Test if tried to connect and failed recently */
  {
    FILETIME now, when ;
    DWORD secs = GetRegDWORD(SERVICE_REGISTRY_KEY, RECONNECT_TIMEOUT_NAME) ;

    if ( secs <= 0 )
      secs = RECONNECT_TIMEOUT ;

    GetSystemTimeAsFileTime(&now) ;
    when = TimeAdd(&server->connecttime, secs) ;
    if ( CompareFileTime(&now, &when) < 0 ) {
      debug(DEBUG_CONNECT, "Failed to connect to %s in last %d seconds\n",
	    server->hostname, secs) ;
      goto connect_exit ;
    }

    /* Set next timeout. Non-debug timeout should be got from registry. */
    server->connecttime = now ;
  }

  if ( (hostentry = GetHostByName(server->hostname)) == NULL ) {
    debug(DEBUG_CONNECT, "Gethostbyname for %s failed\n",
	  server->hostname) ;
    goto connect_exit ;
  }

  /* Set up remote socket address */
  memset(&sremote, 0, sizeof(struct sockaddr_in)) ;
  sremote.sin_family = AF_INET ;
  sremote.sin_port = outport ;
  memcpy(&sremote.sin_addr, hostentry->h_addr, hostentry->h_length);

  /* Check if we've connected there already */
  for ( servind = 0 ; servind < nServers ; servind++ ) {
    Server *thatserv = serverTable[servind] ;
    Connection *thatconn = thatserv->connection ;

    if ( thatconn && sremote.sin_addr.s_addr == thatconn->sin_addr.s_addr ) {
      conn = server->connection = thatconn ;
      debug(DEBUG_CONNECT, "Already connected to %s as host %s\n",
	    server->hostname, thatserv->hostname) ;
      goto connect_exit ;
    }
  }

  if ( INVALID_SOCKET == (sock = socket(PF_INET, SOCK_STREAM, 0)) ) {
    message(EVENTLOG_ERROR_TYPE, MSG_OPENSOCKET,
	    ErrorText(WSAGetLastError()), NULL) ;
    goto connect_exit ;
  }

  if ( SOCKET_ERROR == connect(sock, (struct sockaddr *)&sremote,
			       sizeof(struct sockaddr_in)) ) {
    debug(DEBUG_CONNECT, "Connect call for %s failed\n", server->hostname) ;
    goto connect_exit ;
  }

  /* Suck greeting from server */
  if ( (socks = MakeSocketStream(sock, stream_error)) == NULL ) {
    debug(DEBUG_CONNECT, "Create stream for %s failed\n", server->hostname) ;
    goto connect_exit ;
  }

  if ( (lines = MakeLineStream(socks, "\r\n")) == NULL ) {
    debug(DEBUG_CONNECT, "Create line stream for %s failed\n",
	  server->hostname) ;
    goto connect_exit ;
  }

  IoctlStream(lines, STREAM_DEBUG, LINESTREAM_DEBUG_READ, DEBUG_SERVER) ;
  IoctlStream(lines, STREAM_DEBUG, LINESTREAM_DEBUG_WRITE, DEBUG_SERVER) ;

  if ( (postok = Response(lines)) < 0 ) {
    debug(DEBUG_CONNECT, "Read greeting for %s failed\n", server->hostname) ;
    goto connect_exit ;
  }

  debug(DEBUG_CONNECT, "Server %s greeting response is %d\n",
	server->hostname, postok) ;

  dotquad = inet_ntoa(sremote.sin_addr) ;	/* get canonical name */

  if ( (conn = malloc(sizeof(Connection))) == NULL ||
       (conn->name = malloc(strlen(dotquad) + 1)) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_MEMORY, "connecting server", NULL) ;
    if ( conn ) {
      free(conn) ;
      conn = NULL ;
    }
    goto connect_exit ;
  }

  conn->sin_addr = sremote.sin_addr ;
  strcpy(conn->name, dotquad) ;
  conn->mark = 0 ;
  conn->stream = lines ;
  conn->postok = postok ;

  { /* Look up last download times */
    char keyname[MAX_REGISTRYNAME] ;

    memset(&conn->newgroups, 0, sizeof(FILETIME)) ;

    strcpy(keyname, SERVICE_REGISTRY_KEY) ;
    strcat(keyname, "\\") ;
    strcat(keyname, dotquad) ;

    if ( !GetRegFiletime(keyname, NEWGROUPS_REGNAME, &conn->newgroups) )
      memset(&conn->newgroups, 0, sizeof(FILETIME)) ;
  }

  server->connection = conn ;

  debug(DEBUG_CONNECT, "Connected %s successfully\n", server->hostname) ;

  lines = socks = NULL ; /* Don't destroy these, they're now connected */

connect_exit:
  if ( socks )
    DestroyStream(socks) ;

  if ( lines )
    DestroyStream(lines) ;

  Unlock(SERVER_MUTEX) ;

  return conn ;
}

int DisconnectServer(Server *server)
{
  Connection *conn ;

  if ( !Lock(SERVER_MUTEX, INFINITE) )
    return FALSE ;

  debug(DEBUG_CONNECT, "Disconnecting server %s\n", server->hostname) ;

  conn = server->connection ;

  if ( conn ) {
    int servind ;

    for ( servind = 0 ; servind < nServers ; servind++ ) {
      Server *thisserv = serverTable[servind] ;

      if ( thisserv->connection == conn )
	thisserv->connection = NULL ;
    }

    while ( conn->stream != NULL )
      conn->stream = DestroyStream(conn->stream) ;

    free(conn->name) ;
    free(conn) ;
  }

  Unlock(SERVER_MUTEX) ;

  return FALSE ;
}

int ForAllServers(int (*fn)(Server *server, void *passthrough),
		  void *passthrough,
		  int online)
{
  int servind ;
  int result = TRUE ;

  /*
  if ( !Lock(SERVER_MUTEX, INFINITE) )
    return FALSE ;
    */

  for (servind = 0 ; result && servind < nServers ; servind++) {
    Server *thisserv = serverTable[servind] ;

    if ( online ) {      /* try to connect, continue if failed */
      if ( ConnectServer(thisserv) == NULL )
	continue ;
    }
    result = (*fn)(thisserv, passthrough) ;
  }

  /*
  Unlock(SERVER_MUTEX) ;
  */

  return result ;
}

/*
 * $Log: servers.c,v $
 * Revision 1.3  2000/01/14 21:22:34  angus
 * More changes to logging. Use new streams 1.31 ioctl to specify debugging info
 * for client & server connections. Make fmtout more secure against buffer
 * overruns, and add a new format specifier for binary strings with visible
 * control characters. Change logging interface to add a varargs log, and use
 * that for stream debugging. Modify debug flags, add more status returns.
 *
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
