/* slurp.c $Id: slurp.c,v 1.3 2000/01/14 21:22:34 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Manage downloading of news for NNTP buffer
 */

#include "nntp.h"
#include "messages.h"

static BOOL LinkFile(char *old, char *new)
{
  HANDLE fileh ;
  BOOL   result = FALSE;
  char oldname[MAX_PATH], newname[MAX_PATH] ;

  if ( _fullpath(oldname, old, MAX_PATH) == NULL ||
       _fullpath(newname, new, MAX_PATH) == NULL )
    return FALSE ;

  if ( (fileh = CreateFile(oldname, 0, 0, NULL,
			   OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS,
			   NULL)) != INVALID_HANDLE_VALUE ) {
    int wlen ;

    /* Confusingly, the "alternate" stream name field does not apply
       when restoring a hard link, and instead contains the actual
       stream data for the link (ie. the name of the link to create).
       The WIN32_STREAM_ID structure before the cStreamName field is the
       stream header, which is immediately followed by the stream data.  */

    struct {
      WIN32_STREAM_ID wid ;
      WCHAR wbuffer[MAX_PATH] ;	/* extra space for link name */
    } data;

    if ( (wlen = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, newname, -1,
				     data.wid.cStreamName, MAX_PATH)) != 0 ) {
      LPVOID context = NULL ;
      DWORD wbytes = 0 ;

      data.wid.dwStreamId = BACKUP_LINK ;
      data.wid.dwStreamAttributes = 0 ;
      data.wid.Size.LowPart = wlen * sizeof(WCHAR) ;
      data.wid.Size.HighPart = 0 ;
      data.wid.dwStreamNameSize = 0 ;

      if ( BackupWrite(fileh, (LPBYTE)&data,
		       offsetof(WIN32_STREAM_ID, cStreamName) + data.wid.Size.LowPart,
		       &wbytes, FALSE, FALSE, &context)
	   && BackupWrite(fileh, NULL, 0, &wbytes, TRUE, FALSE, &context) )
	result = TRUE;
    }

    CloseHandle(fileh);
  }

  return result ;
}

static BOOL LinkOrCopyFile(char *old, char *new)
{
  if ( !LinkFile(old, new) )
    return CopyFile(old, new, TRUE) ;
}

typedef struct {
  DWORD interval ;
  long inactivate ;
  long number ;
} SlurpArgs ;

int SlurpGroup(char *group, void *passthrough)
{
  SlurpArgs *args = passthrough ;
  FILETIME now, then ;
  Group data, *gptr ;
  Server *server ;
  Pattern *pattern ;
  Connection *conn ;
  Stream *dots = NULL ;
  int code = 0 ;
  long idalloc = 0, idused = 0 ;
  char *idlist = NULL ;
  int result = FALSE ;

  /* Test if ordered to stop */
  switch ( WaitForMultipleObjects(ALL_CONTROLS, hEvents[SIGNAL_SLURP], FALSE, 0) ) {
  case WAIT_OBJECT_0 + CONTROL_START:
  case WAIT_TIMEOUT:
    break ;
  default: /* Abort on client connect, quit, stop, or error */
    return FALSE ;
  }

  Log(LOG_STATUS, "Downloading group %s\n", group) ;

  if ( !Lock(SLURP_MUTEX, INFINITE) )
    goto slurp_nolocks ;

  if ( !Lock(GROUP_MUTEX, INFINITE) )
    goto slurp_groupfail ;

  if ( !Lock(SERVER_MUTEX, INFINITE) )
    goto slurp_serverfail ;

  GetSystemTimeAsFileTime(&now) ;
  then = TimeAdd(&now, -(long)args->interval) ;

  /* Ignore if group not found, is not active, or if NEWNEWS issued
     recently */
  if ( (gptr = GroupDBGet(group)) == NULL || gptr->state != GROUP_ACTIVE ||
       CompareFileTime(&gptr->newnews, &then) >= 0 ) {
    result = TRUE ;
    goto slurp_exit ;
  }

  data = *gptr ;

  /* Check if we can inactivate group */
  {
    FILETIME client, selected ;

    /* Last time group was read, plus inactivation timout */
    selected = TimeAdd(&gptr->selected, args->inactivate) ;

    /* Last time client was quit, plus inactivation timeout */
    if ( GetRegFiletime(SERVICE_REGISTRY_KEY, LAST_CLIENT, &client) == NULL )
      memset(&client, 0, sizeof(FILETIME)) ;

    /* If been getting new news and client has been used recently, but hasn't
       selected group, inactivate it */

    if ( CompareFileTime(&gptr->newnews, &selected) > 0 &&
	 CompareFileTime(&client, &selected) > 0 ) {
      data.state = GROUP_INACTIVE ;
      message(EVENTLOG_INFORMATION_TYPE, MSG_INACTIVATE, group, NULL) ;
      goto slurp_change ;
    }
  }

  if ( !FindGroupServer(group, &server, &pattern) ) {
    debug(DEBUG_SLURPTHREAD, "No server for %s\n", group) ;
    goto slurp_exit ;
  }

  if ( (conn = ConnectServer(server)) == NULL )
    goto slurp_exit ;

  debug(DEBUG_SLURPTHREAD, "Downloading group %s from %s\n", group,
	server->hostname) ;

  /* Check group exists */
  if ( SendStream(conn->stream, "GROUP %s\r\n", group) >= 0 &&
       (code = Response(conn->stream)) >= 0 ) {
    if ( code == 411 ) {	/* No such group */
      char reason[BUFSIZ] ;
      if ( Reason(conn->stream, reason, BUFSIZ) > 0 )
        debug(DEBUG_SLURPTHREAD, "FROM Server: %s\n", reason) ;
      data.state = GROUP_BOGUS ;
      goto slurp_change ;
    }
  } else {
    debug(DEBUG_SLURPTHREAD, "No server response checking %s from %s exists\n",
	  group, server->hostname) ;
    DisconnectServer(server) ;
    goto slurp_exit ;
  }

  /* Issue NEWNEWS to server, then wait for response, read
     message-ids and store them, then read messages by id, and add to
     group */
  Log(LOG_STATUS, "NEWNEWS group %s\n", group) ;

  if ( SendStream(conn->stream, "NEWNEWS %s %t GMT\r\n", group,
                  &gptr->newnews) < 0 ||
       (code = Response(conn->stream)) <= 0 ) {
    debug(DEBUG_SLURPTHREAD, "No server response getting %s from %s\n",
	  group, server->hostname) ;
    DisconnectServer(server) ;
    goto slurp_exit ;
  }

  if ( code != 230 ) {
    char reason[BUFSIZ] ;

    if ( Reason(conn->stream, reason, BUFSIZ) > 0 )
      debug(DEBUG_SLURPTHREAD, "FROM Server: %s\n", reason) ;

    debug(DEBUG_SLURPTHREAD, "No news in %s, server %s responded %d\n",
	  group, server->hostname, code) ;
    goto slurp_exit ;
  }

  if ( (dots = MakeDotStream(conn->stream)) == NULL ) {
    debug(DEBUG_SLURPTHREAD, "Can't create dot stream for %s from %s\n",
	  group, server->hostname) ;
    goto slurp_exit ;
  }

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_SERVER, DEBUG_DOWNLOAD) ;

  /* Read message-id responses until <CRLF>.<CRLF> */
  for (;;) {
    char *line, *msgid = NULL ;
    long rbytes ;

    if ( (rbytes = ReadStream(dots, &line, 0)) < 0 ) {
      debug(DEBUG_SLURPTHREAD, "Can't read %s message-ids from %s\n",
	    group, server->hostname) ;
      DisconnectServer(server) ;
      goto slurp_free ;
    }

    if ( rbytes == 0 )	/* Dot line */
      break ;

    while ( rbytes ) {
      char ch = *line++ ;

      rbytes-- ;

      switch ( ch ) {
      case '<':
	if ( msgid != NULL )
	  debug(DEBUG_SLURPTHREAD, "Unterminated Message-ID\n") ;
	msgid = line ; /* char after < */
	break ;
      case '>':
	if ( msgid != NULL ) {
	  int len = line - msgid ;

	  /* Add to end of idlist */
	  if ( idlist == NULL || idused + len >= idalloc ) {
	    idalloc += 2048 ;
	    idlist = realloc(idlist, idalloc) ;
	    if ( idlist == NULL ) {
	      message(EVENTLOG_ERROR_TYPE, MSG_MEMORY, "NEWNEWS", NULL) ;
	      goto slurp_exit ;
	    }
	  }

	  /* Save away message-id */
	  strncpy(idlist + idused, msgid, len) ;
	  idused += len ;
	  idlist[idused - 1] = '\0' ;

	  msgid = NULL ; /* Look for next message-id */
	} else
	  debug(DEBUG_SLURPTHREAD, "Uninitialised Message-ID\n") ;
	break ;
      case ' ':
      case '\t':
      case '\r':
      case '\n':
	if ( msgid != NULL )
	  debug(DEBUG_SLURPTHREAD, "Unterminated Message-ID\n") ;
	msgid = NULL ;
      }
    }
  }

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_DOWNLOAD, DEBUG_SERVER) ;

  /* Scan message-id list, looking for existing */
  if ( idlist ) {
    char *msgid = idlist ;

    idlist[idused] = '\0' ; /* Terminate list of ids */

    while ( *msgid ) {
      int len = strlen(msgid) ;

      switch ( WaitForMultipleObjects(ALL_CONTROLS, hEvents[SIGNAL_SLURP], FALSE, 0) ) {
      case WAIT_OBJECT_0 + CONTROL_START:
      case WAIT_TIMEOUT:
        break ;
      default: /* Abort on client connect, quit, stop, or error */
        DisconnectServer(server) ;
        goto slurp_free ; /* Abort */
      }

      if ( Lock(ARTICLE_MUTEX, INFINITE) ) {
	Article *art = ArticleDBGet(msgid) ;

	debug(DEBUG_SLURPTHREAD, "Retrieving %s\n", msgid) ;

	if ( art != NULL ) { /* We have this article already */
	  debug(DEBUG_SLURPTHREAD, "Have %s\n", msgid) ;

	  if ( !ArticleInGroup(art, group) ) {
	    char filename[MAX_PATH], oldname[MAX_PATH] ;

	    debug(DEBUG_SLURPTHREAD, "Not in group %s\n", group) ;

	    /* if not in group, copy or link file to this dir */
	    if ( ArticleFileName(group, data.high + 1, filename, MAX_PATH) &&
		 MessageFileName(art->names, oldname, MAX_PATH) &&
		 MakeArticleDirs(filename) &&
		 LinkOrCopyFile(oldname, filename) &&
		 (art = AddArticleName(msgid, group, data.high + 1, art)) != NULL ) {
	      debug(DEBUG_SLURPTHREAD, "Article <%s> copied from %s\n",
		    msgid, art->names) ;
	      data.high++ ;
	      data.articles++ ;
	    }
	  }
	} else {  /* No such message, download it */
	  Stream *files = NULL ;

          Log(LOG_STATUS, "Downloading group %s article %s\n", group, msgid) ;

	  GetSystemTimeAsFileTime(&server->connecttime) ;
	  if ( IoctlStream(dots, STREAM_CLEAREOF) < 0 ||
	       SendStream(conn->stream, "ARTICLE <%s>\r\n", msgid) < 0 ||
	       (code = Response(conn->stream)) <= 0 ) {
	    debug(DEBUG_SLURPTHREAD, "No server response ARTICLE <%s> from %s\n",
		  msgid, server->hostname) ;
	    DisconnectServer(server) ;
	    goto slurp_free ;
	  }

	  if ( code == 220 ) {
	    char filename[MAX_PATH] ;
	    long artsize = 0 ;

	    debug(DEBUG_SLURPTHREAD, "Article <%s> unknown, downloading from %s\n",
		  msgid, server->hostname) ;
	    if ( ArticleFileName(group, data.high + 1, filename, MAX_PATH) == NULL ||
		 !MakeArticleDirs(filename) ||
		 (files = MakeFileStream(filename, FILESTREAM_WRITE, stream_error)) == NULL ) {
	      debug(DEBUG_SLURPTHREAD, "Can't create download stream for <%s>\n",
		    msgid) ;
	      DisconnectServer(server) ;
	      goto slurp_free ;
	    }

            IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_SERVER, DEBUG_DOWNLOAD) ;

	    /* Read from server to <LF>.<CRLF>, store in posting file. */

	    for (;;) {
	      char *line ;
	      long rbytes ;

	      if ( (rbytes = ReadStream(dots, &line, 0)) < 0 ) {
		debug(DEBUG_SLURPTHREAD, "Download error for <%s>\n",
		      msgid) ;
		DestroyStream(files) ;
		DeleteFile(filename) ;
		DisconnectServer(server) ;
		goto slurp_free ;
	      }

	      if ( rbytes == 0 )	/* Read article */
		break ;

	      artsize += rbytes ;

	      if ( !WriteStreamAll(files, line, rbytes) ) {
		debug(DEBUG_SLURPTHREAD, "Store error for <%s>\n", msgid) ;
		DestroyStream(files) ;
		DeleteFile(filename) ;
		DisconnectServer(server) ;
		goto slurp_free ;
	      }
	    }

	    DestroyStream(files) ;

            IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_DOWNLOAD, DEBUG_SERVER) ;

	    /* got message */
	    if ( (art = NewArticle(msgid, ARTICLE_BODY, artsize)) != NULL &&
		 AddArticleName(msgid, group, data.high + 1, art) != NULL ) {
	      debug(DEBUG_SLURPTHREAD, "Article <%s> downloaded from %s\n",
		    msgid, server->hostname) ;
	      data.articles++ ;
	      data.high++ ;
	      args->number++ ;
	    } else /* All that for nothing */
	      DeleteFile(filename) ;
	  } else {
            char reason[BUFSIZ] ;
            if ( Reason(conn->stream, reason, BUFSIZ) > 0 )
              debug(DEBUG_SLURPTHREAD, "FROM Server: %s\n", reason) ;
          }
	}

	Unlock(ARTICLE_MUTEX) ;
      }
      msgid += len + 1 ;
    }
  }

  /* Write back changes to group */
  data.newnews = now ;

slurp_change:
  (void)GroupDBPut(group, &data) ;
  result = TRUE ;
	
  Log(LOG_STATUS, "Downloaded group %s\n", group) ;

slurp_free:
  if ( idlist )
    free(idlist) ;

  IoctlStream(conn->stream, STREAM_DEBUG, DEBUG_DOWNLOAD, DEBUG_SERVER) ;

  if ( dots )
    DestroyStream(dots) ;

slurp_exit:
  Unlock(SERVER_MUTEX) ;

slurp_serverfail:
  Unlock(GROUP_MUTEX) ;

slurp_groupfail:
  Unlock(SLURP_MUTEX) ;

slurp_nolocks:
  Sleep(0) ; /* Yield control to other processes if possible; be nice */

  return result ;
}

DWORD WINAPI SlurpThread(LPVOID info)
{
  DWORD wait, retry, interval, inactivate ;
  FILETIME then, now ;

  ThreadType(SLURP_THREAD) ;

  if ( (debuglevel & DEBUG_NOSLURP) != 0 )
    return 1 ;

  debug(DEBUG_SLURPTHREAD, "Started download thread\n") ;

  wait = GetRegDWORD(SERVICE_REGISTRY_KEY, DOWNLOAD_WAIT_NAME) ;
  if ( wait <= 0 )	/* Default if silly value */
    wait = DOWNLOAD_WAIT_DEFAULT ;

  retry = GetRegDWORD(SERVICE_REGISTRY_KEY, DOWNLOAD_RETRY_NAME) ;
  if ( retry <= 0 )	/* Default if silly value */
    retry = DOWNLOAD_RETRY_DEFAULT ;
  debug(DEBUG_SLURPTHREAD, "Download retry time is %d seconds\n", retry) ;

  interval = GetRegDWORD(SERVICE_REGISTRY_KEY, DOWNLOAD_INTERVAL_NAME) ;
  if ( interval <= 0 )	/* Default if silly value */
    interval = DOWNLOAD_INTERVAL_DEFAULT ;
  debug(DEBUG_SLURPTHREAD, "Download interval time is %d seconds\n", interval) ;

  if ( GetRegFiletime(SERVICE_REGISTRY_KEY, LAST_DOWNLOAD, &then) ) {
    long diff ;

    GetSystemTimeAsFileTime(&now) ;
    diff = TimeDiff(&then, &now) ;
    if ( diff + wait < interval ) /* Take max of remaining interval & wait */
      wait = interval - diff ;
  }

  debug(DEBUG_SLURPTHREAD, "Download wait time is %d seconds\n", wait) ;

  inactivate = GetRegDWORD(SERVICE_REGISTRY_KEY, DOWNLOAD_INACTIVATE_NAME) ;

  debug(DEBUG_SLURPTHREAD, "Download inactivate interval = %lu\n", inactivate) ;

  if ( inactivate <= 0 ) /* Restrict inactivation wait */
    inactivate = DOWNLOAD_INACTIVATE_DEFAULT ;

  SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL) ;

  for (;;) {
    DWORD delay = wait ;
    SlurpArgs args ;
    char status[20] ;

    args.number = 0 ;
    args.inactivate = inactivate ;
    args.interval = interval ;

    GetSystemTimeAsFileTime(&now) ;
    then = TimeAdd(&now, delay) ;

    do {
      if ( suspended[SIGNAL_SLURP] ) {
        LogString(LOG_STATUS, "Downloading suspended\n") ;
      } else if ( FileTimeToLocalFileTime(&then, &now) ) {
        Log(LOG_STATUS, "Next download is at %t\n", &now) ;
      }

      switch ( WaitForMultipleObjects(LOOP_CONTROLS, hEvents[SIGNAL_SLURP],
                                      FALSE, delay * 1000) ) {
      case WAIT_TIMEOUT:
      case WAIT_OBJECT_0 + CONTROL_START:
        delay = 0 ;
        break ;
      case WAIT_OBJECT_0 + CONTROL_QUIT:
        goto slurp_quit ;
      default:
        GetSystemTimeAsFileTime(&now) ;
        delay = TimeDiff(&now, &then) ;
        break ;
      }
    } while ( delay > 0 ) ;

    if ( suspended[SIGNAL_SLURP] )
      continue ;

    debug(DEBUG_SLURPTHREAD, "Starting download\n") ;

    if ( ForAllGroups(SlurpGroup, &args) ) {
      GetSystemTimeAsFileTime(&now) ;
      SetRegFiletime(SERVICE_REGISTRY_KEY, LAST_DOWNLOAD, &now) ;
      wait = interval ;
      strcpy(status, "OK") ;
    } else {
      wait = retry ;
      sprintf(status, "%d", GetLastError()) ;
    }

    debug(DEBUG_SLURPTHREAD, "Downloaded %d articles, next attempt is in %lu seconds\n",
	  args.number, wait) ;

    if ( args.number != 0 ) {
      char strnum[20] ;
      sprintf(strnum, "%d", args.number) ;
      message(EVENTLOG_INFORMATION_TYPE, MSG_DOWNLOAD_DONE, strnum, status,
	      NULL) ;
    }
  }

slurp_quit:
  LogString(LOG_STATUS, "Download thread terminated\n") ;

  return 0 ;
}

/*
 * $Log: slurp.c,v $
 * Revision 1.3  2000/01/14 21:22:34  angus
 * More changes to logging. Use new streams 1.31 ioctl to specify debugging info
 * for client & server connections. Make fmtout more secure against buffer
 * overruns, and add a new format specifier for binary strings with visible
 * control characters. Change logging interface to add a varargs log, and use
 * that for stream debugging. Modify debug flags, add more status returns.
 *
 * Revision 1.2  2000/01/14 03:40:04  angus
 * Many changes; added named pipe logging, sending output to client debug
 * process for in-situ service debugging. Use this mechanism for queue status,
 * and enable after-the-fact debugging by saving 100 logged messages. Add
 * symbollic names for debug flags, and Mutex debugging.
 *
 * Revision 1.1.1.1  1999/10/25 07:49:57  angus
 * Initial CVS import
 *
 */
