/* Common.c
 *
 * Copyright (C) Angus J. C. Duggan, 1996
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Common functions for registry munging programs
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <search.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <assert.h>

#include "common.h"

/* General allocation function. Error if fail */
void *xalloc(void *ptr, size_t size)
{
  void *result = realloc(ptr, size) ;

  if ( result == NULL ) {
    fprintf(stderr, "%s: out of memory!\n", program) ;
    fflush(stderr) ;
    exit(1) ;
  }

  return result ;
}

/* Key and value allocation and deallocation. Name and data are packed into
   address pointer space if short enough to fit. */
static KeyValue *NewKey(char *name, DWORD namelen, char *class, DWORD classlen,
			Context *parent, KV_State state)
{
  Key *new = xalloc(NULL, sizeof(Key)) ;

  new->base.type = REG_KEY ;
  new->base.parent = parent ;
  new->base.namelen = namelen ;
  if ( !KVNamePacked(new) )
    new->base.name = xalloc(NULL, namelen + 1) ;
  strcpy(KVName(new), name) ;
  new->base.datalen = classlen ;
  if ( !KVDataPacked(new) )
    new->base.data = xalloc(NULL, classlen) ;
  if ( classlen > 0 )
    memcpy(KVData(new), class, classlen) ;
  new->base.next = NULL ;
  new->base.state = state ;

  return (KeyValue *)new ;
}

static KeyValue *NewValue(char *name, DWORD namelen, unsigned char *data,
			  DWORD datalen, Context *parent, DWORD type,
			  KV_State state)
{
  Value *new = xalloc(NULL, sizeof(Value)) ;

  new->base.type = type ;
  new->base.parent = parent ;
  new->base.namelen = namelen ;
  if ( !KVNamePacked(new) )
    new->base.name = xalloc(NULL, namelen + 1) ;
  strcpy(KVName(new), name) ;
  new->base.datalen = datalen ;
  if ( !KVDataPacked(new) )
    new->base.data = xalloc(NULL, datalen) ;
  if ( datalen > 0 )
    memcpy(KVData(new), data, datalen) ;
  new->base.state = state ;
  new->base.next = NULL ;

  return (KeyValue *)new ;
}

void FreeKeyVal(KeyValue *kv)
{
  if ( !KVNamePacked(kv) )
    free(kv->base.name) ;
  if ( !KVDataPacked(kv) )
    free(kv->base.data) ;
  /* parent field belongs to someone else, don't free it */
  free(kv) ;
}

static Context *NewContext(Context *parent, int namesize)
{
  Context *new = xalloc(NULL, sizeof(Context)) ;

  new->parent = parent ;
  if ( namesize > 0 )
    new->name = xalloc(NULL, namesize) ;
  else
    new->name = NULL ;
  new->list = NULL ;
  new->hkey = (HKEY)NULL ;

  return new ;
}

static Context *FreeContext(Context *context)
{
  Context *parent = context->parent ;

  if( context->name )
    free(context->name) ;

  while ( context->list ) {
    KeyValue *next = context->list->base.next ;
    FreeKeyVal(context->list) ;
    context->list = next ;
  }

  if ( context->hkey != (HKEY)NULL )
    RegCloseKey(context->hkey) ;

  free(context) ;

  return parent ;
}

char *KeyValName(KeyValue *kv, char *str, int *size)
{
  char *pname = kv->base.parent->name ;
  int plen = strlen(pname) ;
  int need = plen + 2 + kv->base.namelen ;

  if ( size == NULL || *size < need ) {
    str = xalloc(str, need) ;
    if ( size )
      *size = need ;
  }

  strcpy(str, pname) ;
  if ( kv->base.type != REG_KEY || kv->base.namelen > 0 ) {
    str[plen++] = REGISTRY_SEPARATOR ;
    strcpy(&str[plen], KVName(kv)) ;
  }

  return str ;
}

/* Escaped string output */
static void init_translate(char *ctrans[])
{
  int index ;

  ctrans['\\'] = "\\\\" ;
  ctrans['"'] = "\\\"" ;
  ctrans['\''] = "\\'" ;
  ctrans['\n'] = "\\n" ;
  ctrans['\r'] = "\\r" ;
  ctrans['\t'] = "\\t" ;

  for ( index = 0 ; index < 32 ; index++ ) {
    static char lower[32][5] ;

    sprintf(lower[index], "\\%03o", index) ;
    ctrans[index] = lower[index] ;
  }

  for ( index = 127 ; index < 256 ; index++ ) {
    static char upper[129][5] ;

    sprintf(upper[index - 127], "\\%03o", index) ;
    ctrans[index] = upper[index - 127] ;
  }
}

static void translate(FILE *out, char *string, long length)
{
  static int initialised = 0 ;
  static char *ctrans[256] ;

  if ( !initialised )
    init_translate(ctrans) ;

  while ( length-- > 0 ) {
    int ch = (unsigned char)*string++ ;
    if ( ctrans[ch] )
      fputs(ctrans[ch], out) ;
    else
      fputc(ch, out) ;
  }
}


/* Other output support */
static void write_string(FILE *out, char *string, long length)
{
  char quote = '"' ;

  if ( length == 0 || string[length-1] != '\0' )
    quote = '\'' ;
  else
    length -= 1 ;

  fputc(quote, out) ;
  translate(out, string, length) ;
  fputc(quote, out) ;
}

static void write_name(FILE *out, Context *parent, char *string, long length)
{
  fputc('"', out) ;
  translate(out, parent->name, strlen(parent->name)) ;
  if ( string ) {
    fputc(REGISTRY_SEPARATOR, out) ;
    fputc(REGISTRY_SEPARATOR, out) ;
    translate(out, string, length) ;
  }
  fputc('"', out) ;
}

/* String extension */
void *extend(void *vector, DWORD *allocated, size_t size)
{
  *allocated = *allocated * 2 + 16 ;
  return xalloc(vector, *allocated * size) ;
}

/* Registry support functions */
static int SortKeyValues(const void *one, const void *two)
{
  char *name1 = KVName(*((KeyValue **)one)) ;
  char *name2 = KVName(*((KeyValue **)two)) ;

  return _stricmp(name1, name2) ;
}


/* Output assistance functions */
static char *nlprefix = NULL ;

static void write_reg_binary(FILE *out, unsigned char *value, DWORD valuelen)
{
  long index ;

  static char hextable[] = "0123456789abcdef" ;

  fprintf(out, "%d", valuelen) ;
  for ( index = 0 ; index < valuelen ; index++ ) {
    int byte = value[index] ;

    if ( (index % 38) == 0 ) {
      fputc('\n', out) ;
      if ( nlprefix )
	fputs(nlprefix, out) ;
      fputs("  ", out) ;
    }

    fputc(hextable[byte >> 4], out) ;
    fputc(hextable[byte & 0xf], out) ;
  }
}

static void write_reg_dword(FILE *out, unsigned char *value, DWORD valuelen)
{
  DWORD word = *(DWORD *)value ;
  fprintf(out, "0x%x", word) ;
}

static void write_reg_dword_little_endian(FILE *out, unsigned char *value, DWORD valuelen)
{
  DWORD word = *(DWORD *)value ;
  fprintf(out, "0x%x", word) ;
}

static void write_reg_dword_big_endian(FILE *out, unsigned char *value, DWORD valuelen)
{
  DWORD word = (value[0] << 24) | (value[1] << 16) | (value[2] << 8) | value[3] ;
  fprintf(out, "0x%x", word) ;
}

static void write_reg_expand_sz(FILE *out, unsigned char *value, DWORD valuelen)
{
  write_string(out, (char *)value, valuelen) ;
}

static void write_reg_sz(FILE *out, unsigned char *value, DWORD valuelen)
{
  write_string(out, (char *)value, valuelen) ;
}

static void write_reg_link(FILE *out, unsigned char *value, DWORD valuelen)
{
  write_reg_binary(out, value, valuelen) ;
}

static void write_reg_multi_sz(FILE *out, unsigned char *value, DWORD valuelen)
{
  char *str, *scan ;
  long number = 0 ;
  long len ;
  char last = '\0' ;

  /* check if well formatted */
  for ( str = (char *)value, len = 0 ; len < valuelen ; str++, len++ ) {
    if ( *str == '\0' )
      number++ ;	/* number of strings */
    last = *str ;
  }

  if ( last != '\0' )
    number++ ;

  fprintf(out, "%d", number) ;

  for ( scan = str = (char *)value, len = 0 ; len < valuelen ; str = scan ) {
    while ( len++ < valuelen && *scan++ != '\0' ) /* search for end */ ;

    fputc('\n', out) ;
    if ( nlprefix )
      fputs(nlprefix, out) ;
    fputs("  ", out) ;
    write_string(out, str, scan - str) ;
  }
}

static void write_reg_none(FILE *out, unsigned char *value, DWORD valuelen)
{
  ; /* Do nothing */
}

static void write_reg_resource_list(FILE *out, unsigned char *value, DWORD valuelen)
{
  write_reg_binary(out, value, valuelen) ;
}

static void write_reg_full_resource_descriptor(FILE *out, unsigned char *value, DWORD valuelen)
{
  write_reg_binary(out, value, valuelen) ;
}

static void write_reg_key(FILE *out, unsigned char *value, DWORD valuelen)
{
  write_string(out, (char *)value, valuelen) ;
}

/* Input assistance functions */
static char *need_buffer(FILE *in, char *buffer, long bufsize, char *scan,
			 long needs)
{
  assert(needs < bufsize) ;	/* needs can't exceed buffer size */

  /* Don't get extra if end of line already found */
  if ( scan + needs > buffer + bufsize && strchr(scan, '\n') == NULL ) {
    long len = buffer + bufsize - scan - 1 ; /* -1 to ignore null character */

    assert(len >= 0) ;

    if ( len > 0 )
      memmove(buffer, scan, len) ;

    scan = fgets(buffer + len, bufsize - len, in) ;
  }

  return scan ;
}

static char *skip_white(FILE *in, char *buffer, int bufsize, char *scan)
{
  while ( isspace(*scan) ) {	/* skip optional space */
    if ( ++scan == buffer + bufsize &&
	 (scan = fgets(buffer, bufsize, in)) == NULL ) {
      fprintf(stderr, "%s: malformed line %s\n", program, buffer) ;
      return NULL ;
    }
  }

  return scan ;
}

static char *read_string(FILE *in, char *buffer, int bufsize, char *scan,
			 char **name, DWORD *namelen, DWORD *actuallen)
{
  int len = 0, more = 1 ;
  int ch ;
  char *lname = *name ;
  DWORD lnamelen = *namelen ;
  char quote ;

  if ( (scan = skip_white(in, buffer, BUFSIZ, scan)) == NULL ||
       (*scan != '"' && *scan != '\'') ) {
    fprintf(stderr, "%s: bad string %s\n", program, scan) ;
    return NULL ;
  }

  quote = *scan++ ;

  do {
    /* Reserve enough space for longest escape sequence */
    if ( (scan = need_buffer(in, buffer, bufsize, scan, 4)) == NULL )
      return NULL ;

    switch ( (ch = *scan++) ) {
    case '\0': case '\n':
      fprintf(stderr, "%s: bad string %s\n", program, buffer) ;
      return NULL ;
    case '\'':
    case '"':
      if ( quote != ch ) {
	if ( len + 1 > lnamelen )
	  lname = extend(lname, &lnamelen, sizeof(char)) ;
	lname[len++] = ch ;
      } else
	more = 0 ;	/* Found end of string */
      break ;
    case '\\':		/* Deal with escaped characters */
      switch ( (ch = *scan++) ) {
      case 'n':
	ch = '\n' ;
	break ;
      case 'r':
	ch = '\r' ;
	break ;
      case 't':
	ch = '\t' ;
	break ;
      case '0': case '1': case '2': case '3':
      case '4': case '5': case '6': case '7':
	ch = ch - '0' ;
	if ( *scan >= '0' && *scan <= '7' )
	  ch = (ch << 3) + *scan++ - '0' ;
	if ( *scan >= '0' && *scan <= '7' )
	  ch = (ch << 3) + *scan++ - '0' ;
	break ;
      }
      /* Fallthrough */
    default:
      if ( len + 1 > lnamelen )
	lname = extend(lname, &lnamelen, sizeof(char)) ;
      lname[len++] = ch ;
    }
  } while ( more ) ;

  if ( quote == '"' ) {
    if ( len + 1 > lnamelen )	/* extra one for null char */
      lname = extend(lname, &lnamelen, sizeof(char)) ;
    lname[len++] = '\0' ;
  }

  *name = lname ;
  *namelen = lnamelen ;
  *actuallen = len ;

  return scan ;
}

static char *read_decimal(FILE *in, char *buffer, int bufsize, char *scan,
			  DWORD *value)
{
  DWORD number = 0 ;

  if ( (scan = skip_white(in, buffer, BUFSIZ, scan)) == NULL ||
       !isdigit(*scan) ) {
    fprintf(stderr, "%s: bad decimal number %s\n", program, scan) ;
    return NULL ;
  }

  while ( isdigit(*scan) ) {
    number *= 10 ;
    number += *scan++ - '0' ;

    /* Make sure next character is available */
    if ( (scan = need_buffer(in, buffer, bufsize, scan, 1)) == NULL )
      return NULL ;
  }

  *value = number ;

  return scan ;
}

static int hexdigit[] = {
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, -1, -1, -1, -1, -1, -1,
  -1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, 10, 11, 12, 13, 14, 15, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
} ;

static char *read_hex(FILE *in, char *buffer, int bufsize, char *scan,
		      DWORD *value)
{
  DWORD number = 0 ;

  if ( (scan = skip_white(in, buffer, BUFSIZ, scan)) == NULL ||
       (scan = need_buffer(in, buffer, bufsize, scan, 3)) == NULL ||
       *scan++ != '0' || *scan++ != 'x' ) {
    fprintf(stderr, "%s: bad hex number %s\n", program, scan) ;
    return NULL ;
  }

  while ( hexdigit[*scan] >= 0 ) {
    number <<= 4 ;
    number |= hexdigit[*scan++] ;

    /* Make sure next character is available */
    if ( (scan = need_buffer(in, buffer, bufsize, scan, 1)) == NULL )
      return NULL ;
  }

  *value = number ;

  return scan ;
}

static char *read_reg_binary(FILE *in, char *buffer, int bufsize, char *scan,
			     unsigned char **data, DWORD *datalen,
			     DWORD *actuallen)
{
  DWORD len ;

  if ( (scan = read_decimal(in, buffer, bufsize, scan, &len)) != NULL ) {
    int byte = 0, digits = 0 ;
    unsigned char *fill = *data ;

    *actuallen = len ;

    while ( *datalen < len )
      *data = fill = extend(*data, datalen, sizeof(char)) ;

    while ( len > 0 ) {
      if ( (scan = fgets(buffer, bufsize, in)) == NULL ) {
	fprintf(stderr, "%s: not enough binary data\n", program) ;
	return NULL ;
      }
      while ( *scan && len > 0 ) {
	if ( hexdigit[*scan] >= 0 ) {
	  byte = (byte << 4) | hexdigit[*scan] ;
	  if ( ++digits == 2 ) {
	    *fill++ = byte ;
	    byte = digits = 0 ;
	    len-- ;
	  }
	}
	scan++ ;
      }
    }
  } else
    *actuallen = 0 ;

  return scan ;
}

static char *read_reg_dword(FILE *in, char *buffer, int bufsize, char *scan,
			    unsigned char **data, DWORD *datalen,
			    DWORD *actuallen)
{
  if ( *datalen < 4 )
    *data = extend(*data, datalen, sizeof(char)) ;

  if ( (scan = read_hex(in, buffer, bufsize, scan, (DWORD *)*data)) != NULL ) {
    *actuallen = 4 ;
  } else
    *actuallen = 0 ;

  return scan ;
}

static char *read_reg_dword_little_endian(FILE *in, char *buffer, int bufsize,
					  char *scan, unsigned char **data,
					  DWORD *datalen, DWORD *actuallen)
{
  if ( *datalen < 4 )
    *data = extend(*data, datalen, sizeof(char)) ;

  if ( (scan = read_hex(in, buffer, bufsize, scan, (DWORD *)*data)) != NULL ) {
    *actuallen = 4 ;
  } else
    *actuallen = 0 ;

  return scan ;
}

static char *read_reg_dword_big_endian(FILE *in, char *buffer, int bufsize,
				       char *scan, unsigned char **data,
				       DWORD *datalen, DWORD *actuallen)
{
  if ( *datalen < 4 )
    *data = extend(*data, datalen, sizeof(char)) ;

  if ( (scan = read_hex(in, buffer, bufsize, scan, (DWORD *)*data)) != NULL ) {
    unsigned char temp = (*data)[0] ;
    (*data)[0] = (*data)[3] ;
    (*data)[3] = temp ;
    temp = (*data)[1] ;
    (*data)[1] = (*data)[2] ;
    (*data)[2] = temp ;
    *actuallen = 4 ;
  } else
    *actuallen = 0 ;

  return scan ;
}

static char *read_reg_expand_sz(FILE *in, char *buffer, int bufsize,
				char *scan, unsigned char **data,
				DWORD *datalen, DWORD *actuallen)
{
  return read_string(in, buffer, bufsize, scan, (char **)data, datalen,
		     actuallen) ;
}

static char *read_reg_sz(FILE *in, char *buffer, int bufsize, char *scan,
			 unsigned char **data, DWORD *datalen,
			 DWORD *actuallen)
{
  return read_string(in, buffer, bufsize, scan, (char **)data, datalen,
		     actuallen) ;
}

static char *read_reg_link(FILE *in, char *buffer, int bufsize, char *scan,
			   unsigned char **data, DWORD *datalen,
			   DWORD *actuallen)
{
  return read_reg_binary(in, buffer, bufsize, scan, data, datalen, actuallen) ;
}

static char *read_reg_multi_sz(FILE *in, char *buffer, int bufsize, char *scan,
			       unsigned char **data, DWORD *datalen,
			       DWORD *actuallen)
{
  DWORD nstr ;
  DWORD multilen, totalsize = 0 ;
  static char *multibuf = NULL ;
  static DWORD multibuflen = 0 ;

  if ( (scan = read_decimal(in, buffer, bufsize, scan, &nstr)) != NULL ) {
    while ( nstr-- > 0 ) {
      if ( (scan = fgets(buffer, bufsize, in)) == NULL ) {
	fprintf(stderr, "%s: not enough strings\n", program) ;
	return NULL ;
      }

      if ( *scan == '+' || *scan == '!' || *scan == '-' )
	scan++ ;	/* skip diff type */

      if ( read_string(in, buffer, bufsize, scan, &multibuf, &multibuflen,
		       &multilen) == NULL )
	return NULL ;

      while ( *datalen < totalsize + multilen )
	*data = extend(*data, datalen, sizeof(char)) ;

      memcpy(*data + totalsize, multibuf, multilen) ;
      totalsize += multilen ;
    }

    *actuallen = totalsize ;
  } else
    *actuallen = 0 ;

  return scan ;
}

static char *read_reg_none(FILE *in, char *buffer, int bufsize, char *scan,
			   unsigned char **data, DWORD *datalen,
			   DWORD *actuallen)
{
  *actuallen = 0 ;

  return scan ;
}

static char *read_reg_resource_list(FILE *in, char *buffer, int bufsize,
				    char *scan, unsigned char **data,
				    DWORD *datalen, DWORD *actuallen)
{
  return read_reg_binary(in, buffer, bufsize, scan, data, datalen, actuallen) ;
}

static char *read_reg_full_resource_descriptor(FILE *in, char *buffer,
					       int bufsize, char *scan,
					       unsigned char **data,
					       DWORD *datalen,
					       DWORD *actuallen)
{
  return read_reg_binary(in, buffer, bufsize, scan, data, datalen, actuallen) ;
}

static char *read_reg_key(FILE *in, char *buffer, int bufsize, char *scan,
			  unsigned char **data, DWORD *datalen,
			  DWORD *actuallen)
{
  return read_string(in, buffer, bufsize, scan, (char **)data, datalen,
		     actuallen) ;
}

static struct ValueTable {
  DWORD type ;
  char *name ;
  int namelen ;
  void (*write)(FILE *out, unsigned char *, DWORD) ;
  char *(*read)(FILE *in, char *, int, char *, unsigned char **, DWORD *, DWORD *) ;
} valuetable[] = {
  {REG_BINARY, "REG_BINARY", 10,
   write_reg_binary, read_reg_binary},
  {REG_DWORD, "REG_DWORD", 9,
   write_reg_dword, read_reg_dword},
  {REG_DWORD_LITTLE_ENDIAN, "REG_DWORD_LITTLE_ENDIAN", 23,
   write_reg_dword_little_endian, read_reg_dword_little_endian},
  {REG_DWORD_BIG_ENDIAN, "REG_DWORD_BIG_ENDIAN", 20,
   write_reg_dword_big_endian, read_reg_dword_big_endian},
  {REG_EXPAND_SZ, "REG_EXPAND_SZ", 13,
   write_reg_expand_sz, read_reg_expand_sz},
  {REG_SZ, "REG_SZ", 6,
   write_reg_sz, read_reg_sz},
  {REG_LINK, "REG_LINK", 8,
   write_reg_link, read_reg_link},
  {REG_MULTI_SZ, "REG_MULTI_SZ", 12,
   write_reg_multi_sz, read_reg_multi_sz},
  {REG_NONE, "REG_NONE", 8,
   write_reg_none, read_reg_none},
  {REG_RESOURCE_LIST, "REG_RESOURCE_LIST", 17,
   write_reg_resource_list, read_reg_resource_list},
  {REG_FULL_RESOURCE_DESCRIPTOR, "REG_FULL_RESOURCE_DESCRIPTOR", 28,
   write_reg_full_resource_descriptor, read_reg_full_resource_descriptor},
  {REG_KEY, "REG_KEY", 7,
   write_reg_key, read_reg_key},
} ;

/* ReadKeyVal returns a key or value, with an indication if the key was
   modified, added, or deleted. KV_None is returned when there are no more keys
   to read */
static KeyValue *ReadKeyVal(Iteration *it)
{
  char buffer[BUFSIZ] ;
  int result, lookup ;
  char *scan ;
  Context *context ;
  FILE *in ;

  static char *namebuf = NULL ;
  static DWORD namebuflen = 0 ;
  static char *databuf = NULL ;
  static DWORD databuflen = 0 ;

  assert(it) ;

  in = it->fp ;

  if ( (scan = fgets(buffer, BUFSIZ, in)) != NULL ) {
    KV_State state = KV_Set ;
    DWORD type ;

    switch ( *scan ) {
    case '-':
      state ^= KV_Modify|KV_Delete ;	/* MA^MD = AD^MA = MD^M = D */
    case '!':
      state ^= KV_Modify|KV_Add ;	/* MA^MA = 0^M = M */
    case '+':
      state ^= KV_Modify ;		/* MA^M = A */
      ++scan ;
    }

    if ( (scan = skip_white(in, buffer, BUFSIZ, scan)) == NULL )
      return NULL ;

    type = REG_BOGUS ;
    for ( lookup = 0 ; 
	  lookup < sizeof(valuetable)/sizeof(struct ValueTable) ;
	  lookup++ ) {
      DWORD len = valuetable[lookup].namelen ;
      if ( (scan = need_buffer(in, buffer, BUFSIZ, scan, len)) == NULL ) {
	fprintf(stderr, "%s: malformed line %s\n", program, buffer) ;
	return NULL ;
      }
      if ( isspace(scan[len]) &&
	   _strnicmp(scan, valuetable[lookup].name, len) == 0 ) {
	char *sep ;
	DWORD namelen, datalen ;

	type = valuetable[lookup].type ;	/* Found it! */
	scan += len ;
	if ( (scan = read_string(in, buffer, BUFSIZ, scan, &namebuf,
				 &namebuflen, &namelen)) == NULL ) {
	  fprintf(stderr, "%s: malformed line %s\n", program, buffer) ;
	  return NULL ;
	}

	sep = strrchr(namebuf, REGISTRY_SEPARATOR) ;
	if ( sep == NULL )
	  sep = namebuf + namelen ;
	else {
	  namelen = sep - namebuf ;
	  sep++ ;
	}
	do {
	  context = it->context ;
	  /* Create new context if no existing context, or this is sub-key or
	     value of current one */
	  if ( context == NULL ||
	       (namebuf[strlen(context->name)] == REGISTRY_SEPARATOR &&
		_strnicmp(context->name, namebuf, strlen(context->name)) == 0) ) {
	    context = it->context = NewContext(context, namelen + 1) ;
	    strncpy(context->name, namebuf, namelen) ;
	    context->name[namelen] = '\0' ;
	    break ;
	  } else if ( _stricmp(namebuf, context->name) == 0 ) {
	    /* same context */
	    break ;
	  } else {	/* Different context, pop this one and try again */
	    context = it->context = FreeContext(context) ;
	  }
	} while (1) ;

	/* execute read method for type-specific data */
	if ( (valuetable[lookup].read)(in, buffer, BUFSIZ, scan, &databuf,
				       &databuflen, &datalen) == NULL )
	  return NULL ;

	if ( type == REG_KEY )
	  return NewKey(sep, strlen(sep), databuf, datalen, context, state) ;

	return NewValue(sep, strlen(sep), databuf, datalen, context, type,
			state) ;
      }
    }
    
    return NULL ;
  }

  if ( (result = ferror(in)) != 0 )
    fprintf(stderr, "%s: error %d reading file\n", program, result) ;

  return NULL ;
}

void WriteKeyVal(FILE *out, KeyValue *this)
{
  int lookup ;

  for ( lookup = 0 ;
	lookup < sizeof(valuetable)/sizeof(struct ValueTable) ;
	lookup++ ) {
    if ( valuetable[lookup].type == this->base.type ) {
      switch ( this->base.state ) {
      case KV_Add:
	nlprefix = "+" ;
	break ;
      case KV_Delete:
	nlprefix = "-" ;
	break ;
      case KV_Modify:
	nlprefix = "!" ;
	break ;
      }
      if( nlprefix )
	fputs(nlprefix, out) ;
      fputs(valuetable[lookup].name, out) ;
      fputc(' ', out) ;
      if ( this->base.namelen > 0 || this->base.type != REG_KEY )
	write_name(out, this->base.parent, KVName(this), this->base.namelen) ;
      else
	write_name(out, this->base.parent, NULL, 0) ;
      fputc(' ', out) ;
      (valuetable[lookup].write)(out, KVData(this), this->base.datalen) ;
      fputc('\n', out) ;
      
      break ;
    }
  }
}

/* Opens a key for a context and puts in context structure */
HKEY ContextHKEY(Context *context, REGSAM perms)
{
  HKEY hkey = (HKEY)NULL ;
  char *keyname, *basename ;
  int len, index ;
  LONG result ;

  static struct _keytable {
    char *name ;
    HKEY key ;
  } keytable[] = {
    { "HKEY_CURRENT_USER", HKEY_CURRENT_USER },
    { "HKEY_CLASSES_ROOT", HKEY_CLASSES_ROOT },
    { "HKEY_LOCAL_MACHINE", HKEY_LOCAL_MACHINE },
    { "HKEY_USERS", HKEY_USERS },
  } ;

  /* Re-open and replace using permissions specified */
  if ( context->hkey != (HKEY)NULL ) {
    if ( RegOpenKeyEx(context->hkey, NULL, 0, perms, &hkey) == ERROR_SUCCESS ) {
      RegCloseKey(context->hkey) ;
      context->hkey = hkey ;
    }

    return context->hkey ;
  }

  keyname = context->name ;
  basename = strchr(keyname, REGISTRY_SEPARATOR) ;
  len = basename ? basename - keyname : strlen(keyname) ;

  if ( basename ) /* Adjust to point at basename */
    basename = keyname + len + 1 ;
  else
    basename = keyname + len ;


  for ( index = 0 ; index < sizeof(keytable)/sizeof(struct _keytable) ; index++ ) {
    if ( len == strlen(keytable[index].name) &&
	 _strnicmp(keyname, keytable[index].name, len) == 0 ) {
      hkey = keytable[index].key ;
      break ;
    }
  }

  result = RegOpenKeyEx(hkey, basename, 0, (KEY_READ|perms), &(context->hkey)) ;
  if ( result != ERROR_SUCCESS && (perms & KEY_CREATE_SUB_KEY) != 0 ) {
    DWORD disposition ;
    result = RegCreateKeyEx(hkey, basename, 0, "",
			    REG_OPTION_NON_VOLATILE, (KEY_READ|perms),
			    NULL, &context->hkey, &disposition) ;
  }

  if ( result != ERROR_SUCCESS )
    return NULL ;

  return context->hkey ;
}

/* User visible iteration functions; iterate over registry key or file, use
   NextKeyVal to access elements until NULL returned, after setting up with
   IterateKey or IterateFile */
Iteration *IterateKey(char *keyname, REGSAM perms)
{
  Iteration *it ;
  char *basename = strchr(keyname, REGISTRY_SEPARATOR) ;
  int len = basename ? basename - keyname : strlen(keyname) ;
  Context *context = NewContext(NULL, len + 1) ;
  HKEY testkey = (HKEY)NULL ;
  int index ;
  DWORD classlen ;
  LONG result = ERROR_MORE_DATA ;

  static DWORD classsize = 0 ;
  static char *class = NULL ;

  strncpy(context->name, keyname, len) ;
  context->name[len] = '\0' ;

  if ( basename ) /* Adjust to point at basename */
    basename = keyname + len + 1 ;
  else
    basename = keyname + len ;

  testkey = ContextHKEY(context, perms) ;

  if ( testkey == NULL ||
       RegOpenKeyEx(testkey, basename, 0, (KEY_READ|perms), &testkey) !=
       ERROR_SUCCESS ) {
    (void)FreeContext(context) ;
    return NULL ;
  }

  do {
    if ( result == ERROR_MORE_DATA || class == NULL )
      class = extend(class, &classsize, sizeof(char)) ;
      
    classlen = classsize ;
      
    result = RegQueryInfoKey(testkey, class, &classlen, NULL,
			     NULL, NULL, NULL, NULL,
			     NULL, NULL, NULL, NULL) ;
  } while ( result == ERROR_MORE_DATA ) ;

  RegCloseKey(testkey) ;

  if ( result != ERROR_SUCCESS ) {
    (void)FreeContext(context) ;
    return NULL ;
  }

  context->list = NewKey(basename, strlen(basename), class, classlen,
			 context, KV_Set) ;

  it = xalloc(NULL, sizeof(Iteration)) ;
  it->flags = perms ;
  it->context = context ;
  it->fp = NULL ;

  return it ;
}

/* Set up iteration for file, so that GetKeyVal can return next key/value
   entry */
Iteration *IterateFile(char *keyname)
{
  Iteration *it = xalloc(NULL, sizeof(Iteration)) ;

  if ( (it->fp = fopen(keyname, "r")) != NULL ) {
    it->flags = 0 ;
    it->context = NULL ;

    return it ;
  }

  fprintf(stderr, "%s: can't open file \"%s\", error %d\n", program,
	  keyname, GetLastError()) ;
  free(it) ;

  return NULL ;
}

/* Returns next key or value from iteration, returning NULL when there
   are no more keys or values. Caller is responsible for freeing returned
   structure using FreeKeyVal. */
static KeyValue *GetKeyVal(Iteration *it)
{
  Context *context ;
  LONG result ;

  assert(it) ;

  while ( (context = it->context) != NULL ) {
    if ( context->list != NULL ) { /* take first element from list */
      KeyValue *head = context->list ;

      context->list = head->base.next ;
      head->base.next = NULL ;

      if ( head->base.type == REG_KEY ) {	/* open a new context */
	Context *new = NewContext(context, 0) ;

	new->name = KeyValName(head, NULL, NULL) ;

	if ( (result = RegOpenKeyEx(context->hkey, KVName(head), 0,
				    (KEY_READ|it->flags), &(new->hkey))) != ERROR_SUCCESS ) {
	  fprintf(stderr, "%s: can't open key %s, error %d\n", program,
		  new->name, result) ;
	  (void)FreeContext(new) ;
	} else
	  it->context = new ;
      }

      if ( context->list == NULL ) {	/* End of list */
	RegCloseKey(context->hkey) ;
	context->hkey = (HKEY)NULL ;
      }

      return head ;
    } else if ( context->hkey != (HKEY)NULL ) { /* enumerate list, return key */
      LONG result, index ;
      DWORD subkeys, maxsubkeylen, maxclasslen, values, maxvalnamelen,
	maxvaluelen ;

      result = RegQueryInfoKey(context->hkey, NULL, NULL, NULL,
			       &subkeys, &maxsubkeylen,
			       &maxclasslen, &values,
			       &maxvalnamelen, &maxvaluelen,
			       NULL, NULL) ;

      if ( result != ERROR_SUCCESS &&
	   /* Hack for what appears to be error in NT 4.0 SP 3 */
	   result != ERROR_INSUFFICIENT_BUFFER ) {
	fprintf(stderr, "%s: can't query key %s, error %d\n", program,
		context->name, result) ;
	return NULL ;
      }

      if ( subkeys + values > 0 ) {
	KeyValue **array = xalloc(NULL, (subkeys + values) * sizeof(KeyValue *)) ;
	DWORD namebufsize, databufsize ;
	char *namebuf, *databuf ;
	int last = 0 ;

	namebufsize = maxsubkeylen + 1 ;	/* Allow for NULL */
	if ( maxvalnamelen >= namebufsize )
	  namebufsize = maxvalnamelen + 1 ;
	databufsize = maxclasslen + 1 ;	/* Allow for NULL */
	if ( maxvaluelen > databufsize )
	  databufsize = maxvaluelen ;

	namebuf = xalloc(NULL, namebufsize) ;
	databuf = xalloc(NULL, databufsize) ;

	for ( index = 0 ; index < subkeys ; index++ ) {
	  DWORD namesize = namebufsize ;
	  DWORD datasize = databufsize ;
	  FILETIME lastwrite ;

	  result = RegEnumKeyEx(context->hkey, index, namebuf, &namesize,
				NULL, databuf, &datasize, &lastwrite) ;

	  if ( result != ERROR_SUCCESS ) {
	    fprintf(stderr, "%s: can't enumerate key %d in %s, error %d\n",
		    program, index, context->name, result) ;
	  } else
	    array[last++] = NewKey(namebuf, namesize, databuf, datasize,
				   context, KV_Set) ;
	}

	for ( index = 0 ; index < values ; index++ ) {
	  DWORD namesize = namebufsize ;
	  DWORD datasize = databufsize ;
	  DWORD type ;

	  result = RegEnumValue(context->hkey, index, namebuf, &namesize,
				NULL, &type, databuf, &datasize) ;

	  if ( result != ERROR_SUCCESS ) {
	    fprintf(stderr, "%s: can't enumerate value %d in %s, error %d\n",
		    program, index, context->name, result) ;
	  } else
	    array[last++] = NewValue(namebuf, namesize, databuf,
				     datasize, context, type, KV_Set) ;
	}

	qsort(array, last, sizeof(KeyValue *), &SortKeyValues);

	while ( --last >= 0 ) {
	  array[last]->base.next = context->list ;
	  context->list = array[last] ;
	}

	free(array) ;
	free(databuf) ;
	free(namebuf) ;
      } else {	/* No subkeys/values, pop context and loop */
	RegCloseKey(context->hkey) ;
	context->hkey = (HKEY)NULL ;
      }
    } else { /* Done with this list, destroy context and pop to parent */
      it->context = FreeContext(context) ;
    }
  }

  return NULL ;
}

/* Generic iterator get next key/value from file or registry. Returns next key
   or value from iteration, returning NULL when there are no more keys or
   values. Caller is responsible for freeing returned structure using
   FreeKeyVal. */
KeyValue *NextKeyVal(Iteration *it)
{
  if ( it->fp )
    return ReadKeyVal(it) ;

  if ( it->context )
    return GetKeyVal(it) ;

  return NULL ;
}

void EndIteration(Iteration *it)
{
  assert(it) ;

  if ( it->fp )
    fclose(it->fp) ;

  while ( it->context )
    it->context = FreeContext(it->context) ;

  free(it) ;
}
