/* Regdiff.c
 *
 * Copyright (C) Angus J. C. Duggan, 1996
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Diff registry keys with output of regdump, ready for regset.
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "common.h"

char *program ;

static void usage(void)
{
  fprintf(stderr, "Copyright (C) Angus J. C. Duggan, 1996-1997\n") ;
  fprintf(stderr, "Usage:\n  %s [-new] [-modify] [-delete] [-all] [file...]\n", program) ;
  fflush(stderr) ;
  exit(1) ;
}

KV_State Compare(KeyValue *kva, KeyValue *kvb)
{
  int result ;
  char *pnamea, *pnameb ;

  if ( kva == NULL ) {
    if ( kvb == NULL )
      return KV_Set ;
    return KV_Add ;
  } else if ( kvb == NULL )
    return KV_Delete ;

  pnamea = kva->base.parent->name ;
  pnameb = kvb->base.parent->name ;

  if ( _stricmp(pnamea, pnameb) != 0 ) {
    static char *namea = NULL, *nameb = NULL ;
    static int sizea = 0, sizeb = 0 ;

    namea = KeyValName(kva, namea, &sizea) ;
    nameb = KeyValName(kvb, nameb, &sizeb) ;

    result = _stricmp(namea, nameb) ;
  } else {
    result = _stricmp(KVName(kva), KVName(kvb)) ;
  }

  if ( result == 0 ) {	/* Same key name, now test type &c */
    if ( kva->base.type == kvb->base.type &&
	 kva->base.datalen == kvb->base.datalen ) {
      int datalen = kva->base.datalen ;
      unsigned char *dataa = KVData(kva) ;
      unsigned char *datab = KVData(kvb) ;

      while ( datalen-- ) {
	if ( *dataa++ != *datab++ )
	  return KV_Modify ;
      }

      return KV_Set ;	/* Values are exactly the same */
    }
    return KV_Modify ;
  } else 

  if ( result < 0 )
    return KV_Delete ;

  if ( result > 0 )
    return KV_Add ;

  return KV_None ;
}

/* Work out what state to report. state is actual state of key/value, options
   is reporting options, reverse indicates if reversing diff */
KV_State DiffState(KV_State state, KV_State options, int reverse)
{
  if ( state != KV_Modify && reverse )
    state ^= KV_Add|KV_Delete ;

  return (state & options) ;
}

/* Diff keys from file with registry. File determines which keys to iterate
   over. */
void DiffKeys(char *infile, KV_State options, int reverse)
{
  Iteration *itf, *itr = NULL ;
  char *rname = NULL ;
  int rlen = 0 ;
  KeyValue *kvf = NULL, *kvr = NULL ;

  if ( (itf = IterateFile(infile)) == NULL ) {
    fprintf(stderr, "%s: can't iterate file %s, skipping\n", program, infile) ;
    return ;
  }

  do {
    KV_State result ;

    if ( itf != NULL && kvf == NULL ) {
      if ( (kvf = NextKeyVal(itf)) != NULL ) {
	char *cname = NULL ;
	int clen = 0 ;

	cname = KeyValName(kvf, NULL, NULL) ;
	clen = strlen(cname) ;

	if ( _strnicmp(cname, rname, rlen) != 0 ||
	     (cname[rlen] != REGISTRY_SEPARATOR &&
	      cname[rlen] != '\0') ) {
	  /* Changed section. Close old registry iteration and start new one */
	  if ( itr ) {	/* All remaining registry keys & values are extra */
	    while ( kvr != NULL || (kvr = NextKeyVal(itr)) != NULL ) {
	      if ( (kvr->base.state = DiffState(KV_Add, options, reverse)) != KV_None )
		WriteKeyVal(stdout, kvr) ;
	      FreeKeyVal(kvr) ;
	      kvr = NULL ;
	    }
	    EndIteration(itr) ;
	  }
	  if ( kvf->base.type == REG_KEY ) {
	    itr = IterateKey(cname, KEY_READ) ;
	  } else
	    itr = NULL ;
	  free(rname) ;
	  rname = cname ;
	  rlen = clen ;
	} else
	  free(cname) ;
      } else {	/* kvf returned null, itf must be finished */
	EndIteration(itf) ;
	itf = NULL ;
      }
    }

    if ( itr != NULL && kvr == NULL ) {
      if ( (kvr = NextKeyVal(itr)) == NULL ) {
	EndIteration(itr) ;
	itr = NULL ;
      }
    }

    result = Compare(kvf, kvr) ;

    switch ( result ) {
    case KV_Add:
      if ( (result = DiffState(result, options, reverse)) != KV_None ) {
	kvr->base.state = result ;
	WriteKeyVal(stdout, kvr) ;
      }
      if ( kvr )	/* dispose of registry key */
	FreeKeyVal(kvr) ;
      kvr = NULL ;
      break ;
    case KV_Delete:
      if ( (result = DiffState(result, options, reverse)) != KV_None ) {
	kvf->base.state = result ;
	WriteKeyVal(stdout, kvf) ;
      }
      if ( kvf )	/* dispose of file key */
	FreeKeyVal(kvf) ;
      kvf = NULL ;
      break ;
    case KV_Modify:	/* Changed key or value */
      if ( (result = DiffState(result, options, reverse)) != KV_None ) {
	KeyValue *which = reverse ? kvf : kvr ;
	which->base.state = result ;
	WriteKeyVal(stdout, which) ;
      }
      /* Fallthrough, dispose of both keys/values */
    case KV_Set:	/* Same key/value, don't show anything */
    case KV_None:
      if ( kvf )	/* dispose of both keys */
	FreeKeyVal(kvf) ;
      if ( kvr )
	FreeKeyVal(kvr) ;
      kvr = kvf = NULL ;
      break ;
    }
  } while ( itr != NULL || itf != NULL ) ;
}

int main (int argc, char *argv[])
{
  KV_State options = KV_None ;
  int reverse = 0 ;

  for (program = *argv++ ; --argc ; argv++) {
    if (argv[0][0] == '-') {
      switch (argv[0][1]) {
      case 'r': /* reverse */
	if ( argv[0][2] != '\0' && strcmp("-reverse", *argv) != 0 )
	  usage() ;
	reverse = !reverse ;
	break ;
      case 'a':	/* all */
	if ( argv[0][2] != '\0' && strcmp("-all", *argv) != 0 )
	  usage() ;
	options |= KV_Add|KV_Delete|KV_Modify ;
	break;
      case 'n':	/* new */
	if ( argv[0][2] != '\0' && strcmp("-new", *argv) != 0 )
	  usage() ;
	options |= KV_Add ;
	break;
      case 'd':	/* delete */
	if ( argv[0][2] != '\0' && strcmp("-delete", *argv) != 0 )
	  usage() ;
	options |= KV_Delete ;
	break;
      case 'm':	/* modify */
	if ( argv[0][2] != '\0' && strcmp("-modify", *argv) != 0 )
	  usage() ;
	options |= KV_Modify ;
	break;
      default:
	usage();
      }
    } else { /* dump key */
      DiffKeys(*argv, (options == KV_None ? (KV_Add|KV_Modify|KV_Delete) :
		       options), reverse) ;
    }
  }

  return 0 ;
}
