/* Regset.c
 *
 * Copyright (C) Angus J. C. Duggan, 1996
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Set registry keys from output of regdump or regdiff.
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "common.h"

char *program ;

static void usage(void)
{
  fprintf(stderr, "Copyright (C) Angus J. C. Duggan, 1996-1997\n") ;
  fprintf(stderr, "Usage:\n  %s [-new] [-modify] [-delete] [-all] [file...]\n", program) ;
  fflush(stderr) ;
  exit(1) ;
}

void SetKeyVal(KeyValue *kv, KV_State state)
{
  HKEY hkey ;
  LONG result = ERROR_SUCCESS ;

  switch ( state & kv->base.state ) {
  case KV_Set:
  case KV_Add:
  case KV_Modify:
    hkey = ContextHKEY(kv->base.parent, KEY_ALL_ACCESS) ;
    if ( kv->base.type == REG_KEY ) { /* add/modify subkeys */
      DWORD disposition ;
      result = RegCreateKeyEx(hkey, KVName(kv), 0, KVData(kv),
			      REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS,
			      NULL, &hkey, &disposition) ;
      if ( result == ERROR_SUCCESS )
	RegCloseKey(hkey) ;
    } else {
      result = RegSetValueEx(hkey, KVName(kv), 0, kv->base.type,
			     KVData(kv), kv->base.datalen) ;
    }
    break ;
  case KV_Delete:
    hkey = ContextHKEY(kv->base.parent, KEY_SET_VALUE|KEY_ENUMERATE_SUB_KEYS) ;
    if ( kv->base.type == REG_KEY ) { /* remove subkeys; should enumerate */
      /* will fail on nested keys; needs enumeration */
      result = RegDeleteKey(hkey, KVName(kv)) ;
    } else {
      result = RegDeleteValue(hkey, KVName(kv)) ;
    }
    break ;
  default:
    fprintf(stderr, "%s: key/value state %x not recognised\n", program,
	    (state & kv->base.state)) ;
  }

  if ( result != ERROR_SUCCESS ) {
    fprintf(stderr, "%s: SetKeyVal failed on %s, error %d\n", program,
	    KVName(kv), result) ;
  }
}

int main (int argc, char *argv[])
{
  KV_State options = KV_None ;

  for (program = *argv++ ; --argc ; argv++) {
    if (argv[0][0] == '-') {
      switch (argv[0][1]) {
      case 'a':	/* all */
	if ( argv[0][2] != '\0' && strcmp("-all", *argv) != 0 )
	  usage() ;
	options |= KV_Add|KV_Delete|KV_Modify ;
	break;
      case 'n':	/* new */
	if ( argv[0][2] != '\0' && strcmp("-new", *argv) != 0 )
	  usage() ;
	options |= KV_Add ;
	break;
      case 'd':	/* delete */
	if ( argv[0][2] != '\0' && strcmp("-delete", *argv) != 0 )
	  usage() ;
	options |= KV_Delete ;
	break;
      case 'm':	/* modify */
	if ( argv[0][2] != '\0' && strcmp("-modify", *argv) != 0 )
	  usage() ;
	options |= KV_Modify ;
	break;
      default:
	usage();
      }
    } else { /* dump key */
      Iteration *it ;

      if ( (it = IterateFile(*argv)) != NULL ) {
	KeyValue *kv ;

	while ( (kv = NextKeyVal(it)) != NULL ) {
	  SetKeyVal(kv, (options == KV_None ? KV_Set : options)) ;
	  FreeKeyVal(kv) ;
	}
	EndIteration(it) ;
      } else {
	fprintf(stderr, "%s: can't open file %s, skipping\n", program, *argv) ;
      }
    }
  }

  return 0 ;
}
