/* relay.c
 *
 * Copyright (C) Angus J. C. Duggan, 1997-1998.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Relay connections from port to telnet monitor, or on to third party,
 * logging connection.
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

static char *program = NULL ;

void usage(void)
{
  fprintf(stderr, "Copyright Angus J. C. Duggan, 1997-1998\n"
	  "Usage: %s [-l[file]] [-r] inport[:outport] [host:port]\n"
          "\tinport[:outport]\tlocal ports in and out (numeric or service name)\n"
          "\thost:port\tremote host and port relayed\n"
          "\t-l[file]\tlog connection to file (default stdout)\n"
          "\t-r\traw mode (no line buffering)\n", program) ;
  exit(1) ;
}

void error(char *format, ...)
{
  va_list args ;
  char *scan = format, ch ;

  fprintf(stderr, "%s: ", program) ;
  va_start(args, format) ;
  while ( (ch = *scan++) != '\0' ) {
    if ( ch == '%' ) {
      switch ( *scan++ ) {
      case 'd':
	fprintf(stderr, "%d", va_arg(args, long)) ;
	break ;
      case 's':
	fprintf(stderr, "%s", va_arg(args, char *)) ;
	break ;
      case '%':
	fputc(ch, stderr) ;
	break ;
      default:
	fprintf(stderr, "Bad format string\n") ;
	exit(1) ;
      }
    } else
      fputc(ch, stderr) ;
  }
  fputc('\n', stderr) ;
  fflush(stderr) ;
  va_end(args) ;

  exit(1) ;
}

CRITICAL_SECTION log_lock ;

#define LOCK_LOG() EnterCriticalSection(&log_lock)
#define RELEASE_LOG() LeaveCriticalSection(&log_lock)

typedef struct _Relay {
  DWORD threadid ;
  SOCKET from, to ;
  FILE *log ;
  char *prefix ;
} Relay ;

/* Thread to relay info from socket to socket, via file */
DWORD WINAPI RelayThread(LPVOID info)
{
  Relay *relay = (Relay *)info ;
  char buffer[BUFSIZ], line[BUFSIZ] ;
  int size, linesize = 0 ;

  /* read from stdin, send to client */
  while ( (size = recv(relay->from, buffer, BUFSIZ, 0)) > 0 ) {
    if ( relay->log ) {
      LOCK_LOG() ;
      if ( relay->prefix ) { /* Cooked output: visible forms, line buffered */
	int scan, outdex ;

	for ( scan = 0 ; scan < size ; scan++ ) {
	  line[linesize++] = buffer[scan] ;
	  if ( buffer[scan] == '\n' || linesize == BUFSIZ ) {
	    fprintf(relay->log, relay->prefix, relay->threadid) ;
	    for ( outdex = 0 ; outdex < linesize ; outdex++ )
	      switch ( line[outdex] ) {
	      case '\r':
		fputs("<CR>", relay->log) ;
		break ;
	      case '\n':
		fputs("<LF>", relay->log) ;
		break ;
	      case 127:
		fputs("<DEL>", relay->log) ;
		break ;
	      case '\b':
		fputs("<BS>", relay->log) ;
		break ;
	      case '\t':
		fputs("<TAB>", relay->log) ;
		break ;
	      default:
		if ( isprint(line[outdex]) ) {
		  fputc(line[outdex], relay->log) ;
		  break ;
		}
	      case '<':
	      case '>':
		fprintf(relay->log, "<%d%d%d>",
			(line[outdex] & 0300) >> 6,
			(line[outdex] & 070) >> 3, (line[outdex] & 0x7)) ;
		break ;
	      }
	    fputc('\n', relay->log) ;
	    linesize = 0 ;
	  }
	}
      } else
	fwrite(buffer, size, sizeof(char), relay->log) ;
      fflush(relay->log) ;
      RELEASE_LOG() ;
    }
    if ( send(relay->to, buffer, size, 0) != size )
      break ;
  }

  if ( relay->log && relay->prefix ) {
    LOCK_LOG() ;
    fprintf(relay->log, relay->prefix, relay->threadid) ;
    if ( size > 0 )
      fputs("Receiver disconnected\n", relay->log) ;
    else
      fputs("Sender disconnected\n", relay->log) ;
    fflush(relay->log) ;
    RELEASE_LOG() ;
  }

  return size ;
}

typedef struct _Connection {
  Relay in, out ;	/* two symmetric relays */
  SOCKET server, client ;
  HANDLE threads[2] ;
} Connection ;

DWORD WINAPI ConnectionThread(LPVOID info)
{
  Connection *conn = (Connection *)info ;
  DWORD result = WaitForMultipleObjects(2, conn->threads, FALSE, INFINITE) ;

  closesocket(conn->client) ;
  closesocket(conn->server) ;

  switch ( result ) {
  case WAIT_OBJECT_0:
  case WAIT_ABANDONED_0:
    if ( WaitForSingleObject(conn->threads[1], 5000) == WAIT_TIMEOUT )
      TerminateThread(conn->threads[1], 1) ;
    break ;
  case WAIT_OBJECT_0 + 1:
  case WAIT_ABANDONED_0 + 1:
    if ( WaitForSingleObject(conn->threads[0], 5000) == WAIT_TIMEOUT )
      TerminateThread(conn->threads[0], 1) ;
    break ;
  default:
    error("WaitForMultipleObjects returned strange error %d",
	  GetLastError()) ;
  }

  CloseHandle(conn->threads[0]) ;
  CloseHandle(conn->threads[1]) ;

  if ( conn->in.log ) {
    LOCK_LOG() ;
    fprintf(conn->in.log, "Relay thread pair %d, %d terminated\n",
	    conn->in.threadid, conn->out.threadid) ;
    fflush(conn->in.log) ;
    RELEASE_LOG() ;
  }

  free((char *)conn) ;

  return 0 ;
}

typedef struct _Monitor {
  SOCKET telnet, instance ;
  HANDLE ready, used ;
  FILE *log ;
} Monitor ;

DWORD WINAPI TelnetAcceptor(LPVOID info)
{
  Monitor *tmon = info ;
  struct servent *serviceentry = NULL ;
  int port ;

  if ( (serviceentry = getservbyname("telnet", "tcp")) == NULL )
    error("can't get service entry for telnet service, error %d",
	  GetLastError()) ;

  port = ntohs(serviceentry->s_port) ;

  for (;;) {
    struct sockaddr_in accept_in ;
    int accept_len = sizeof(struct sockaddr_in) ;

    fprintf(stderr, "Waiting for monitor connection on port %d...\n",
	    ntohs(serviceentry->s_port)) ;
    fflush(stderr) ;

    if ( (tmon->instance = accept(tmon->telnet, (struct sockaddr *)&accept_in,
				  &accept_len)) == INVALID_SOCKET )
      error("can't accept connection: %d", WSAGetLastError()) ;

    SetEvent(tmon->ready) ; /* Note that it is valid */

    fprintf(stderr, "Accepted monitor connection from %s on port %d\n",
	    inet_ntoa(accept_in.sin_addr), port) ;
    fflush(stderr) ;
    if ( tmon->log ) {
      LOCK_LOG() ;
      fprintf(tmon->log, "Accepted monitor connection from %s on port %d\n",
	      inet_ntoa(accept_in.sin_addr), port) ;
      fflush(tmon->log) ;
      RELEASE_LOG() ;
    }

    /* Wait for client to take connection from me */
    WaitForSingleObject(tmon->used, INFINITE) ;
  }

  return 0 ;
}

int main(int argc, char *argv[])
{
  WSADATA wsaData;
  struct sockaddr_in server_in ;
  SOCKET server, telnet = INVALID_SOCKET ;
  BOOL reuse = TRUE ;
  char *host = NULL ;
  int server_port = 0 ;
  int outgoing_port = 0 ;
  int client_port = 0 ;
  FILE *log = NULL ;
  int raw = FALSE ;
  Monitor tmon ;
  HANDLE tmonh = NULL ;

  if ( WSAStartup(MAKEWORD(1,1), &wsaData) != 0 )
    error("winsock can't start") ;

  if ( LOBYTE(wsaData.wVersion) != 1 || HIBYTE(wsaData.wVersion) != 1 )
    error("invalid winsock version") ;

  for (program = *argv++ ; --argc ; argv++ ) {
    if ( argv[0][0] == '-' ) {
      switch ( argv[0][1] ) {
      case 'l':
	if ( argv[0][2] != '\0' ) {
	  if ( (log = fopen(&argv[0][2], "wb")) == NULL )
	    error("can't open log file %s", &argv[0][2]) ;
	} else
	  log = stdout ;
	break ;
      case 'r':
	raw = TRUE ;
	break ;
      default:
	usage() ;
      }
    } else if ( server_port == 0 ) {
      struct servent *serviceentry = NULL ;
      char *colon = strchr(*argv, ':') ;

      if ( colon != NULL )
	*colon++ = '\0' ;

      if ( isdigit(argv[0][0]) ) {
	server_port = atoi(argv[0]) ;
      } else if ( (serviceentry = getservbyname(argv[0], "tcp")) != NULL ) {
	server_port = ntohs(serviceentry->s_port) ;
      } else
	error("invalid service name %s", argv[0]) ;

      if ( colon != NULL ) {
	if ( isdigit(*colon) ) {
	  outgoing_port = atoi(colon) ;
	} else if ( (serviceentry = getservbyname(colon, "tcp")) != NULL ) {
	  outgoing_port = ntohs(serviceentry->s_port) ;
	} else
	  error("invalid service name %s", colon) ;
      }
    } else if ( host == NULL ) {
      char *colon = strchr(*argv, ':') ;

      host = *argv ;

      if ( colon != NULL ) {
	struct servent *serviceentry = NULL ;

	*colon++ = '\0' ;

	if ( isdigit(*colon) ) {
	  client_port = atoi(colon) ;
	} else if ( (serviceentry = getservbyname(colon, "tcp")) != NULL ) {
	  client_port = ntohs(serviceentry->s_port) ;
	} else
	  error("invalid service name %s", argv[0]) ;
      }
    } else if ( client_port == 0 ) {
      struct servent *serviceentry = NULL ;

      if ( isdigit(argv[0][0]) ) {
	client_port = atoi(argv[0]) ;
      } else if ( (serviceentry = getservbyname(argv[0], "tcp")) != NULL ) {
	client_port = ntohs(serviceentry->s_port) ;
      } else
	error("invalid service name %s", argv[0]) ;
    } else
      usage() ;
  }

  if ( server_port <= 0 )
    error("invalid server port %d", server_port) ;
  else if ( host != NULL && client_port <= 0 )
    error("invalid client port %d", client_port) ;

  fprintf(stderr, "Relaying port %d to %s:%d as %d\n", server_port,
	  host ? host : "localhost",
	  client_port ? client_port : 23,
	  outgoing_port ? outgoing_port : 0) ;

  if ( INVALID_SOCKET == (server = socket(PF_INET, SOCK_STREAM, 0)) )
    error("can't create server socket, error %d", WSAGetLastError()) ;

  if ( setsockopt(server, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse,
		  sizeof(reuse)) != 0 )
    error("can't set server socket options, error %d", WSAGetLastError()) ;

  /* Set up listener socket address */
  memset(&server_in, 0, sizeof(struct sockaddr_in)) ;
  server_in.sin_family = AF_INET ;
  server_in.sin_port = htons((unsigned short)server_port) ;
  server_in.sin_addr.s_addr = INADDR_ANY ;

  if ( bind(server, (struct sockaddr *)&server_in,
	    sizeof(struct sockaddr_in)) != 0 )
    error("can't bind server socket, error %d", WSAGetLastError()) ;

  if ( listen(server, 2) == SOCKET_ERROR )
    error("can't listen on server socket, error %d", WSAGetLastError()) ;

  /* If no host specified, listen for telnet monitor connection */
  if ( host == NULL ) {
    struct servent *serviceentry = NULL ;
    struct sockaddr_in telnet_in ;
    DWORD threadid ;

    if ( (serviceentry = getservbyname("telnet", "tcp")) == NULL )
      error("can't get service entry for telnet service, error %d",
	    GetLastError()) ;

    if ( INVALID_SOCKET == (telnet = socket(PF_INET, SOCK_STREAM, 0)) )
      error("can't create telnet socket, error %d", WSAGetLastError()) ;

    if ( setsockopt(telnet, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse,
		    sizeof(reuse)) != 0 )
      error("can't set telnet socket options, error %d", WSAGetLastError()) ;

    /* Set up listener socket address */
    memset(&telnet_in, 0, sizeof(struct sockaddr_in)) ;
    telnet_in.sin_family = AF_INET ;
    telnet_in.sin_port = serviceentry->s_port ;
    telnet_in.sin_addr.s_addr = INADDR_ANY ;

    if ( bind(telnet, (struct sockaddr *)&telnet_in,
	      sizeof(struct sockaddr_in)) != 0 )
      error("can't bind telnet socket, error %d", WSAGetLastError()) ;

    if ( listen(telnet, 1) == SOCKET_ERROR )
      error("can't listen on telnet socket, error %d", WSAGetLastError()) ;

    /* Fill in monitor structure and start accepter thread */
    tmon.telnet = telnet ;
    tmon.log = log ;

    if ( (tmon.ready = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL )
      error("can't create ready event for telnet monitor, error %d",
	    GetLastError()) ;

    if ( (tmon.used = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL )
      error("can't create used event for telnet monitor, error %d",
	    GetLastError()) ;

    tmonh = CreateThread(NULL,	/* Security same as me */
			 0,		/* Same stack too */
			 TelnetAcceptor, /* Entry point */
			 (LPVOID)&tmon, /* parameter */
			 0,		/* flags */
			 &threadid) ;

    if ( tmonh == NULL )
      error("can't create telnet accepter thread, error %d", GetLastError()) ;
  }

  if ( log )
    InitializeCriticalSection(&log_lock) ;

  for (;;) {
    Connection *conn ;
    HANDLE connh ;
    DWORD threadid ;
    SOCKET servinst, clientinst ;
    struct sockaddr_in accept_in ;
    int accept_len = sizeof(struct sockaddr_in) ;

    if ( (conn = (Connection *)malloc(sizeof(Connection))) == NULL )
      error("can't allocate memory for connection\n") ;

    /* Setup client socket instance (which I am server for) */
    if ( (servinst = accept(server, (struct sockaddr *)&accept_in,
			    &accept_len)) != INVALID_SOCKET ) {
      fprintf(stderr, "Accepted client connection from %s on %d\n",
	      inet_ntoa(accept_in.sin_addr), server_port) ;
      fflush(stderr) ;
      if ( log ) {
	LOCK_LOG() ;
	fprintf(log, "Accepted client connection from %s on %d\n",
		inet_ntoa(accept_in.sin_addr), server_port) ;
        fflush(log) ;
	RELEASE_LOG() ;
      }
    } else
      error("can't accept client connection, error %d", WSAGetLastError()) ;

    /* Setup server socket instance (which I am a client to) */
    if ( host != NULL ) {      /* Connect to client */
      struct servent *serviceentry = NULL ;
      struct hostent *hostentry = NULL ;
      struct sockaddr_in remote_in ;

      if ( INVALID_SOCKET == (clientinst = socket(PF_INET, SOCK_STREAM, 0)) )
	error("can't create socket %s:%d, error %d", host, client_port,
	      WSAGetLastError()) ;

      /* Set up local socket address */
      if ( outgoing_port != 0 ) {
	char buffer[BUFSIZ] ;

	if ( SOCKET_ERROR == gethostname(buffer, BUFSIZ) )
	  error("can't get own host name, error %d\n", WSAGetLastError()) ;

	if ( (hostentry = gethostbyname(buffer)) == NULL)
	  error("can't get own host entry for %s, error %d\n", buffer,
		WSAGetLastError()) ;

	memset(&remote_in, 0, sizeof(remote_in)) ;
	remote_in.sin_family = AF_INET ;
	remote_in.sin_port = htons((unsigned short)outgoing_port) ;
	memcpy(&remote_in.sin_addr, hostentry->h_addr, hostentry->h_length);

	if ( bind(clientinst, (struct sockaddr *)&remote_in, sizeof(remote_in)) != 0 )
	  error("bind to outgoing port %d failed, error %d\n", outgoing_port,
		WSAGetLastError()) ;
      }

      if ( (hostentry = gethostbyname(host)) == NULL )
	error("client host %s not reachable, error %d", host, GetLastError()) ;

      /* Set up remote socket address */
      memset(&remote_in, 0, sizeof(struct sockaddr_in)) ;
      remote_in.sin_family = AF_INET ;
      remote_in.sin_port = htons((unsigned short)client_port) ;
      memcpy(&remote_in.sin_addr, hostentry->h_addr, hostentry->h_length);
      
      if ( SOCKET_ERROR == connect(clientinst, (struct sockaddr *)&remote_in,
				   sizeof(struct sockaddr_in)) )
	error("can't connect socket %s:%d, error %d", host, client_port,
	      WSAGetLastError()) ;

      fprintf(stderr, "Connected to client %s on port %d as %d\n",
	      inet_ntoa(remote_in.sin_addr), client_port, outgoing_port) ;
      fflush(stderr) ;
      if ( log ) {
	LOCK_LOG() ;
	fprintf(log, "Connected to client %s on port %d as %d\n",
		inet_ntoa(remote_in.sin_addr), client_port, outgoing_port) ;
        fflush(log) ;
	RELEASE_LOG() ;
      }

    } else { /* Using telnet monitor */
      WaitForSingleObject(tmon.ready, INFINITE) ; /* wait for telnet */

      clientinst = tmon.instance ; /* copy instance */

      SetEvent(tmon.used) ; /* signal accepter to resume listening */
    }

    conn->server = servinst ;
    conn->client = clientinst ;

    conn->in.from = servinst ;
    conn->in.to = clientinst ;
    conn->in.log = log ;
    conn->in.prefix = raw ? NULL : "C [%05d]->" ;
    conn->threads[0] = CreateThread(NULL,	/* Security same as me */
				    0,		/* Same stack too */
				    RelayThread, /* Entry point */
				    (LPVOID)&conn->in, /* parameter */
				    0,		/* flags */
				    &conn->in.threadid) ;
	
    if ( conn->threads[0] == NULL )
      error("can't spawn input relay, error %d", GetLastError()) ;

    conn->out.from = clientinst ;
    conn->out.to = servinst ;
    conn->out.log = log ;
    conn->out.prefix = raw ? NULL : "S [%05d]->" ;
    conn->threads[1] = CreateThread(NULL,	/* Security same as me */
				    0,		/* Same stack too */
				    RelayThread, /* Entry point */
				    (LPVOID)&conn->out, /* parameter */
				    0,		/* flags */
				    &conn->out.threadid) ;
	
    if ( conn->threads[1] == NULL )
      error("can't spawn output relay, error %d", GetLastError()) ;

    /* Pass off info to ConnectionThread to monitor connection */
    connh = CreateThread(NULL,	/* Security same as me */
			 0,		/* Same stack too */
			 ConnectionThread, /* Entry point */
			 (LPVOID)conn, /* parameter */
			 0,		/* flags */
			 &threadid) ;

    if ( connh == NULL )
      error("can't spawn connection monitor, error %d", GetLastError()) ;

    CloseHandle(connh) ;
  }

  closesocket(server) ;

  if ( telnet != INVALID_SOCKET )	/* Terminate monitor */
    closesocket(telnet) ;

  if ( tmonh != NULL ) {
    CloseHandle(tmon.ready) ;
    CloseHandle(tmon.used) ;
    CloseHandle(tmonh) ;
  }

  if ( log != NULL ) {
    DeleteCriticalSection(&log_lock) ;
    fclose(log) ;
  }

  if ( WSACleanup() != 0 )
    error("cleanup failed") ;

  return 0 ;
}
