/* common.c
 *
 * (C) Angus J. C. Duggan, 1997-1998
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Common parts for SMTP service for Windows NT.
 */

#include "smtp.h"
#include "messages.h"

#define MAX_FORMAT 10 /* Max format specification length */

char *fmtout(char *buffer, int bufsize, char *format, va_list args)
{
  int c ;
  char *here = buffer ;

  assert(format != NULL) ;

  while ( (c = (unsigned char)*format++) != 0 ) {
    if ( c == '%' ) {
      int done, index = 0, longform = 0 ;
      char fmtbuf[MAX_FORMAT] ;

      fmtbuf[index++] = c ;
      do {
	done = 1 ;
	fmtbuf[index++] = c = *format++ ;
	fmtbuf[index] = '\0' ;
	switch (c) {
	case '%':
	  *here++ = c ;
	case '\0':
	  break ;
	case 'e': case 'E': case 'f': case 'g': case 'G':
	  {
	    double d = va_arg(args, double) ;
	    sprintf(here, fmtbuf, d) ;
	    here += strlen(here) ;
	  }
	  break ;
	case 'c': case 'd': case 'i': case 'o':
	case 'p': case 'u': case 'x': case 'X':
	  if ( longform ) {
	    long l = va_arg(args, long) ;
	    sprintf(here, fmtbuf, l) ;
	    here += strlen(here) ;
	  } else {
	    int i = va_arg(args, int) ;
	    sprintf(here, fmtbuf, i) ;
	    here += strlen(here) ;
	  }
	  break ;
	case 's':
	  {
	    char *s = va_arg(args, char *) ;
	    sprintf(here, fmtbuf, s) ;
	    here += strlen(here) ;
	  }
	  break ;
	case 't': /* FILETIME struct */
	  {
	    SYSTEMTIME systime ;
	    if ( FileTimeToSystemTime(va_arg(args, FILETIME *), &systime) ) {
	      static char *month[] = { "",
				       "Jan", "Feb", "Mar",
				       "Apr", "May", "Jun",
				       "Jul", "Aug", "Sep",
				       "Oct", "Nov", "Dec" } ;
	      sprintf(here, "%d %s %04d %02d:%02d:%02d", systime.wDay,
		      month[systime.wMonth],
		      systime.wYear < 1000 ? systime.wYear + 1900 : systime.wYear,
		      systime.wHour, systime.wMinute, systime.wSecond) ;
	      here += strlen(here) ;
	    }
	  }
	  break ;
	case 'l':
	  longform = 1 ;
	  /* FALLTHRU */
	default:
	  done = 0 ;
	}
      } while ( !done ) ;
    } else	/* Not %, write out character */
      *here++ = c ;
  }

  *here = '\0' ;

  return buffer ;
}

void debug(int level, char *format, ...)
{
  va_list args ;

  va_start(args, format) ;
  if ( (debuglevel & level) != 0 ) {
    char buffer[1024] ;
    char *str = fmtout(buffer, 1024, format, args) ;
    fputs(str, stderr) ;
    fflush(stderr) ;
  }
  va_end(args) ;
}

void debugstr(int level, char *string, long len)
{
  if ( (debuglevel & level) != 0 ) {
    int nl = 1 ;

    while ( len-- > 0 ) {
      char ch = *string++ ;

      nl = 0 ; /* Assume not a Newline */

      switch (ch) {
      case '\r': fputs("<CR>", stderr) ; break ;
      case '\n': fputs("<NL>\n", stderr) ; nl = 1 ; break ;
      case '\t': fputs("<TAB>", stderr) ; break ;
      default: fputc(ch, stderr) ; break ;
      }
    }
    if ( !nl )
      fputc('\n', stderr) ;
    fflush(stderr) ;
  }
}

char *formatstr(char *buffer, int bufsize, char *format, ...)
{
  char *result ;
  va_list args ;

  va_start(args, format) ;
  result = fmtout(buffer, bufsize, format, args) ;
  va_end(args) ;

  return result ;
}

/* Type is
   EVENTLOG_ERROR_TYPE	Error event
   EVENTLOG_WARNING_TYPE	Warning event
   EVENTLOG_INFORMATION_TYPE	Information event
   */
void message(WORD type, DWORD id, ...)
{
  va_list args ;

  va_start(args, id) ;
  if ( debuglevel == 0 ) {
    HANDLE hEventSource = RegisterEventSource(NULL, SERVICE_NAME);
    char *strings[20], *arg ;
    WORD inserts = 0 ;

    while ( (arg = va_arg(args, char*)) != NULL && inserts < 20 )
      strings[inserts++] = arg ;

    if ( hEventSource != NULL ) {
      ReportEvent(hEventSource,	/* handle of event source */
		  type,		/* event type */
		  0,		/* event category */
		  id,		/* event ID */
		  NULL,		/* current user's SID */
		  inserts,	/* number of strings */
		  0,		/* no bytes of raw data */
		  strings,	/* array of insert strings */
		  NULL) ;	/* raw data */
    
      (VOID)DeregisterEventSource(hEventSource) ;
    }
  } else {	/* Use FormatMessage to write out to stderr */
    char buffer[2000] ;

    if ( FormatMessage(FORMAT_MESSAGE_FROM_HMODULE,
		       GetModuleHandle(NULL),	/* Current module */
		       id,	/* From parameters */
		       LANG_NEUTRAL,	/* Default language */
		       buffer,	/* buffer address */
		       2000,	/* buffer size */
		       &args) != 0 ) {
      fputs(buffer, stderr) ;
    } else {
      fprintf(stderr, "FormatMessage error %s\n", ErrorText(GetLastError())) ;
    }
    fflush(stderr) ;
  }
  va_end(args) ;
}

char *ErrorText(DWORD err)
{
  DWORD dwRet;
  char *string = NULL ;

  static char *longest = NULL ;

  dwRet = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
			FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			err,
			LANG_NEUTRAL,
			(LPTSTR)&string,
			0,
			NULL);

  /* Save longest buffer to copy results into */
  if ( dwRet == 0 ) {      /* No error string; what to do? */
    static char message[64] ;

    sprintf(message, "error %d (%x)", err, err) ;
    return message ;
  } else if ( longest == NULL ) {
    longest = string ;
  } else if ( dwRet > strlen(longest) ) {
    LocalFree((HLOCAL)longest) ;
    longest = string ;
  } else {
    strcpy(longest, string) ;
    LocalFree((HLOCAL)string);
  }

  assert(longest != NULL) ;

  return longest ;
}

/* Trawl over all files with names consisting solely of digits in directory */
int ForAllFiles(char *directory, char *extn,
		int (*fn)(char *filename, void *passthrough),
		void *passthrough)
{
  char pattern[MAX_PATH] ;
  WIN32_FIND_DATA fdata ;
  HANDLE it ;
  int dirlen = strlen(directory) ;
  int result = TRUE ;

  strcpy(pattern, directory) ;
  strcpy(pattern + dirlen, "\\*") ;
  strcpy(pattern + dirlen + 2, extn) ;

  if ( (it = FindFirstFile(pattern, &fdata)) != INVALID_HANDLE_VALUE ) {
    do {
      char *scan = fdata.cFileName ;

      while (isdigit(*scan)) scan++ ;	/* Skip digits in filename */

      debug(DEBUG_FILES, "Pattern=%s, file=%s\n", pattern, fdata.cFileName) ;

      if ( scan != fdata.cFileName && Match(extn, scan) ) {
	strcpy(pattern + dirlen + 1, fdata.cFileName) ;
	if ( !(*fn)(pattern, passthrough) ) {
	  result = FALSE ;
	  break ;
	}
      }
    } while (FindNextFile(it, &fdata)) ;
    if ( GetLastError() != ERROR_NO_MORE_FILES )
      result = FALSE ;
    FindClose(it) ;
  } else	/* FindFirstFile failed */
    result = FALSE ;

  return result ;
}

/* Get queue files, creating linked list. Files must match <int>.extn, if
   lastfile pointer is not NULL, highest file number found is returned in
   lastfile. */
int ReadQueue(char *filename, void *data)
{
  FileQueue **fileq = data ;
  char *tail = strrchr(filename, '\\') ;
  int num = atoi(tail + 1) ;
  FileQueue *this ;

  if ( (this = (FileQueue *)malloc(sizeof(FileQueue))) != NULL ) {
    strcpy(this->name, filename) ;
    this->next = *fileq ;
    *fileq = this ;

    debug(DEBUG_QUEUE, "Debug: ReadQueue added %s\n", this->name) ;

    return 1 ;
  }

  return 0 ;
}

char *Summary(char *file, char *summary, long sumsize)
{
  Stream *files = NULL, *smtps = NULL ;
  char *sumptr = NULL ;

  static SMTPCallbacks summary_callbacks = {
    NULL, /* Open */
    NULL, /* Command */
    NULL, /* Error */
    NULL, /* Close */
  } ;

  assert(sumsize >= 4) ;

  summary[0] = '\0' ;

  if ( (files = MakeFileStream(file, FILESTREAM_READ)) != NULL ) {
    if ( (smtps = MakeSMTPStream(&summary_callbacks, NULL)) != NULL ) {
      char buffer[BUFSIZ] ;
      long rbytes ;

      if ( gethostname(buffer, BUFSIZ) == 0 &&
           SendStream(smtps, "HELO %s\r\n", buffer) >= 0 ) {
        while ( (rbytes = ReadStream(files, &buffer, BUFSIZ)) > 0 ) {
          if ( !WriteStreamAll(smtps, buffer, rbytes) )
            break ;
        }
        if ( IoctlStream(smtps, SMTPSTREAM_SUMMARY, &sumptr) == 0 &&
             sumptr != NULL ) {
          int len = strlen(sumptr) ;

          if ( len >= sumsize ) { /* truncate summary */
            strncpy(summary, sumptr, sumsize - 4) ;
            strcpy(summary + sumsize - 4, "...") ; /* terminate */
          } else
            strcpy(summary, sumptr) ; /* copy whole summary */
        }
      }
      
      DestroyStream(smtps) ;
    }
    DestroyStream(files) ;
  }

  if ( summary[0] == '\0' && strlen(file) < sumsize )
    strcpy(summary, file) ;

  return summary ;
}

int DefaultDirectories(void)
{
  /* Look up registry values for mailhost, queuedir and postmaster if not
     given on command line. Registry key is
     HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\SERVICE_NAME 
     */

  /* Default mailhost and queue directory */
  if ( mailhost == NULL &&
       (mailhost = GetRegString(SERVICE_REGISTRY_KEY, "MAILHOST")) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_NOMAILHOST, NULL) ;
    return 0 ;
  }

  /* should default to subdirectory of system dir. Create dir and set registry
     key if installing service, delete dir if removing service */
  if ( queuedir == NULL &&
       (queuedir = GetRegString(SERVICE_REGISTRY_KEY, "QUEUEDIR")) == NULL ) {
    char cwd[MAX_PATH] ;
    DWORD len = GetModuleFileName(NULL, cwd, MAX_PATH) ;

    if ( len ) {
      char *slash = strrchr(cwd, '\\') ;

      if ( slash ) {
	len = slash - cwd ;
	*slash = '\0' ;
      }
    }

    if ( len == 0 ||
	 (queuedir = malloc(len + strlen(Q_BASENAME) + 1)) == NULL ) {
      message(EVENTLOG_ERROR_TYPE, MSG_NOQUEUEDIR, NULL) ;
      return 0 ;
    }
    strcpy(queuedir, cwd) ;
    strcat(queuedir, Q_BASENAME) ;
  }

  if ( postmaster == NULL &&
       (postmaster = GetRegString(SERVICE_REGISTRY_KEY, "POSTMASTER")) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_NOPOSTMASTER, NULL) ;
    return 0 ;
  }

  if ( inport == 0 ) {
    DWORD port = GetRegDWORD(SERVICE_REGISTRY_KEY, "ClientPort") ;

    if ( port )
      inport = htons((unsigned short)port) ;
  }

  if ( outport == 0 ) {
    DWORD port = GetRegDWORD(SERVICE_REGISTRY_KEY, "ServerPort") ;

    if ( port )
      outport = htons((unsigned short)port) ;
  }

  return 1 ;
}

/* recursive pattern match with * wildchars */
int Match(char *pattern, char *string)
{
  char *star = strchr(pattern, '*') ;

  if ( star == NULL )	/* No globs, must match exactly */
    return strcmp(pattern, string) == 0 ;

  if ( star != pattern ) { /* match and remove stuff before star */
    if ( strncmp(pattern, string, star - pattern) != 0 )
      return FALSE ;
    string += star - pattern ;
  }

  while (*star == '*') star++ ;	/* find char after star */

  if ( *star != '\0' ) {  /* loop searching for thing after star */
    while ( (string = strchr(string, *star)) != NULL ) {
      if ( Match(star, string) )
	return TRUE ;
      string++ ;	/* skip over initial match char */
    }

    return FALSE ;
  }

  /* Nothing after star, so it must match */
  return TRUE ;
}

/* end of common.c */
