/* connects.c
 *
 * Copyright (C) Angus J. C. Duggan, 1998.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Connection handling for SMTP buffer.
 */

#include "smtp.h"
#include "messages.h"
#include "streami.h"

typedef struct {
  Stream *base ;
  Stream *server ; /* Server stream */
  Stream *client ; /* Client stream */
  QueueNameFn *getname ; /* Callback for new filename */
  int status ;     /* Status of current connection */
  Stream *save ;   /* File stream for saving data */
  Stream *smtp ;   /* SMTP filter stream */
  char filename[MAX_PATH] ; /* Buffer for save filename */
  enum { CONNECTION_STREAM_COOKIE = 0x434f4e4e } cookie ; /* 'CONN' */
} ConnectionStream ;

static int ConnectionResponse(Stream *server, Stream *client, int *status)
{
  int response = -1 ;

  if ( server && (*status & STATUS_ESERVER) == 0 ) {
    do { /* More response to read */
      char *line ;
      long rbytes = ReadStream(server, &line, 0) ;

#ifdef DEBUG
      stream_debugstr(CONNECTION_DEBUG_RESPONSE, line, rbytes) ;
#endif

      /* Parse for response number */
      if ( rbytes < 4 || /* Includes rbytes < 0 */
           !isdigit(line[0]) || !isdigit(line[1]) || !isdigit(line[2]) ||
           (line[3] != ' ' && line[3] != '-') ) {
        *status |= STATUS_ESERVER ;
        break ;
      } else if ( line[3] == ' ' ) { /* Finished response */
        response = (line[0] - '0') * 100 + (line[1] - '0') * 10 + (line[2] - '0') ;
      }

      /* Forward response to client */
      if ( client && (*status & STATUS_ECLIENT) == 0 ) {
        if ( !WriteStreamAll(client, line, rbytes) )
          *status |= STATUS_ECLIENT ;
      }
    } while ( response < 0 ) ; 
  } else /* No server or bad server, must be an error */
    *status |= STATUS_ESERVER ;

  return response ;
}

/* SMTP stream callbacks */

/* Open callback */
static int ConnectsOpen(void *data)
{
  ConnectionStream *conns = data ;

  conns->status &= ~(STATUS_EMESSAGE|STATUS_EAGAIN) ;

  return 1 ;
}

/* Command callback */
static int ConnectsCommand(void *data, int type, char *line, long linelen,
                           char *args, long argslen)
{
  ConnectionStream *conns = data ;
  int response = -1 ; /* Invalid response */

  switch ( type ) { /* New message? */
  case SMTP_MAIL:
  case SMTP_SAML:
  case SMTP_SOML:
  case SMTP_SEND:
  case SMTP_RSET:
    if ( conns->save ) { /* Abort old save file */
      do {
        conns->save = DestroyStream(conns->save) ;
      } while ( conns->save ) ;
      if ( !DeleteFile(conns->filename) ) { /* Oops */
        message(EVENTLOG_ERROR_TYPE, MSG_DELETEFILE, conns->filename,
                ErrorText(GetLastError()), NULL) ;
      }
    }
    conns->status &= ~(STATUS_EMESSAGE|STATUS_EAGAIN|STATUS_ESAVE) ;
    conns->filename[0] = '\0' ;
    if ( type != SMTP_RSET && conns->getname ) { /* Start new save file */
      if ( !(conns->getname)(conns->filename, MAX_PATH) ||
           (conns->save = MakeFileStream(conns->filename, FILESTREAM_WRITE)) == NULL ) {
        message(EVENTLOG_ERROR_TYPE, MSG_OPENTRICKLE, conns->filename,
                ErrorText(GetLastError()), NULL) ;
        conns->status |= STATUS_ESAVE ;
      }
    }
  }

  /* Forward to server */
  switch ( type ) {
  case SMTP_HELO: /* Don't forward these */
    if ( conns->client && (conns->status & STATUS_ECLIENT) == 0 &&
         SendStream(conns->client, "250 How do you do?\r\n") < 0 )
      conns->status |= STATUS_ECLIENT ;
    response = 250 ;
    break ;
  case SMTP_QUIT:
    if ( conns->client && (conns->status & STATUS_ECLIENT) == 0 &&
         SendStream(conns->client, "221 Goodbye, and have a nice day.\r\n") < 0 )
      conns->status |= STATUS_ECLIENT ;
    response = 221 ;
    break ;
  case SMTP_MESSAGE: /* Don't forward if message error */
  case SMTP_RCPT:
  case SMTP_DOT:
  case SMTP_DATA:
    if ( (conns->status & (STATUS_EAGAIN|STATUS_EMESSAGE)) != 0 ) {
      if ( type == SMTP_DATA )
        response = 354 ;
      else
        response = 250 ;
      break ;
    } /* FALLTHRU */
  default: /* Forward everything else */
    if ( conns->server && (conns->status & STATUS_ESERVER) == 0 ) {
      if ( !WriteStreamAll(conns->server, line, linelen) )
        conns->status |= STATUS_ESERVER ;

      if ( type != SMTP_MESSAGE ) /* Don't need response if in message */
        if ( (conns->status & STATUS_ESERVER) == 0 ) /* Get server response */
          response = ConnectionResponse(conns->server, conns->client, &conns->status) ;
    }
    
    /* Fabricate response? */
    if ( type == SMTP_MESSAGE ) {
      response = 250 ; /* OK response for message */
    } else if ( response < 0 ) {
      char *responsestr = NULL ;

      switch ( type ) {
      case SMTP_MAIL:
        responsestr = "250 From address not verified.\r\n" ;
        break ;
      case SMTP_RCPT:
        responsestr = "251 User not local.\r\n" ;
        break ;
      case SMTP_DATA:
        responsestr = "354 Send mail data, end with <CRLF>.<CRLF>\r\n" ;
        break ;
      case SMTP_DOT:
        responsestr = "250 Burp! Yum!\r\n" ;
        break ;
      case SMTP_RSET:
        responsestr = "250 Resetting stored parameters\r\n" ;
        break ; 
      case SMTP_SAML:
        responsestr = "502 SAML command not supported locally\r\n" ;
        break ; 
      case SMTP_SOML:
        responsestr = "502 SOML command not supported locally\r\n" ;
        break ; 
      case SMTP_SEND:
        responsestr = "502 SEND command not supported locally\r\n" ;
        break ; 
      case SMTP_VRFY:
        responsestr = "502 VRFY command not supported locally\r\n" ;
        break ; 
      case SMTP_EXPN:
        responsestr = "502 EXPN command not supported locally\r\n" ;
        break ; 
      case SMTP_HELP:
        responsestr = "214 Why do you feel that you need HELP?\r\n" ;
        break ; 
      case SMTP_NOOP:
        responsestr = "250 Well, what did you expect a NOOP to do?\r\n" ;
        break ; 
      case SMTP_TURN:
        responsestr = "502 TURN command not supported.\r\n" ;
        break ; 
      default:
        responsestr = "502 Command not recognised.\r\n" ;
        break ;
      }

      assert(isdigit(responsestr[0]) && isdigit(responsestr[1]) &&
             isdigit(responsestr[2]) && responsestr[3] == ' ') ;
          
      response = (responsestr[0] - '0') * 100 + (responsestr[1] - '0') * 10 + (responsestr[2] - '0') ;

#ifdef DEBUG
      stream_debugstr(CONNECTION_DEBUG_RESPONSE, responsestr, strlen(responsestr)) ;
#endif

      /* Send response to client */
      if ( conns->client &&
           !WriteStreamAll(conns->client, responsestr, strlen(responsestr)) )
        conns->status |= STATUS_ECLIENT ;
    }

    break ;
  }

  /* Is this line saveable? */
  switch ( type ) {
  case SMTP_MAIL:
  case SMTP_SAML:
  case SMTP_SOML:
  case SMTP_SEND:
  case SMTP_RCPT:
  case SMTP_DATA:
  case SMTP_MESSAGE:
  case SMTP_DOT:
    if ( conns->save && (conns->status & STATUS_ESAVE) == 0 ) { /* Save command */
      if ( !WriteStreamAll(conns->save, line, linelen) )
        conns->status |= STATUS_ESAVE ;
    }

    /* What was the outcome of the command? */
    if ( response >= 400 && response < 500 ) /* Re-tryable error */
      conns->status |= STATUS_EAGAIN ;
    else if ( response >= 500 && response < 600 ) /* Fatal error */
      conns->status |= STATUS_EMESSAGE ;
  }

  /* End of message? */
  if ( type == SMTP_DOT ) {
    char *summary ;

    if ( IoctlStream(conns->smtp, SMTPSTREAM_SUMMARY, &summary) < 0 ||
         summary == NULL || *summary == '\0' )
      summary = conns->filename ;

    if ( conns->save ) { /* Did we save it? */
      do { /* Close save stream */
        conns->save = DestroyStream(conns->save) ;
      } while ( conns->save ) ;
    }

    /* Sent successfully? */
    if ( conns->server && (conns->status & STATUS_ESERVER) == 0 ) {
      /* Sent message successfully */
      message(EVENTLOG_INFORMATION_TYPE, MSG_SENT, summary, mailhost, NULL) ;

      if ( conns->filename[0] != '\0' &&
           !DeleteFile(conns->filename) ) { /* Oops */
        message(EVENTLOG_ERROR_TYPE, MSG_DELETEFILE, conns->filename,
                ErrorText(GetLastError()), NULL) ;
      }
    } else if ( (conns->status & STATUS_ESAVE) == 0 ) {
      /* Queued message successfully */
      message(EVENTLOG_INFORMATION_TYPE, MSG_QUEUED, summary, mailhost, NULL) ;
    } else {
      /* Didn't queue or send message successfully */
      message(EVENTLOG_INFORMATION_TYPE, MSG_LOST, summary, NULL) ;
    }
  }

  return response ;
}

/* Error callback */
static int ConnectsError(void *data, int type, char *line, long linelen)
{
  return 1 ; /* Carry on from all bogus state transitions */
}

/* Close callback */
static void ConnectsClose(void *data)
{
  ConnectionStream *conns = data ;

  if ( conns->save ) {
    do { /* Abort save file */
      conns->save = DestroyStream(conns->save) ;
    } while ( conns->save ) ;
    if ( !DeleteFile(conns->filename) ) { /* Remove incomplete file */
      message(EVENTLOG_ERROR_TYPE, MSG_DELETEFILE, conns->filename,
              ErrorText(GetLastError()), NULL) ;
    }
  }
}

/* Connection stream. This controls the association between servers,
   clients, and queueing. A connection stream deals with only one server,
   which may be a null server. Clients may be changed using an IOCTL, which
   also returns the connection status. */

/* Change client connected and/or save status */
static int connectionioctl(Stream *this, int op, va_list args)
{
  ConnectionStream *conns = (ConnectionStream *)this ;
  int result = 0 ;

  /* Sanity check on stream type */
  if ( conns->cookie != CONNECTION_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  switch ( op ) {
  case CONNECTION_CHANGE: {
    Stream *client = va_arg(args, Stream *) ;
    QueueNameFn *getname = va_arg(args, QueueNameFn *) ;

    /* Tell previous client we're through. */
    if ( client != conns->client ) {
      char hostname[MAX_HOSTLEN] ;
      
      if ( conns->client && (conns->status & STATUS_ECLIENT) != 0 ) {
        /* Shut down early, inform client */
        if ( gethostname(hostname, MAX_HOSTLEN) != 0 ||
             SendStream(conns->client, "441 %s We're through. I've found somebody else.\r\n", hostname) < 0 ) {
          conns->status |= STATUS_ECLIENT ;
        }
      }
    }

    /* Were we partway through saving something? */
    if ( conns->save ) {
      do {
        conns->save = DestroyStream(conns->save) ;
      } while ( conns->save ) ;
      conns->status |= STATUS_ESAVE ; /* Didn't close off save properly */
      if ( !DeleteFile(conns->filename) ) {
        message(EVENTLOG_ERROR_TYPE, MSG_DELETEFILE, conns->filename,
                ErrorText(GetLastError()), NULL) ;
      }
    }

    (void)IoctlStream(this, STREAM_CLEAREOF) ;
    (void)IoctlStream(this, STREAM_CLEARERR) ;

    result = conns->status ;
    conns->status &= ~(STATUS_ECLIENT|STATUS_ESAVE|STATUS_EMESSAGE) ;
    conns->client = client ;
    conns->getname = getname ;

    /* Greet new client */
    if ( client ) {
      char hostname[MAX_HOSTLEN] ;

      if ( gethostname(hostname, MAX_HOSTLEN) != 0 ||
           SendStream(client,
                      "220 %s Hello, this is SMTP Buffer "
                      VERSION_STRING
                      " Copyright (C) AJCD 1997-1998\r\n", hostname) < 0 )
        conns->status |= STATUS_ECLIENT ;
    }
    break ;
  }
  case STREAM_CLEAREOF:
  case STREAM_CLEARERR:
    break ;
  default:
    SetLastError(ERROR_INVALID_PARAMETER) ;
    result = -1 ;
    break ;
  }

  return result ;
}

/* Read from client */
static long connectionread(Stream *this, void *buffer, long bufsize)
{
  ConnectionStream *conns = (ConnectionStream *)this ;
  long result = -1 ;

  /* Sanity check on stream type */
  if ( conns->cookie != CONNECTION_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  if ( conns->client && (conns->status & STATUS_ECLIENT) == 0 ) {
    result = ReadStream(conns->client, buffer, bufsize) ;
    if ( result == 0 ) /* client closed connection */
      conns->client = NULL ;
    else if ( result < 0 )
      conns->status |= STATUS_ECLIENT ;
  }

#ifdef DEBUG
  stream_debug(CONNECTION_DEBUG_READ, "[CR %04d] ", result) ;
  stream_debugstr(CONNECTION_DEBUG_READ, buffer, result) ;
#endif

  return result ;
}

/* Write to SMTP filter */
static long connectionwrite(Stream *this, void *buffer, long bufsize)
{
  ConnectionStream *conns = (ConnectionStream *)this ;
  long result = -1 ;

  /* Sanity check on stream type */
  if ( conns->cookie != CONNECTION_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

#ifdef DEBUG
  stream_debug(CONNECTION_DEBUG_WRITE, "[CW %04d] ", bufsize) ;
  stream_debugstr(CONNECTION_DEBUG_WRITE, buffer, bufsize) ;
#endif

  return WriteStream(conns->smtp, buffer, bufsize) ;
}

/* Close connection */
static void connectiondestroy(Stream *this)
{
  ConnectionStream *conns = (ConnectionStream *)this ;

  /* Sanity check on stream type */
  if ( conns->cookie != CONNECTION_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return ;
  }

  /* Send server QUIT, get response, terminate and free connection */
  if ( conns->server && (conns->status & STATUS_ESERVER) == 0 ) {
    if ( WriteStreamAll(conns->server, "QUIT\r\n", 6) )
      (void)ConnectionResponse(conns->server, NULL, &conns->status) ;
    else
      conns->status |= STATUS_ESERVER ;
  }

  while ( conns->smtp ) /* Destroy SMTP filter */
    conns->smtp = DestroyStream(conns->smtp) ;

  while ( conns->server ) /* Destroy server connection */
    conns->server = DestroyStream(conns->server) ;
}

/* Start a new server-client-save association */
Stream *MakeConnectionStream(char *hostname, Stream *client, QueueNameFn *getname)
{
  Stream *this ;

  static SMTPCallbacks ConnectionCallbacks = {
    ConnectsOpen, ConnectsCommand, ConnectsError, ConnectsClose
  } ;

  if ( (this = NewStream(NULL, NULL, sizeof(ConnectionStream),
                         connectionread, connectionwrite,
                         connectionioctl, connectiondestroy)) != NULL ) {
    ConnectionStream *conns = (ConnectionStream *)this ;
    int response = -1 ;

    /* Init fully in case of SMTP stream failure */
    conns->server = NULL ;
    conns->client = NULL ;
    conns->getname = NULL ;
    conns->save = NULL ;
    conns->smtp = NULL ;
    conns->status = 0 ;
    conns->cookie = CONNECTION_STREAM_COOKIE ;

    if ( (conns->smtp = MakeSMTPStream(&ConnectionCallbacks, this)) != NULL ) {
      if ( !hostname || (conns->server = MakeServerStream(hostname)) == NULL )
        conns->status |= STATUS_ESERVER ;
      conns->client = client ;
      conns->getname = getname ;

      if ( conns->server ) { /* Pass greeting through */
        response = ConnectionResponse(conns->server, client, &conns->status) ;

        if ( response >= 200 && response < 300 &&
             (conns->status & STATUS_ESERVER) == 0 ) {
          char hostname[MAX_HOSTLEN] ;

          /* It's OK, try HELO and response */
          if ( gethostname(hostname, MAX_HOSTLEN) != 0 ||
               SendStream(conns->server, "HELO %s\r\n", hostname) < 0 ) {
            conns->status |= STATUS_ESERVER ;
            response = 421 ; /* Closing channel */
          } else
            response = ConnectionResponse(conns->server, NULL, &conns->status) ;
        }
      }

      /* Send greeting ourself if no server or error reading from server. */
      if ( response < 0 && client && (conns->status & STATUS_ECLIENT) == 0 ) {
        char hostname[MAX_HOSTLEN] ;

        if ( getname == NULL ||
             gethostname(hostname, MAX_HOSTLEN) != 0 ||
             SendStream(client,
                        "220 %s Hello, this is SMTP Buffer "
                        VERSION_STRING
                        " Copyright (C) AJCD 1997-1998\r\n", hostname) < 0 )
          conns->status |= STATUS_ECLIENT ;
        else
          response = 220 ;
      }
    
      if ( response < 200 || response >= 300 || /* Doesn't want to talk :-( */
           (conns->status & STATUS_ECLIENT) != 0 ) { /* Problem with client */
        do {
          this = DestroyStream(this) ;
        } while ( this ) ;
      }
    } else {
      do {
        this = DestroyStream(this) ;
      } while ( this ) ;
    }
  }

  return this ;
}

/* end of connects.c */
