/* servstr.c
 *
 * (C) Angus J. C. Duggan, 1998
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Server stream for SMTP service for Windows NT.
 */

#include "smtp.h"
#include "messages.h"
#include "streami.h"

/* Call async gethostbyname, timing out after 5 secs */
static struct hostent *GetHostByName(char *name)
{
  MSG msg ;
  HANDLE hcall ;
  DWORD timeout ;

  static char buffer[MAXGETHOSTSTRUCT] ;
  static HWND hWnd ;
  static ATOM hClass ;

#ifndef WSAGETASYNCERROR
#define WSAGETASYNCERROR(lParam)            HIWORD(lParam)
#endif
#ifndef WSAGETASYNCBUFLEN
#define WSAGETASYNCBUFLEN(lParam)           LOWORD(lParam)
#endif
#ifndef MWMO_INPUTAVAILABLE
#define MWMO_INPUTAVAILABLE 0
#endif

  if ( hClass == 0 &&
       (hClass = GlobalAddAtom("Async Message Class")) == 0 ) {
    message(EVENTLOG_ERROR_TYPE, MSG_ATOMFAILED, name, ErrorText(GetLastError()), NULL) ;
    return NULL ;
  }

  if ( hWnd == NULL &&
       (hWnd = CreateWindow((LPCTSTR)hClass, /* class name */
                            NULL, /* window name */
                            WS_DISABLED, /* style */
                            0, /* x */
                            0, /* y */
                            0, /* width */
                            0, /* height */
                            NULL, /* parent */
                            NULL, /* menu */
                            NULL, /* instance */
                            NULL /* param */
                            )) == NULL ) {
    message(EVENTLOG_ERROR_TYPE, MSG_WINDOWFAILED, name, ErrorText(GetLastError()), NULL) ;
    return NULL ;
  }

  timeout = GetRegDWORD(SERVICE_REGISTRY_KEY, CONNECT_TIMEOUT_NAME) ;
  if ( timeout <= 0 ) /* Default if silly value */
    timeout = CONNECT_TIMEOUT ;

  if ( PeekMessage(&msg, (HWND)hWnd, WM_USER, WM_USER, PM_REMOVE) ) {
    debug(DEBUG_CONNECT, "Message ready before WSAAsyncGetHostByName\n") ;
  }

  if ( (hcall = WSAAsyncGetHostByName((HWND)hWnd, WM_USER, name, buffer, MAXGETHOSTSTRUCT)) != NULL ) {
  again:
    debug(DEBUG_CONNECT, "Waiting for message from WSAAsyncGetHostByName for %s\n", name) ;
    switch ( MsgWaitForMultipleObjectsEx(0, NULL, timeout * 1000, QS_ALLINPUT,
                                         MWMO_INPUTAVAILABLE)) {
    case WAIT_OBJECT_0: /* Message waiting */
      debug(DEBUG_CONNECT, "Message received...") ;
      if ( PeekMessage(&msg, (HWND)hWnd, WM_USER, WM_USER, PM_REMOVE) ) {
        debug(DEBUG_CONNECT, "is WM_USER (%d)\n", WSAGETASYNCERROR(msg.lParam)) ;
	if ( WSAGETASYNCERROR(msg.lParam) == 0 )
	  return (struct hostent *)buffer ;
	/* Error returned by gethostbyname */
	return NULL ;
      }
      debug(DEBUG_CONNECT, "Not WM_USER\n") ;
      /* Other message */      
      if ( PeekMessage(&msg, (HWND)hWnd, 0, 0, PM_REMOVE) )
	goto again ;
      /* Strange error, fallthrough, cancel & quit */
    default:
      debug(DEBUG_CONNECT,
            "Strange error from MsgWaitForMultipleObjectsEx %d\n",
            GetLastError()) ;
    case WAIT_TIMEOUT:      /* Timed out, cancel request and return */
      (void)WSACancelAsyncRequest(hcall) ;
      break ;
    }
  } else
    message(EVENTLOG_ERROR_TYPE, MSG_ASYNCFAILED, name, ErrorText(GetLastError()), NULL) ;

  return NULL ;
}

/*****************************************************************************/

/* Server stream for connections to and from SMTP server. Writes to socket are
   raw, reads from socket are line-buffered. */
typedef struct {
  Stream *base ;
  Stream *socks ; /* bypass line stream */
  enum { SERVER_STREAM_COOKIE = 0x53455256 } cookie ; /* 'SERV' */
} ServerStream ;

/* Write direct to under-underlying socket stream */
static long serverwriter(Stream *this, void *buffer, long bufsize)
{
  register ServerStream *sserver = (ServerStream *)this ;

  /* Sanity check on stream */
  if ( sserver == NULL || sserver->cookie != SERVER_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

#ifdef DEBUG
  stream_debug(SERVERSTREAM_DEBUG_WRITE, "[SW %04d] ", bufsize) ;
  stream_debugstr(SERVERSTREAM_DEBUG_WRITE, buffer, bufsize) ;
#endif

  return WriteStream(sserver->socks, buffer, bufsize) ;
}

/* Read from line-buffered underlying stream */
static long serverreader(Stream *this, void *buffer, long bufsize)
{
  register ServerStream *sserver = (ServerStream *)this ;
  long rbytes ;

  /* Sanity check on stream */
  if ( sserver == NULL || sserver->cookie != SERVER_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  rbytes = ReadStream(UnderlyingStream(this), buffer, bufsize) ;

#ifdef DEBUG
  stream_debug(SERVERSTREAM_DEBUG_READ, "[SR] %04d ", rbytes) ;
  stream_debugstr(SERVERSTREAM_DEBUG_READ, buffer, rbytes) ;
#endif

  return rbytes ;
}

Stream *MakeServerStream(char *hostname)
{
  Stream *socks, *waste = NULL ;
  struct hostent *hostentry = NULL ;
  struct sockaddr_in sremote ;
  SOCKET outsock ;

  if ( (hostentry = GetHostByName(hostname)) == NULL ) {
    /* Don't log failure, this one is quite probable */
#ifdef DEBUG
    stream_debug(SERVERSTREAM_DEBUG_CONNECT,
                 "Debug: gethostbyname(%s) failed, %s\n", hostname,
                 ErrorText(WSAGetLastError())) ;
#endif
    return NULL ;
  }

  /* Set up remote socket address */
  memset(&sremote, 0, sizeof(struct sockaddr_in)) ;
  sremote.sin_family = AF_INET ;
  sremote.sin_port = outport ;
  memcpy(&sremote.sin_addr, hostentry->h_addr, hostentry->h_length);
      
  if ( INVALID_SOCKET == (outsock = socket(PF_INET, SOCK_STREAM, 0)) ) {
    message(EVENTLOG_ERROR_TYPE, MSG_OPENSOCKET,
            ErrorText(WSAGetLastError()), NULL) ;
    return NULL ;
  }

  if ( SOCKET_ERROR == connect(outsock, (struct sockaddr *)&sremote,
                               sizeof(struct sockaddr_in)) ) {
    /* Don't log failure, this one is quite probable */
#ifdef DEBUG
    stream_debug(SERVERSTREAM_DEBUG_CONNECT,
                 "Debug: connect to server failed, %s\n",
                 ErrorText(WSAGetLastError())) ;
#endif
    return NULL ;
  }

  if ( (socks = MakeSocketStream(outsock)) != NULL ) {
    Stream *lines ;

    waste = socks ;
    if ( (lines = MakeLineStream(socks, "\r\n")) != NULL ) {
      Stream *this ;

      waste = lines ;
      
      if ( (this = NewStream(NULL, lines, sizeof(ServerStream),
                             serverreader, serverwriter, NULL, NULL)) != NULL ) {
        ServerStream *sserver = (ServerStream *)this ;

        sserver->socks = socks ;
        sserver->cookie = SERVER_STREAM_COOKIE ;

        return this ;
      }
    }
  }

  message(EVENTLOG_ERROR_TYPE, MSG_CREATESTREAM,
          ErrorText(GetLastError()), NULL) ;
  closesocket(outsock) ;

  /* Dispose of any partially-constructed streams */
  while ( waste )
    waste = DestroyStream(waste) ;

  return NULL ;
}

/* end of servstr.c */
