/* smtp.c
 *
 * Copyright (C) Angus J. C. Duggan, 1997-1998
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * SMTP outgoing buffer for NT; connects machine specified by registry entry
 * MAILHOST if possible, otherwise buffers outgoing mail, passing OK response
 * to local SMTP client.
 */

#include "smtp.h"
#include "messages.h"

#define MAXDATALEN 1024
#define MAXREPLYLEN 512

char *program ;
int debuglevel ;

char *mailhost = NULL ;
char *queuedir = NULL ;
char *postmaster = NULL ;
unsigned short inport = 0, outport = 0 ;

int code = 0 ;

static void usage(void)
{
  message(EVENTLOG_INFORMATION_TYPE, MSG_USAGE, SERVICE_DISPLAYNAME,
	  VERSION_STRING, SERVICE_NAME, NULL) ;
  if ( !debuglevel++ ) { /* repeat message for stderr */
    message(EVENTLOG_INFORMATION_TYPE, MSG_USAGE, SERVICE_DISPLAYNAME,
	    VERSION_STRING, SERVICE_NAME, NULL) ;
  }
  ExitProcess(0) ;
}

static int MailKill(char *filename, void *passthrough)
{
  char *arg = passthrough ;
  char *subject = NULL, summary[MAX_SUMMARY] ;
  char *tail = strrchr(filename, '\\') ; /* tail end of filename */

  Summary(filename, summary, MAX_SUMMARY) ;

  if ( (subject = strstr(summary, "Subject:")) != NULL ) {
    for ( subject += 8 ; *subject && isspace(*subject) ; ++subject ) ;
  }

  if ( atoi(arg) == atoi(tail + 1) || /* Same message number */
       (subject != NULL && Match(arg, subject)) ) {
    if ( !DeleteFile(filename) )
      printf("Failed to remove message %s\n", summary) ;
    else
      printf("Removed message %s\n", summary) ;
  }

  return TRUE ;
}

/* Bounce a message, by looking at name and renaming to other queue. Messages
   transition through nnn.out, nnn.err, nnn.
   */
typedef struct {
  Stream *outs ;
  int state ;
  int looped ;
  char from[MAX_PATH] ;
} BounceData ;

int BounceCommand(void *data, int type, char *line, long linelen,
                  char *args, long argslen)
{
  BounceData *bounce = data ;
  int response = -1 ;

  switch ( type ) {
  case SMTP_HELO:
    if ( bounce->state != SMTP_OPENED )
      return -1 ;
    bounce->state = SMTP_HELO ;
    return 220 ;
  case SMTP_MAIL:
    if ( bounce->state == SMTP_HELO &&
         argslen >= 5 && _strnicmp(args, "FROM:", 5) == 0 ) {
      args += 5 ;
      argslen -= 5 ;

      while ( argslen > 0 && isspace(*args) ) {
        --argslen; ++args ; /* Drop leading whitespace */
      }

      assert(argslen < MAX_PATH - 1) ;
      strncpy(bounce->from, args, argslen) ;
      bounce->from[argslen] = '\0' ;

      if ( SendStream(bounce->outs, "MAIL FROM: <%s>\r\n", postmaster) < 0 )
        return -1 ;

      bounce->state = SMTP_MAIL ;

      return 250 ;
    }
    return -1 ;
  case SMTP_RCPT:
    if ( (bounce->state == SMTP_MAIL || bounce->state == SMTP_RCPT) &&
         argslen >= 3 && _strnicmp(args, "TO:", 3) == 0 ) {
      args += 3 ;
      argslen -= 3 ;

      while ( argslen > 0 && isspace(*args) ) {
        --argslen; ++args ; /* Drop leading whitespace */
      }

      if ( argslen - 2 != strlen(postmaster) ||
           _strnicmp(args + 1, postmaster, argslen - 2) != 0  )
        bounce->looped = 0 ; /* Wasn't send to postmaster only */

      bounce->state = SMTP_RCPT ;

      return 251 ;
    }
    return -1 ;
  case SMTP_DATA:
    if ( bounce->state == SMTP_RCPT && !bounce->looped ) {
      if ( SendStream(bounce->outs, "RCPT TO: %s\r\n", bounce->from) < 0 )
        return -1 ;

      response = 354 ;
      break ;
    }
    return -1 ;
  case SMTP_MESSAGE:
    if ( bounce->state == SMTP_DATA ) { /* First time in message */
      if ( SendStream(bounce->outs, "X-Bounced-by: SMTP Buffer "
                      VERSION_STRING
                      " Copyright (C) AJCD 1997-1998\r\nX-Bounced-To: %s\r\n",
                      bounce->from) < 0 )
        return -1 ;
    } else if ( bounce->state != SMTP_MESSAGE )
      return -1 ;
    response = 250 ;
    break ;
  case SMTP_DOT:
    if ( bounce->state != SMTP_MESSAGE )
      return -1 ;
    response = 250 ; /* Write to outs */
    break ;
  case SMTP_QUIT:
    if ( bounce->state == SMTP_DOT )
      return 221 ;
    return -1 ;
  default: /* Any other command is an error */
    return -1 ;
  }

  if ( !WriteStreamAll(bounce->outs, line, linelen) )
    return -1 ;

  bounce->state = type ;

  return response ;
}

void BounceMessage(char *filename, int how, QueueNameFn *getname)
{
  char bouncename[MAX_PATH] ;

  assert(filename != NULL) ;

  /* Open FileStream on message, pass to SMTPStream with callbacks:

     BOUNCE_RETURN changes MAIL to RCPT, uses postmaster for MAIL. Adds data
     after header to indicate problem.

     BOUNCE_FATAL moves file aside, and generates bounce report to postmaster.

     BOUNCE_RESEND does nothing.
     */
     
  debug(DEBUG_BOUNCE, "Debug: Bouncing message %s type %d\n", filename, how) ;

  switch ( how ) {
  case BOUNCE_RETURN: { /* Return to sender, or to postmaster */
    Stream *instr = NULL, *outstr = NULL, *smtps = NULL ;
    int bounceok = 0 ;
    BounceData bounce ;

    static SMTPCallbacks bounce_callbacks = {
      NULL, /* Open */
      BounceCommand, /* Command */
      NULL, /* Error */
      NULL, /* Close */
    } ;

    if ( (*getname)(bouncename, MAX_PATH) ) {
      if ( (instr = MakeFileStream(filename, FILESTREAM_READ)) != NULL &&
           (outstr = MakeFileStream(bouncename, FILESTREAM_WRITE)) != NULL &&
           (smtps = MakeSMTPStream(&bounce_callbacks, &bounce)) != NULL ) {
        long rbytes ;
        char buffer[BUFSIZ] ;

        bounce.outs = outstr ;
        bounce.looped = 1 ;
        bounce.state = SMTP_OPENED ;
        strcpy(bounce.from, "nobody") ;

        if ( gethostname(buffer, BUFSIZ) == 0 &&
             SendStream(smtps, "HELO %s\r\n", buffer) >= 0 ) {
          while ( (rbytes = ReadStream(instr, &buffer, BUFSIZ)) > 0 ) {
            if ( !WriteStreamAll(smtps, buffer, rbytes) )
              break ;
          }
          if ( rbytes == 0 &&
               WriteStreamAll(smtps, "QUIT\r\n", 6) ) { /* Wrote through to EOF */
            char *summary ;

            bounceok = 1 ;

            if ( IoctlStream(smtps, SMTPSTREAM_SUMMARY, &summary) != 0 ||
                 summary == NULL )
              summary = filename ;

            message(EVENTLOG_INFORMATION_TYPE, MSG_RETURNED, summary,
                    bounce.from, NULL) ;
          }
        }
      }

      while ( smtps )
        smtps = DestroyStream(smtps) ;
    
      while ( outstr )
        outstr = DestroyStream(outstr) ;

      while ( instr )
        instr = DestroyStream(instr) ;

      /* Quit if we bounced successfully */
      if ( bounceok && DeleteFile(filename) )
        return ; /* Successful bounce */

      (void)DeleteFile(bouncename) ;
    }

    /* FALLTHRU */
  }
  case BOUNCE_FATAL: { /* Move to alternate file name */
    char summary[MAX_SUMMARY], *extn ;

    strcpy(bouncename, filename) ;
    if ( (extn = strrchr(bouncename, '.')) == NULL )
      extn = bouncename + strlen(bouncename) ;

    assert(extn + strlen(Q_ERROR_EXTN) < bouncename + MAX_PATH) ;
    strcpy(extn, Q_ERROR_EXTN) ;

    message(EVENTLOG_WARNING_TYPE, MSG_BOUNCE,
	    Summary(filename, summary, MAX_SUMMARY), NULL) ;
    if ( !MoveFile(filename, bouncename) ) {
      message(EVENTLOG_ERROR_TYPE, MSG_MOVEFILE, filename, bouncename,
	      ErrorText(GetLastError()), NULL) ;
      if ( !DeleteFile(filename) )
	message(EVENTLOG_ERROR_TYPE, MSG_DELETEFILE, filename,
		ErrorText(GetLastError()), NULL) ;
      else
	message(EVENTLOG_INFORMATION_TYPE, MSG_DELETED, filename, NULL) ;
    } else
      message(EVENTLOG_INFORMATION_TYPE, MSG_MOVED, filename, bouncename, NULL) ;

    break ;
  }
  case BOUNCE_RESEND: { /* Tell them we had transient problems */
    char summary[MAX_SUMMARY] ;

    message(EVENTLOG_INFORMATION_TYPE, MSG_TRANSIENT,
            Summary(filename, summary, MAX_SUMMARY), NULL) ;
    break ;
  }
  default:
    assert(0) ;
  }
}

int parse_args(int argc, char *argv[])
{
  int action = ACTION_START ;
  char *arg ;

  for ( program = *argv++ ; --argc ; argv++ ) {
    if (argv[0][0] == '-') {
      char ch = argv[0][1] ;

      /* Get argument if necessary */
      if ( strchr("KFpmdDcs", ch) != NULL ) {
	if ( argv[0][2] != '\0' )
	  arg = &argv[0][2] ;
	else if ( --argc )
	  arg = *++argv ;
	else
	  return ACTION_USAGE ;
      } else if ( argv[0][2] != '\0' )
        return ACTION_USAGE ;

      switch (ch) {
      case 'i':	/* install service */
	action = ACTION_INSTALL ;
	break;
      case 'x':	/* stop service */
        code = SERVICE_CONTROL_STOP ;
	action = ACTION_CONTROL ;
	break;
      case 'r':	/* remove service */
	action = ACTION_UNINSTALL ;
	break;
      case 'q':	{ /* report on queue */
        FileQueue *queue = NULL ;
        char summary[MAX_SUMMARY] ;
        char *status ;
        WSADATA wsaData;

        if ( !DefaultDirectories() )
          return ACTION_USAGE ;

        /* Initialise winsock for gethostname */
        if ( WSAStartup(MAKEWORD(1,1), &wsaData) != 0) {
          message(EVENTLOG_ERROR_TYPE, MSG_WSASTARTUP,
                  ErrorText(WSAGetLastError()), NULL) ;
          return ACTION_DONE ;
        }

        if ( (status = GetRegString(SERVICE_REGISTRY_KEY, SMTP_STATUS)) != NULL ) {
          printf("%s\n", status) ;
          free(status) ;
        }

        ForAllFiles(queuedir, Q_OK_EXTN, ReadQueue, &queue) ;

        while ( queue ) {
          FileQueue *next = queue->next ;
          printf("Queued %s\n", Summary(queue->name, summary, MAX_SUMMARY)) ;
          free(queue) ;
          queue = next ;
        }

        ForAllFiles(queuedir, Q_ERROR_EXTN, ReadQueue, &queue) ;

        while ( queue ) {
          FileQueue *next = queue->next ;
          printf("Bounced %s\n", Summary(queue->name, summary, MAX_SUMMARY)) ;
          free(queue) ;
          queue = next ;
        }

        WSACleanup() ;

        action = ACTION_DONE ;
        break;
      }
      case 'K': { /* kill email */
        WSADATA wsaData;

        if ( !DefaultDirectories() )
          return ACTION_USAGE ;

        /* Initialise winsock for gethostname */
        if ( WSAStartup(MAKEWORD(1,1), &wsaData) != 0) {
          message(EVENTLOG_ERROR_TYPE, MSG_WSASTARTUP,
                  ErrorText(WSAGetLastError()), NULL) ;
          return ACTION_DONE ;
        }

        ForAllFiles(queuedir, Q_OK_EXTN, MailKill, arg) ;
        ForAllFiles(queuedir, Q_ERROR_EXTN, MailKill, arg) ;

        WSACleanup() ;

        action = ACTION_DONE ;
        break ;
      }
      case 'm':	/* set mailhost */
        mailhost = arg ;
	break;
      case 'd':	/* set queue directory */
        queuedir = arg ;
	break;
      case 'c':	/* client port number */
        inport = htons((unsigned short)atoi(arg)) ;
	break;
      case 's':	/* server port number */
        outport = htons((unsigned short)atoi(arg)) ;
	break;
      case 'p':	/* set postmaster */
        postmaster = arg ;
	break;
      case 'D':	/* debug level service */
        debuglevel |= strtol(arg, &arg, 0) ;
	if ( debuglevel == 0 )
	  return ACTION_USAGE ;
	action = ACTION_DEBUG ;
	break;
      case 'F':	/* email forwarding control */
        if ( strcmp(arg, "start") == 0 ) {
          code = SMTP_CONTROL_START_MAIL ;
        } else if ( strcmp(arg, "abort") == 0 ) {
          code = SMTP_CONTROL_ABORT_MAIL  ;
        } else if ( strcmp(arg, "stop") == 0 ) {
          code = SMTP_CONTROL_STOP_MAIL;
        } else
          return ACTION_USAGE ;
        action = ACTION_CONTROL ;
        break ;
      default:
	return ACTION_USAGE ;
      }
    } else
      return ACTION_USAGE ;
  }

  if ( program == NULL )
    return ACTION_USAGE ;

  if ( !DefaultDirectories() )
    return ACTION_USAGE ;

  return action ;
}

int main(int argc, char *argv[])
{
  int action = ACTION_START ;

  SERVICE_TABLE_ENTRY dispatchTable[] = {
    { TEXT(SERVICE_NAME), (LPSERVICE_MAIN_FUNCTION)service_main },
    { NULL, NULL }
  } ;

  switch ( parse_args(argc, argv) ) {
  case ACTION_DONE:
    break ;
  case ACTION_CONTROL:
    if ( !UserService(code) )
      return 1 ;
    break ;
  case ACTION_INSTALL:
    if ( !InstallService() )
      return 1 ;
    break ;
  case ACTION_UNINSTALL:
    if ( !RemoveService() )
      return 2 ;
    break ;
  case ACTION_START:
    if ( !StartServiceCtrlDispatcher(dispatchTable) ) {
      if ( GetLastError() != ERROR_FAILED_SERVICE_CONTROLLER_CONNECT ||
           !RestartService() ) {
        message(EVENTLOG_ERROR_TYPE, MSG_STARTFAILED,
                ErrorText(GetLastError()), NULL) ;
        return 3 ;
      }
    }
    break ;
  case ACTION_DEBUG:
    SetConsoleCtrlHandler(ControlHandler, TRUE) ;
    service_main((DWORD)argc, (LPTSTR *)argv) ;
    break ;
  case ACTION_USAGE:
    usage() ;
  }

  return 0 ;
}

/* end of smtp.c */
