/* smtpstr.c
 *
 * Copyright (C) Angus J. C. Duggan, 1997-1998.
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * SMTP parser stream. Writes to this stream cause callbacks on functions
 * which can forward or file the mail as appropriate.
 */

#include "smtp.h"
#include "messages.h"
#include "streami.h"

/*****************************************************************************/

/* SMTP streams parse SMTP input, calling back functions to dispose of each
   parsed unit separately. They can only be written to, and are self-buffered.
   */

typedef struct {
  Stream *base ;
  SMTPCallbacks *callbacks ;
  void *data ;
  char *buffer ;
  long bufsize, bytes ;
  char *sbuffer ; /* summary */
  long sbufsize, sbytes ;
  enum { STATE_INIT, STATE_READY, STATE_FROM, STATE_RCPT,
         STATE_HEADER, STATE_BODY, STATE_QUIT, STATE_UNCHANGED } state ;
  enum { SMTP_STREAM_COOKIE = 0x534d5450 } cookie ; /* 'SMTP' */
} SMTPStream ;

static int addsummary(SMTPStream *ssmtp, char *summary, long sumlen,
                      char *prefix)
{
  long prefixlen = strlen(prefix) ;
  long size = ssmtp->sbytes + sumlen + prefixlen + 1 ;

  assert(ssmtp->sbytes >= 0) ;
  assert(ssmtp->sbytes <= ssmtp->sbufsize) ;

  /* Add buffered data to current data */
  if ( size > ssmtp->sbufsize ) {
    ssmtp->sbufsize = size ;
    ssmtp->sbuffer = realloc(ssmtp->sbuffer, size) ;

    if ( ssmtp->sbuffer == NULL ) {
      SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
      return 0 ;
    }
  }

  memcpy(ssmtp->sbuffer + ssmtp->sbytes, prefix, prefixlen) ;
  ssmtp->sbytes += prefixlen ;

  memcpy(ssmtp->sbuffer + ssmtp->sbytes, summary, sumlen) ;
  ssmtp->sbytes += sumlen ;

  ssmtp->sbuffer[ssmtp->sbytes] = '\0' ;

  return 1 ;
}

static long smtpwriter(Stream *this, void *buffer, long bufsize)
{
  register SMTPStream *ssmtp = (SMTPStream *)this ;
  register char *term ;			/* Remaining terminator chars */
  long nbytes = 0 ;			/* number of bytes written */
  char *scan ;				/* Cache buffer pointer */
  char *line ;				/* Start of command */
  long lbytes = 0 ;			/* Length to write */
  int matched = 0 ;
  char *smtpterm = "\r\n" ;
  SMTPCallbacks *callbacks = ssmtp->callbacks ;

  static struct command_entry {
    char *name ;
    int len, type ; /* name length, command */
    int nfrom, from[8] ; /* Predecessor states */
    int ok_low, ok_high ; /* Successful response codes */
    int ok_state, err_state ; /* State transitions */
  } smtp_commands[] = {
#define MESSAGE_INDEX 0
    { "", 0, SMTP_MESSAGE, 2, {STATE_HEADER, STATE_BODY},
      200, 300, STATE_UNCHANGED, STATE_UNCHANGED},
#define DOT_INDEX 1
    { ".\r\n", 3, SMTP_DOT, 2, {STATE_HEADER, STATE_BODY},
      200, 300, STATE_READY, STATE_READY},
#define UNKNOWN_INDEX 2
    { "", 0, SMTP_UNKNOWN,
      5, {STATE_INIT, STATE_READY, STATE_FROM, STATE_RCPT, STATE_QUIT},
      200, 300, STATE_UNCHANGED, STATE_UNCHANGED},
#define COMMAND_INDEX 3
    { "HELO ", 5, SMTP_HELO, 1, {STATE_INIT},
      200, 300, STATE_READY, STATE_QUIT},
    { "MAIL ", 5, SMTP_MAIL, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_FROM, STATE_READY},
    { "SEND ", 5, SMTP_SEND, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_FROM, STATE_READY},
    { "SOML ", 5, SMTP_SOML, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_FROM, STATE_READY},
    { "SAML ", 5, SMTP_SAML, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_FROM, STATE_READY},
    { "RCPT ", 5, SMTP_RCPT, 2, {STATE_FROM, STATE_RCPT},
      200, 300, STATE_RCPT, STATE_RCPT},
    { "DATA\r\n", 6, SMTP_DATA, 1, {STATE_RCPT},
      300, 400, STATE_HEADER, STATE_READY},
    { "RSET\r\n", 6, SMTP_RSET, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_READY, STATE_READY},
    { "VRFY ", 5, SMTP_VRFY, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_UNCHANGED, STATE_UNCHANGED},
    /* VRFY should return 252. See RFC1123 */
    { "EXPN ", 5, SMTP_EXPN, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_UNCHANGED, STATE_UNCHANGED},
    { "HELP ", 5, SMTP_HELP, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_UNCHANGED, STATE_UNCHANGED},
    { "HELP\r\n", 5, SMTP_HELP, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_UNCHANGED, STATE_UNCHANGED},
    { "NOOP\r\n", 5, SMTP_NOOP, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_UNCHANGED, STATE_UNCHANGED},
    { "QUIT\r\n", 5, SMTP_QUIT, 3, {STATE_READY, STATE_FROM, STATE_RCPT},
      200, 300, STATE_QUIT, STATE_QUIT},
    { "TURN\r\n", 5, SMTP_TURN, 1, {STATE_READY},
      200, 300, STATE_READY, STATE_READY},
  } ;

  /* Must have stream and valid buffer */
  if ( this == NULL || bufsize < 0 || buffer == NULL ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( ssmtp->cookie != SMTP_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Trivial write succeeds */
  if ( bufsize == 0 )
    return 0 ;

  term = smtpterm ;

  assert(ssmtp->bytes >= 0) ;
  assert(ssmtp->bytes <= ssmtp->bufsize) ;

  /* Add buffered data to current data */
  if ( ssmtp->bytes + bufsize > ssmtp->bufsize ) {
    ssmtp->bufsize = ssmtp->bytes + bufsize ;
    ssmtp->buffer = realloc(ssmtp->buffer, ssmtp->bufsize) ;

    if ( ssmtp->buffer == NULL ) {
      SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
      return -1 ;
    }
  }

  line = scan = ssmtp->buffer ; /* Cache buffer pointer */

  assert(scan) ;

  memcpy(scan + ssmtp->bytes, buffer, bufsize) ;
  ssmtp->bytes += bufsize ;

  /* Deliver commands to callbacks */
  while ( nbytes < ssmtp->bytes ) {
    nbytes++ ;
    if ( *scan++ == *term ) {
      if ( *++term == '\0' ) {	/* Set number of bytes to write */
        int index, cmdindex = UNKNOWN_INDEX ;

        lbytes = scan - line ;
#ifdef DEBUG
        stream_debug(SMTPSTREAM_DEBUG, "[SMTP %04d] ", lbytes) ;
        stream_debugstr(SMTPSTREAM_DEBUG, line, lbytes) ;
#endif
        if ( ssmtp->state == STATE_HEADER ) {
          if ( lbytes == 3 && strncmp(line, ".\r\n", 3) == 0 ) {
            cmdindex = DOT_INDEX ;
          } else { /* Header of message */
            char *start = line ;

            cmdindex = MESSAGE_INDEX ;

            do { /* Add to summary, search for body */
              char *end = start ;

              while ( end < scan && *end++ != '\n' ) ; /* Find newline */

              /* Add Subject to summary */
              if ( end - start >= 8 && _strnicmp("Subject:", line, 8) == 0 ) {
                while ( end > start && isspace(end[-1]) )
                  --end ; /* Drop trailing space */
                if ( !addsummary(ssmtp, start, end - start, "  ") )
                  return -1 ;
                break ;
              } else if ( end - start == 1 && start[0] == '\n' ) {
                ssmtp->state = STATE_BODY ;
                break ; /* Empty line separating body */
              } else if ( end - start == 2 &&
                          start[0] == '\r' && start[1] == '\n' ) {
                ssmtp->state = STATE_BODY ;
                break ; /* Empty line separating body */
              }

              start = end ;
            } while ( start < scan ) ;
          }
        } else if ( ssmtp->state == STATE_BODY ) {
          if ( lbytes == 3 && strncmp(line, ".\r\n", 3) == 0 ) {
            cmdindex = DOT_INDEX ;
          } else { /* Body of message */
            cmdindex = MESSAGE_INDEX ;
          }
        } else {
          for ( index = COMMAND_INDEX ;
                index < sizeof(smtp_commands)/sizeof(struct command_entry) ;
                ++index ) {
            if ( lbytes >= smtp_commands[index].len &&
                 _strnicmp(line,
                           smtp_commands[index].name,
                           smtp_commands[index].len) == 0 ) {
              cmdindex = index ;
              break ;	/* Found command */
            }
          }
        }

        /* Now have cmdindex set. Check that state is valid. */
        for ( index = 0 ; index < smtp_commands[cmdindex].nfrom ; ++index ) {
          if ( smtp_commands[cmdindex].from[index] == ssmtp->state )
            break ;
        }

        /* If state is valid, or error callback exists and copes, allow
           command callback */
        if ( index < smtp_commands[cmdindex].nfrom ||
             (callbacks->error != NULL &&
              (callbacks->error)(ssmtp->data, smtp_commands[cmdindex].type,
                                 line, lbytes)) ) {
          int response = smtp_commands[cmdindex].ok_low ;
          char *arg = line + smtp_commands[cmdindex].len ;
          long abytes = lbytes - smtp_commands[cmdindex].len ;

          while ( abytes > 0 && isspace(*arg) ) {
            --abytes ; ++arg ; /* Skip leading whitespace */
          }

          while ( abytes > 0 && isspace(arg[abytes - 1]) ) {
            --abytes ; /* Skip trailing whitespace */
          }

          if ( callbacks->command ) {
            response = (callbacks->command)(ssmtp->data,
                                            smtp_commands[cmdindex].type,
                                            line, lbytes,
                                            arg, abytes) ;
            if ( response < 0 ) /* Fatal error */
              return -1 ;
          }

          if ( response >= smtp_commands[cmdindex].ok_low &&
               response < smtp_commands[cmdindex].ok_high ) {
            if ( smtp_commands[cmdindex].ok_state != STATE_UNCHANGED )
              ssmtp->state = smtp_commands[cmdindex].ok_state ;
          } else {
            if ( smtp_commands[cmdindex].err_state != STATE_UNCHANGED )
              ssmtp->state = smtp_commands[cmdindex].err_state ;
          }

          switch ( ssmtp->state ) { /* Check new state */
          case STATE_FROM: /* Reset summary for new message */
            ssmtp->sbytes = 0 ;
            break ;
          case STATE_RCPT: /* Add To: to summary */
            if ( abytes >= 3 && _strnicmp(arg, "TO:", 3) == 0 ) {
              abytes -= 3 ; arg += 3 ; /* Skip TO: */

              while ( abytes > 0 && isspace(*arg) ) {
                --abytes ; ++arg ; /* Skip leading whitespace */
              }

              if ( !addsummary(ssmtp, arg, abytes,
                               (ssmtp->sbytes == 0 ? "To: " : ", ")) )
                return -1 ;
            }
            break ;
          }
        } else { /* Failure to change state causes error */
          SetLastError(ERROR_ACCESS_DENIED) ;
          return -1 ;
        }

        /* Set up for next command */
        lbytes = nbytes ;
        line = scan ;
	term = smtpterm ;
	matched = 0 ;
      } else
	matched++ ;
    } else
      while ( matched > 0 ) { /* Reset search string */
	if ( strncmp(scan - matched, term - matched, matched) == 0 )
	  break ;
	term-- ;
	matched-- ;
      }
  }

  /* Remove written bytes from buffer */
  ssmtp->bytes -= lbytes ;
  if ( ssmtp->bytes > 0 )
    memmove(ssmtp->buffer, line, ssmtp->bytes) ;

  return bufsize ;
}

static void smtpdestroy(Stream *this)
{
  SMTPStream *ssmtp = (SMTPStream *)this ;

  assert(this) ;

  /* Sanity check on stream type */
  if ( ssmtp->cookie != SMTP_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return ;
  }

  if ( ssmtp->callbacks->close )
    (ssmtp->callbacks->close)(ssmtp->data) ;

  if ( ssmtp->buffer )
    free(ssmtp->buffer) ;

  if ( ssmtp->sbuffer )
    free(ssmtp->sbuffer) ;
}

/* Just one SMTP ioctl currently, which returns a summary of the last/current
   message */
static long smtpioctl(Stream *this, int op, va_list args)
{
  SMTPStream *ssmtp = (SMTPStream *)this ;
  char **sumptr ;

  /* Sanity check on stream type */
  if ( ssmtp->cookie != SMTP_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  switch ( op ) {
  case SMTPSTREAM_SUMMARY:
    if ( (sumptr = va_arg(args, char **)) == NULL ) {
      SetLastError(ERROR_INVALID_PARAMETER) ;
      return -1 ;
    }
    *sumptr = ssmtp->sbuffer ;
    break ;
  case STREAM_CLEAREOF:
  case STREAM_CLEARERR:
    break ;
  default:
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  return 0 ;
}

/* Fill in details to make a stream into a smtp stream object */
Stream *MakeSMTPStream(SMTPCallbacks *callbacks, void *passthrough)
{
  Stream *this ;

  if ( callbacks == NULL ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return NULL ;
  }

  this = NewStream(NULL, NULL, sizeof(SMTPStream),
                   NULL, smtpwriter, smtpioctl, smtpdestroy) ;

  if ( this ) {
    SMTPStream *ssmtp = (SMTPStream *)this ;
    
    ssmtp->callbacks = callbacks ;
    ssmtp->data = passthrough ;
    ssmtp->bufsize = ssmtp->bytes = 0 ;
    ssmtp->buffer = NULL ;
    ssmtp->cookie = SMTP_STREAM_COOKIE ;
    ssmtp->sbufsize = ssmtp->sbytes = 0 ;
    ssmtp->sbuffer = NULL ;
    ssmtp->state = STATE_INIT ;

    if ( ssmtp->callbacks->open &&
         !(ssmtp->callbacks->open)(ssmtp->data) ) {
      do {
        this = DestroyStream(this) ;
      } while ( this ) ;
    }
  }

  return this ;
}

/* end of smtpstr.c */
