/* dotstr.c
 *
 * Copyright (C) Angus J. C. Duggan, 1997
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Dot-terminated streams
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>
#include <assert.h>

#include "streams.h"
#include "streami.h"

/*****************************************************************************/

typedef struct {
  Stream *base ;
  char *rbuffer ;
  long rbytes, rused ;
  enum { DOT_STREAM_COOKIE = 0x444f5453 } cookie ; /* 'DOTS' */
} DotStream ;

/* Dot stream is for reading line-buffered streams up to .<CRLF>
   Note there is no destructor because the pointer is to the underlying stream
   buffer, not this buffer. There is no writer for dot streams, as we can't
   determine when the stream is finished. */
static long dotreader(Stream *this, void *buffer, long bufsize)
{
  DotStream *sdot = (DotStream *)this ;
  char *line ;
  long nbytes ;

  if ( this == NULL || buffer == NULL || bufsize < 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( sdot->cookie != DOT_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  assert(sdot->rbytes >= 0) ;
  assert(sdot->rused >= 0) ;
  assert(sdot->rused <= sdot->rbytes) ;

  /* Didn't deliver full amount last time, deliver more now */
  nbytes = sdot->rbytes - sdot->rused ;
  if ( nbytes > 0 ) {
    assert(sdot->rbuffer != NULL) ;

    if ( bufsize == 0 ) { /* Self buffered, return rest of line */
      bufsize = nbytes ;
      *((char **)buffer) = sdot->rbuffer + sdot->rused ;
    } else {		/* Externally buffered, copy to buffer */
      if ( nbytes > bufsize )
	nbytes = bufsize ;
      memcpy(buffer, sdot->rbuffer + sdot->rused, nbytes) ;
    }

    sdot->rused += nbytes ;

    return bufsize ;
  }

  /* Nothing left to deliver, get another line from underlying stream. */
  nbytes = ReadStream(UnderlyingStream(this), &line, 0) ;

  if ( nbytes < 0 )	/* Error in underlying stream */
    return nbytes ;
  
  if ( nbytes == 0 ) {	/* EOF in underlying stream before dot found */
    SetLastError(ERROR_IO_DEVICE) ;
    return -1 ;
  }

  if ( nbytes == 3 && line[0] == '.' && line[1] == '\r' && line[2] == '\n' )
    return 0 ;	/* Found .<CRLF> */

  if ( bufsize == 0 ) {	/* Return whole buffered line */
    *((char **)buffer) = line ;
  } else {		/* Return as much as possible */
    sdot->rbuffer = line ;	/* Note underlying self-buffer address */
    sdot->rbytes = nbytes ;	/* And line length */

    if ( nbytes > bufsize )
      nbytes = bufsize ;
    memcpy(buffer, line, nbytes) ;

    sdot->rused = nbytes ;	/* Set amount used */
  }

  return nbytes ;
}

Stream *MakeDotStream(Stream *under)
{
  Stream *this = NewStream(NULL, under, sizeof(DotStream),
			   dotreader, NULL, NULL, NULL) ;

  if ( this ) {
    DotStream *sdot = (DotStream *)this ;

    sdot->rbytes = sdot->rused = 0 ;
    sdot->rbuffer = NULL ;
    sdot->cookie = DOT_STREAM_COOKIE ;
  }

  return this ;
}

