/* filestr.c $Id: filestr.c,v 1.2 2000/01/14 21:20:04 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * File streams
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <assert.h>

#ifdef WIN32
#include <fcntl.h>
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#else
#include "not_yet_implemented"
#endif

#include "streams.h"
#include "streami.h"

/*****************************************************************************/

/* File streams replicate the semantics of the underlying files quite
   closely. They can be called in self-buffered or non-buffered mode to read.
   */

typedef struct {
  Stream *base ;
  HANDLE handle ;
  char *rbuffer ;	/* Read buffer; writes are passed through */
  long rbufsize ;
#ifndef NDEBUG
  long debug ;
#endif
  enum { FILE_STREAM_COOKIE = 0x46494c45 } cookie ; /* 'FILE' */
} FileStream ;

static long filereader(Stream *this, void *buffer, long bufsize)
{
  register FileStream *sfile = (FileStream *)this ;
  DWORD result ;

  if ( this == NULL || buffer == NULL || bufsize < 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( sfile->cookie != FILE_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Create buffer if self buffered and not done already */
  if ( bufsize == 0 ) {
    if ( sfile->rbuffer == NULL ) {
      sfile->rbufsize = BUFSIZ ;
      sfile->rbuffer = malloc(sfile->rbufsize) ;
      if ( sfile->rbuffer == NULL ) {
	SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
	return -1 ;
      }
    }

    *((char **)buffer) = sfile->rbuffer ;
    bufsize = sfile->rbufsize ;
    buffer = sfile->rbuffer ;
  }

  if ( !ReadFile(sfile->handle, buffer, (DWORD)bufsize, &result, NULL) ) {
    SetLastError(ERROR_IO_DEVICE) ;
    return -1 ;
  }

#ifndef NDEBUG
  stream_debug(sfile->debug,
	       "Debug: Got %d bytes from HANDLE %d\n", result,
	       sfile->handle) ;
#endif

  return (long)result ;
}

static long filewriter(Stream *this, void *buffer, long bufsize)
{
  register FileStream *sfile = (FileStream *)this ;
  DWORD wbytes ;

  if ( this == NULL || buffer == NULL || bufsize <= 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( sfile->cookie != FILE_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  if ( !WriteFile(sfile->handle, buffer, (DWORD)bufsize, &wbytes,
		  NULL) ) {
    SetLastError(ERROR_IO_DEVICE) ;
    return -1 ;
  }

#ifndef NDEBUG
  stream_debug(sfile->debug,
	       "Debug: Wrote %d bytes to HANDLE %d\n", wbytes,
	       sfile->handle) ;
#endif

  return (long)wbytes ;
}

static void filedestroy(Stream *this)
{
  FileStream *sfile = (FileStream *)this ;

  assert(this) ;

#ifndef NDEBUG
  stream_debug(sfile->debug, "Debug: closing HANDLE %d\n", sfile->handle) ;
#endif

  /* Sanity check on stream type */
  if ( sfile->cookie != FILE_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return ;
  }

  if ( !CloseHandle(sfile->handle) )
    ErrorStream(this, FILESTREAM_ERROR_CLOSE, (long)sfile->handle,
                (long)GetLastError()) ;

  if ( sfile->rbuffer )
    free(sfile->rbuffer) ;
}

/* Just one file ioctl currently, which rewinds the file to the start */
static long fileioctl(Stream *this, long op, va_list args)
{
  FileStream *sfile = (FileStream *)this ;
  DWORD where = 0 ;

  /* Sanity check on stream type */
  if ( sfile->cookie != FILE_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  switch ( op ) {
  case STREAM_POSITION:
    where = va_arg(args, long) ;
  case STREAM_REWIND:
    if ( SetFilePointer(sfile->handle, where, NULL,
			where < 0 ? FILE_END : FILE_BEGIN) == 0xFFFFFFFF ) {
      SetLastError(ERROR_IO_DEVICE) ;
      return -1 ;
    }
    break ;
  case STREAM_CLEAREOF:
  case STREAM_CLEARERR:
    break ;
  case STREAM_DEBUG:
#ifndef NDEBUG
    {
      long arg1 = va_arg(args, long), arg2 = va_arg(args, long) ;

      if ( sfile->debug == arg1 )
        sfile->debug = arg2 ;
    }
#endif
    break ;
  default: /* Base stream, don't forward to any other callers */
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  return 0 ;
}

/* Fill in details to make a stream into a file stream object */
Stream *MakeFileStream(char *filename, int access, Error fileerror)
{
  Stream *this ;
  HANDLE handle ;
  DWORD disp = 0 ;
  Handler *reader = NULL, *writer = NULL ;
  DWORD flags = 0 ;

  switch ( access ) {
  case FILESTREAM_READ:
    reader = filereader ;
    disp = OPEN_EXISTING ;
    flags = GENERIC_READ ;
    break ;
  case FILESTREAM_WRITE:
    writer = filewriter ;
    disp = CREATE_ALWAYS ;
    flags = GENERIC_WRITE ;
    break ;
  case FILESTREAM_READWRITE:
    reader = filereader ;
    writer = filewriter ;
    disp = OPEN_ALWAYS ;
    flags = GENERIC_READ|GENERIC_WRITE ;
    break ;
  default:
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return NULL ;
  }

  if ( (handle = CreateFile(filename, flags, 0, NULL,
			    disp, FILE_FLAG_SEQUENTIAL_SCAN, NULL)) ==
       INVALID_HANDLE_VALUE ) {
    SetLastError(ERROR_ACCESS_DENIED) ;
    return NULL ;
  }

  
  this = NewStream(NULL, NULL, sizeof(FileStream),
		   reader, writer, fileioctl, fileerror, filedestroy) ;

  if ( this ) {
    FileStream *sfile = (FileStream *)this ;

    sfile->handle = handle ;
    sfile->rbufsize = 0 ;
    sfile->rbuffer = NULL ;
#ifndef NDEBUG
    sfile->debug = FILESTREAM_DEBUG ;
#endif
    sfile->cookie = FILE_STREAM_COOKIE ;
  } else
    CloseHandle(handle) ;

  return this ;
}

/*
 * $Log: filestr.c,v $
 * Revision 1.2  2000/01/14 21:20:04  angus
 * Add STREAM_DEBUG ioctl, to modify returned debugging information. Update
 * release number to 1.31.
 *
 * Revision 1.1.1.1  1999/10/24 07:06:53  angus
 * Initial CVS import
 *
 */
