/* linestr.c
 *
 * Copyright (C) Angus J. C. Duggan, 1997
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Line-buffered streams
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>
#include <assert.h>

#include "streams.h"
#include "streami.h"

/*****************************************************************************/

/* Line streams are inherently buffered; they return a buffer with a whole line
   of input, or buffer up a whole line before flushing it to the output. They
   cannot read into supplied buffers. The underlying stream for a line stream
   must support supplied-buffer reads. */

typedef struct {
  Stream *base ;
  char *terminator ;				/* End conditions */
  char *rbuffer, *wbuffer ;
  long rbufsize, rbytes, rused, wbufsize, wbytes ;
  enum { LINE_STREAM_COOKIE = 0x4c494e45 } cookie ; /* 'LINE' */
} LineStream ;

static long linereader(Stream *this, void *buffer, long bufsize)
{
  register LineStream *sline = (LineStream *)this ;
  register char *term ;			/* Remaining terminator chars */
  register long nbytes = 0 ;		/* number of bytes read */
  int matched = 0 ;

  /* Must have stream and buffer return pointer, but can't use caller's
     buffer */
  if ( this == NULL || bufsize != 0 || buffer == NULL ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( sline->cookie != LINE_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  term = sline->terminator ;

  assert(sline->rbytes >= 0) ;
  assert(sline->rbytes <= sline->rbufsize) ;
  assert(sline->rused >= 0) ;
  assert(sline->rused <= sline->rbytes) ;

  /* Remove initial used portion of buffer */
  if ( sline->rused > 0 ) {
    long left = sline->rbytes - sline->rused ;

    assert(sline->rbuffer != NULL) ;
    assert(left >= 0 && left <= sline->rbytes) ;

    if ( left > 0 )
      memmove(sline->rbuffer, sline->rbuffer + sline->rused, left) ;

    sline->rbytes -= sline->rused ;
    sline->rused = 0 ;
  }

  for (;;) {
    register char *scan = sline->rbuffer + nbytes ; /* Cache buffer pointer */
    long result ;

    /* Scan buffer contents for terminator */
    while ( nbytes < sline->rbytes ) {
      nbytes++ ;
      if ( *scan++ == *term ) {
	if ( *++term == '\0' ) {
	  /* OK, we've found the end of the termination string */
	  sline->rused = nbytes ;	/* Mark used portion as invalid */
	  *((char **)buffer) = sline->rbuffer ;

#ifdef DEBUG
	  stream_debugstr(LINESTREAM_DEBUG_WRITE, sline->rbuffer, nbytes) ;
#endif

	  return nbytes ;	/* Return number of bytes available */
	} else
	  matched++ ;
      } else
	while ( matched > 0 ) { /* Reset search string */
	  if ( strncmp(scan - matched, term - matched, matched) == 0 )
	    break ;
	  term-- ;
	  matched-- ;
	}
    }

    /* Need more, run out of bytes to scan */

    /* Extend buffer if there is no more space to read into */
    if ( sline->rbytes == sline->rbufsize ) {
      sline->rbufsize += BUFSIZ ;
      sline->rbuffer = realloc(sline->rbuffer, sline->rbufsize) ;

      if ( sline->rbuffer == NULL ) {
	SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
	return -1 ;
      }
    }


    assert(sline->rbuffer != NULL) ;

    /* read bytes from underlying stream */
    result = ReadStream(UnderlyingStream(this), sline->rbuffer + sline->rbytes,
			sline->rbufsize - sline->rbytes) ;

    if ( result < 0 )	/* Error in underlying stream */
      return result ;

    if ( result == 0 ) {
      if ( sline->rbytes > 0 ) { /* EOF in underlying stream before terminator found */
	SetLastError(ERROR_IO_DEVICE) ;
	return -1 ;
      }
      return 0 ;
    }

    sline->rbytes += result ;
  }

  return -1 ;	/* Code not accessible */
}

static long linewriter(Stream *this, void *buffer, long bufsize)
{
  register LineStream *sline = (LineStream *)this ;
  register char *term ;			/* Remaining terminator chars */
  long nbytes = 0 ;			/* number of bytes written */
  char *scan ;				/* Cache buffer pointer */
  long lbytes = 0 ;			/* Length to write */
  int matched = 0 ;

  /* Must have stream and valid buffer */
  if ( this == NULL || bufsize < 0 || buffer == NULL ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( sline->cookie != LINE_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Trivial write succeeds */
  if ( bufsize == 0 )
    return 0 ;

  term = sline->terminator ;

  assert(sline->wbytes >= 0) ;
  assert(sline->wbytes <= sline->wbufsize) ;

  /* Add buffered data to current data */
  if ( sline->wbytes + bufsize > sline->wbufsize ) {
    sline->wbufsize = sline->wbytes + bufsize ;
    sline->wbuffer = realloc(sline->wbuffer, sline->wbufsize) ;

    if ( sline->wbuffer == NULL ) {
      SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
      return -1 ;
    }
  }

  scan = sline->wbuffer ; /* Cache buffer pointer */

  assert(scan) ;

  memcpy(scan + sline->wbytes, buffer, bufsize) ;
  sline->wbytes += bufsize ;

  /* Deliver as many complete lines as possible to underlying stream */
  while ( nbytes < sline->wbytes ) {
    nbytes++ ;
    if ( *scan++ == *term ) {
      if ( *++term == '\0' ) {	/* Set number of bytes to write */
	lbytes = nbytes ;
	term = sline->terminator ;
	matched = 0 ;
      } else
	matched++ ;
    } else
      while ( matched > 0 ) { /* Reset search string */
	if ( strncmp(scan - matched, term - matched, matched) == 0 )
	  break ;
	term-- ;
	matched-- ;
      }
  }

  /* Write completed lines */
  nbytes = 0 ;
  scan = sline->wbuffer ;

#ifdef DEBUG
  stream_debugstr(LINESTREAM_DEBUG_READ, scan, lbytes) ;
#endif

  while ( nbytes < lbytes ) {
    long result = WriteStream(UnderlyingStream(this), scan, lbytes) ;

    if ( result < 0 )	/* Error in underlying stream */
      return result ;

    if ( result == 0 ) { /* EOF in underlying stream before written */
      SetLastError(ERROR_IO_DEVICE) ;
      return -1 ;
    }

    scan += result ;
    nbytes += result ;
  }

  /* Remove written bytes from buffer */
  sline->wbytes -= lbytes ;
  if ( sline->wbytes > 0 )
    memmove(sline->wbuffer, scan, sline->wbytes) ;

  return bufsize ;
}

/* Destructor for line stream, frees allocated buffers */
static void linedestroy(Stream *this)
{
  LineStream *sline = (LineStream *)this ;

  /* Sanity check on stream type */
  if ( sline->cookie != LINE_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return ;
  }

  if ( sline->rbuffer )
    free(sline->rbuffer) ;

  if ( sline->wbuffer )
    free(sline->wbuffer) ;
}

/* Just one line ioctl currently, which resets the terminator */
static long lineioctl(Stream *this, int op, va_list args)
{
  LineStream *sline = (LineStream *)this ;
  char *term ;

  /* Sanity check on stream type */
  if ( sline->cookie != LINE_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  switch ( op ) {
  case LINESTREAM_AGAIN:
    if ( sline->rbuffer == NULL || sline->rbufsize == 0 ) {
      SetLastError(ERROR_INVALID_PARAMETER) ;
      return -1 ;
    }
    sline->rused = 0 ;
    break ;
  case LINESTREAM_TERM:
    if ( (term = va_arg(args, char *)) == NULL ) {
      SetLastError(ERROR_INVALID_PARAMETER) ;
      return -1 ;
    }
    sline->terminator = term ;
    break ;
  case STREAM_POSITION:
    sline->rused = sline->rbytes = 0 ;
    return IoctlStream(UnderlyingStream(this), STREAM_POSITION,
		       va_arg(args, long)) ;
  case STREAM_REWIND:
    sline->rused = sline->rbytes = 0 ;
    return IoctlStream(UnderlyingStream(this), STREAM_REWIND) ;
  case STREAM_CLEAREOF:
  case STREAM_CLEARERR:
    break ;
  default:
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  return 0 ;
}

/* Constructor for LineStreams */
Stream *MakeLineStream(Stream *under, char *term)
{
  Stream *this = NewStream(NULL, under, sizeof(LineStream),
			   linereader, linewriter, lineioctl, linedestroy) ;
  if ( this ) {
    LineStream *sline = (LineStream *)this ;

    sline->terminator = term ;
    sline->rbufsize = sline->rbytes = 0 ;
    sline->wbufsize = sline->wbytes = 0 ;
    sline->rbuffer = sline->wbuffer = NULL ;
    sline->cookie = LINE_STREAM_COOKIE ;
  }

  return this ;
}

/*****************************************************************************/

/* Response function is for reading response code from line-buffered stream */
long Response(Stream *this)
{
  char *buffer = NULL ;
  long result = ReadStream(this, &buffer, 0) ;
  long code = 0 ;

  if ( result < 0 || buffer == NULL || !isdigit(*buffer) )
    return -1 ;

  while ( isdigit(*buffer) ) {	/* Get response code */
    code *= 10 ;
    code += *buffer++ - '0' ;
  }

  return code ;
}

/* Get reason for last response */
long Reason(Stream *this, char *buffer, int bufsize)
{
  long rbytes ;
  char *line ;

  if ( buffer == NULL || bufsize <= 1 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  buffer[0] = '\0' ;

  if ( IoctlStream(this, LINESTREAM_AGAIN) < 0 )
    return -1 ;

  if ( (rbytes = ReadStream(this, &line, 0)) <= 0 )
    return -1 ;

  if ( rbytes >= bufsize )
    rbytes = bufsize - 1 ;

  strncpy(buffer, line, rbytes) ;
  buffer[rbytes] = '\0' ;
    
  return rbytes ;
}

/* end of linestr.c */
