/* sockstr.c
 *
 * Copyright (C) Angus J. C. Duggan, 1997
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Socket streams
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>
#include <assert.h>

#ifdef WIN32
#include <winsock2.h>
#include <windows.h>
#else
#error not yet implemented
#endif

#include "streams.h"
#include "streami.h"

/*****************************************************************************/

/* Socket streams replicate the semantics of the underlying socket quite
   closely. They can be called in self-buffered or non-buffered mode to read.
   */
typedef struct {
  Stream *base ;
  SOCKET socket ;
  char *rbuffer ;	/* Read buffer; writes are passed through */
  long rbufsize ;
  enum { SOCKET_STREAM_COOKIE = 0x534f434b } cookie ; /* 'SOCK' */
} SocketStream ;

static long socketreader(Stream *this, void *buffer, long bufsize)
{
  register SocketStream *ssock = (SocketStream *)this ;
  long result ;

  if ( this == NULL || buffer == NULL || bufsize < 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( ssock->cookie != SOCKET_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Create buffer if self buffered and not done already */
  if ( bufsize == 0 ) {
    if ( ssock->rbuffer == NULL ) {
      ssock->rbufsize = BUFSIZ ;
      ssock->rbuffer = malloc(ssock->rbufsize) ;
      if ( ssock->rbuffer == NULL ) {
	SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
	return -1 ;
      }
    }

    *((char **)buffer) = ssock->rbuffer ;
    bufsize = ssock->rbufsize ;
    buffer = ssock->rbuffer ;
  }

  result = recv(ssock->socket, buffer, (int)bufsize, 0) ;

  if ( result == SOCKET_ERROR )
    return -1 ;

#ifdef DEBUG
  stream_debug(SOCKSTREAM_DEBUG,
	       "Debug: Got %d bytes from SOCKET %d\n", result,
	       ssock->socket) ;
#endif

  return result ;
}

static long socketwriter(Stream *this, void *buffer, long bufsize)
{
  register SocketStream *ssock = (SocketStream *)this ;
  int wbytes ;

  if ( this == NULL || buffer == NULL || bufsize <= 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( ssock->cookie != SOCKET_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  wbytes = send(ssock->socket, buffer, (int)bufsize, 0) ;

#ifdef DEBUG
  stream_debug(SOCKSTREAM_DEBUG,
	       "Debug: Sent %d bytes to SOCKET %d\n", wbytes,
	       ssock->socket) ;
#endif

  if ( wbytes == SOCKET_ERROR )
    return -1 ;

  return wbytes ;
}

static void socketdestroy(Stream *this)
{
  SocketStream *ssock = (SocketStream *)this ;

  /* Sanity check on stream type */
  if ( ssock->cookie != SOCKET_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return ;
  }

#ifdef DEBUG
  stream_debug(SOCKSTREAM_DEBUG,
	       "Debug: closing SOCKET %d\n", ssock->socket) ;
#endif

  if ( closesocket(ssock->socket) != 0 )
    stream_log(STREAM_ERROR_CLOSE_SOCKET, (long)WSAGetLastError()) ;

  if ( ssock->rbuffer )
    free(ssock->rbuffer) ;
}

/* Fill in details to make a stream into a socket stream object */
Stream *MakeSocketStream(StreamSocket socket)
{
  Stream *this = NewStream(NULL, NULL, sizeof(SocketStream),
			   socketreader, socketwriter, NULL, socketdestroy) ;

  if ( this ) {
    SocketStream *ssock = (SocketStream *)this ;
    ssock->socket = socket ;
    ssock->rbufsize = 0 ;
    ssock->rbuffer = NULL ;
    ssock->cookie = SOCKET_STREAM_COOKIE ;
  }

  return this ;
}

/* End of sockstr.c */
