/* sockstr.c $Id: sockstr.c,v 1.2 2000/01/14 21:20:04 angus Exp $
 *
 * Copyright (C) Angus J. C. Duggan, 1997
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Socket streams
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>
#include <assert.h>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#else
#error not yet implemented
#endif

#include "streams.h"
#include "streami.h"

/*****************************************************************************/

/* Socket streams replicate the semantics of the underlying socket quite
   closely. They can be called in self-buffered or non-buffered mode to read.
   */
typedef struct {
  Stream *base ;
  SOCKET socket ;
  char *rbuffer ;	/* Read buffer; writes are passed through */
  long rbufsize ;
#ifndef NDEBUG
  long debug ;
#endif
  enum { SOCKET_STREAM_COOKIE = 0x534f434b } cookie ; /* 'SOCK' */
} SocketStream ;

static long socketreader(Stream *this, void *buffer, long bufsize)
{
  register SocketStream *ssock = (SocketStream *)this ;
  long result ;

  if ( this == NULL || buffer == NULL || bufsize < 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( ssock->cookie != SOCKET_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Create buffer if self buffered and not done already */
  if ( bufsize == 0 ) {
    if ( ssock->rbuffer == NULL ) {
      ssock->rbufsize = BUFSIZ ;
      ssock->rbuffer = malloc(ssock->rbufsize) ;
      if ( ssock->rbuffer == NULL ) {
	SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
	return -1 ;
      }
    }

    *((char **)buffer) = ssock->rbuffer ;
    bufsize = ssock->rbufsize ;
    buffer = ssock->rbuffer ;
  }

  result = recv(ssock->socket, buffer, (int)bufsize, 0) ;

  if ( result == SOCKET_ERROR )
    return -1 ;

#ifndef NDEBUG
  stream_debug(ssock->debug,
	       "Debug: Got %d bytes from SOCKET %d\n", result,
	       ssock->socket) ;
#endif

  return result ;
}

static long socketwriter(Stream *this, void *buffer, long bufsize)
{
  register SocketStream *ssock = (SocketStream *)this ;
  int wbytes ;

  if ( this == NULL || buffer == NULL || bufsize <= 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( ssock->cookie != SOCKET_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  wbytes = send(ssock->socket, buffer, (int)bufsize, 0) ;

#ifndef NDEBUG
  stream_debug(ssock->debug,
	       "Debug: Sent %d bytes to SOCKET %d\n", wbytes,
	       ssock->socket) ;
#endif

  if ( wbytes == SOCKET_ERROR )
    return -1 ;

  return wbytes ;
}

static void socketdestroy(Stream *this)
{
  SocketStream *ssock = (SocketStream *)this ;

  /* Sanity check on stream type */
  if ( ssock->cookie != SOCKET_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return ;
  }

#ifndef NDEBUG
  stream_debug(ssock->debug,
	       "Debug: closing SOCKET %d\n", ssock->socket) ;
#endif

  if ( closesocket(ssock->socket) != 0 )
    ErrorStream(this, SOCKSTREAM_ERROR_CLOSE, (long)ssock->socket,
                (long)WSAGetLastError()) ;

  if ( ssock->rbuffer )
    free(ssock->rbuffer) ;
}

/* No socket-specific ioctls */
static long socketioctl(Stream *this, long op, va_list args)
{
  SocketStream *ssock = (SocketStream *)this ;

  /* Sanity check on stream type */
  if ( ssock->cookie != SOCKET_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  switch ( op ) {
  case STREAM_CLEAREOF:
  case STREAM_CLEARERR:
    break ;
  case STREAM_DEBUG:
#ifndef NDEBUG
    {
      long arg1 = va_arg(args, long), arg2 = va_arg(args, long) ;

      if ( ssock->debug == arg1 )
        ssock->debug = arg2 ;
    }
#endif
    break ;
  case STREAM_POSITION:
  case STREAM_REWIND:
  default: /* Base stream, don't forward to any other callers */
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  return 0 ;
}

/* Fill in details to make a stream into a socket stream object */
Stream *MakeSocketStream(StreamSocket socket, Error socketerror)
{
  Stream *this = NewStream(NULL, NULL, sizeof(SocketStream),
			   socketreader, socketwriter, socketioctl,
                           socketerror, socketdestroy) ;

  if ( this ) {
    SocketStream *ssock = (SocketStream *)this ;
    ssock->socket = socket ;
    ssock->rbufsize = 0 ;
    ssock->rbuffer = NULL ;
#ifndef NDEBUG
    ssock->debug = SOCKSTREAM_DEBUG ;
#endif
    ssock->cookie = SOCKET_STREAM_COOKIE ;
  }

  return this ;
}

/*
 * $Log: sockstr.c,v $
 * Revision 1.2  2000/01/14 21:20:04  angus
 * Add STREAM_DEBUG ioctl, to modify returned debugging information. Update
 * release number to 1.31.
 *
 * Revision 1.1.1.1  1999/10/24 07:06:53  angus
 * Initial CVS import
 *
 */
