/* srchstr.c
 *
 * Copyright (C) Angus J. C. Duggan, 1997
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Search streams
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>
#include <assert.h>

#include "streams.h"
#include "streami.h"

/* Search stream; fast incremental search for occurrence of byte pattern in
   stream. Returns EOF on iteration after found, Ioctls provided to reset
   search and tell whether search was success or failure. */
typedef struct {
  Stream *base ;
  unsigned char *search ;
  int *prefix, len, index ;
  enum { SEARCH_STREAM_COOKIE = 0x53524348 } cookie ;
} SearchStream ;

static long search_reader(Stream *this, void *buffer, long bufsize)
{
  SearchStream *sstr = (SearchStream *)this ;
  long rbytes ;

  if ( this == NULL || buffer == NULL || bufsize < 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( sstr->cookie != SEARCH_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  if ( sstr->index == sstr->len ) /* Found it! */
    return 0 ;

  rbytes = sstr->len - sstr->index ;

  assert(rbytes > 0) ; /* Otherwise would have returned EOF */

  /* At least rbytes bytes remaining in pattern, read this number if
     possible */
  if ( rbytes > bufsize )
    rbytes = bufsize ;

  if ( (rbytes = ReadStream(UnderlyingStream(this), buffer, rbytes)) > 0 ) {
    long i ;
    unsigned char *got = buffer ;

    for ( i = 0 ; i < rbytes ; ++i ) {
      for (;;) {
	if ( got[i] == sstr->search[sstr->index] ) {
	  sstr->index++ ;
	  break ;
	}
	if ( sstr->index == 0 )
	  break ;
	sstr->index = sstr->prefix[sstr->index] ;
      }
    }
  }
  
  return rbytes ;
}

static long search_ioctl(Stream *this, int op, va_list args)
{
  SearchStream *sstr = (SearchStream *)this ;

  /* Sanity check on stream type */
  if ( sstr->cookie != SEARCH_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  switch ( op ) {
  case SEARCHSTREAM_AGAIN:
    sstr->index = 0 ;
    break ;
  case SEARCHSTREAM_RESULT:
    return (sstr->index == sstr->len) ;
  case STREAM_REWIND:
    sstr->index = 0 ;
    return IoctlStream(UnderlyingStream(this), STREAM_REWIND) ;
  case STREAM_POSITION:
    sstr->index = 0 ;
    return IoctlStream(UnderlyingStream(this), STREAM_POSITION,
		       va_arg(args, long)) ;
  case STREAM_CLEAREOF:
  case STREAM_CLEARERR:
    break ;
  default:
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  return 0 ;
}

static void search_destroy(Stream *this)
{
  SearchStream *sstr = (SearchStream *)this ;

  /* Sanity check on stream type */
  if ( sstr->cookie != SEARCH_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return ;
  }

  if ( sstr->prefix )
    free(sstr->prefix) ;
}

static Stream *MakeSearchStream(Stream *under, void *search, int len)
{
  int *prefix ;
  Stream *this ;

  if ( !under || !search || len < 1 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return NULL ;
  }

  if ( (prefix = malloc(sizeof(int) * len)) == NULL ) {
    SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
    return NULL ;
  }

  if ( (this = NewStream(NULL, under, sizeof(SearchStream), search_reader,
			 NULL, search_ioctl, search_destroy)) != NULL ) {
    SearchStream *sstr = (SearchStream *)this ;
    unsigned char *find = search ;
    int i, last = 0 ;

    /* Set prefix; these indicate how far back to shift the search focus
       (index) if the character does not match. This ensures that each
       character is read only once. */
    for ( i = 0 ; i < len ; ++i ) {
      prefix[i] = last ;
      while ( i > 0 ) {
	if ( find[i] == find[last] ) {
	  ++last ;
	  break ;
	}
	if ( last == 0 )
	  break ;
	last = prefix[last] ;
      }
    }

    sstr->search = search ;
    sstr->prefix = prefix ;
    sstr->len = len ;
    sstr->index = 0 ;
    sstr->cookie = SEARCH_STREAM_COOKIE ;
  } else
    free(prefix) ;

  return this ;
}

/* End of srchstr.c */
