/* stdiostr.c
 *
 * Copyright (C) Angus J. C. Duggan, 1998
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Standard I/O streams
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>
#include <assert.h>

#include "streams.h"
#include "streami.h"

/*****************************************************************************/

/* Standard IO streams copy data to and from standard I/O files. They can be
   called in self-buffered or non-buffered mode to read. */

typedef struct {
  Stream *base ;
  FILE *fp ;
  char *rbuffer ;	/* Read buffer; writes are passed through */
  long rbufsize ;
  enum { STDIO_STREAM_COOKIE = 0x5354494f } cookie ; /* 'STIO' */
} StdioStream ;

static long stdioreader(Stream *this, void *buffer, long bufsize)
{
  register StdioStream *sstdio = (StdioStream *)this ;
  long result ;

  if ( this == NULL || buffer == NULL || bufsize < 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( sstdio->cookie != STDIO_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Create buffer if self buffered and not done already */
  if ( bufsize == 0 ) {
    if ( sstdio->rbuffer == NULL ) {
      sstdio->rbufsize = BUFSIZ ;
      sstdio->rbuffer = malloc(sstdio->rbufsize) ;
      if ( sstdio->rbuffer == NULL ) {
	SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
	return -1 ;
      }
    }

    *((char **)buffer) = sstdio->rbuffer ;
    bufsize = sstdio->rbufsize ;
    buffer = sstdio->rbuffer ;
  }

  if ( feof(sstdio->fp) )
    result = 0 ;
  else if ( (result = fread(buffer, 1, bufsize, sstdio->fp)) == 0 &&
	    ferror(sstdio->fp) ) {
    SetLastError(ERROR_IO_DEVICE) ;
    return -1 ;
  }

#ifdef DEBUG
  stream_debug(STDIOSTREAM_DEBUG,
	       "Debug: Got %d bytes from FILE %d\n", result,
	       sstdio->fp) ;
#endif

  return result ;
}

static long stdiowriter(Stream *this, void *buffer, long bufsize)
{
  register StdioStream *sstdio = (StdioStream *)this ;

  if ( this == NULL || buffer == NULL || bufsize <= 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  /* Sanity check on stream type */
  if ( sstdio->cookie != STDIO_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  if ( fwrite(buffer, 1, bufsize, sstdio->fp) != bufsize ) {
    SetLastError(ERROR_IO_DEVICE) ;
    return -1 ;
  }

#ifdef DEBUG
  stream_debug(STDIOSTREAM_DEBUG,
	       "Debug: Wrote %d bytes to FILE %d\n", bufsize,
	       sstdio->fp) ;
#endif

  return bufsize ;
}

static void stdiodestroy(Stream *this)
{
  StdioStream *sstdio = (StdioStream *)this ;

  assert(this) ;

  /* Sanity check on stream type */
  if ( sstdio->cookie != STDIO_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return ;
  }

  if ( sstdio->rbuffer )
    free(sstdio->rbuffer) ;
}

/* Just one file ioctl currently, which rewinds the file to the start */
static long stdioioctl(Stream *this, int op, va_list args)
{
  StdioStream *sstdio = (StdioStream *)this ;
  long where = 0 ;

  /* Sanity check on stream type */
  if ( sstdio->cookie != STDIO_STREAM_COOKIE ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  switch ( op ) {
  case STREAM_POSITION:
    where = va_arg(args, long) ;
  case STREAM_REWIND:
    if ( fseek(sstdio->fp, where, where < 0 ? SEEK_END : SEEK_SET) != 0 ) {
      SetLastError(ERROR_IO_DEVICE) ;
      return -1 ;
    }
    break ;
  case STREAM_CLEAREOF:
  case STREAM_CLEARERR:
    break ;
  default:
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  return 0 ;
}

/* Fill in details to make a stream into a file stream object */
Stream *MakeStdioStream(FILE *fp)
{
  Stream *this = NewStream(NULL, NULL, sizeof(StdioStream),
			   stdioreader, stdiowriter, stdioioctl,
			   stdiodestroy) ;

  if ( this ) {
    StdioStream *sstdio = (StdioStream *)this ;

    sstdio->fp = fp ;
    sstdio->rbufsize = 0 ;
    sstdio->rbuffer = NULL ;
    sstdio->cookie = STDIO_STREAM_COOKIE ;
  }

  return this ;
}

/* End of stdiostr.c */
