/* streams.c
 *
 * Copyright (C) Angus J. C. Duggan, 1997-1998
 *
 * Freely redistributable software. See file LICENSE for details.
 *
 * Streams support for network programs
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>
#include <assert.h>

#include "streams.h"
#include "streami.h"

/* Bitmask for flag values */
enum {STREAM_MYSTREAM = 1, STREAM_EOF = 2, STREAM_ERROR = 4} ;

/* Definition of opaque Stream type. All member access is through function
   calls. */
struct _Stream {
  struct _Stream *self ;/* Pointer to self allows opaque overlaying */
  short size ;		/* Structure size */
  short flags ;		/* What did we allocate and other info */
  Stream *under ;	/* underlying stream */
  Handler *reader ;	/* handler for reading */
  Handler *writer ;	/* handler for writing */
  Ioctl *ioctl ;	/* handler for IO control */
  Destructor *destructor ; /* handler for deletion */
#ifdef WIN32
  CRITICAL_SECTION lock ; /* mutual exclusion */
#endif
} ;

/* Streams use the self pointer to preserve opaqueness and allow sub-classing.
   In the actual stream structure, self will point directly to its own address.
   For sub-classes (determined by extrabytes > 0 in the NewStream call), self
   will point to the stream data, which will be sizeof(Stream) before the
   sub-class address. */
#define IS_A_STREAM(_s) \
  ((_s)->self == ((Stream *)(_s) - ((_s)->self->size != sizeof(Stream))))

/* Access underlying stream. */
Stream *UnderlyingStream(Stream *this)
{
  if ( this == NULL || !IS_A_STREAM(this) ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return NULL ;
  }

  SetLastError(ERROR_SUCCESS) ;

  return this->self->under ;
}

/* Create a new stream. If buffer and size supplied, use them, otherwise
   allocate a buffer. */
Stream *NewStream(Stream *this, Stream *under, short extrabytes,
		  Handler reader, Handler writer, Ioctl ioctl,
		  Destructor destructor)
{
  short flags = 0 ;
  Stream *self ;

  if ( extrabytes < sizeof(Stream *) && extrabytes != 0 ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return NULL ;
  }

  if ( this == NULL ) { /* Allocate a stream object if not supplied */
    if ( (self = malloc(sizeof(Stream) + extrabytes)) == NULL ) {
      SetLastError(ERROR_NOT_ENOUGH_MEMORY) ;
      return NULL ;
    }
    self->self = self ; /* Stream header pointer points to self */
    flags |= STREAM_MYSTREAM ;
  } else { /* Verify that supplied stream is correct size */
    self = this->self ;
    if ( !IS_A_STREAM(this) ||
	 self->size != sizeof(Stream) + extrabytes ) {
      SetLastError(ERROR_INVALID_PARAMETER) ;
      return NULL ;
    }
    flags = (self->flags & STREAM_MYSTREAM) ;
  }

  self->flags = flags ;		/* Whatever flags we've given it */
  self->under = under ;		/* underlying stream */
  self->size = sizeof(Stream) + extrabytes ; /* size of struct */
  self->reader = reader ;
  self->writer = writer ;
  self->ioctl = ioctl ;
  self->destructor = destructor ;
#ifdef WIN32
  InitializeCriticalSection(&self->lock) ;
#endif

  if ( extrabytes != 0 ) { /* Clear extra area, and adjust this pointer */
    this = self + 1 ;
    memset((void *)(this), 0, extrabytes) ;
    this->self = self ;
  } else
    this = self ;

  return this ;
}

/* Destroy stream object after closing stream, return underlying stream */
Stream *DestroyStream(Stream *this)
{
  Stream *under, *self ;

  if ( this == NULL || !IS_A_STREAM(this) ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return NULL ;
  }

  self = this->self ; /* Real stream data */
  under = self->under ;

  if ( self->destructor != NULL )
    (self->destructor)(this) ;

#ifdef WIN32
  DeleteCriticalSection(&self->lock) ;
#endif

  if ( (self->flags & STREAM_MYSTREAM) != 0 )
    free(self) ;

  return under ;
}

/* Lock stream object for exclusive access */
void LockStream(Stream *this)
{
#ifdef WIN32
  if ( this != NULL && IS_A_STREAM(this) )
    EnterCriticalSection(&this->self->lock) ;
#endif
}

/* Unlock stream object when finished with exclusive access */
void UnlockStream(Stream *this)
{
#ifdef WIN32
  if ( this != NULL && IS_A_STREAM(this) )
    LeaveCriticalSection(&this->lock) ;
#endif
}

/* Ioctl is for miscellaneous operations on streams, such as rewinding,
   resetting error and EOF conditions, or changing parameters. */
long IoctlStream(Stream *this, int op, ...)
{
  Stream *self ;
  va_list args ;
  long result = 0 ;

  if ( this == NULL || !IS_A_STREAM(this) ) {
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  self = this->self ;

  va_start(args, op) ;

  switch ( op ) {
  case STREAM_CLEAREOF:
    if ( (self->flags & STREAM_EOF) != 0 ) {	/* Needs clearing */
      if ( self->ioctl == NULL ||		/* Check Ioctl */
	   (result = (self->ioctl)(this, op, args)) >= 0 ) {
	if ( self->under )			/* Clear underlying */
	  result = IoctlStream(self->under, op, args) ;
      }
      if ( result >= 0 )
	self->flags &= ~STREAM_EOF ;
    }
    break ;
  case STREAM_CLEARERR:	/* Doesn't refer to underlying */
    if ( (self->flags & STREAM_ERROR) != 0 ) {	/* Needs clearing */
      if ( self->ioctl == NULL ||		/* Check Ioctl */
	   (result = (self->ioctl)(this, op, args)) >= 0 ) {
	if ( self->under )			/* Clear underlying */
	  result = IoctlStream(self->under, op, args) ;
      }
      if ( result >= 0 )
	self->flags &= ~STREAM_ERROR ;
    }
    break ;
  case STREAM_POSITION:	/* Must be serviced */
  case STREAM_REWIND:	/* Must be serviced */
    if ( self->ioctl == NULL )
      result = IoctlStream(self->under, op, args) ;
    else
      result = (self->ioctl)(this, op, args) ;
    if ( result >= 0 )
      self->flags &= ~(STREAM_ERROR|STREAM_EOF) ;
    break ;
  default:		/* Must be serviced */
    if ( self->ioctl == NULL ) {
      SetLastError(ERROR_INVALID_PARAMETER) ;
      return -1 ;
    }

    result = (self->ioctl)(this, op, args) ;
    break ;
  }

  va_end(args) ;

  return result ;
}

/* Read bytes from stream. Buffer pointer must always be specified, but if
   bufsize is zero, stream will buffer data itself. In this case, it assumes
   that the next read has consumed all of the bytes delivered previously. */
long ReadStream(Stream *this, void *buffer, long bufsize)
{
  Stream *self ;
  long result ;

  if ( this == NULL || !IS_A_STREAM(this) ) { /* Can't read from bad stream */
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  self = this->self ;

  if ( self->reader == NULL ) { /* No read permission */
    SetLastError(ERROR_ACCESS_DENIED) ;
    return -1 ;
  }

#ifdef WIN32
  EnterCriticalSection(&self->lock) ;
#endif

  if ( (self->flags & STREAM_ERROR) != 0 ) {
    SetLastError(ERROR_ACCESS_DENIED) ;
    result = -1 ;
  } else if ( (self->flags & STREAM_EOF) != 0 )
    result = 0 ;
  else {
    result = (self->reader)(this, buffer, bufsize) ;
    if ( result < 0 )
      self->flags |= STREAM_ERROR ;
    else if ( result == 0 )
      self->flags |= STREAM_EOF ;
  }

#ifdef WIN32
  LeaveCriticalSection(&self->lock) ;
#endif

  return result ;
}

long WriteStream(Stream *this, void *buffer, long bufsize)
{
  Stream *self ;
  long result ;

  if ( this == NULL || !IS_A_STREAM(this) ) { /* Can't write to bad stream */
    SetLastError(ERROR_INVALID_PARAMETER) ;
    return -1 ;
  }

  self = this->self ;

  if ( self->writer == NULL ) { /* No write permission */
    SetLastError(ERROR_ACCESS_DENIED) ;
    return -1 ;
  }

#ifdef WIN32
  EnterCriticalSection(&self->lock) ;
#endif

  if ( (self->flags & STREAM_ERROR) != 0 ) {
    SetLastError(ERROR_ACCESS_DENIED) ;
    result = -1 ;
  } else {
    result = (self->writer)(this, buffer, bufsize) ;
    if ( result < 0 )
      self->flags |= STREAM_ERROR ;
  }

#ifdef WIN32
  LeaveCriticalSection(&self->lock) ;
#endif

  return result ;
}

/* WriteStreamAll forces output to a stream; it returns true or false values
   for success or failure */
long WriteStreamAll(Stream *this, char *buffer, long bufsize)
{
  while ( bufsize > 0 ) {
    long result = WriteStream(this, buffer, bufsize) ;

    if ( result <= 0 )
      return 0 ;

    bufsize -= result ;
    buffer += result ;
  }

  return 1 ;
}

/*****************************************************************************/

/* Formatted output to stream. Limited to BUFSIZ characters */

long SendStream(Stream *this, char *format, ...)
{
  va_list args ;
  long result = 0 ;

  va_start(args, format) ;

  if ( strchr(format, '%') != NULL ) {
    char buffer[BUFSIZ] ;
    char ch, *here = buffer ;
    
    while ( (ch = *format++) != '\0' ) {
      long len = here - buffer ;

      if ( len >= BUFSIZ - 15 ) {
	if ( WriteStreamAll(this, buffer, len) ) {
	  here = buffer ;	/* Reset output buffer */
          result += len ;
	} else
	  return -1 ;	/* Write failed */
      }

      if ( ch == '%' ) {
	switch ( *format++ ) {
	case 'd':	/* decimal */
	  sprintf(here, "%d", va_arg(args, long)) ;
	  here += strlen(here) ;
	  break ;
	case 's':	/* string */
	  strcpy(here, va_arg(args, char *)) ;
	  here += strlen(here) ;
	  break ;
	case 'c':	/* char */
	  *here++ = va_arg(args, char) ;
	  break ;
	case 't':	/* date time in NNTP format */
	  {
#ifdef WIN32
	    LPFILETIME when = va_arg(args, LPFILETIME) ;
	    SYSTEMTIME stime ;

	    if ( !FileTimeToSystemTime(when, &stime) )
	      return -1 ;

	    sprintf(here, "%02d%02d%02d %02d%02d%02d",
		    (stime.wYear % 100), stime.wMonth, stime.wDay,
		    stime.wHour, stime.wMinute, stime.wSecond) ;
	    here += strlen(here) ;
#else
	    return -1 ; /* NYI */
#endif
	  }
	  break ;
	case '%':
	  *here++ = '%' ;
	  break ;
	default:
	  assert(0) ;
	  return -1 ;
	}
      } else
	*here++ = ch ;
    }

    if ( !WriteStreamAll(this, buffer, here - buffer) )
      return -1 ;

    result += here - buffer ;
  } else {	/* no escape codes, write out directly */
    result = strlen(format) ;
    if ( !WriteStreamAll(this, format, result) )
      return -1 ;
  }

  va_end(args) ;

  return result ;
}

/* end of streams.c */
